Require Import BakerLemma.
Require Import Baker.
Require Import VGC.
Require Cheney.


Set Implicit Arguments.


(* various lemmas *)

Definition fieldMapped f h x :=
  lexists v', !(exists v, f v = v' /\ h ||- x |-> v ** top) ** x |-> v'.

Definition objMapped f h x :=
  fieldMapped f h x ** fieldMapped f h (4 + x).


Lemma fieldMappedEqv : forall h S,
  h ||- (foralls x in S, x |->__) -> forall f h',
  h' ||- (foralls x in S, objMapped f h x) ->
  Heap.heap_eqv h' (Heap.HeapMap.map_mapRange f h).
Proof.
  unfold objMapped; unfold fieldMapped.
  intros h S Hp f h' Hp'.
  cut (forall x, Heap.get h' x = 
    match Heap.get h x with
    | Some y => Some (f y)
    | None => None
    end).

  clear S Hp Hp'; unfold Heap.get; intros Eq x; hyr Eq (Eq (4 * x)).
  rewrite div44eq in Eq; rewrite isPtrEq in Eq.
  assert (X := Heap.HeapMap.map_mapRangeGetEq h x f).
  destruct (Heap.HeapMap.map_get h x).
  transitivity (Some (f b)); [assumption | symmetry; assumption].
  transitivity (None (A := nat)); [assumption | symmetry; assumption].
  eapply ptrP_intro; reflexivity.

  intros x; destruct (set_In_dec x (S |_| set_map (fun x => 4 + x) S)) as [I | I].

  (* case: x is in the heap *)
  assert (E : exists y, set_In y S /\ (x = y \/ x = 4 + y)).
  destruct (set_union_elim I) as [I' | I']; clear I.
  exists x; auto.
  destruct (set_map_elim I') as [y [xEq Iy]]; clear I'.
  exists y; auto.
  destruct E as [y [Iy Xeq]]; clear I.
  clear Hp.
  hyr Hp' (fasElim Iy Hp'); lle Hp';
    destruct H5 as [v0 [x2Eq HpV0]]; destruct H6 as [v1 [x3Eq HpV1]].
  assert (E0 := contGetEqL HpV0); assert (E0' := contGetEqL Hp'); 
    sconjLift2 Hp'; assert (E1 := contGetEqL HpV1); assert (E1' := contGetEqL Hp');
      clear Hp'; subst x0 x1.
  destruct Xeq; subst x; 
    [rewrite E0; rewrite E0' | rewrite E1; rewrite E1']; reflexivity.

  (* case: x is not in the heap *)
  rewrite notInHeapDomNone.
  rewrite notInHeapDomNone.
  reflexivity.

  (* show x is not in h *)
  intros I'; apply I; clear I.
  assert (I : set_In x (set_mmap (fun x => set_add x (set_add (x + 4) empty_set)) S)).
  assert (X := objBufferDom Hp).
  rewrite <- X; assumption.
  clear I'; destruct (set_mmap_elim _ _ _ I) as [y [Iy Ix]];
    simpl in Ix; decompose [or] Ix; clear Ix; subst.
  sui1; assumption.
  sui2; apply set_map_intro with y; intuition.
  elf; assumption.

  (* show x is not in h' (similar to above) *)
  intros I'; apply I; clear I.
  assert (I : set_In x (set_mmap (fun x => set_add x (set_add (x + 4) empty_set)) S)).
  assert (Hp'' : h' ||- (foralls x in S, x |->__)).
  llMatch; lleim.
  assert (X := objBufferDom Hp'').
  rewrite <- X; assumption.
  clear I'; destruct (set_mmap_elim _ _ _ I) as [y [Iy Ix]];
    simpl in Ix; decompose [or] Ix; clear Ix; subst.
  sui1; assumption.
  sui2; apply set_map_intro with y; intuition.
  elf; assumption.
Qed.

Lemma eqvMapRangeImpl : forall h h',
  h ||- Heap.heap_eqv h' -> forall phi,
  Heap.HeapMap.map_mapRange phi h ||- Heap.heap_eqv (Heap.HeapMap.map_mapRange phi h').
Proof.
  intros h h' Hp phi.
  hyr Hp (eqvPredEqv Hp).
  apply eqvPredEqv'.
  apply Heap.HeapMap.map_mapRangeEqv; assumption.
Qed.


Lemma objCopiedHCopied : forall h x h' rf y,
  h ||- objCopied x (h', rf) y ->
  h ||- objCopiedH x h' y.
Proof.
  hpunfold; intros h x h' rf y Hp; lleim.
Qed.


Lemma fwdedObjHpOk : forall h x phi h0,
  h ||- objIsFwdedH phi h0 x -> forall X X',
  iso_set phi X X' ->
  h ||- okObjAt X' x.
Proof.
  hpunfold; intros h x phi h0 Hp X X' Iso; lle Hp.
  destruct H1 as [x'' [v0 [v1 [_ [Hp0 [Meq0 Meq1]]]]]];
    rename x into x'; rename x'' into x.
  simpl in Hp; simpl in Hp0; lle Hp0.
  lli.
  2:apply Hp.
  clear Hp Hp0.
  
  assert (IO := isOddOk v0); unfold mapX in Meq0; 
    destruct (isOdd v0); [left | right].
  injection Meq0; intro; subst v0; assumption.
  destruct Iso as [_ [Iso _]]; apply Iso with v0; assumption.

  assert (IO := isOddOk v1); unfold mapX in Meq1; 
    destruct (isOdd v1); [left | right].
  injection Meq1; intro; subst v1; assumption.
  destruct Iso as [_ [Iso _]]; apply Iso with v1; assumption.

Qed.


Lemma rrlistMapElim : forall v (f:rnum->nat) rr,
  set_In v (List.map f rr) ->
  exists r, In r rr /\ f r = v.
Proof.
  induction rr; intros I.
  elf; assumption.
  simpl in I; destruct I as [I | I].
  exists a; intuition.
  destruct (IHrr I) as [r [Ir Er]]; exists r; intuition.
Qed.


Lemma objIsFwdedHMerge : forall h phi h0 X Y,
  h ||- (foralls x in X, objIsFwdedH phi h0 x) ** (foralls x in Y, objIsFwdedH phi h0 x) ->
  h ||- (foralls x in X |_| Y, objIsFwdedH phi h0 x).
Proof.
  intros h phi h0 X Y Hp.
  apply fasMerge.
  assumption.
  assert (Hp' : h ||- fasd X ** fasd Y).
  hpunfoldIn Hp; llMatch; lleim; simpl in Hp; scm.
  apply (fasFasDisj' Hp'); auto.
Qed.


Lemma mapDomObjHp : forall h X M Y,
  h ||- objHp (M |_| X) Y -> forall phi G,
  iso_set phi M G ->
  set_disj X M ->
  (forall x, set_In x M -> even x) ->
  Heap.HeapMap.map_mapRange (fun x => match phi x with | Some y => y | None => x end) h ||-
    objHp (G |_| X) Y.
Proof.
  unfold objHp; hpunfold; intros h X M Y Hp phi G Iso Disj Meven.
  apply (fasMapRangeImpl Hp); clear h Hp; intros h x Hp;
    apply (sconjMapRangeImpl Hp); clear h Hp; intros h Hp;
      (lle Hp; lli; [idtac | apply (contMapRangeImpl Hp)]);
      cbv beta; clear x h Hp.
  (* first field *)
  assert (IO := isOddOk x0); destruct (isOdd x0); [left | right].
  destruct Iso as [Iso _]; destruct (Iso x0) as [_ Q]; destruct (phi x0).
  elf; apply (not_even_and_odd x0); auto; apply Meven; apply Q; discriminate.
  assumption.
  destruct H1 as [okX | okX0].
  elf; apply (not_even_and_odd x0); assumption.
  destruct (set_union_elim okX0) as [Ix | Ix]; clear okX0; [sui1 | sui2].
  destruct Iso as [Iso [Iso' _]]; destruct (Iso x0) as [Q _]; hyr Iso' (Iso' x0);
    destruct (phi x0); [eauto | elf; intuition].
  destruct Iso as [Iso _]; destruct (Iso x0) as [_ Q]; destruct (phi x0).
  elf; apply (set_disj_elim Disj x0); [assumption | apply Q; discriminate].
  assumption.

  (* second field, same as first *)
  assert (IO := isOddOk x0); destruct (isOdd x0); [left | right].
  destruct Iso as [Iso _]; destruct (Iso x0) as [_ Q]; destruct (phi x0).
  elf; apply (not_even_and_odd x0); auto; apply Meven; apply Q; discriminate.
  assumption.
  destruct H1 as [okX | okX0].
  elf; apply (not_even_and_odd x0); assumption.
  destruct (set_union_elim okX0) as [Ix | Ix]; clear okX0; [sui1 | sui2].
  destruct Iso as [Iso [Iso' _]]; destruct (Iso x0) as [Q _]; hyr Iso' (Iso' x0);
    destruct (phi x0); [eauto | elf; intuition].
  destruct Iso as [Iso _]; destruct (Iso x0) as [_ Q]; destruct (phi x0).
  elf; apply (set_disj_elim Disj x0); [assumption | apply Q; discriminate].
  assumption.
Qed.


Lemma mapDomObjHpDisj : forall h X M Y,
  h ||- objHp X Y -> forall phi G,
  iso_set phi M G ->
  set_disj X M ->
  (forall x, set_In x M -> even x) ->
  Heap.HeapMap.map_mapRange (fun x => match phi x with | Some y => y | None => x end) h ||-
    objHp X Y.
Proof.
  unfold objHp; hpunfold; intros h X M Y Hp phi G Iso Disj Meven.
  apply (fasMapRangeImpl Hp); clear h Hp; intros h x Hp;
    apply (sconjMapRangeImpl Hp); clear h Hp; intros h Hp;
      (lle Hp; lli; [idtac | apply (contMapRangeImpl Hp)]);
      cbv beta; clear x h Hp.
  
  (* first field *)
  assert (IO := isOddOk x0); destruct (isOdd x0); [left | right].
  destruct Iso as [Iso _]; destruct (Iso x0) as [_ Q]; destruct (phi x0).
  elf; apply (not_even_and_odd x0); auto; apply Meven; apply Q; discriminate.
  assumption.
  destruct H1 as [okX | okX0].
  elf; apply (not_even_and_odd x0); assumption.
  destruct Iso as [Iso _]; destruct (Iso x0) as [_ Q]; destruct (phi x0).
  elf; apply (set_disj_elim Disj x0); [assumption | apply Q; discriminate].
  assumption.

  (* second field, same as the first *)
  assert (IO := isOddOk x0); destruct (isOdd x0); [left | right].
  destruct Iso as [Iso _]; destruct (Iso x0) as [_ Q]; destruct (phi x0).
  elf; apply (not_even_and_odd x0); auto; apply Meven; apply Q; discriminate.
  assumption.
  destruct H1 as [okX | okX0].
  elf; apply (not_even_and_odd x0); assumption.
  destruct Iso as [Iso _]; destruct (Iso x0) as [_ Q]; destruct (phi x0).
  elf; apply (set_disj_elim Disj x0); [assumption | apply Q; discriminate].
  assumption.

Qed.

Lemma heapMapSetEqv : forall h f h',
  Heap.heap_eqv h (Heap.HeapMap.map_mapRange f h') -> forall x y,
  Heap.heap_eqv (Heap.uset h x (f y))
    (Heap.HeapMap.map_mapRange f (Heap.uset h' x y)).
Proof.
  intros h f h' Heq x y.
  apply heap_eqvIntro; intros z.
  unfold Heap.get.
  destruct (isPtr z).

  compare x z; intro NE.
  subst z; simpl;
    destruct (POnat.POnat.eqDec (div4 x) (div4 x)); [auto | elf; auto].
  simpl; destruct (POnat.POnat.eqDec (div4 z) (div4 x)).
  reflexivity.
  apply Heq.
  
  reflexivity.
Qed.


(* Baker write barrier *)

Definition bakerWrite k :=
  sw a0 k rroot;
  jr ra.


Definition bakerWritePre : nat -> state -> Prop := VGC.writePre repr rroot.
Definition bakerWriteGuar : nat -> state -> state -> Prop := 
  VGC.writeGuar repr rroot BMGI.gcGuar BMGI.presReg.
Notation bakerWriteSpec := (fun k => bspeci (bakerWritePre k) (bakerWriteGuar k)).


Lemma bakerWriteOk : forall k P,
  k = 0 \/ k = 4 ->
  iseqOkS P (bakerWritePre k) (bakerWriteGuar k) (bakerWrite k).
Proof.
  intros k P K.
  unfold bakerWrite.
  iseqWeakSolver fail.
  intros s [vs [Hp Vhp]]; unfold repr in Hp.
  lle Vhp. lle Hp.
  rename x2 into toSt; rename x4 into scan; rename x5 into free; rename x6 into alloc;
    rename x3 into toEnd; rename x7 into phi; rename x8 into W; rename x9 into M; 
      rename x0 into frSt; rename x1 into frEnd; rename x10 into vh.
  destruct H3 as [_ [Al [Al' [WMeq [Iso [Rok Vhp']]]]]].
  set (B := rangeObjs toSt scan) in * ;
    set (G := rangeObjs scan free) in * ;
      set (B' := rangeObjs alloc toEnd) in * .

  assert (rootEq : rget vs rroot = rget s rroot).
    destruct (Rok rroot).
    simpl; auto.
    assumption.

  assert (I : set_In (rget s rroot) (B |_| G |_| B') /\ even (rget s rroot)).
    destruct (Rok rroot).
    simpl; auto.
    assert (X := isOddOk (rget s rroot)).
    destruct (isOdd (rget s rroot)).
    elf; apply (not_even_and_odd (rget s rroot)).
    apply ptrPeven.
    rewrite H in Vhp; assert (Y := contElim1L Vhp).
    destruct K; subst k.
    assumption.
    inversion Y; assumption.
    assumption.
    constructor.
    destruct H0.
    elf; apply (not_even_and_odd (rget s rroot)); assumption.
    assumption.
    assumption.

  destruct I as [I rootEven].

  assert (Hp' : s |||- (Heap.heap_eqv vh && 
    (k + rget vs rroot |->_ ** top)) ** top).
    scm.
    destruct (aconjElim Hp) as [Hp1 Hp2]; apply aconjIntro.
    assumption.
    assert (Hp' : h ||- (foralls x in B |_| G |_| B', x |->__) ** top).
      sconjLift3of3 Hp2; sconjLift3of3 Hp2; hyr Hp2 (sconjAssocR Hp2); scm.
      assert (Hp' : h0 ||- (foralls x in B |_| B', x |->__) ** (foralls x in G, x |->__)).
        scm; exact (objHpIsObjs Hp0).
      clear Hp0 Hp.
      unfold hpredTrue; rewrite (set_eqvUnionSymm G B'); 
        rewrite <- (set_eqvUnionAssoc B B' G); apply fasMerge.
      assumption.
      apply (fasFasDisj' Hp'); illeim.
    assert (Hp'' : h ||- (rget s rroot |->__ ** top) ** top).
      scm; hyr Hp0 (fasElim I Hp0); scm.
    clear Hp Hp1 Hp2 Hp'; lle Hp''; destruct K; subst k.
    rewrite rootEq; simpl in Hp''; lli; scm.
    rewrite rootEq; simpl in Hp''; sconjLift2 Hp''; lli; scm.
  
  lle Hp'.
  destruct (aconjElim (sconjDistrib2 Hp')) as [_ Hp1]; clear Hp'; lle Hp1;
    rewrite rootEq in Hp1.
  rewrite (swStep Hp1); try reflexivity.
  constructor.
  trivial.

  intros s' S'eq; subst s'.
  clear frSt frEnd toSt toEnd scan free alloc phi W M vh Al Al' WMeq B G Iso B' Rok
    Vhp' Hp I Hp1 vs x Vhp rootEq x0.

  constructor.
  2:intros r X; stSimpl; reflexivity.

  (* show the guarantee holds *)
  unfold repr.
  assert (x := 0). (* this is just a shim to avoid some renaming *)
  intros vs A rr Hp II; lle Hp.
  rename x2 into toSt; rename x4 into scan; rename x5 into free; rename x6 into alloc;
    rename x3 into toEnd; rename x7 into phi; rename x8 into W; rename x9 into M; 
      rename x0 into frSt; rename x1 into frEnd; rename x10 into vh.
  clear x.
  destruct H3 as [NIR [Al [Al' [WMeq [Iso [Rok Vhp']]]]]].
  set (B := rangeObjs toSt scan) in * ;
    set (G := rangeObjs scan free) in * ;
      set (B' := rangeObjs alloc toEnd) in * .

  assert (rootEq : rget vs rroot = rget s rroot).
    destruct (Rok rroot).
    apply II; simpl; auto.
    assumption.

  assert (I : set_In (rget s rroot) (B |_| G |_| B')).
    destruct (Rok rroot).
    apply II; simpl; auto.
    destruct H0.
    elf; apply (not_even_and_odd (rget s rroot)); assumption.
    assumption.

  assert (a0Eq : rget vs a0 = rget s a0 /\ okFieldVal (B |_| G |_| B') (rget s a0)).
    destruct (Rok a0); [apply II; simpl; auto | constructor; assumption].
  destruct a0Eq as [a0Eq a0Ok].

  set (root := rget s rroot) in * .
  set (newf := rget s a0) in * .

  exists (uhset vs (k + root) newf); constructor.
  rewrite rootEq; rewrite a0Eq; apply Cheney.gcGuarRefl.

  assert (HpZ := heapEqvImplR Hp).
  destruct (aconjElim (sconjDistrib2 Hp)) as [Hp' _];
    clear Hp; rename Hp' into Hp.

  assert (toFrDisj : set_disj (W |_| M) ((B |_| B') |_| G)).
    hyr Hp (sconjInsert Hp HpZ); lle Hp.
    sconjLift4 Hp; sconjLift2 Hp; apply sconjAssocR in Hp; sconjLift3 Hp; sconjLift3 Hp; 
      apply sconjAssocR in Hp; sconjLift2 Hp.
    assert (HX : s |||- fasd (W |_| M) ** fasd ((B |_| B') |_| G) ** top).
      scm; apply fasdMerge.
      unfold objHp in Hp; hpunfoldIn Hp; llMatch.
      lleim.
      hyr Hp0 (mapHpIsObjs Hp); llMatch; lleim.
      scm; hyr Hp0 (objHpIsObjs Hp); llMatch; lleim.
    apply (fasFasDisj HX); auto.

  lli.
  constructor.
  assumption.
  constructor.
  apply Al.
  constructor.
  apply Al'.
  constructor.
  2:eauto.
  eauto.
  constructor.
  eauto.
  constructor.
  rename II into II0.
  intros r II.
  replace (RFile.get (rfileOf (uhset s (k + root) newf)))
    with (rget (uhset s (k + root) newf)); [stSimpl | reflexivity].
  exact (Rok _ II).

  instantiate (1 := Heap.uset vh (k + root) newf).

  replace (heapOf (uhset vs (k + root) newf))
    with (Heap.uset (heapOf vs) (k + root) newf) by reflexivity.

  set (pmap := fun x : POnat.OneTypeNat.B => match phi x with Some y => y | None => x end) in * .
  apply Heap.heap_eqvTrans with (Heap.uset (heapOf vs) (k + root) (pmap newf)).
  2:apply heapMapSetEqv; assumption.
  replace (pmap newf) with newf.
  apply Heap.heap_eqvRefl.
  subst pmap; destruct Iso as [Iso _]; destruct (Iso newf) as [_ Q]; destruct (phi newf).
  elf; destruct a0Ok as [fOdd | IQ].
  apply (not_even_and_odd newf).
  apply ptrPeven; sconjLift2 Hp.
  assert (HQ : s |||- (newf |->_ ** top) ** top).
  assert (IM : set_In newf M).
  apply Q; discriminate.
  scm; hyr Hp (fasElim IM (mapHpIsObjs Hp)); lleim.
  lle HQ; apply (contElim1L HQ).
  assumption.
  apply (set_disj_elim toFrDisj newf).
  sui2; apply Q; discriminate.
  rewrite (set_eqvUnionAssoc B B' G); rewrite (set_eqvUnionSymm B' G); assumption.
  reflexivity.

  (* step forward the state *)


  assert (SEq : Heap.set vh (k + root) newf =
    Some (Heap.uset vh (k + root) newf)).
  assert (HX : vh ||- k + root |->_ ** top).
  assert (HX : vh ||- (root |->__ ** top) ** top).
  assert (IX : set_In root ((B |_| B') |_| G)).
  set_InElim; [sui1; sui1 | sui2 | sui1; sui2]; assumption.
  destruct (set_union_elim IX) as [Ix | Ix]; clear IX; [sconjLift2 HpZ | sconjLift3of3 HpZ];
    scm; hyr Hp0 (fasElim Ix (objHpIsObjs Hp0)); lleim.
  clear HpZ; lle HX; destruct K; subst k; [idtac | sconjLift2 HX]; lli; sconjMatch.
  lle HX; assert (X := contHsetEqL HX (vh, rfileOf s)).
  assert (E : heapOf (vh, rfileOf s) = vh).
  reflexivity.
  hyr X (X E newf); unfold uhset in X; unfold hset in X; clear E; simpl in X.
  destruct (Heap.set vh (k + root) newf).
  injection X; intro E; rewrite E; reflexivity.
  discriminate X.

  hyr Hp (sconjHpSet Hp (k + root) newf SEq).
  replace (Heap.uset (heapOf s) (k + root) newf) with 
    (heapOf (uhset s (k + root) newf)) in Hp; [idtac | reflexivity].
  scm.
  apply aconjIntro.
  assumption.
  eapply heapEqvTrue.
  apply Heap.heap_eqvSymm; apply eqvPredEqv; apply Hp.
  clear h Hp SEq vs Vhp' Rok rootEq a0Eq II rr NIR.
  replace (rangeObjs toSt scan) with B; [idtac | reflexivity];
    replace (rangeObjs scan free) with G; [idtac | reflexivity];
      replace (rangeObjs alloc toEnd) with B'; [idtac | reflexivity].
  (* showing the heap update is type safe on the concrete heap *)
  assert (I' : set_In root ((B |_| B') |_| G)).
  set_InElim; [sui1; sui1 | sui2 | sui1; sui2]; assumption.
  clear I; destruct (set_union_elim I') as [I | I]; clear I'.
  (* case: root in B |_| B' *)
  assert (Hp : vh ||- (okObjAt (B |_| G |_| B') root ** 
    objHp (B |_| G |_| B') (set_remove root (B |_| B'))) ** objHp (W |_| M) W **
    objHp ((W |_| M) |_| B |_| G |_| B') G).
  sconjLift2 HpZ; scm; hyr Hp (objHpElim I Hp); scm.
  clear HpZ; hpunfoldIn Hp; lle Hp.
  destruct K; subst k; [idtac | sconjLift2 Hp];
    hyr Hp (contUpdateSt Hp newf); hyr Hp (sconjAssocR (sconjAssocR Hp)); 
      sconjLift2 Hp; scm;
        unfold hpredTrue; rewrite <- (addRemoveEqv I); apply objHpIntro; scm;
	  [idtac | sconjLift2of2 Hp];
	  (hpunfold; lli; [idtac | apply Hp | idtac]; assumption).
  (* case: root in G *)
  assert (Hp : vh ||- (okObjAt ((W |_| M) |_| B |_| G |_| B') root ** 
    objHp ((W |_| M) |_| B |_| G |_| B') (set_remove root G)) ** objHp (W |_| M) W **
    objHp (B |_| G |_| B') (B |_| B')).
  sconjLift3of3 HpZ; scm; hyr Hp (objHpElim I Hp); scm.
  clear HpZ; hpunfoldIn Hp; lle Hp.
  assert (newfOk' : okFieldVal ((W |_| M) |_| B |_| G |_| B') newf).
  apply (okFieldValWeaken a0Ok); intros z Iz; sui2; assumption.
  clear a0Ok.
  destruct K; subst k; [idtac | sconjLift2 Hp];
    hyr Hp (contUpdateSt Hp newf); hyr Hp (sconjAssocR (sconjAssocR Hp)); 
      sconjLift3of3 Hp; sconjLift3of3 Hp; scm;
        unfold hpredTrue; rewrite <- (addRemoveEqv I); 
	  apply objHpWeaken with ((W |_| M) |_| B |_| G |_| B').
  apply objHpIntro; scm; hpunfold; lli; [idtac | apply Hp | idtac]; assumption.
  intros z Iz; rewrite (addRemoveEqv I); assumption.
  apply objHpIntro; scm; sconjLift2of2 Hp; hpunfold; lli; [idtac | apply Hp | idtac]; assumption.
  intros z Iz; rewrite (addRemoveEqv I); assumption.
Qed.




(* Baker read barrier *)

Definition rdTmp1 := t3.
Definition rdTmp2 := t6.


Definition BAKERREADRET := 38.


Definition bakerRead k :=
  (* load the registers we need for scanField *)
  lw rfrStart 0 rgcInfo;
  lw rfrEnd 4 rgcInfo;
  lw rtoStart 8 rgcInfo;
  lw rtoEnd 12 rgcInfo;
  lw rfree 20 rgcInfo;
  lw ralloc 24 rgcInfo;

  (* call scanField on the field we are reading *)
  addiu rdTmp1 ra 0;
  addiu a0 rroot k;
  addiu rdTmp2 a0 0;
  jal SCANFIELD BAKERREADRET.

Definition bakerReadRet :=
  (* load the forwarded value and return *)
  lw v0 0 rdTmp2;
  sw rfree 20 rgcInfo;
  addiu ra rdTmp1 0;
  jr ra.


Definition bakerReadPre : nat -> state -> Prop := VGC.readPre repr rroot.
Definition bakerReadGuar : nat -> state -> state -> Prop := 
  VGC.readGuar repr rroot BMGI.gcGuar BMGI.presReg.
Notation bakerReadSpec := (fun k => bspeci (bakerReadPre k) (bakerReadGuar k)).


Definition bakerReadRetPre s := 
  s |||- rget s rdTmp2 |->_ ** 20 + rget s rgcInfo |->_ ** top.

Definition bakerReadRetGuar s s' :=
  (forall A x y,
    s |||- rget s rdTmp2 |-> x ** 20 + rget s rgcInfo |-> y ** A ->
    s' |||- rget s rdTmp2 |-> x ** 20 + rget s rgcInfo |-> rget s rfree ** A /\
      rget s' v0 = x) /\
  regPres (regNotInList (v0::ra::nil)) s s' /\
  rget s' ra = rget s rdTmp1.

Notation bakerReadRetSpec := (bspeci bakerReadRetPre bakerReadRetGuar).


Lemma finalReadReprOk : forall k s s' s'' toSt toEnd scan free alloc frSt frEnd,
  forall rr A x h0 h00 x0 x1 M W M' W' G' F' phi phi' vs,
  k = 0 \/ k = 4 ->
  let k' := if beq_nat k 0 then 4 else 0 in
  let root := rget s rroot in
  let free' := rget s' rfree in
  toSt <= free' <= alloc ->
  alloc <= toEnd ->
  rget s'' v0 = x ->
  rget s'' ra = rget s ra ->
  aligned8 free' alloc ->
  aligned8 frSt frEnd ->
  aligned8 free free' ->
  aligned8s (toSt :: scan :: free :: alloc :: toEnd :: nil) ->
  let GCI := rget s rgcInfo |->, (frSt::frEnd::toSt::toEnd::scan::free::alloc::nil) in
  let GCI1 := rget s rgcInfo |->, (frSt :: frEnd :: toSt :: toEnd :: scan :: nil) in
  let GCI' := rget s rgcInfo |->, (frSt::frEnd::toSt::toEnd::scan::free'::alloc::nil) in
  even root ->
  regPres
    (regNotInList
       (rfree::a0::v0::ast::rtemp0::rtemp1::rtemp2::raSave0::nil))
    (rset (rset (rset (rset (rset (rset (rset (rset 
       (rset (rset s rfrStart frSt) rfrEnd frEnd) rtoStart toSt) rtoEnd toEnd) rfree free)
     ralloc alloc) rdTmp1 (rget s ra)) a0 
    (k + root)) rdTmp2 (k + root)) ra BAKERREADRET) s' ->
  regPres (regNotInList (v0 :: ra :: nil)) s' s'' ->
  rget vs rroot = root ->
  In rroot rr ->
  (forall r : rnum, In r rr -> r = rroot \/ calleeSaved r) ->
  ~ In rgcInfo rr ->
  let mphi := (fun x => match map_union phi phi' x with Some y => y | None => x end) in
  let mphi1 := (fun x => match phi x with Some y => y | None => x end) in
  let mphi2 := (fun x => match phi' x with Some y => y | None => x end) in
  Heap.heap_eqv (heapOf vs) (Heap.HeapMap.map_mapRange mphi1 h0) ->

  let B := rangeObjs toSt scan in
  let G := rangeObjs scan free in
  let F := rangeObjs free alloc in
  let B' := rangeObjs alloc toEnd in
  set_eqv (W |_| M) (rangeObjs frSt frEnd) ->
  set_eqv (rangeObjs toSt toEnd) (B |_| G |_| F |_| B') ->
  set_eqv W (M' |_| W') ->
  set_eqv F (G' |_| F') ->
  set_eqv F' (rangeObjs free' alloc) ->
  iso_set phi M (B |_| G) ->
  iso_set phi' M' G' ->

  (forall r, In r rr ->
    rget vs r = rget s r /\ okFieldVal (B |_| G |_| B') (rget s r)) ->
  set_In root (B |_| G |_| B') ->

  mapX (map_unionnn (spaceIdMap toSt toEnd) (map_union phi phi')) x0 = Some x ->

  s |||- (Heap.heap_eqv h0 && objHp (W |_| M) W **
    objHp (B |_| G |_| B') (B |_| B') ** objHp ((W |_| M) |_| B |_| G |_| B') G) **
    mapHp M phi ** (foralls x in F, x |->__) ** GCI ** A ->

  s |||- k + root |-> x0 ** top ->
  s |||- k' + root |-> x1 ** top ->

  s'' |||- k + root |-> x ** Heap.heap_eqv h00 **
    mapHp (M |_| M') (map_union phi phi') ** (foralls x in F', x |->__) ** GCI' ** A ->

  okFieldVal ((B |_| G |_| B') |_| if set_mem root G then W |_| M else empty_set) x0 ->
  okFieldVal ((B |_| G |_| B') |_| if set_mem root G then W |_| M else empty_set) x1 ->

  h00 ||- (foralls x in set_remove root (B |_| B' |_| G) |_| W', objCopied x s x) **
    k' + root |-> x1 ** (foralls x in G', objCopiedPhi phi' s x) ->
  h00 ||- k' + root |-> x1 ** objHp (B |_| G |_| B') (set_remove root (B |_| B')) **
    objHp ((W |_| M) |_| B |_| G |_| B') (set_remove root G) **
    objHp (W' |_| M |_| M') W' **
    (foralls x in G', objCopiedPhi phi' s x) ->

  set_disj (B |_| B' |_| G) G' ->
  set_disj (B |_| B') G ->
  set_disj (setRange frSt frEnd) (setRange toSt toEnd) ->

  let h1 := Heap.uset h00 (k + root) x in
  let vh' := Heap.HeapMap.map_mapRange mphi h1 in

  (s'' |||- repr (vh', rfileOf s'') (rfileOf s'') (v0 :: rr) ** A) /\
  vh' ||- k + rget s'' rroot |-> rget s'' v0 ** top.
Proof.
  intros k s s' s'' toSt toEnd scan free alloc frSt frEnd
    rr A x h0 h00 x0 x1 M W M' W' G' F' phi phi' vs.
  intros K k' root free' LE1 LE2 xEq raEq al1 al2 al3 al4 GCI GCI1 GCI'.
  intros rootEven Req Req' rootEq rootIn RReq RRNi mphi mphi1 mphi2
    vsEq B0 G F B' frEq toEq Weq Feq F'eq Iso Iso' rootOk Iroot Meq Hp
    HpX0 HpX1 Hp'' x0Ok x1Ok Hp00 Hp00' BBGGdisj BBGdisj frToDisj h1 vh'.

  assert (MMdisj : set_disj (M |_| M') ((B0 |_| G |_| B') |_| G')).
    apply set_unionDisjElimL with W'; rewrite (set_eqvUnionAssoc M M' W'); rewrite <- Weq.
    apply set_disj_intro; intros z Iz Iz'; apply (set_disj_elim frToDisj z);
      apply objRangeSetRangeSubset.
    rewrite <- frEq; rewrite (set_eqvUnionSymm W M); assumption.
    rewrite toEq; clear Iz; set_InElim; [sui1 | sui2; sui1 | sui2; sui2; sui2 |
      sui2; sui2; sui1; rewrite Feq; sui1]; assumption.

  assert (CSO : calleeSavedOk s s'').
    intros z Ze.
    unfold calleeSaved in Ze; simpl in Ze; decompose sum Ze; clear Ze; subst z;
      try (rewrite <- Req'; [rewrite <- Req; [stSimpl; reflexivity | regNotInListSolver] | 
        regNotInListSolver]).
    auto.

  assert (MMW'disj : set_disj (M |_| M') W').
    assert (HpW : h00 ||- fasd W' ** top).
      sconjLift4 Hp00'; scm; hyr Hp0 (objHpIsObjs Hp0); llMatch; lleim.
    assert (HpX : s'' |||- fasd (M |_| M') ** fasd W' ** top).
      sconjLift2 Hp''; hyr Hp'' (sconjInsert Hp'' HpW); lle Hp''.
      sconjLift3 Hp''; scm; hyr Hp0 (mapHpIsObjs Hp0); llMatch; lleim.
    apply (fasFasDisj HpX); auto.

  assert (Iso12 : iso_set (map_union phi phi') (M |_| M') ((B0 |_| G) |_| G')).
    apply iso_union.
    assumption.
    assumption.
    (* set_disj M M' *)
    apply set_disjSymm; apply set_unionDisjElimL with W'; rewrite <- Weq.
    destruct (aconjElim (sconjDistrib2 Hp)) as [_ HpX].
    lle HpX; sconjLift4 HpX; sconjLift2 HpX; unfold objHp in HpX; unfold mapHp in HpX;
      lle HpX; apply (fasFasDisj HpX); uilleim.
    apply set_unionDisjElimL with B'.
    setoid_replace ((B0 |_| G) |_| B') with (B0 |_| B' |_| G).
    assumption.
    rewrite (set_eqvUnionSymm B' G); rewrite (set_eqvUnionAssoc B0 G B'); apply set_eqvRefl.

  assert (GGeq : set_eqv (rangeObjs scan free') (G |_| G')).
    assert (G'eq : set_eqv G' (rangeObjs free free')).
      apply rangeSplit with alloc.
      rewrite <- F'eq.
      assert (Hp00Z : h00 ||- top ** top ** fasd G').
        clear Hp00'; scm.
        llMatch; hpunfoldIn Hp0; lleim.
      sconjLift2 Hp''; hyr Hp'' (sconjInsert Hp'' Hp00Z); lle Hp''.
      sconjLift4 Hp''; sconjLift2 Hp''; apply (fasFasDisj Hp''); illeim.
      destruct al1; omega.
      rewrite <- F'eq; rewrite <- Feq; apply set_eqvRefl.
      assumption.
    assert (LT : free <= free').
      destruct al3; omega.
    destruct al4 as [_ [X _]].
    rewrite <- (rangeObjsTransEqv X LT).
    rewrite <- G'eq; apply set_eqvRefl.

  assert (xOk : okFieldVal ((B0 |_| G |_| B') |_| G') x).
    unfold mapX in Meq; assert (IO := isOddOk x0); destruct (isOdd x0); [left | right].
    injection Meq; intro; subst x0; assumption.
    destruct x0Ok as [Q | Ix0].
    elf; apply (not_even_and_odd x0); assumption.
    unfold map_unionnn in Meq; unfold map_union at 1 in Meq;
      rewrite spaceIdMapEq in Meq.
    destruct (set_union_elim Ix0) as [Ix | Ix]; clear Ix0.
    (* x0 in to space *)
    rewrite set_mem_true in Meq.
    injection Meq; intro; subst x0; sui1; assumption.
    apply objRangeSetRangeSubset; rewrite toEq.
    rewrite (set_eqvUnionSymm F B'); rewrite <- (set_eqvUnionAssoc G B' F);
      rewrite <- (set_eqvUnionAssoc B0 (G |_| B') F); sui1; assumption.
    destruct (set_mem root G).
    2:elf; assumption.
    (* x0 in from space *)
    rewrite set_mem_false in Meq.
    assert (I' : set_In x0 ((M |_| M') |_| W')).
    rewrite (set_eqvUnionAssoc M M' W'); rewrite <- Weq; rewrite (set_eqvUnionSymm M W);
      assumption.
    destruct (set_union_elim I') as [Ix' | Ix']; clear Ix I'.
    (* case: x0 in M |_| M' *)
    destruct Iso12 as [_ [Iso12 _]]; hyr Iso12 (Iso12 _ _ Meq).
    set_InElim; [sui1; sui1 | sui1; sui2; sui1 | sui2]; assumption.
    (* case: x0 in W'.  contradiction. *)
    elf; destruct Iso12 as [Iso12 _]; destruct (Iso12 x0) as [_ Q];
      destruct (map_union phi phi' x0).
    apply (set_disj_elim MMW'disj x0).
    apply Q; discriminate.
    assumption.
    discriminate Meq.
    intro I'; apply (set_disj_elim frToDisj x0).
    apply objRangeSetRangeSubset; rewrite <- frEq; assumption.
    assumption.

  constructor.

  unfold repr.
  hyr Hp'' (sconjAssocR (sconjAssocR (sconjAssocR (sconjAssocR Hp'')))); scm.
  existsl frSt; existsl frEnd; existsl toSt; existsl toEnd; existsl scan;
    existsl free'; existsl alloc; existsl (map_union phi phi');
    existsl W'; existsl (M |_| M'); existsl h1.
  stSimpl.
  replace (RFile.get (rfileOf s'')) with (rget s'') by reflexivity.
  replace (rangeObjs toSt scan) with B0 by reflexivity.
  replace (rangeObjs alloc toEnd) with B' by reflexivity.
  lli.
  repeat (apply conj; [auto | ]).
  intros [Z | Z]; [discriminate Z | auto].
  simpl; simpl in al4; decompose [and] al4;
    repeat (constructor; try assumption);
      apply aligned8Trans with free; assumption.
  rewrite (set_eqvUnionSymm M M'); rewrite <- (set_eqvUnionAssoc W' M' M);
    rewrite (set_eqvUnionSymm W' M'); rewrite <- Weq; assumption.
  rewrite GGeq; rewrite <- (set_eqvUnionAssoc B0 G G'); assumption.
  intros r Irr; constructor.
  reflexivity.
  rewrite GGeq. simpl in Irr; destruct Irr. subst r; rewrite xEq.
  apply (okFieldValWeaken xOk); intros z Iz; set_InElim;
    [sui1 | sui2; sui1; sui1 | sui2; sui2 | sui2; sui1; sui2]; assumption.
  replace (rget s'' r) with (rget s r).
  destruct (rootOk _ H) as [_ rOk]; apply (okFieldValWeaken rOk); intros z Iz;
    set_InElim; [sui1 | sui2; sui1; sui1 | sui2; sui2]; assumption.
  destruct (RReq _ H).
  subst r; rewrite <- Req'; [idtac | regNotInListSolver];
    rewrite <- Req; [idtac | regNotInListSolver];
      stSimpl; reflexivity.
  rewrite <- (CSO _ H0); reflexivity.
  apply Heap.heap_eqvRefl.

  (* show the concrete heap is well-formed *)
  lle Hp0; hyr Hp0 (sconjAssocR Hp0); scm.
  2:unfold hpredTrue; rewrite <- F'eq; assumption.
  2:rewrite <- Req'; [ | regNotInListSolver];
    rewrite <- Req; [stSimpl | regNotInListSolver]; assumption.
  subst h1; apply aconjIntro.
  hyr Hp0 (sconjAssocL (unitIR Hp0)); apply unitER; apply heapEqvAdd; assumption.
  sconjLift2of2 Hp0; hyr Hp0 (sconjInsert Hp0 Hp00'); clear Hp00 Hp00'; lle Hp0;
    sconjLift6of6 Hp0; hyr Hp0 (sconjAssocR Hp0).
  assert (Hp0' : h ||- okObjAt ((B0 |_| (G |_| G') |_| B')
    |_| (if set_mem root G then W |_| M else empty_set)) root **
    objHp (B0 |_| G |_| B') (set_remove root (B0 |_| B')) **
    objHp ((W |_| M) |_| B0 |_| G |_| B') (set_remove root G) **
    objHp (W' |_| M |_| M') W' ** objHp ((W |_| M) |_| B0 |_| G |_| B') G').
    scm.
    (* show root is well-typed *)
    assert (x1Ok' : okFieldVal ((B0 |_| (G |_| G') |_| B')
          |_| (if set_mem root G then W |_| M else empty_set)) x1).
      apply (okFieldValWeaken x1Ok); intros z Iz; set_InElim;
        [sui1; sui1 | sui1; sui2; sui1; sui1 | sui1; sui2; sui2 | sui2]; assumption.
    assert (xOk' : okFieldVal ((B0 |_| (G |_| G') |_| B')
      |_| (if set_mem root G then W |_| M else empty_set)) x).
      apply (okFieldValWeaken xOk); intros z Iz; set_InElim;
        sui1; [sui1 | sui2; sui1; sui1 | sui2; sui2 | sui2; sui1; sui2]; assumption.
    hpunfold; destruct K; subst k' k; [idtac | sconjLift2of2 Hp0]; scm;
      (lli; [idtac | apply Hp0]; assumption).
    (* done showing root, now show G' *)
    apply (objHpIso _ _ Hp0) with M' top.
    destruct (aconjElim (sconjDistrib2 Hp)) as [_ Hp'].
    assert (HpS : s |||- (objHp (W |_| M) M' ** objHp (W |_| M) W') ** top).
      clear HpX0 HpX1 Hp; lle Hp'; scm; apply objHpSplit.
      unfold hpredTrue; rewrite <- Weq; assumption.
      apply set_unionDisjElimR with M; assumption.
    clear HpX0 HpX1; lle HpS; scm.
    apply (objHpWeaken Hp1); intros z Iz; sui1; assumption.
    assumption.
  (* done showing heap has some type *)
  sconjLift4 Hp0'; apply (sconjImplR Hp0'); clear Hp0'; intros h' Hp'.
  assert (SubB : set_subset (B0 |_| G |_| B')
    (B0 |_| rangeObjs scan free' |_| B')).
    rewrite GGeq; intros z Iz; set_InElim;
      [sui1 | sui2; sui1; sui1 | sui2; sui2]; assumption.
  assert (SubG : set_subset ((W |_| M) |_| B0 |_| G |_| B')
    ((W' |_| M |_| M') |_| B0 |_| rangeObjs scan free' |_| B')).
    rewrite GGeq; intros z Iz; destruct (set_union_elim Iz) as [Iz' | Iz'].
    sui1; rewrite (set_eqvUnionSymm M M'); rewrite <- (set_eqvUnionAssoc W' M' M);
    rewrite (set_eqvUnionSymm W' M'); rewrite <- Weq; assumption.
    sui2; clear Iz; set_InElim; [sui1 | sui2; sui1; sui1 | sui2; sui2]; assumption.
  assert (Iroot' := set_memSound root G); destruct (set_mem root G).
  (* case: root in G *)
  sconjLift2 Hp'; apply (sconjImpl Hp'); clear h' Hp'; intros h' Hp'.
  rewrite set_removeEq in Hp'.
  apply (objHpWeaken Hp'); assumption.
  intros Iroot''; apply (set_disj_elim BBGdisj root); assumption.
  unfold hpredTrue; rewrite GGeq; apply objHpMerge;
    hyr Hp' (sconjAssocR Hp'); scm.
  unfold hpredTrue; rewrite <- (addRemoveEqv Iroot'); apply objHpIntro; scm;
    unfold hpredTrue; rewrite (addRemoveEqv Iroot').
  apply (okObjAtWeaken Hp1); intros z Iz; destruct (set_union_elim Iz) as [Iz' | Iz'].
  sui2; assumption.
  sui1; rewrite (set_eqvUnionSymm M M'); rewrite <- (set_eqvUnionAssoc W' M' M);
    rewrite (set_eqvUnionSymm W' M'); rewrite <- Weq; assumption.
  apply (objHpWeaken Hp1); rewrite <- GGeq; assumption.
  apply (objHpWeaken Hp1); rewrite <- GGeq; assumption.
  (* case: root not in G *)
  hyr Hp' (sconjAssocR Hp'); scm.
  assert (Ix : set_In root (B0 |_| B')).
  set_InElim; [sui1 | elf; auto | sui2]; assumption.
  unfold hpredTrue; rewrite <- (addRemoveEqv Ix); apply objHpIntro; scm.
  apply (okObjAtWeaken Hp1); rewrite GGeq; intros z Iz; 
    destruct (set_union_elim Iz) as [Iz' | Iz']; [assumption | elf; assumption].
  apply (objHpWeaken Hp1); assumption.
  rewrite (set_removeEq Iroot') in Hp1.
  unfold hpredTrue; rewrite GGeq; apply objHpMerge; scm;
    apply (objHpWeaken Hp1); rewrite <- GGeq; assumption.
  (* done showing the heap has the repr thing *)

  (* show that we've read the right value *)
  rewrite xEq.
  rewrite <- Req'; [ | regNotInListSolver];
    rewrite <- Req; [ | regNotInListSolver]; stSimpl; subst h1 vh'.
  assert (HH : Heap.uset h00 (k + root) x ||- k + root |-> x ** top).
    apply heapAddOk. lli. exact (contElim1L Hp'').
    intros HQ; sconjLift2 Hp''; hyr Hp'' (sconjInsert Hp'' HQ); lle Hp''.
  apply (sconjMapRangeImpl HH); clear HH; intros hh HH.
  replace x with (mphi x).
  apply (contMapRangeImpl HH).
  subst mphi; destruct Iso12 as [Iso12 _]; destruct (Iso12 x) as [_ Q];
    destruct (map_union phi phi' x).
  2:reflexivity.
  assert (IO := isOddOk x); destruct (isOdd x).
  (* case: x is odd *)
  elf; apply (not_even_and_odd x).
  apply ptrPeven.
  assert (HX : s'' |||- (x |->__ ** top) ** top).
  sconjLift3 Hp''; scm.
  assert (II : set_In x (M |_| M')).
  apply Q; discriminate.
  hyr Hp0 (fasElim II (mapHpIsObjs Hp0)); scm.
  lle HX; exact (contElim1L HX).
  assumption.
  elf; destruct xOk as [X | InX].
  apply (not_even_and_odd x); assumption.
  apply (set_disj_elim MMdisj x).
  apply Q; discriminate.
  assumption.
  lli.
Qed.



Lemma gcGuarOk : forall k s s' s'' rr vs phi phi' x0 x x1 h00 M W M' W' G'
  toSt toEnd scan alloc free F' h0 frSt frEnd,
  k = 0 \/ k = 4 ->
  let root := rget s rroot in
  let B := rangeObjs toSt scan in
  let G := rangeObjs scan free in
  let B' := rangeObjs alloc toEnd in
  let F := rangeObjs free alloc in
  regPres
    (regNotInList
       (rfree::a0::v0::ast::rtemp0::rtemp1::rtemp2::raSave0::nil))
    (rset (rset (rset (rset (rset (rset (rset (rset 
       (rset (rset s rfrStart frSt) rfrEnd frEnd) rtoStart toSt) rtoEnd toEnd) rfree free)
     ralloc alloc) rdTmp1 (rget s ra)) a0 
    (k + root)) rdTmp2 (k + root)) ra BAKERREADRET) s' ->
  regPres (regNotInList (v0 :: ra :: nil)) s' s'' ->
  rget s'' ra = rget s ra ->
  In rroot rr ->
  (forall r, In r rr -> r = rroot \/ calleeSaved r) ->
  ~ In rgcInfo rr ->
  Heap.heap_eqv (heapOf vs)
    (Heap.HeapMap.map_mapRange (fun x =>
               match phi x with Some y => y | None => x end) h0) ->
  set_eqv (W |_| M) (rangeObjs frSt frEnd) ->
  iso_set phi M (B |_| G) ->
  (forall r, In r rr -> rget vs r = rget s r /\ okFieldVal (B |_| G |_| B') (rget s r)) ->
  (s |||- (Heap.heap_eqv h0 &&
    objHp (W |_| M) W **
    objHp (B |_| G |_| B') (B |_| B') **
    objHp ((W |_| M) |_| B |_| G |_| B') G) **
    mapHp M phi ** top) ->
  set_eqv (rangeObjs toSt toEnd) (B |_| G |_| F |_| B') ->
  let k' := if beq_nat k 0 then 4 else 0 in
  k <> k' ->
  set_disj (B |_| B') G ->
  set_disj (B |_| B' |_| G) G' ->
  set_disj (setRange frSt frEnd) (setRange toSt toEnd) ->
  let free' := rget s' rfree in
  set_eqv W (M' |_| W') ->
  set_eqv F (G' |_| F') ->
  iso_set phi' M' G' ->
  aligned8 free free' ->
  set_eqv F' (rangeObjs free' alloc) ->
  s |||- k + root |-> x0 ** top ->
  mapX (map_unionnn (spaceIdMap toSt toEnd) (map_union phi phi')) x0 = Some x ->
  rget s'' v0 = x ->
  s'' |||- k + root |-> x ** Heap.heap_eqv h00 **
    mapHp (M |_| M') (map_union phi phi') ** top ->
  okFieldVal ((B |_| G |_| B') |_| if set_mem root G then W |_| M else empty_set) x0 ->
  okFieldVal ((B |_| G |_| B') |_| if set_mem root G then W |_| M else empty_set) x1 ->
  s |||- k' + root |-> x1 ** top ->
  h00 ||- (foralls x in set_remove root (B |_| B' |_| G) |_| W',
    objCopied x s x) **
    k' + root |-> x1 ** (foralls x in G', objCopiedPhi phi' s x) ->
  h00 ||- k' + root |-> x1 **
    objHp (B |_| G |_| B') (set_remove root (B |_| B')) **
    objHp ((W |_| M) |_| B |_| G |_| B') (set_remove root G) **
    objHp (W' |_| M |_| M') W' **
    (foralls x in G', objCopiedPhi phi' s x) ->
  let h1 := Heap.uset h00 (k + root) x in
  let mphi := (fun x => match map_union phi phi' x with Some y => y | None => x end) in
  let vh' := Heap.HeapMap.map_mapRange mphi h1 in
    BMGI.gcGuar vs (vh', rfileOf s'') rr.
Proof.
  intros k s s' s'' rr vs phi phi' x0 x x1 h00 M W M' W' G'
    toSt toEnd scan alloc free F' h0 frSt frEnd.
  intros K root B0 G B' F Req Req' ReqRa.
  intros RRroot NRR NRRgci vsEq WMeq Iso rootOk Hp toEq k'.
  intros NEk DisjBBG DisjBGG frToDisj free' Weq Feq Iso' 
    AlFr F'eq Hp0 Meq v0eq Hp'' x0Ok x1Ok Hp1.
  intros Hp00 Hp00' h1 mphi vh'.

  assert (WMdisj : set_disj W M).
    assert (HpX : s |||- (fasd W ** top) ** fasd M ** top).
      clear Hp0 Hp1; scm.
      destruct (aconjElim Hp) as [_ Hp']; clear Hp; scm; apply objHpIsObjs in Hp;
        llMatch; lleim.
      hyr Hp (mapHpIsObjs Hp); llMatch; lleim.
    lle HpX; apply (fasFasDisj HpX); auto.

  assert (WM'disj : set_disj (M |_| M') W').
    assert (HpW : h00 ||- fasd W' ** top).
      sconjLift4 Hp00'; scm; hyr Hp2 (objHpIsObjs Hp2); llMatch; lleim.
    assert (HpX : s'' |||- fasd (M |_| M') ** fasd W' ** top).
      sconjLift2 Hp''; hyr Hp'' (sconjInsert Hp'' HpW); lle Hp''.
      sconjLift3 Hp''; scm; hyr Hp2 (mapHpIsObjs Hp2); llMatch; lleim.
    apply (fasFasDisj HpX); auto.

  clear Hp00'.
  assert (Hp00' : h00 ||- (foralls x in set_remove root (B0 |_| B' |_| G) |_| W',
    objCopiedH x h0 x) **
    k' + root |-> x1 ** (foralls x in G', objCopiedPhi phi' (h0, rfileOf s'') x)).
    clear Req Req' vsEq rootOk Hp'' Hp0 Hp1; rename Hp into Hp0.

    assert (HX : h0 ||- foralls x in W |_| (B0 |_| B') |_| G, x |->__).
      assert (HX := heapEqvImplR Hp0).
      apply objsMerge; scm.
      exact (objHpIsObjs Hp).
      apply objsMerge; scm; exact (objHpIsObjs Hp).

    scm; llMatch.
    assert (X : h ||- objCopied x2 (h0, rfileOf s) x2).
      eapply objCopiedTrans2.
      2:apply Hp.
      assert (I' : set_In x2 (W |_| (B0 |_| B') |_| G)).
        rewrite (set_eqvUnionAssoc B0 B' G); rewrite (set_eqvUnionSymm W (B0 |_| B' |_| G)).
        destruct (set_union_elim I) as [I' | I']; clear I.
      destruct (set_remove_elim _ _ _ I') as [I Ix]; clear I'.
      sui1; assumption.
      sui2; rewrite Weq; sui2; assumption.
      hyr HX (fasElim I' HX); lle HX.
      destruct (aconjElim (sconjDistrib2 Hp0)) as [HY _]; clear Hp0.
      hyr HY (sconjInsert HY HX); lle HY; hpunfold; lli.
      scm.
      scm.
      sconjLift2 HX; scm.
    exact (objCopiedHCopied X).
    eapply objCopiedPhiCopied; eauto.
    intros y Iy.
    clear I.
    assert (I : set_In y (W |_| (B0 |_| B') |_| G)).
    rewrite Weq; sui1; sui1; assumption.
    clear Iy.
    hyr HX (fasElim I HX); lle HX.
    destruct (aconjElim (sconjDistrib2 Hp0)) as [HY _]; clear Hp0.
    hyr HY (sconjInsert HY HX); lle HY; hpunfold; lli.
    scm.
    scm.
    sconjLift2 HX; scm.
  (* done showing Hp00' *)

  assert (frToDisj' : set_disj (W |_| M) ((B0 |_| B' |_| G) |_| F)).
    apply set_disj_intro; intros z Iz Iz'; apply (set_disj_elim frToDisj z);
      apply objRangeSetRangeSubset.
    rewrite <- WMeq; assumption.
    rewrite toEq; clear Iz; set_InElim; [sui1 | sui2; sui2; sui2 | sui2; sui1 | sui2; sui2; sui1];
      assumption.

  assert (DisjMWG' : set_disj (M |_| W) G').
    apply set_disj_intro; intros y Iy Iy'; apply (set_disj_elim frToDisj' y).
    rewrite (set_eqvUnionSymm W M); assumption.
    sui2; rewrite Feq; sui1; assumption.
    
  assert (Iso12 : iso_set (map_union phi phi') (M |_| M') ((B0 |_| G) |_| G')).
    apply iso_union.
    assumption.
    assumption.
    (* set_disj M M' *)
    apply set_disjSymm; apply set_unionDisjElimL with W'; rewrite <- Weq.
    destruct (aconjElim (sconjDistrib2 Hp)) as [_ HpX].
    lle HpX; sconjLift4 HpX; sconjLift2 HpX; unfold objHp in HpX; unfold mapHp in HpX;
    lle HpX; apply (fasFasDisj HpX); uilleim.
    apply set_unionDisjElimL with B'.
    setoid_replace ((B0 |_| G) |_| B') with (B0 |_| B' |_| G).
    assumption.
    rewrite (set_eqvUnionSymm B' G); rewrite (set_eqvUnionAssoc B0 G B'); apply set_eqvRefl.

  clear Hp00.
  assert (X : ~ (h00 ||- k + root |->_ ** top)).
  intro X; sconjLift2 Hp''; hyr Hp'' (sconjInsert Hp'' X); lle Hp''.
  assert (Vhp := heapEqvImplR Hp).
  assert (Vhp' := heapAddOk _ Hp00' (contElim1L Hp'') X x); clear X Hp00'.
  replace (Heap.uset h00 (k + root) x) with h1 in Vhp'; [idtac | reflexivity].
  clear Hp''.

  intros objs h VhpM.
  stSimpl.
  sconjLift3 Vhp'.
  set (pmap' := fun x => match phi' x with | Some y => y | None => x end) in * .
  assert (Vhp'' : h1 ||- k' + root |-> x1 ** k + root |-> x **
    ((foralls x in set_inter (set_remove root (B0 |_| B' |_| G) |_| W') objs, 
      objCopiedH x h0 x) ** top) **
    ((foralls x in set_map pmap' (set_inter M' objs),
      objCopiedPhi phi' (h0, rfileOf s'') x) ** top)).
  clear Req Req' vsEq.
  scm.
  assert (Hp'' : h2 ||- (foralls x2 in set_inter (set_remove root (B0 |_| B' |_| G) |_| W') objs,
        objCopiedH x2 h0 x2) ** 
  (foralls x2 in set_diff (set_remove root (B0 |_| B' |_| G) |_| W') objs, objCopiedH x2 h0 x2)).
  2:scm.
  apply fasSplit.
  unfold hpredTrue; rewrite <- (set_inter_split (set_remove root (B0 |_| B' |_| G) |_| W') objs);
    assumption.
  apply interDiffDisj.

  assert (Hp'' : h2 ||- (foralls x2 in set_map pmap' (set_inter M' objs), 
    objCopiedPhi phi' (h0, rfileOf s'') x2) **
     (foralls x2 in set_map pmap' (set_diff M' objs), 
       objCopiedPhi phi' (h0, rfileOf s'') x2)).
  2:scm.
  apply fasSplit.
  unfold hpredTrue; 
    rewrite <- (set_mapUnionDistrib pmap' (set_inter M' objs) (set_diff M' objs));
    rewrite <- (set_inter_split M' objs).
  setoid_replace (set_map pmap' M') with G'.
  assumption.
  subst pmap'; constructor; intro I.
  destruct (set_map_elim I) as [b [Aeq Ib]]; subst a; clear I.
  destruct Iso' as [Iso1 [Iso2 _]]; hyr Iso2 (Iso2 b); destruct (Iso1 b) as [Q _].
  hyr Q (Q Ib); destruct (phi' b).
  apply Iso2; reflexivity.
  elf; apply Q; reflexivity.
  destruct Iso' as [Iso1 [_ [_ Iso2]]].
  destruct (Iso2 a I) as [b X].
  apply set_map_intro with b.
  rewrite X; reflexivity.
  destruct (Iso1 b) as [_ Q]; apply Q; rewrite X; discriminate.

  apply set_disj_intro; intros z Iz Iz'.
  destruct (set_map_elim Iz) as [b [zeq Ib]]; clear Iz;
    destruct (set_map_elim Iz') as [b' [zeq' Ib']]; clear Iz'.
  subst z pmap'; simpl in zeq'.
  destruct (set_inter_elim Ib) as [Ib1 IZ]; clear Ib.
  destruct (set_diff_elim Ib') as [Ib1' IZ']; clear Ib'.
  destruct Iso' as [Iso1 [Iso2 [Iso3 _]]].
  hyr Iso3 (Iso3 b b'); destruct (Iso1 b) as [X _];
    destruct (Iso1 b') as [X' _]; clear Iso1.
  hyr X (X Ib1); hyr X' (X' Ib1'); clear Ib1 Ib1'; destruct (phi' b).
  destruct (phi' b').
  subst a0; assert (b = b'); [apply (Iso3 a); reflexivity | subst b'].
  auto.
  auto.
  auto.

  clear Vhp'; lle Vhp''.
  assert (Vhp''2 : h1 ||- 
    (if set_mem root objs then (k' + root |-> x1 ** k + root |-> x) else unit) **
    (foralls x in set_inter (set_remove root (B0 |_| B' |_| G) |_| W') objs,
      objCopiedH x h0 x) **
    (foralls x in set_map pmap' (set_inter M' objs),
      objCopiedPhi phi' (h0, rfileOf s'') x) ** top).
  destruct (set_In_dec root objs) as [I | NI].
  rewrite (set_mem_true I); lli; scm.
  rewrite (set_mem_false NI); lli; sconjLift4 Vhp''; sconjLift4 Vhp''; scm.
  clear Vhp''; rename Vhp''2 into Vhp''.

  sconjLift4of4 Vhp''; hyr Vhp'' (sconjComm Vhp''); destruct (sconjExtract Vhp'')
    as [h2 [Hp' Vhp']]; clear Vhp''.
  set (vh'' :=  Heap.HeapMap.map_mapRange mphi h2).
  exists vh''; constructor.

  (* show that vh'' is a subset of vh' *)
  subst vh' vh''.
  apply (sconjMapRangeImpl Hp').
  intros h3 Hp3.
  apply eqvMapRangeImpl; assumption.
  intros; lli.

  (* prove the two virtual states are isomorphic *)
  exists objs; exists (set_inter (set_add root empty_set) objs |_| 
    (set_inter (set_remove root (B0 |_| B' |_| G) |_| W') objs) |_| 
      set_map pmap' (set_inter M' objs)).
  stSimpl.
  set (phi2 := fun x => if set_mem x (set_inter objs ((B0 |_| B' |_| G) |_| M' |_| W'))
    then Some (pmap' x) else None).
  exists phi2.
  constructor.
  eapply minObjHpElim1; apply (heapEqvImplL VhpM).

  assert (IsoObjs : iso_set phi2 objs
     (set_inter (set_add root empty_set) objs
      |_| set_inter (set_remove root (B0 |_| B' |_| G) |_| W') objs
          |_| set_map pmap' (set_inter M' objs))).
    clear vh'' Vhp' h2 Hp'.
    subst phi2 pmap'; cbv beta.
    constructor.
    (* domain is correct *)
    intros y; constructor; intro Iy.
    assert (HX : vs |||- objHp (W |_| (B0 |_| B' |_| G)) (W |_| (B0 |_| B' |_| G))).

      assert (BGBdisj : set_disj (B0 |_| G |_| B') M).
        apply set_unionDisjElimL with F.
        rewrite (set_eqvUnionAssoc B0 (G |_| B') F).
        rewrite (set_eqvUnionAssoc G B' F); rewrite (set_eqvUnionSymm B' F).
        rewrite <- toEq.
        apply set_disjSymm; apply set_unionDisjElimR with W; rewrite WMeq;
          apply set_disjSymm; apply set_disj_intro; intros z Iz Iz';
            apply (set_disj_elim frToDisj z); apply objRangeSetRangeSubset; assumption.

      eapply heapEqvTrue; eauto.
      apply objHpMerge; apply (sconjMapRangeImpl Vhp); clear Vhp; intros hh HH.
      apply objHpWeaken with ((B0 |_| G) |_| W).
      apply mapDomObjHp with M.
      apply objHpWeaken with (W |_| M).
      assumption.
      intros z Iz; rewrite (set_eqvUnionSymm M W); assumption.
      assumption.
      assumption.
      intros z Iz; apply ptrPeven; sconjLift2 Hp.
      assert (HQ : s |||- (z |->_ ** top) ** top).
        clear Hp0 Hp1; scm; hyr Hp (fasElim Iz (mapHpIsObjs Hp)); lleim.
      lle HQ; exact (contElim1L HQ).
      intros z Iz; set_InElim.
      sui2; sui1; assumption.
      sui2; sui2; sui2; assumption.
      sui1; assumption.
      unfold hpredTrue; rewrite <- (set_eqvUnionAssoc B0 B' G).
      apply objHpMerge; apply (sconjMapRangeImpl HH); clear hh HH; intros hh HH.
      apply objHpWeaken with (B0 |_| G |_| B').
      apply mapDomObjHpDisj with M (B0 |_| G); auto.
      intros z Iz; apply ptrPeven; sconjLift2 Hp.
      assert (HQ : s |||- (z |->_ ** top) ** top).
        clear Hp0 Hp1; scm; hyr Hp (fasElim Iz (mapHpIsObjs Hp)); lleim.
      lle HQ; exact (contElim1L HQ).
      intros z Iz; set_InElim; sui2; [sui1; sui1 | sui2 | sui1; sui2]; assumption.
      apply objHpWeaken with ((B0 |_| G) |_| B0 |_| G |_| W |_| B').
      apply mapDomObjHp with M; auto.
      apply objHpWeaken with ((W |_| M) |_| B0 |_| G |_| B').
      assumption.
      intros z Iz; set_InElim; [sui2; sui2; sui2; sui1 | sui1 | sui2; sui1 | sui2; sui2; sui1 |
        sui2; sui2; sui2; sui2]; assumption.
      rewrite (set_eqvUnionSymm W B'); rewrite <- (set_eqvUnionAssoc G B' W);
        rewrite <- (set_eqvUnionAssoc B0 (G |_| B') W); apply set_disj_intro; intros z Iz Iz'.
      destruct (set_union_elim Iz).
      apply (set_disj_elim BGBdisj z); assumption.
      apply (set_disj_elim WMdisj z); assumption.
      intros z Iz; apply ptrPeven; sconjLift2 Hp; 
        assert (HQ : s |||- (z |->_ ** top) ** top).
      clear Hp0 Hp1; scm; hyr Hp (fasElim Iz (mapHpIsObjs Hp)); lleim.
      lle HQ; exact (contElim1L HQ).
      intros z Iz; set_InElim;
        [sui2; sui1; sui1 | sui2; sui2 | sui2; sui1; sui1 | sui2; sui2 | sui1 | sui2; sui1; sui2]; 
        assumption.

    assert (Z : set_subset objs (W |_| B0 |_| B' |_| G)).
      apply minObjHpSub with (heapOf vs) unit (List.map (rget vs) rr).
      lli; assumption.
      destruct (aconjElim (sconjDistrib2 VhpM)); assumption.
      intros v Iv.
      destruct (rrlistMapElim _ _ _ Iv) as [r [Ir Veq]];
        destruct (rootOk _ Ir) as [E Z].
      rewrite E in Veq; subst v; apply (okFieldValWeaken Z).
      intros z Iz; set_InElim; [sui2; sui1 | sui2; sui2; sui2 | sui2; sui2; sui1]; assumption.
      (* finished showing Z *)

    assert (U := Z _ Iy).
    rewrite set_mem_true.
    discriminate.
    apply set_inter_intro.
    assumption.
    set_InElim; [sui2; rewrite <- Weq | sui1; sui1 | sui1; sui2; sui1 | sui1; sui2; sui2]; assumption.
    assert (X := set_memSound y (set_inter objs ((B0 |_| B' |_| G) |_| M' |_| W'))).
    destruct (set_mem y (set_inter objs ((B0 |_| B' |_| G) |_| M' |_| W'))).
    destruct (set_inter_elim X); assumption.
    elf; auto.

  (* next part *)
  constructor.
  intros y z Peq.
  assert (X := set_memSound y (set_inter objs ((B0 |_| B' |_| G) |_| M' |_| W')));
    destruct (set_mem y (set_inter objs ((B0 |_| B' |_| G) |_| M' |_| W'))).
  2:discriminate Peq.
  injection Peq; clear Peq; intro; subst z.
  destruct (set_inter_elim X) as [Iy Iy']; clear X.
  destruct Iso' as [Iso' _]; destruct (Iso' y) as [_ Q].
  destruct (set_union_elim Iy') as [Iy'' | Iy'']; clear Iy'.
  (* case: y in to space *)
  destruct (phi' y).
  elf; apply (set_disj_elim frToDisj' y).
  rewrite Weq; sui1; sui1; apply Q; discriminate.
  sui1; assumption.
  compare y root; intro NE.
  sui1; apply set_inter_intro; [left | idtac]; auto.
  sui2; sui1; apply set_inter_intro.
  sui1; apply set_remove_intro; assumption.
  assumption.
  (* case: y in from space *)
  destruct (set_union_elim Iy'') as [Iy' | Iy']; clear Iy''.
  (* case: y in M' *)
  sui2; sui2; apply set_map_intro with y.
  destruct (phi' y); reflexivity.
  apply set_inter_intro; assumption.
  (* case: y in W' *)
  destruct (phi' y).
  elf; apply (set_disj_elim WM'disj y); [sui2; apply Q; discriminate | assumption].
  sui2; sui1; apply set_inter_intro.
  sui2; assumption.
  assumption.

  (* next part of isomorphism *)
  constructor.
  intros y y' z Peq Peq'.
  assert (Iy := set_memSound y (set_inter objs ((B0 |_| B' |_| G) |_| M' |_| W'))).
  assert (Iy' := set_memSound y' (set_inter objs ((B0 |_| B' |_| G) |_| M' |_| W'))).
  destruct (set_mem y (set_inter objs ((B0 |_| B' |_| G) |_| M' |_| W'))).
  2:discriminate Peq.
  destruct (set_mem y' (set_inter objs ((B0 |_| B' |_| G) |_| M' |_| W'))).
  2:discriminate Peq'.
  injection Peq'; injection Peq; clear Peq Peq'; intros Peq Peq';
    rewrite <- Peq' in Peq; clear Peq'.
  destruct Iso' as [Iso1 [Iso2 [Iso3 _]]]; 
    destruct (Iso1 y) as [_ Q]; destruct (Iso1 y') as [_ Q']; 
      assert (X := Iso2 y); assert (X' := Iso2 y'); hyr Iso3 (Iso3 y y').
  destruct (set_inter_elim Iy) as [Iy1 Iy2]; clear Iy.
  destruct (set_inter_elim Iy') as [Iy1' Iy2']; clear Iy'.
  destruct (phi' y).
  destruct (phi' y').
  apply (Iso3 a); auto.
  subst a; destruct (set_union_elim Iy2') as [Iy' | Iy']; clear Iy2'.
  elf; apply (set_disj_elim DisjBGG y'); eauto.
  elf; apply (set_disj_elim DisjMWG' y').
  sui2; rewrite Weq; assumption.
  eauto.
  clear Iso3 X Q Iy1 Iy1'.
  destruct (phi' y').
  subst a; destruct (set_union_elim Iy2) as [Iy' | Iy']; clear Iy2.
  elf; apply (set_disj_elim DisjBGG y); eauto.
  elf; apply (set_disj_elim DisjMWG' y).
  sui2; rewrite Weq; assumption.
  eauto.
  assumption.
  (* final part of isomorphism *)
  intros y Iy0.
  destruct (set_union_elim Iy0) as [Iy | Iy0']; clear Iy0.
  (* case: y is root *)
  destruct (set_inter_elim Iy) as [Iy1 Iy2]; clear Iy; exists root.
  simpl in Iy1; destruct Iy1.
  subst y; destruct (rootOk _ RRroot) as [_ Z].
  replace (rget s rroot) with root in Z; [idtac | reflexivity].
  destruct Z as [Z | Z].
  elf; apply (not_even_and_odd root).
  apply ptrPeven; assert (HH := minObjHpElim1 (heapEqvImplL VhpM));
    hyr HH (objHpElim Iy2 HH); hpunfoldIn HH; lle HH; apply (contElim1L HH).
  assumption.
  rewrite set_mem_true.
  destruct Iso' as [Iso' _]; destruct (Iso' root) as [_ Q]; destruct (phi' root).
  elf; apply (set_disj_elim frToDisj' root).
  sui1; rewrite Weq; sui1; apply Q; discriminate.
  sui1; rewrite (set_eqvUnionSymm B' G); assumption.
  reflexivity.
  apply set_inter_intro.
  assumption.
  sui1; rewrite (set_eqvUnionSymm B' G); assumption.
  elf; assumption.

  destruct (set_union_elim Iy0') as [Iy0 | Iy]; clear Iy0'.
  (* case: y not copied *)
  destruct (set_inter_elim Iy0) as [Iy0' IyObjs]; clear Iy0.
  exists y.
  rewrite set_mem_true.
  destruct Iso' as [Iso' _]; destruct (Iso' y) as [_ Q]; destruct (phi' y).
  elf; destruct (set_union_elim Iy0') as [Iy0 | Iy]; clear Iy0'.
  destruct (set_remove_elim _ _ _ Iy0) as [Iy NE]; clear Iy0.
  apply (set_disj_elim frToDisj' y).
  rewrite Weq; sui1; sui1; apply Q; discriminate.
  sui1; assumption.
  apply (set_disj_elim WM'disj y); [sui2; apply Q; discriminate | assumption].
  reflexivity.
  apply set_inter_intro.
  assumption.
  destruct (set_union_elim Iy0') as [Iy0 | Iy]; clear Iy0'.
  destruct (set_remove_elim _ _ _ Iy0) as [_ Iy]; sui1; assumption.
  sui2; sui2; assumption.
  (* case: y in G' *)
  destruct (set_map_elim Iy) as [z [Zeq Iy']]; clear Iy.
  destruct (set_inter_elim Iy') as [Iy IyObjs]; clear Iy'.
  exists z; rewrite set_mem_true.
  subst y; destruct (phi' z); reflexivity.
  apply set_inter_intro; [idtac | sui2; sui1]; assumption.
  (* done showing isomorphism *)

  assert (Disj : set_disj (M |_| W) ((B0 |_| B') |_| G)).
    assert (DD : s |||- fasd (M |_| W) ** fasd ((B0 |_| B') |_| G) ** top).
      destruct (aconjElim (sconjDistrib2 Hp)) as [_ HpX].
      clear Hp0 Hp1.
      lle HpX; sconjLift4 HpX; hyr HpX (sconjAssocR HpX); 
      sconjLift3 HpX; sconjLift3 HpX; hyr HpX (sconjAssocR HpX); 
      sconjLift2 HpX; scm.
      apply fasdMerge; scm; [hyr Hp0 (mapHpIsObjs Hp0) | hyr Hp0 (objHpIsObjs Hp0)];
        llMatch; lleim.
      apply fasdMerge; scm; hyr Hp0 (objHpIsObjs Hp0); llMatch; lleim.
    apply (fasFasDisj DD); auto.

  assert (Meven : forall x, set_In x (M |_| M') -> even x).
    intros z Iz0; apply ptrPeven; clear Hp0 Hp1.
    assert (Z : s |||- ((z |->_ ** top) ** top) ** top).
      destruct (set_union_elim Iz0) as [Iz | Iz]; clear Iz0.
      sconjLift2 Hp; scm; hyr Hp (mapHpElim1 Iz Hp); lleim.
      assert (Iz' : set_In z W) by (rewrite Weq; sui1; assumption).
      destruct (aconjElim (sconjDistrib2 Hp)) as [_ HQ];
        scm; hyr Hp0 (objHpElim Iz' Hp0); hpunfoldIn Hp0; lleim.
    lle Z; apply (contElim1L Z).

  assert (M'even : forall x, set_In x M' -> even x).
    intros z Iz; apply ptrPeven.
    assert (Iz' : set_In z W) by (rewrite Weq; sui1; assumption).
    assert (Z : h0 ||- ((z |->_ ** top) ** top) ** top).
      scm; hyr Hp2 (objHpElim Iz' Hp2); hpunfoldIn Hp2; lleim.
    lle Z; apply (contElim1L Z).

  assert (BGeven : forall x, set_In x (B0 |_| G) -> even x).
    intros z Iz; apply ptrPeven.
    assert (Z : h0 ||- ((z |->_ ** top) ** top) ** top).
      destruct (set_union_elim Iz) as [Iz' | Iz'].
      clear Iz; assert (Iz : set_In z (B0 |_| B')) by (sui1; assumption).
      sconjLift2 Vhp; scm; hyr Hp2 (objHpElim Iz Hp2); hpunfoldIn Hp2; lleim.
      sconjLift3of3 Vhp; scm; hyr Hp2 (objHpElim Iz' Hp2); hpunfoldIn Hp2; lleim.
    lle Z; apply (contElim1L Z).

  assert (xOk : okFieldVal ((B0 |_| G |_| B') |_| 
    (if set_mem root G then (G' |_| W') else empty_set)) x).
    unfold mapX in Meq; assert (IO := isOddOk x0); destruct (isOdd x0); [left | right].
    injection Meq; intro; subst x0; assumption.
    destruct x0Ok as [x0Ok | x0Ok].
    elf; apply (not_even_and_odd x0); assumption.
    destruct Iso12 as [Iso12 [Iso12' _]]; destruct (Iso12 x) as [_ Q].
    clear Hp0 Hp1.
    destruct (set_union_elim x0Ok) as [Ix0 | Ix0]; clear x0Ok.
    (* case: x0 to space *)
    cut (x = x0).
    intro E; subst x0; clear Meq; sui1; assumption.
    unfold map_unionnn in Meq; unfold map_union in Meq; rewrite spaceIdMapEq in Meq.
    rewrite set_mem_true in Meq.
    injection Meq; auto.
    apply objRangeSetRangeSubset; rewrite toEq.
    rewrite (set_eqvUnionSymm F B'); rewrite <- (set_eqvUnionAssoc G B' F);
    rewrite <- (set_eqvUnionAssoc B0 (G |_| B') F); sui1; assumption.
    (* case: x0 in from space *)
    destruct (set_mem root G).
    2:elf; assumption.
    unfold map_unionnn in Meq; unfold map_union at 1 in Meq; rewrite spaceIdMapEq in Meq;
      rewrite set_mem_false in Meq.
    hyr Iso12' (Iso12' _ _ Meq).
    destruct (set_union_elim Iso12') as [Ix' | Ix'].
    sui1; rewrite <- (set_eqvUnionAssoc B0 G B');
      sui1; assumption.
    sui2; sui1; assumption.
    intro Ix'; apply (set_disj_elim frToDisj x0).
    apply objRangeSetRangeSubset; rewrite <- WMeq; assumption.
    assumption.
  rename pmap' into mphi2.
  set (mphi1 := fun x => match phi x with Some y => y | None => x end) in * .
  assert (mphi1Eq : forall x, set_In x ((B0 |_| G |_| B') |_| W) -> 
    mphi1 x = x).
    intros z Iz; subst mphi1 phi2 mphi2; cbv beta; 
      destruct Iso as [Iso _]; destruct (Iso z) as [_ Q].
    destruct (phi z).
    elf; destruct (set_union_elim Iz) as [Iz' | Iz']; clear Iz.
    apply (set_disj_elim Disj z).
    sui1; apply Q; discriminate.
    rewrite (set_eqvUnionAssoc B0 B' G); rewrite (set_eqvUnionSymm B' G); assumption.
    apply (set_disj_elim WMdisj z).
    assumption.
    apply Q; discriminate.
    reflexivity.
  assert (mphi2Eq : forall x, set_In x ((B0 |_| G |_| B') |_| W') -> 
    mphi2 x = x /\ (set_In x objs -> phi2 x = Some x)).
    intros z Iz; subst phi2 mphi2; cbv beta;
      destruct Iso' as [Iso' _]; destruct (Iso' z) as [_ Q].
    destruct (phi' z).
    elf; destruct (set_union_elim Iz) as [Iz' | Iz']; clear Iz.
    apply (set_disj_elim Disj z).
    rewrite Weq; sui2; sui1; apply Q; discriminate.
    rewrite (set_eqvUnionAssoc B0 B' G); rewrite (set_eqvUnionSymm B' G); assumption.
    apply (set_disj_elim WM'disj z).
    sui2; apply Q; discriminate.
    assumption.
    constructor.
    reflexivity.
    intros Iz'; rewrite set_mem_true.
    reflexivity.
    apply set_inter_intro.
    assumption.
    clear Iz'; rewrite (set_eqvUnionSymm B' G); set (X := B0 |_| G |_| B') in *; set_InElim;
      [sui1 | sui2; sui2]; assumption.

  assert (mphiEq : forall x, set_In x ((B0 |_| G |_| B') |_| W') -> 
    mphi x = x).
    intros z Iz; subst mphi; cbv beta; unfold map_union.
    assert (Iz' : set_In z ((B0 |_| G |_| B') |_| W)).
      rewrite Weq; destruct (set_union_elim Iz) as [Iz' | Iz'];
        [sui1 | sui2; sui2]; assumption.
    assert (E1 := mphi1Eq _ Iz').
    destruct (mphi2Eq _ Iz) as [E2 _].
    subst mphi1 mphi2; cbv beta in E1, E2.
    destruct (phi z); assumption.

  assert (mphiMeq : forall x, set_In x M -> 
    exists y, phi x = Some y /\ mphi x = y /\ mphi1 x = y /\ set_In y (B0 |_| G) /\
      mphi2 x = x /\ phi2 x = None).
    intros z Iz.
    subst mphi phi2 mphi1 mphi2; cbv beta; unfold map_union;
      destruct Iso as [Iso [IsoX _]]; destruct (Iso z) as [Q _]; 
        destruct Iso' as [Iso' _]; destruct (Iso' z) as [_ Q']; 
	  hyr IsoX (IsoX z); destruct (phi z).
    exists b; constructor; [reflexivity | constructor; eauto].
    constructor.
    reflexivity.
    constructor.
    eauto.
    destruct (phi' z).
    elf; apply (set_disj_elim WMdisj z).
    rewrite Weq; sui1; apply Q'; discriminate.
    assumption.
    constructor.
    reflexivity.
    rewrite set_mem_false.
    reflexivity.
    intro Iz0; destruct (set_inter_elim Iz0) as [_ Iz2]; 
      destruct (set_union_elim Iz2) as [Iz' | Iz']; clear Iz0 Iz2.
    apply (set_disj_elim Disj z).
    sui1; assumption.
    rewrite (set_eqvUnionAssoc B0 B' G); assumption.
    apply (set_disj_elim WMdisj z); try (rewrite Weq); assumption.
    elf; apply Q; auto.

  assert (mphiMeq' : forall x, set_In x M' -> 
    exists y, phi' x = Some y /\ mphi x = y /\ mphi1 x = x /\ set_In y G' /\
      mphi2 x = y /\ (set_In x objs -> phi2 x = Some y)).
    intros z Iz.
    subst mphi phi2 mphi2 mphi1; unfold map_union; cbv beta.
    assert (I3 : set_In z ((B0 |_| G |_| B') |_| W)).
    sui2; rewrite Weq; sui1; assumption.
    rewrite (mphi1Eq _ I3); clear I3.
    unfold map_union; destruct Iso as [Iso _]; destruct (Iso z) as [_ X1];
      destruct (phi z).
    elf; apply (set_disj_elim WMdisj z).
    rewrite Weq; sui1; assumption.
    apply X1; discriminate.
    destruct Iso' as [Iso' [IsoX _]]; destruct (Iso' z) as [Q _]; 
      hyr IsoX (IsoX z); destruct (phi' z).
    exists a; constructor; [reflexivity | constructor; eauto].
    constructor; [reflexivity | constructor; [eauto | constructor; auto]].
    intros Iz'; rewrite set_mem_true.
    reflexivity.
    apply set_inter_intro; [assumption | sui2; sui1; assumption].
    elf; apply Q; auto.
    
  assert (mphiOddEq : forall x, odd x -> mphi1 x = x /\ mphi x = x).
    intros z OddZ; subst mphi1 mphi; cbv beta; unfold map_union.
    destruct Iso as [Iso _]; destruct (Iso z) as [_ Q]; destruct (phi z).
    elf; apply (not_even_and_odd z); [apply Meven; sui1; apply Q; discriminate | assumption].
    destruct Iso' as [Iso' _]; clear Q; destruct (Iso' z) as [_ Q]; destruct (phi' z).
    elf; apply (not_even_and_odd z); [apply Meven; sui2; apply Q; discriminate | assumption].
    constructor; reflexivity.

  assert (Ximp : forall x, okFieldVal ((B0 |_| G |_| B') |_| (W |_| M)) x -> 
    okFieldVal objs (mphi1 x) ->
    mapX phi2 (mphi1 x) = Some (mphi x)).
    intros z okZ okZ'; unfold mapX; assert (IO := isOddOk z); destruct (isOdd z).
    (* case: z is odd *)
    destruct (mphiOddEq _ IO) as [Eq1 Eq2]; rewrite Eq1; rewrite Eq1 in okZ'; rewrite Eq2;
      rewrite (isOddTrue IO); reflexivity.
    (* case: z is even *)
    destruct okZ as [okZ | okZ]; [elf; apply (not_even_and_odd z); assumption | idtac].
    destruct (set_union_elim okZ) as [okZ0' | Iz']; clear okZ.
    (* case: z in to-space *)
    rewrite mphi1Eq; [idtac | sui1; assumption].
    rewrite mphi1Eq in okZ'; [idtac | sui1; assumption].
    rewrite (isOddFalse IO).
    destruct (mphi2Eq z) as [_ Peq]; [sui1; assumption | rewrite Peq].
    rewrite mphiEq; [reflexivity | sui1; assumption].
    destruct okZ'; [elf; apply (not_even_and_odd z); assumption | assumption].
    (* case: z in from-space *)
    destruct (set_union_elim Iz') as [okZ0 | Iz]; clear Iz'.
    (* case: z in W *)
    assert (Iz0 : set_In z (M' |_| W')); [rewrite <- Weq; assumption | clear okZ0].
    destruct (set_union_elim Iz0) as [Iz | Iz]; clear Iz0.
    (* case: z in M' *)
    destruct (mphiMeq' _ Iz) as [y [Phi'eq [Meq0 [Meq1 [Iy [Meq2 Phi2eq]]]]]].
    rewrite Meq1; rewrite Meq1 in okZ'; rewrite Meq0;
      destruct okZ' as [Iz' | Iz']; [elf; apply (not_even_and_odd z); assumption | idtac].
    rewrite (isOddFalse IO); apply Phi2eq; assumption.
    (* case: z in W' *)
    generalize dependent okZ';
      rewrite mphi1Eq; [idtac | sui2; rewrite Weq; sui2; assumption];
        rewrite (mphiEq z); [idtac | sui2; assumption];
	  destruct (mphi2Eq z) as [Peq2 Peq3]; [sui2; assumption | idtac];
	    rewrite (isOddFalse IO); intro.
    destruct okZ' as [Iz' | Iz']; [elf; apply (not_even_and_odd z); assumption | idtac].
    auto.
    (* case: z in M *)
    destruct (mphiMeq _ Iz) as [y [Peq0 [Peq1 [Peq2 [Iy [Phi2Eq Eq]]]]]].
    rewrite Peq2; rewrite Peq2 in okZ'; rewrite Peq1.
    destruct okZ' as [Z | Iz']; [elf; apply (not_even_and_odd y); auto | idtac].
    clear z IO Iz Peq0 Peq1 Peq2 Phi2Eq Eq.
    destruct (isOdd y).
    reflexivity.
    destruct (mphi2Eq y) as [Eq1 Eq2].
    sui1; rewrite <- (set_eqvUnionAssoc B0 G B'); sui1; assumption.
    auto.

  (* first show what happens from h2 to vh'' *)
  assert (Vhp'' : vh'' ||- (if set_mem root objs then k' + root |-> mphi x1 ** 
    k + root |-> x else unit) **
  (foralls x in set_inter (set_remove root (B0 |_| B' |_| G) |_| W') objs,
    objIsFwdedH phi2 h x) **
  (foralls x in set_map mphi2 (set_inter M' objs), objIsFwdedH phi2 h x)).
  subst vh''.
  apply (sconjMapRangeImpl Vhp'); clear Vhp'; intros h' Hp''.
  (* show the root object is okay *)
  destruct (set_mem root objs).

  assert (MPhiX : mphi x = x).
  destruct xOk as [xOk | xOk].
  destruct (mphiOddEq _ xOk) as [_ E]; exact E.
  destruct (set_union_elim xOk) as [Ix | Ix]; clear xOk.
  apply mphiEq; sui1; assumption.
  destruct (set_mem root G).
  2:elf; assumption.
  destruct (set_union_elim Ix) as [Ix' | Ix']; clear Ix.
  2:apply mphiEq; sui2; assumption.
  subst mphi; cbv beta; destruct Iso12 as [Iso12 _]; destruct (Iso12 x) as [_ Q]; 
    destruct (map_union phi phi' x).
  elf; apply (set_disj_elim DisjMWG' x).
  rewrite Weq; rewrite <- (set_eqvUnionAssoc M M' W'); sui1; apply Q; discriminate.
  assumption.
  reflexivity.

  rewrite <- MPhiX.
  apply (sconjMapRangeImpl Hp''); clear h' Hp''; intros h' Hp'';
    apply (contMapRangeImpl Hp'').
  
  apply (unitMapRangeImpl Hp'').
  (* done showing scanned object is okay *)

  apply (sconjMapRangeImpl Hp''); clear h' Hp''; intros h' Hp'';
    apply (fasMapRangeImplI Hp''); clear h' Hp''; intros h' y Iy Hp'';
      hpunfold; hpunfoldIn Hp''; lle Hp''.
  (* unfwded objects *)
  lli; [idtac | 
    simpl; apply (sconjMapRangeImpl Hp'');
      intros h'x Hp''x; apply (contMapRangeImpl Hp''x)].
  clear h' Hp''.
  destruct (set_inter_elim Iy) as [Iy1 Iy2]; clear Iy.
  rename Iy1 into Iy.
  exists y; exists (mphi1 x2); exists (mphi1 x3).
  constructor.
  assert (Iy' : set_In y ((B0 |_| G |_| B') |_| W')).
  rewrite (set_eqvUnionSymm G B'); set (X := B0 |_| B' |_| G) in *;
    set_InElim; [sui1 | sui2]; assumption.
  destruct (mphi2Eq _ Iy') as [_ X]; apply X; assumption.
  constructor.
  lli.
  cut (vs |||- y |-> mphi1 x2 ** 4 + y |-> mphi1 x3 ** top).
  intro HpS.
  assert (HX := objHpElim Iy2 (minObjHpElim1 (heapEqvImplL VhpM)));
    destruct (aconjElim (sconjDistrib2 VhpM)) as [_ VhpM']; clear VhpM.
  hpunfoldIn HX; lle HX; hyr VhpM' (sconjInsert VhpM' HX); lle VhpM'.
  assert (X1 := contContEq HpS VhpM'); sconjLift2 HpS; sconjLift2 VhpM';
    assert (X2 := contContEq HpS VhpM'); subst x4 x5; scm.
  apply (heapEqvTrue vsEq).
  assert (HH : h0 ||- y |-> x2 ** 4 + y |-> x3 ** top).

  apply contContCombine; auto; omega.
  apply (sconjMapRangeImpl HH); clear HH; intros h' HH.
  apply (contMapRangeImpl HH).
  apply (sconjMapRangeImpl HH); clear h' HH; intros h' HH.
  apply (contMapRangeImpl HH).
  lli.

  assert (HpV : vs |||- (okObjAt objs y ** top) ** top).
  scm; destruct (aconjElim Hp2) as [Hp3 _]; hyr Hp3 (minObjHpElim1 Hp3);
    hyr Hp3 (objHpElim Iy2 Hp3); scm.
  hpunfoldIn HpV; lle HpV.
  assert (HpX : h0 ||- (okObjAt ((W |_| M) |_| B0 |_| G |_| B') y ** top) ** top).
  assert (Iy'' : set_In y (((B0 |_| B') |_| G) |_| W')).
  destruct (set_union_elim Iy).
  destruct (set_remove_elim _ _ _ H); sui1;
    rewrite (set_eqvUnionAssoc B0 B' G); assumption.
  sui2; assumption.
  clear H3 H4.
  destruct (set_union_elim Iy'') as [IyZ | IyZ]; clear Iy.
  destruct (set_union_elim IyZ) as [Iy | Iy]; clear Iy'' IyZ;
    [sconjLift2 Vhp | sconjLift3of3 Vhp]; scm; hyr Hp2 (objHpElim Iy Hp2); scm;
    apply (okObjAtWeaken Hp2); intros z Iz; sui2; assumption.
  assert (HX : h0 ||- (objHp (W |_| M) W' ** objHp (W |_| M) M') ** top).
  scm.
  apply sconjComm; apply objHpSplit.
  unfold hpredTrue; rewrite <- Weq; assumption.
  apply set_unionDisjElimR with M; assumption.
  lle HX; scm; hyr Hp2 (objHpElim IyZ Hp2); scm;
  apply (okObjAtWeaken Hp2); intros z Iz; sui1; assumption.
  hpunfoldIn HpX; lle HpX.
  assert (X1 := contContEq HpX H3); sconjLift2 HpX; assert (X2 := contContEq HpX H4);
    subst x6 x7; clear HpX Iy2.

  constructor.
  (* show fwd okay for field 1 *)
  assert (x4 = mphi1 x2).
  assert (HpV' : vs |||- y |-> mphi1 x2 ** top).
  eapply heapEqvTrue; eauto.
  apply (sconjMapRangeImpl H3); intros hx HpX.
  apply (contMapRangeImpl HpX).
  lli.
  exact (contContEq HpV HpV').
  subst x4; clear HpV.
  unfold mapX; assert (IO := isOddOk x2); destruct (isOdd x2).
  (* case: x2 is odd *)
  destruct (mphiOddEq _ IO) as [Eq1 Eq2]; rewrite Eq1; rewrite Eq2; 
    rewrite (isOddTrue IO); reflexivity.
  (* case: x2 is even *)
  destruct H6 as [okF | Ix0].
  elf; apply (not_even_and_odd x2); assumption.
  assert (Ix : set_In x2 (((B0 |_| G |_| B') |_| W') |_| M |_| M')).
  rewrite (set_eqvUnionAssoc (B0 |_| G |_| B') W' (M |_| M')).
  destruct (set_union_elim Ix0).
  sui2; rewrite (set_eqvUnionSymm M M'); rewrite <- (set_eqvUnionAssoc W' M' M);
    rewrite (set_eqvUnionSymm W' M'); rewrite <- Weq; assumption.
  sui1; assumption.
  destruct (set_union_elim Ix) as [Ix' | Ix']; clear Ix.
  (* case: x2 uncopied *)
  cut (mphi1 x2 = x2).
  intro E1; rewrite E1.
  rewrite (isOddFalse IO); rewrite (mphiEq _ Ix').
  rewrite E1 in H5; destruct H5 as [X | okF].
  elf; apply (not_even_and_odd x2); assumption.
  destruct (mphi2Eq _ Ix') as [_ E]; auto.
  apply mphi1Eq; destruct (set_union_elim Ix') as [Ix'' | Ix''];
    [sui1 | rewrite Weq; sui2; sui2]; assumption.
  (* case: x2 copied *)
  destruct (set_union_elim Ix') as [Ix | Ix]; clear Ix'.
  (* case: x2 in M *)
  destruct (mphiMeq _ Ix) as [z [Peq [Mpeq [Mpeq1 [Iz [Mpeq2 P2eq]]]]]].
  rewrite Mpeq1; rewrite Mpeq; rewrite Mpeq1 in H5.
  assert (IOz := isOddOk z); destruct (isOdd z); [reflexivity | idtac].
  destruct H5 as [okZ | okZ].
  elf; apply (not_even_and_odd z); assumption.
  assert (Iz' : set_In z ((B0 |_| G |_| B') |_| W')).
  sui1; rewrite <- (set_eqvUnionAssoc B0 G B'); sui1; assumption.
  destruct (mphi2Eq _ Iz') as [_ E]; apply E; assumption.
  (* case: x2 in M' *)
  destruct (mphiMeq' _ Ix) as [z [Peq' [Peq [Peq1 [Iz [Peq2 Phi2Eq]]]]]].
  rewrite Peq1; rewrite Peq; rewrite Peq1 in H5.
  assert (IOz := M'even _ Ix).
  rewrite (isOddFalse IOz).
  apply Phi2Eq.
  destruct H5 as [okF | okF].
  elf; apply (not_even_and_odd x2); assumption.
  assumption.

  (* show fwd okay for field 2 *)
  clear H5; rename H8 into H5; rename x4 into x00; rename x5 into x4;
    rename x2 into x20; rename x3 into x2; clear H3; rename H4 into H3;
      clear H6; rename H10 into H6.
  assert (x4 = mphi1 x2).
  assert (HpV' : vs |||- 4 + y |-> mphi1 x2 ** top).
  eapply heapEqvTrue; eauto.
  apply (sconjMapRangeImpl H3); intros hx HpX.
  apply (contMapRangeImpl HpX).
  lli.
  sconjLift2 HpV; exact (contContEq HpV HpV').
  subst x4; clear HpV.
  unfold mapX; assert (IO := isOddOk x2); destruct (isOdd x2).
  (* case: x2 is odd *)
  destruct (mphiOddEq _ IO) as [Eq1 Eq2]; rewrite Eq1; rewrite Eq2; 
    rewrite (isOddTrue IO); reflexivity.
  (* case: x2 is even *)
  destruct H6 as [okF | Ix0].
  elf; apply (not_even_and_odd x2); assumption.
  assert (Ix : set_In x2 (((B0 |_| G |_| B') |_| W') |_| M |_| M')).
  rewrite (set_eqvUnionAssoc (B0 |_| G |_| B') W' (M |_| M')).
  destruct (set_union_elim Ix0).
  sui2; rewrite (set_eqvUnionSymm M M'); rewrite <- (set_eqvUnionAssoc W' M' M);
    rewrite (set_eqvUnionSymm W' M'); rewrite <- Weq; assumption.
  sui1; assumption.
  destruct (set_union_elim Ix) as [Ix' | Ix']; clear Ix.
  (* case: x2 uncopied *)
  cut (mphi1 x2 = x2).
  intro E1; rewrite E1.
  rewrite (isOddFalse IO); rewrite (mphiEq _ Ix').
  rewrite E1 in H5; destruct H5 as [X | okF].
  elf; apply (not_even_and_odd x2); assumption.
  destruct (mphi2Eq _ Ix') as [_ E]; auto.
  apply mphi1Eq; destruct (set_union_elim Ix') as [Ix'' | Ix''];
    [sui1 | rewrite Weq; sui2; sui2]; assumption.
  (* case: x2 copied *)
  destruct (set_union_elim Ix') as [Ix | Ix]; clear Ix'.
  (* case: x2 in M *)
  destruct (mphiMeq _ Ix) as [z [Peq [Mpeq [Mpeq1 [Iz [Mpeq2 P2eq]]]]]].
  rewrite Mpeq1; rewrite Mpeq; rewrite Mpeq1 in H5.
  assert (IOz := isOddOk z); destruct (isOdd z); [reflexivity | idtac].
  destruct H5 as [okZ | okZ].
  elf; apply (not_even_and_odd z); assumption.
  assert (Iz' : set_In z ((B0 |_| G |_| B') |_| W')).
  sui1; rewrite <- (set_eqvUnionAssoc B0 G B'); sui1; assumption.
  destruct (mphi2Eq _ Iz') as [_ E]; apply E; assumption.
  (* case: x2 in M' *)
  destruct (mphiMeq' _ Ix) as [z [Peq' [Peq [Peq1 [Iz [Peq2 Phi2Eq]]]]]].
  rewrite Peq1; rewrite Peq; rewrite Peq1 in H5.
  assert (IOz := M'even _ Ix).
  rewrite (isOddFalse IOz).
  apply Phi2Eq.
  destruct H5 as [okF | okF].
  elf; apply (not_even_and_odd x2); assumption.
  assumption.

  (* case: y in G' *)
  lli; [idtac | 
    simpl; apply (sconjMapRangeImpl Hp'');
      intros h'x Hp''x; apply (contMapRangeImpl Hp''x)].
  clear h' Hp''.
  destruct (set_map_elim Iy) as [z [yeq Iz]]; clear Iy.
  destruct (set_inter_elim Iz) as [Iz1 Iz2]; clear Iz.
  destruct (mphiMeq' _ Iz1) as [y' [Peq' [Meq0 [Meq1 [Iy [Meq2 Peq2]]]]]];
    rewrite Meq2 in yeq; subst y' y; hyr Peq2 (Peq2 Iz2).
  assert (z = x2);
    [destruct Iso' as [_ [_ [Iso' _]]]; apply Iso' with (mphi z); assumption |
     subst x2].
  exists z; exists (mphi1 x3); exists (mphi1 x4).
  constructor.
  auto.
  rename x3 into x2; rename x4 into x3.
  constructor.
  lli.
  cut (vs |||- z |-> mphi1 x2 ** 4 + z |-> mphi1 x3 ** top).
  intro HpS.
  assert (HX := objHpElim Iz2 (minObjHpElim1 (heapEqvImplL VhpM)));
    destruct (aconjElim (sconjDistrib2 VhpM)) as [_ VhpM']; clear VhpM.
  hpunfoldIn HX; lle HX; hyr VhpM' (sconjInsert VhpM' HX); lle VhpM'.
  assert (X1 := contContEq HpS VhpM'); sconjLift2 HpS; sconjLift2 VhpM';
    assert (X2 := contContEq HpS VhpM'); subst x4 x5; scm.
  apply (heapEqvTrue vsEq).
  assert (HH : h0 ||- z |-> x2 ** 4 + z |-> x3 ** top).
  apply contContCombine; auto; omega.
  apply (sconjMapRangeImpl HH); clear HH; intros h' HH.
  apply (contMapRangeImpl HH).
  apply (sconjMapRangeImpl HH); clear h' HH; intros h' HH.
  apply (contMapRangeImpl HH).
  lli.

  rename Iz2 into Iy2.
  assert (HpV : vs |||- (okObjAt objs z ** top) ** top).
  scm; destruct (aconjElim Hp2) as [Hp3 _]; hyr Hp3 (minObjHpElim1 Hp3);
    hyr Hp3 (objHpElim Iy2 Hp3); scm.
  hpunfoldIn HpV; lle HpV.
  assert (HpX : h0 ||- (okObjAt (W |_| M) z ** top) ** top).
  assert (Iz' : set_In z W); [rewrite Weq; sui1; assumption | idtac].
  clear H6 H7; scm; hyr Hp2 (objHpElim Iz' Hp2); scm.
  hpunfoldIn HpX; lle HpX.
  assert (X1 := contContEq HpX H6); sconjLift2 HpX; assert (X2 := contContEq HpX H7);
    subst x6 x7; clear HpX Iy2.

  constructor.
  (* show fwd okay for field 1 *)
  assert (x4 = mphi1 x2).
  assert (HpV' : vs |||- z |-> mphi1 x2 ** top).
  eapply heapEqvTrue; eauto.
  apply (sconjMapRangeImpl H6); intros hx HpX.
  apply (contMapRangeImpl HpX).
  lli.
  exact (contContEq HpV HpV').
  subst x4; clear HpV.
  apply Ximp.
  apply (okFieldValWeaken H5); intros q Iq; sui2; assumption.
  assumption.
  (* show fwd okay for field 2 *)
  assert (x5 = mphi1 x3).
  assert (HpV' : vs |||- 4 + z |-> mphi1 x3 ** top).
  eapply heapEqvTrue; eauto.
  apply (sconjMapRangeImpl H7); intros hx HpX.
  apply (contMapRangeImpl HpX).
  lli.
  sconjLift2 HpV; exact (contContEq HpV HpV').
  subst x5; clear HpV.
  apply Ximp.
  apply (okFieldValWeaken H11); intros q Iq; sui2; assumption.
  assumption.
  (* done showing the forwarding *)

  clear Vhp'.

  (* show that if the root is in objs, its fields are okay in objs *)
  assert (rootFieldsOk : set_In root objs -> 
    okFieldVal objs (mphi1 x1) /\ okFieldVal objs (mphi1 x0)).
  intro Ir.
  destruct (aconjElim (sconjDistrib2 Hp)) as [HpX _].
  assert (HpY := heapEqvImplR Hp).
  destruct (rootOk _ RRroot) as [_ okRoot].
  replace (rget s rroot) with root in okRoot; [idtac | reflexivity].
  destruct okRoot as [X | okRoot].
  elf; apply (not_even_and_odd root).
  apply ptrPeven; destruct K; clear vh' Hp'; subst k' k;
    [apply (contElim1L Hp0) |
     rewrite beq_false in Hp1; [apply (contElim1L Hp1) | omega]].
  assumption.
  assert (HH : h0 ||- (root |->__ ** top) ** top).
  assert (okRoot' : set_In root ((B0 |_| B') |_| G)).
  set_InElim.
  sui1; sui1; assumption.
  sui2; assumption.
  sui1; sui2; assumption.
  clear okRoot; destruct (set_union_elim okRoot') as [okRoot | okRoot]; clear okRoot';
    [sconjLift2 HpY | sconjLift3of3 HpY]; scm;
    hyr Hp2 (objHpElim okRoot Hp2); scm; hpunfoldIn Hp2; lleim.
  clear HpY; lle HH.
  assert (Z : okFieldVal objs (mphi1 x2) /\ okFieldVal objs (mphi1 x3)).
  assert (HV : vs |||- (okObjAt objs root ** top) ** top).
  scm; destruct (aconjElim Hp2) as [HpY _]; clear Hp2;
    hyr HpY (objHpElim Ir (minObjHpElim1 HpY)); scm.
  assert (HV' : vs |||- root |-> mphi1 x2 ** 4 + root |-> mphi1 x3 ** top).
  eapply heapEqvTrue; eauto.
  apply (sconjMapRangeImpl HH); clear HH; intros h3 HH.
  apply (contMapRangeImpl HH).
  apply (sconjMapRangeImpl HH); clear h3 HH; intros h3 HH.
  apply (contMapRangeImpl HH).
  lli.
  hpunfoldIn HV; lle HV; assert (X1 := contContEq HV HV'); 
    sconjLift2 HV; sconjLift2 HV'; assert (X2 := contContEq HV HV');
      subst x4 x5; constructor; assumption.
  hyr HpX (sconjInsert HpX HH); lle HpX; destruct K;
    [idtac | sconjLift2 HpX]; clear Hp'; subst vh' k' k;
      assert (X1 := contContEq Hp0 HpX); sconjLift2 HpX;
	assert (X2 := contContEq Hp1 HpX);
	  subst x3 x2; destruct Z; constructor; assumption.

  constructor.

  (* show the virtual heap is closed *)
  apply objHpMerge; scm.
  destruct (set_In_dec root objs) as [Ir | NI].
  (* case: root in objs *)
  rewrite (set_mem_true Ir) in Hp2.
  assert (Sub : set_subset (set_add root empty_set) objs).
  intros z Iz; simpl in Iz; destruct Iz; [subst z | elf]; assumption.
  unfold hpredTrue; rewrite (subsetInterEq Sub).
  apply objHpIntro; hyr Hp2 (unitIR Hp2).
  scm.

  assert (HX : h3 ||- okFieldAt
    (set_add root empty_set |_| set_inter (set_remove root (B0 |_| B' |_| G) |_| W') objs
      |_| set_map mphi2 (set_inter M' objs)) (k' + root) **
    okFieldAt (set_add root empty_set |_| 
      set_inter (set_remove root (B0 |_| B' |_| G) |_| W') objs
      |_| set_map mphi2 (set_inter M' objs)) (k + root)).
  destruct (rootFieldsOk Ir) as [oko1 oko2]; clear rootFieldsOk.
  hpunfold; scm; lli; eauto.
  (* first field *)
  clear h3 Hp2 oko2.
  assert (IO := isOddOk x1); destruct (isOdd x1); [left | right].
  (* case: field is odd *)
  destruct (mphiOddEq _ IO) as [_ Eq]; rewrite Eq; assumption.
  (* case: field is even *)
  destruct x1Ok as [X | x1Ok]; [elf; apply (not_even_and_odd x1); assumption | idtac].
  destruct (set_union_elim x1Ok) as [Ix1 | Ix1]; clear x1Ok.
  (* case: field is in to-space *)
  assert (IX : set_In x1 ((B0 |_| G |_| B') |_| W)).
  sui1; auto.
  assert (IX' : set_In x1 ((B0 |_| G |_| B') |_| W')).
  sui1; auto.
  rewrite (mphi1Eq _ IX) in oko1; rewrite (mphiEq _ IX').
  compare x1 root; intro NE.
  sui1; left; subst x1; reflexivity.
  sui2; sui1; apply set_inter_intro.
  sui1; apply set_remove_intro.
  rewrite (set_eqvUnionSymm B' G); assumption.
  assumption.
  destruct oko1; [elf; apply (not_even_and_odd x1); assumption | assumption].
  (* case: field in from-space *)
  destruct (set_mem root G); [idtac | elf; assumption].
  assert (Ix1' : set_In x1 (W' |_| M |_| M')).
  rewrite (set_eqvUnionSymm M M'); rewrite <- (set_eqvUnionAssoc W' M' M);
    rewrite (set_eqvUnionSymm W' M'); rewrite <- Weq; assumption.
  clear Ix1; destruct (set_union_elim Ix1') as [Ix | Ix]; clear Ix1'.
  (* case: field in W' *)
  assert (X1 : set_In x1 ((B0 |_| G |_| B') |_| W)).
  sui2; rewrite Weq; sui2; assumption.
  assert (X : set_In x1 ((B0 |_| G |_| B') |_| W')).
  sui2; assumption.
  rewrite (mphi1Eq _ X1) in oko1; rewrite (mphiEq _ X); clear X X1.
  destruct oko1; [elf; apply (not_even_and_odd x1); assumption | idtac].
  sui2; sui1; apply set_inter_intro.
  sui2; assumption.
  assumption.
  (* case: field has been fwded *)
  destruct (set_union_elim Ix) as [Ix' | Ix']; clear Ix.
  (* case: field in M *)
  clear mphi1Eq mphiEq.
  destruct (mphiMeq _ Ix') as [y' [Peq' [Mxeq' [Meq0 [Iy' _]]]]].
  rewrite Mxeq'; rewrite Meq0 in oko1; clear Mxeq' Meq0.
  destruct oko1 as [oko1 | oko1].
  elf; apply (not_even_and_odd y'); [apply BGeven; eauto | assumption].
  compare y' root; intro NE.
  sui1; left; auto.
  sui2; sui1; apply set_inter_intro.
  sui1; apply set_remove_intro.
  rewrite (set_eqvUnionSymm B' G); rewrite <- (set_eqvUnionAssoc B0 G B');
    sui1; eauto.
  assumption.
  assumption.
  (* case: field in M' *)
  destruct (mphiMeq' _ Ix') as [y [Meqy [Mp [Mp' [Iy _]]]]];
    rewrite Mp; rewrite Mp' in oko1; clear Mp Mp'.
  sui2; sui2; apply set_map_intro with x1.
  subst mphi2; cbv beta; rewrite Meqy; reflexivity.
  destruct oko1; [elf; apply (not_even_and_odd x1); assumption | idtac].
  apply set_inter_intro; assumption.
  (* other field (the one that was forwarded) *)
  clear oko1 h3 Hp2 Sub Ir xOk.
  assert (IO := isOddOk x0); unfold mapX in Meq; destruct (isOdd x0).
  (* case: x0 is odd *)
  injection Meq; intro; subst x0; left; assumption.
  (* case: x0 is even *)
  right; unfold map_unionnn in Meq; unfold map_union at 1 in Meq; rewrite spaceIdMapEq in Meq.
  destruct x0Ok as [xok | xok]; [elf; apply (not_even_and_odd x0); assumption | idtac].
  compare x root; intro NE.
  sui1; left; auto.
  sui2.
  destruct (set_union_elim xok) as [okX | okX]; clear xok.
  (* case: x0 in to-space *)
  rewrite set_mem_true in Meq.
  injection Meq; clear Meq; intro; subst x0.
  sui1; apply set_inter_intro.
  sui1; apply set_remove_intro.
  rewrite (set_eqvUnionSymm B' G); assumption.
  assumption.
  rewrite mphi1Eq in oko2.
  destruct oko2; [elf; apply (not_even_and_odd x) | idtac]; assumption.
  sui1; assumption.
  apply objRangeSetRangeSubset; rewrite toEq.
  set_InElim.
  sui1; assumption.
  sui2; sui1; assumption.
  sui2; sui2; sui2; assumption.
  (* case: x0 in from-space *)
  destruct (set_mem root G).
  2:elf; assumption.
  rewrite set_mem_false in Meq.
  assert (III : set_In x0 (W' |_| M |_| M')).
  rewrite (set_eqvUnionSymm M M'); rewrite <- (set_eqvUnionAssoc W' M' M);
    rewrite (set_eqvUnionSymm W' M'); rewrite <- Weq; assumption.
  destruct (set_union_elim III) as [Ix | Ix]; clear okX III.
  (* case: x0 in W'.  contradiction. *)
  elf; apply (set_disj_elim WM'disj x0).
  destruct Iso12 as [Iso12 _]; destruct (Iso12 x0) as [_ Q]; apply Q;
    rewrite Meq; discriminate.
  assumption.
  (* case: x0 in M |_| M' *)
  unfold map_union in Meq; destruct (set_union_elim Ix) as [Ix' | Ix']; clear Ix.
  (* case: x0 in M *)
  destruct (mphiMeq _ Ix') as [y [Peq [Mx [Mx1 [Iy [Mx2 Peq2]]]]]];
    rewrite Peq in Meq; injection Meq; clear Meq; intro; rewrite Mx1 in oko2.
  sui1; apply set_inter_intro.
  sui1; apply set_remove_intro.
  rewrite H in Iy; set_InElim; [sui1 | 
    sui2; sui2]; assumption.
  assumption.
  rewrite H in oko2.
  destruct oko2; [elf; rewrite H in Iy; apply (not_even_and_odd x); eauto | assumption].
  (* case: x0 in M' *)
  destruct (mphiMeq' _ Ix') as [y [Peq [Mx [Mx1 [Iy [Mx2 Peq2]]]]]];
    rewrite Mx1 in oko2; sui2; rewrite Peq in Meq.
  destruct oko2 as [oko2 | Ix0']; [elf; apply (not_even_and_odd x0); assumption | idtac].
  hyr Peq2 (Peq2 Ix0').
  destruct Iso as [Iso _]; destruct (Iso x0) as [_ Q]; destruct (phi x0).
  elf; apply (set_disj_elim WMdisj x0).
  rewrite Weq; sui1; assumption.
  apply Q; discriminate.
  injection Meq; clear Meq; intro; subst y.
  apply set_map_intro with x0.
  rewrite Mx2; rewrite H; reflexivity.
  apply set_inter_intro; assumption.

  intro Ix0'.
  apply (set_disj_elim frToDisj x0).
  apply objRangeSetRangeSubset; rewrite <- WMeq; assumption.
  assumption.

  (* done with both fields.  show that it is okay no matter what K is *)
  clear vh' Hp'; destruct K; subst k' k; [sconjLift2of2 HX | idtac]; assumption.

  (* case: root not in objs *)
  rewrite (set_mem_false NI) in Hp2.
  unfold hpredTrue; setoid_replace (set_inter (set_add root empty_set) objs) with empty_set.
  apply objHpIntroE; assumption.
  simpl; rewrite (set_mem_false NI); apply set_eqvRefl.

  (* now to deal with the rest of the objects *)
  assert (HX : h3 ||- foralls x in (set_inter (set_remove root (B0 |_| B' |_| G) |_| W') objs) |_|
    set_map mphi2 (set_inter M' objs), objIsFwdedH phi2 h x).
  apply fasMerge.
  assumption.
  apply (fasFasDisj' Hp2); hpunfold; simpl; illeim.
  clear Hp2; unfold objHp; llMatch.
  apply (fwdedObjHpOk Hp2) with objs.
  assumption.

  constructor.
  assumption.
  constructor.

  (* show the roots are okay *)
  assert (rootOk' : forall r, In r rr -> mapX phi2 (rget vs r) = Some (rget s r)).
  intros r Ir.
  destruct (rootOk _ Ir) as [E1 okRoot]; rewrite E1.
  unfold mapX; assert (IO := isOddOk (rget s r)); destruct (isOdd (rget s r)).
  reflexivity.
  destruct okRoot as [X | Irr].
  elf; apply (not_even_and_odd (rget s r)); assumption.
  destruct (mphi2Eq (rget s r)) as [_ Peq].
  sui1; assumption.
  apply Peq.
  assert (HX := heapEqvImplL VhpM).
  destruct (minObjHpRootOk HX (rget s r)) as [X1 | X1].
  rewrite <- E1.
  apply rrlistMapIntro; assumption.
  elf; apply (not_even_and_odd (rget s r)); assumption.
  assumption.

  intros r II'.
  replace (rget (h, rfileOf vs) r) with (rget vs r); [idtac | reflexivity].
  transitivity (Some (rget s'' r)).
  2:reflexivity.
  destruct (regCompare r ra).
  subst r; rewrite ReqRa.
  exact (rootOk' _ II').
  rewrite (rootOk' _ II').
  destruct (NRR _ II') as [Eq | X].
  subst r; rewrite <- Req'; [idtac | regNotInListSolver];
    rewrite <- Req; [stSimpl; reflexivity | regNotInListSolver].
  Ltac rcompStep H r r' := 
    destruct (regCompare r r'); [elf; subst r; calleeSavedSolver H | idtac].
  rcompStep X r rfrStart; rcompStep X r rfrEnd;
  rcompStep X r rtoStart; rcompStep X r rtoEnd; rcompStep X r rfree;
    rcompStep X r ralloc; rcompStep X r rdTmp1; rcompStep X r a0;
    rcompStep X r rdTmp2;
    rcompStep X r v0; rcompStep X r ast; rcompStep X r raSave0; 
    rcompStep X r rtemp0; rcompStep X r rtemp1; rcompStep X r rtemp2.
  rewrite <- Req'.
  rewrite <- Req.
  stSimpl; reflexivity.
  simpl; repeat (constructor; [assumption | idtac]); trivial.
  simpl; repeat (constructor; [assumption | idtac]); trivial.

  (* show the virtual state is forwarded properly *)

  assert (Iroot : set_In root (B0 |_| G |_| B')).
  destruct (rootOk _ RRroot) as [_ okRoot].
  assert (IO := isOddOk root); destruct (isOdd root).
  elf; apply (not_even_and_odd root).
  apply ptrPeven; destruct K as [E | E]; subst h1 vh' k' k;
    [apply (contElim1L Hp0) | apply (contElim1L Hp1)].
  assumption.
  destruct okRoot.
  elf; apply (not_even_and_odd root); assumption.
  assumption.

  apply objIsFwdedHMerge; scm.

  (* case: root object *)
  assert (X := set_memSound root objs); destruct (set_mem root objs).
  assert (Sub : set_subset (set_add root empty_set) objs).
  intros z Iz; simpl in Iz; destruct Iz; [subst z | elf]; assumption.
  unfold hpredTrue; rewrite (subsetInterEq Sub).
  apply fasIntro.
  hyr Hp2 (unitIR Hp2); scm; hpunfold.
  2:scm.

  assert (E : x = mphi x0).
  unfold mapX in Meq; assert (IO := isOddOk x0); destruct (isOdd x0).
  (* case: x0 is odd *)
  injection Meq; clear Meq; intro; subst x0.
  destruct (mphiOddEq _ IO) as [_ R]; rewrite R; reflexivity.
  (* case: x0 is even *)
  destruct x0Ok as [Q | Ix0]; [elf; apply (not_even_and_odd x0); assumption | idtac].
  unfold map_unionnn in Meq; unfold map_union at 1 in Meq; rewrite spaceIdMapEq in Meq.
  set (Fr := B0 |_| G |_| B') in * .
  assert (Ix0' : set_In x0 ((Fr |_| W') |_| M' |_| M)).
  rewrite (set_eqvUnionAssoc Fr W' (M' |_| M));
    rewrite <- (set_eqvUnionAssoc W' M' M); rewrite (set_eqvUnionSymm W' M'); rewrite <- Weq;
      set_InElim.
  sui1; assumption.
  destruct (set_mem root G); [sui2 | elf]; assumption.
  destruct (set_union_elim Ix0') as [Ix | Ix]; clear Ix0 Ix0'.
  (* case: x0 is unforwarded *)
  rewrite (mphiEq _ Ix).
  destruct (set_union_elim Ix) as [Ix' | Ix']; clear Ix.
  (* case: x0 in to-space *)
  rewrite set_mem_true in Meq.
  injection Meq; auto.
  apply objRangeSetRangeSubset; rewrite toEq; rewrite (set_eqvUnionSymm F B'); 
    rewrite <- (set_eqvUnionAssoc G B' F); rewrite <- (set_eqvUnionAssoc B0 (G |_| B') F);
      sui1; assumption.
  (* case: x0 in W' *)
  rewrite set_mem_false in Meq.
  elf; destruct Iso12 as [Iso12 _]; destruct (Iso12 x0) as [_ Q]; 
    destruct (map_union phi phi' x0).
  apply (set_disj_elim WM'disj x0); [apply Q; discriminate | assumption].
  discriminate Meq.
  intro Ix; apply (set_disj_elim frToDisj x0).
  apply objRangeSetRangeSubset; rewrite <- WMeq; rewrite Weq; sui1; sui2; assumption.
  assumption.
  (* case: x0 has been forwarded *)
  rewrite set_mem_false in Meq.
  unfold map_union in Meq.
  destruct (set_union_elim Ix) as [Ix' | Ix']; clear Ix.
  (* case: x0 in M' *)
  destruct (mphiMeq' _ Ix') as [y [Peq [MPeq _]]].
  rewrite MPeq; rewrite Peq in Meq.
  destruct Iso as [Iso _]; destruct (Iso x0) as [_ Q]; destruct (phi x0).
  elf; apply (set_disj_elim WMdisj x0).
  rewrite Weq; sui1; assumption.
  apply Q; discriminate.
  injection Meq; intro; symmetry; assumption.
  (* case: x0 in M *)
  destruct (mphiMeq _ Ix') as [y [Peq [MPeq _]]];
    rewrite MPeq; rewrite Peq in Meq; injection Meq; auto.
  intro Iq; apply (set_disj_elim frToDisj x0).
  apply objRangeSetRangeSubset; rewrite <- WMeq; rewrite Weq; set_InElim;
    [sui1; sui1 | sui2]; assumption.
  assumption.
  (* done showing E *)

  destruct K as [keq | keq]; clear Hp'; subst vh' k' k.
  rewrite beq_true in Hp1; rewrite beq_true in Hp2;
    sconjLift2of2 Hp2.
  existsl x; existsl (mphi x1); lli.
  2:assumption.
  exists root; exists (mphi1 x0); exists (mphi1 x1); constructor.
  destruct (mphi2Eq root) as [_ Xeq].
  destruct (rootOk _ RRroot) as [_ OkR].
  destruct OkR as [okR | okR].
  elf; apply (not_even_and_odd root).
  2:assumption.
  apply ptrPeven; apply (contElim1L Hp2).
  sui1; assumption.
  auto.
  clear Hp2 h1 NEk Req.
  constructor.
  lli.
  assert (HV : vs |||- root |-> mphi1 x0 ** 4 + root |-> mphi1 x1 ** top).
  assert (Hp00 : h0 ||- root |->__ ** top).
  assert (Iroot' : set_In root (B0 |_| B') \/ set_In root G).
  destruct (set_union_elim Iroot) as [Iroot' | Iroot']; clear Iroot.
  left; sui1; assumption.
  destruct (set_union_elim Iroot') as [Iroot | Iroot]; clear Iroot'.
  right; assumption.
  left; sui2; assumption.
  clear Iroot; destruct Iroot' as [Iroot | Iroot].
  assert (HX : h0 ||- top ** (root |->__ ** top) ** top).
  scm; hyr Hp2 (objHpElim Iroot Hp2); hpunfoldIn Hp2; lleim.
  lleim.
  assert (HX : h0 ||- top ** top ** (root |->__ ** top)).
  scm; hyr Hp2 (objHpElim Iroot Hp2); hpunfoldIn Hp2; lleim.
  lleim.
  lle Hp00; destruct (aconjElim (sconjDistrib2 Hp)) as [HpY _];
    hyr HpY (sconjInsert HpY Hp00); lle HpY.
  assert (E1 := contContEq Hp0 HpY); sconjLift2 HpY; assert (E2 := contContEq Hp1 HpY); 
    subst x3 x2; clear Hp0 Hp1 HpY.
  apply (heapEqvTrue vsEq); apply (sconjMapRangeImpl Hp00); clear Hp00; intros h4 Hp00.
  apply (contMapRangeImpl Hp00).
  apply (sconjMapRangeImpl Hp00); clear h4 Hp00; intros h4 Hp00.
  apply (contMapRangeImpl Hp00).
  lli.
  clear Sub h3 Iroot.
  destruct (aconjElim (sconjDistrib2 VhpM)) as [_ VhpM'];
    assert (HH := objHpElim X (minObjHpElim1 (heapEqvImplL VhpM)));
      hpunfoldIn HH; lle HH.
  hyr VhpM' (sconjInsert VhpM' HH); lle VhpM'.
  assert (E1 := contContEq HV VhpM'); sconjLift2 HV; sconjLift2 VhpM';
    assert (E2 := contContEq HV VhpM'); subst x2 x3.
  scm.
  destruct (rootFieldsOk X) as [x1okObjs x0okObjs].
  constructor.
  (* x0 *)
  rewrite E; apply Ximp.
  apply (okFieldValWeaken x0Ok).
  intros z Iz; destruct (set_union_elim Iz) as [Iq | Iq].
  sui1; assumption.
  destruct (set_mem root G); [assumption | elf; assumption].
  assumption.
  (* x1 *)
  apply Ximp.
  apply (okFieldValWeaken x1Ok).
  intros z Iz; destruct (set_union_elim Iz) as [Iq | Iq].
  sui1; assumption.
  destruct (set_mem root G); [assumption | elf; assumption].
  assumption.
  (* done showing first case of K *)

  (* second case of K, mostly copied from above (hopefully) *)
  rewrite beq_false in Hp1; rewrite beq_false in Hp2.
  2:omega.
  2:omega.
  2:omega.
  existsl (mphi x1); existsl x; lli.
  2:assumption.
  exists root; exists (mphi1 x1); exists (mphi1 x0); constructor.
  destruct (mphi2Eq root) as [_ Xeq].
  destruct (rootOk _ RRroot) as [_ OkR].
  destruct OkR as [okR | okR].
  elf; apply (not_even_and_odd root).
  2:assumption.
  apply ptrPeven; apply (contElim1L Hp2).
  sui1; assumption.
  auto.
  clear Hp2 h1 NEk Req.
  constructor.
  lli.
  assert (HV : vs |||- root |-> mphi1 x1 ** 4 + root |-> mphi1 x0 ** top).
  assert (Hp00 : h0 ||- root |->__ ** top).
  assert (Iroot' : set_In root (B0 |_| B') \/ set_In root G).
  destruct (set_union_elim Iroot) as [Iroot' | Iroot']; clear Iroot.
  left; sui1; assumption.
  destruct (set_union_elim Iroot') as [Iroot | Iroot]; clear Iroot'.
  right; assumption.
  left; sui2; assumption.
  clear Iroot; destruct Iroot' as [Iroot | Iroot].
  assert (HX : h0 ||- top ** (root |->__ ** top) ** top).
  scm; hyr Hp2 (objHpElim Iroot Hp2); hpunfoldIn Hp2; lleim.
  lleim.
  assert (HX : h0 ||- top ** top ** (root |->__ ** top)).
  scm; hyr Hp2 (objHpElim Iroot Hp2); hpunfoldIn Hp2; lleim.
  lleim.
  lle Hp00; destruct (aconjElim (sconjDistrib2 Hp)) as [HpY _];
    hyr HpY (sconjInsert HpY Hp00); lle HpY.
  assert (E1 := contContEq Hp1 HpY); sconjLift2 HpY; assert (E2 := contContEq Hp0 HpY); 
    subst x3 x2; clear Hp0 Hp1 HpY.
  apply (heapEqvTrue vsEq); apply (sconjMapRangeImpl Hp00); clear Hp00; intros h4 Hp00.
  apply (contMapRangeImpl Hp00).
  apply (sconjMapRangeImpl Hp00); clear h4 Hp00; intros h4 Hp00.
  apply (contMapRangeImpl Hp00).
  lli.
  clear Sub h3 Iroot.
  destruct (aconjElim (sconjDistrib2 VhpM)) as [_ VhpM'];
    assert (HH := objHpElim X (minObjHpElim1 (heapEqvImplL VhpM)));
      hpunfoldIn HH; lle HH.
  hyr VhpM' (sconjInsert VhpM' HH); lle VhpM'.
  assert (E1 := contContEq HV VhpM'); sconjLift2 HV; sconjLift2 VhpM';
    assert (E2 := contContEq HV VhpM'); subst x2 x3.
  scm.
  destruct (rootFieldsOk X) as [x1okObjs x0okObjs].
  constructor.

  (* show the x1 case *)
  apply Ximp.
  apply (okFieldValWeaken x1Ok).
  intros z Iz; destruct (set_union_elim Iz) as [Iq | Iq].
  sui1; assumption.
  destruct (set_mem root G); [assumption | elf; assumption].
  assumption.

  (* x0 case *)
  rewrite E; apply Ximp.
  apply (okFieldValWeaken x0Ok).
  intros z Iz; destruct (set_union_elim Iz) as [Iq | Iq].
  sui1; assumption.
  destruct (set_mem root G); [assumption | elf; assumption].
  assumption.
  (* done showing second case of K *)

  (* case: root not in objs *)
  assert (DD : set_disj (set_add root empty_set) objs).
  apply set_disj_intro; intros z Iz Iz'.
  simpl in Iz; destruct Iz; [subst z; auto | assumption].
  unfold hpredTrue; rewrite (set_disjInterEmpty DD); apply fasIntroE; assumption.

  (* show the rest of the objects are forwarded properly *)
  apply objIsFwdedHMerge; scm.

Qed.



Lemma bakerReadOk : forall k P, 
  P SCANFIELD = SomeT (bspeci scanFieldPre scanFieldGuar) ->
  P BAKERREADRET = SomeT bakerReadRetSpec ->
  k = 0 \/ k = 4 ->
  iseqOkS P (bakerReadPre k) (bakerReadGuar k) (bakerRead k).
Proof.
  intros k P Peq Peq' K.
  unfold bakerRead.
  iseqWeakSolver fail; eauto.
  exact scanFieldGuarPresRa.

  clear P Peq Peq'.
  intros s [vs [Hp Vhp]].
  unfold repr in Hp; lle Vhp; lle Hp.
  rename x2 into toSt; rename x4 into scan; rename x5 into free; rename x6 into alloc;
    rename x3 into toEnd; rename x7 into phi; rename x8 into W; rename x9 into M; 
      rename x0 into frSt; rename x1 into frEnd; rename x10 into vh.
  destruct H3 as [_ [Al [Al' [WMeq [Iso [Rok VsEq]]]]]].
  set (B := rangeObjs toSt scan) in * ;
    set (G := rangeObjs scan free) in * ;
      set (B' := rangeObjs alloc toEnd) in * .

  assert (rootEq : rget vs rroot = rget s rroot).
  destruct (Rok rroot).
  simpl; auto.
  assumption.

  assert (I : set_In (rget s rroot) (B |_| G |_| B') /\ even (rget s rroot)).
  destruct (Rok rroot).
  simpl; auto.
  assert (X := isOddOk (rget s rroot)).
  destruct (isOdd (rget s rroot)).
  elf; apply (not_even_and_odd (rget s rroot)).
  apply ptrPeven.
  rewrite H in Vhp; assert (Y := contElim1L Vhp).
  destruct K; subst k.
  assumption.
  inversion Y; assumption.
  assumption.
  constructor.
  destruct H0.
  elf; apply (not_even_and_odd (rget s rroot)); assumption.
  assumption.
  assumption.

  destruct I as [I rootEven].
  replace (RFile.get (rfileOf s)) with (rget s) in Hp by reflexivity.

  assert (Hp' : s |||- 
    (Heap.heap_eqv vh &&
      objHp (W |_| M) W **
      objHp (B |_| G |_| B') (B |_| B') **
      objHp ((W |_| M) |_| B |_| G |_| B') G) **
    mapHp M phi **
    (foralls x in rangeObjs free alloc, x |->__) **
    rget s rgcInfo |->, (frSt::frEnd::toSt::toEnd::scan::free::alloc::nil) ** top).
  lli; scm.
  clear Hp; rename Hp' into Hp.

  (* first, step through the block *)
  assert (Hp' : s |||-
    rget s rgcInfo |->, (frSt::frEnd::toSt::toEnd::scan::free::alloc::nil) **
      top).
  sconjLift4 Hp; scm.
  lle Hp'.
  rewrite (lwStep Hp'); try reflexivity.
  sconjLift2 Hp'.
  rewrite (lwStep Hp'); try reflexivity.
  sconjLift3 Hp'.
  rewrite (lwStep Hp'); try reflexivity.
  sconjLift4 Hp'.
  rewrite (lwStep Hp'); try reflexivity.
  sconjLift6 Hp'.
  rewrite (lwStep Hp'); try reflexivity.
  sconjLift7 Hp'.
  rewrite (lwStep Hp'); try reflexivity.
  rewrite addiuStep0; stSimpl.
  rewrite addiuStep; stSimpl.
  rewrite addiuStep0; stSimpl.

  set (F := rangeObjs free alloc) in * .
  apply callChain.
  clear Hp'.

  assert (toEq : set_eqv (rangeObjs toSt toEnd) (B |_| G |_| F |_| B')).
  clear Hp K; simpl in Al; intuition.
  assert (LE1 : alloc <= toEnd).
  destruct H3; omega.
  subst F B'; rewrite (rangeObjsTransEqv H0 LE1).
  assert (LE2 : free <= toEnd).
  destruct H0; omega.
  subst G; rewrite (rangeObjsTransEqv H1 LE2).
  assert (LE3 : scan <= toEnd).
  destruct H1; omega.
  subst B; rewrite (rangeObjsTransEqv H LE3).
  apply set_eqvRefl.
  
  set (sp := (rset (rset (rset (rset (rset (rset (rset
    (rset (rset (rset s rfrStart frSt) rfrEnd frEnd) rtoStart toSt) rtoEnd toEnd) 
    rfree free) ralloc alloc) rdTmp1 (rget s ra)) a0 
    (k + rget s rroot)) rdTmp2 (k + rget s rroot)) ra BAKERREADRET)).

  (* show that we can call the return function *)
  assert (SCok : scanFieldStOk phi W M F
     (okFieldAt ((W |_| M) |_| setRange toSt toEnd) (k + rget s rroot) ** 
       20 + rget s rgcInfo |->_ ** top) sp).
  unfold scanFieldStOk, fromSpace, toSpace; subst sp; stSimpl.
  destruct (aconjElim (sconjDistrib2 Hp)) as [_ Hp']; clear Hp;
    rename Hp' into Hp.
  clear Vhp VsEq rootEq.
  assert (HH : s |||- (foralls z in rangeObjs frSt frEnd, z |->__) **
    (foralls z in rangeObjs toSt toEnd, z |->__) ** top).
  lle Hp; sconjLift4 Hp; apply sconjAssocR in Hp;
    sconjLift3 Hp; sconjLift3 Hp; sconjLift4 Hp;
    hyr Hp (sconjAssocR (sconjAssocR Hp)); sconjLift2 Hp; scm.
  unfold hpredTrue; rewrite <- WMeq; apply objsMerge; apply sconjComm; scm.
  exact (mapHpIsObjs Hp).
  exact (objHpIsObjs Hp).
  lle Hp.
  unfold hpredTrue; rewrite toEq.
  rewrite (set_eqvUnionSymm B (G |_| F |_| B'));
    rewrite <- (set_eqvUnionAssoc G F B');
      rewrite (set_eqvUnionAssoc (G |_| F) B' B);
	rewrite (set_eqvUnionAssoc G F (B' |_| B));
	  rewrite (set_eqvUnionSymm B' B).
  sconjLift3of3 Hp; apply objsMerge; scm.
  exact (objHpIsObjs Hp).
  apply objsMerge; scm.
  exact (objHpIsObjs Hp).
  assert (frToDisj : set_disj (setRange frSt frEnd) (setRange toSt toEnd)).
  apply disjObjsDisjRange with (heapOf s).
  assumption.
  assumption.
  clear K; simpl in Al; intuition; repeat (eapply aligned8Trans; eauto).
  constructor.
  assumption.
  constructor.
  rewrite WMeq; apply objRangeSetRangeSubset.
  constructor.
  subst; apply set_eqvRefl.
  set (GCI := rget s rgcInfo
    |->, (frSt :: frEnd :: toSt :: toEnd :: scan :: free :: alloc :: nil)) in Hp; lle Hp.
  constructor.
  intros y z Peq NSP.
  destruct Iso as [_ [Iso _]]; assert (Y := Iso _ _ Peq); clear Peq Iso.
  assert (I' : set_In z (rangeObjs toSt toEnd)).
  rewrite toEq.
  set_InElim.
  sui1; assumption.
  sui2; sui1; assumption.
  clear Y I; rename I' into I.
  destruct NSP.
  apply (not_even_and_odd z).
  apply ptrPeven.
  assert (Y : s |||- (z |->__ ** top) ** top).
  clear Hp; sconjLift2 HH; scm.
  hyr Hp (fasElim I Hp); hpunfoldIn Hp; scm.
  lle Y; apply (contElim1L Y).
  assumption.
  apply H; apply objRangeSetRangeSubset; assumption.
  constructor.
  clear K; simpl in Al; intuition;
    destruct H; destruct H1; destruct H0; omega.
  constructor.
  clear K; simpl in Al; intuition; destruct H; destruct H1; 
    destruct H0; destruct H3; omega.
  constructor.
  clear K; simpl in Al; intuition; destruct H; destruct H1; 
    destruct H0; destruct H3; omega.
  sconjLift5 Hp; sconjLift2 Hp; sconjLift5 Hp.
  apply (sconjImplR Hp); clear Hp; intros h Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  clear rootEven Rok.
  assert (Hp' : h ||- objHp ((W |_| M) |_| B |_| G |_| B') (B |_| G |_| B') ** GCI ** top).
  apply sconjAssocR in Hp; scm.
  set (Q := (W |_| M) |_| B |_| G |_| B') in *;
  unfold hpredTrue; rewrite (set_eqvUnionSymm G B'); 
    rewrite <- (set_eqvUnionAssoc B B' G); apply objHpMerge; scm.
  subst Q; apply (objHpWeaken Hp); intros z Iz; sui2; auto.
  assert (Hp'' : h ||- ((okFieldAt ((W |_| M) |_| B |_| G |_| B') (k + rget s rroot) ** 
    top) ** top) ** (20 + rget s rgcInfo |->_ ** top) ** top).
  scm.
  hyr Hp0 (objHpElim I Hp0); scm.
  unfold okObjAt in Hp0; destruct K; subst k; [idtac | sconjLift2of2 Hp0]; scm.
  subst GCI; sconjLift6 Hp0; lli; scm.
  clear Hp Hp'; lle Hp''; scm.
  apply (okFieldAtWeaken Hp); intros z Iz.
  clear I; destruct (set_union_elim Iz) as [I | I]; clear Iz.
  sui1; assumption.
  sui2; apply objRangeSetRangeSubset; rewrite toEq.
  rewrite (set_eqvUnionSymm F B').
  rewrite <- (set_eqvUnionAssoc B G (B' |_| F));
    rewrite <- (set_eqvUnionAssoc (B |_| G) B' F); sui1.
  rewrite (set_eqvUnionAssoc B G B'); assumption.
  lli; eauto.

  constructor.
  
  unfold scanFieldPre, scanFieldStOk, toSpace, fromSpace;
    stSimpl; exists phi; exists W; exists M; exists F.
  unfold scanFieldStOk in SCok; clear K; intuition; scm.

  (* show that we can call the return function *)
  unfold scanFieldGuar, scanFieldPre, scanFieldStOk, toSpace, fromSpace;
    subst sp; stSimpl.
  intros s' [Guar Req].
  constructor.
  unfold bakerReadRetPre; rewrite <- Req; [stSimpl | regNotInListSolver].
  assert (Z := Guar phi W M F (20 + rget s rgcInfo |->_ ** top) SCok); clear Guar SCok.
  generalize dependent Z.
  set (free' := rget s' rfree) in * .
  repeat (rewrite <- Req; [stSimpl | regNotInListSolver]).
  intros [M' [W' [F' [G' [phi' X]]]]].
  destruct X as [_ [_ [_ [_ [_ X]]]]].
  destruct X as [_ [_ [_ [_ [_ [_ [_ X]]]]]]].
  sconjLift6 X; sconjLift5 X; scm; hpunfoldIn Hp0; lleim.


  (* finally, show the guarantee *)

  (* first, a little clean up *)
  assert (Hp' : s |||-
    rget s rgcInfo |->, (frSt::frEnd::toSt::toEnd::scan::free::alloc::nil) **
      top).
  sconjLift4 Hp; scm.
  clear Hp F I rootEq vs Vhp Rok VsEq B G B' Iso WMeq vh Al Al' x W M toEq SCok.
  intros s'' Guar'.
  constructor.
  Focus 1.
  intros vs A rr Hp Irr NIrr.
  unfold repr in Hp.
  assert (x := 0); lle Hp; clear x.
  clear phi.
  rename x2 into toSt'; rename x4 into scan'; rename x5 into free'; rename x6 into alloc';
    rename x3 into toEnd'; rename x7 into phi; rename x8 into W; rename x9 into M; 
      rename x0 into frSt'; rename x1 into frEnd'; rename x10 into vh.
  destruct H3 as [NIrrG [Al [Al' [WMeq [Iso [Rok VsEq]]]]]].

  assert (Hp'' : s |||- 
    top ** top ** top **
    rget s rgcInfo |->, (frSt'::frEnd'::toSt'::toEnd'::scan'::free'::alloc'::nil) ** top).
  apply (sconjImpl Hp); clear Hp; intros h Hp.
  lli.
  apply (sconjImpl Hp); clear h Hp; intros h Hp.
  lli.
  apply (sconjImpl Hp); clear h Hp; intros h Hp.
  lli.
  replace (RFile.get (rfileOf s)) with (rget s) in Hp; [idtac | reflexivity].
  lli; scm.

  lle Hp'; lle Hp''.
  assert (X1 := contContEq Hp' Hp'').
  sconjLift2 Hp'; sconjLift2 Hp''; assert (X2 := contContEq Hp' Hp'').
  sconjLift3 Hp'; sconjLift3 Hp''; assert (X3 := contContEq Hp' Hp'').
  sconjLift4 Hp'; sconjLift4 Hp''; assert (X4 := contContEq Hp' Hp'').
  sconjLift5 Hp'; sconjLift5 Hp''; assert (X5 := contContEq Hp' Hp'').
  sconjLift6 Hp'; sconjLift6 Hp''; assert (X6 := contContEq Hp' Hp'').
  sconjLift7 Hp'; sconjLift7 Hp''; assert (X7 := contContEq Hp' Hp'').
  subst frSt' frEnd' toSt' toEnd' scan' free' alloc'; clear Hp' Hp''.

  set (B := rangeObjs toSt scan) in * ;
    set (G := rangeObjs scan free) in * ;
      set (B' := rangeObjs alloc toEnd) in * ;
	set (F := rangeObjs free alloc) in * .

  assert (Hp' : s |||- 
    (Heap.heap_eqv vh &&
      objHp (W |_| M) W **
      objHp (B |_| G |_| B') (B |_| B') **
      objHp ((W |_| M) |_| B |_| G |_| B') G) **
    mapHp M phi **
    (foralls x in rangeObjs free alloc, x |->__) **
    rget s rgcInfo |->, (frSt::frEnd::toSt::toEnd::scan::free::alloc::nil) ** A).
  lli; scm.
  clear Hp; rename Hp' into Hp.

  (* done with cleanup. start building up the base again. *)

  assert (rootEq : rget vs rroot = rget s rroot).
  destruct (Rok rroot).
  simpl; auto.
  assumption.

  assert (I : set_In (rget s rroot) (B |_| G |_| B')).
  destruct (Rok rroot).
  simpl; auto.
  assert (X := isOddOk (rget s rroot)).
  destruct (isOdd (rget s rroot)).
  elf; apply (not_even_and_odd (rget s rroot)); assumption.
  destruct H0.
  elf; apply (not_even_and_odd (rget s rroot)); assumption.
  assumption.

  assert (toEq : set_eqv (rangeObjs toSt toEnd) (B |_| G |_| F |_| B')).
  clear Hp K; simpl in Al; intuition.
  assert (LE1 : alloc <= toEnd).
  destruct H3; omega.
  subst F B'; rewrite (rangeObjsTransEqv H0 LE1).
  assert (LE2 : free <= toEnd).
  destruct H0; omega.
  subst G; rewrite (rangeObjsTransEqv H1 LE2).
  assert (LE3 : scan <= toEnd).
  destruct H1; omega.
  subst B; rewrite (rangeObjsTransEqv H LE3).
  apply set_eqvRefl.

  set (GCI := rget s rgcInfo
    |->, (frSt :: frEnd :: toSt :: toEnd :: scan :: free :: alloc :: nil)) in * .

  set (k' := if beq_nat k 0 then 4 else 0).
  assert (NEk : k <> k').
  destruct K; subst k' k; [rewrite beq_true | rewrite beq_false]; omega.

  assert (disjBBG : set_disj (B |_| B') G).
  destruct (aconjElim (sconjDistrib2 Hp)) as [_ Hp']; lle Hp'; clear Hp Guar Req;
    sconjLift3 Hp'; sconjLift3 Hp'.
  unfold objHp in Hp'; eapply fasFasDisj.
  apply Hp'.
  uilleim.
  uilleim.
 
  (* show that the heap is okay *)
  assert (SCok : scanFieldStOk phi W M F
     (okFieldAt ((W |_| M) |_| setRange toSt toEnd) (k + rget s rroot) ** 
      (lexists vh, Heap.heap_eqv vh &&
      okFieldAt (if set_mem (rget s rroot) G then ((W |_| M) |_| B |_| G |_| B')
        else (B |_| G |_| B')) (k' + rget s rroot) **
      objHp (B |_| G |_| B') (set_remove (rget s rroot) (B |_| B')) **
      objHp ((W |_| M) |_| B |_| G |_| B') (set_remove (rget s rroot) G)) **
      GCI ** A)
     (rset (rset (rset (rset (rset (rset (rset
       (rset (rset (rset s rfrStart frSt) rfrEnd frEnd) rtoStart toSt) rtoEnd toEnd) 
     rfree free) ralloc alloc) rdTmp1 (rget s ra)) a0 (k + rget s rroot)) rdTmp2
           (k + rget s rroot)) ra BAKERREADRET)).
  unfold scanFieldStOk; unfold fromSpace; unfold toSpace; stSimpl.
  destruct (aconjElim (sconjDistrib2 Hp)) as [_ Hp']; clear Hp;
    rename Hp' into Hp.
  clear VsEq rootEq.
  assert (HH : s |||- (foralls z in rangeObjs frSt frEnd, z |->__) **
    (foralls z in rangeObjs toSt toEnd, z |->__) ** top).
  lle Hp; sconjLift4 Hp; apply sconjAssocR in Hp;
    sconjLift3 Hp; sconjLift3 Hp; sconjLift4 Hp;
    hyr Hp (sconjAssocR (sconjAssocR Hp)); sconjLift2 Hp; scm.
  unfold hpredTrue; rewrite <- WMeq; apply objsMerge; apply sconjComm; scm.
  exact (mapHpIsObjs Hp).
  exact (objHpIsObjs Hp).
  lle Hp.
  unfold hpredTrue; rewrite toEq.
  rewrite (set_eqvUnionSymm B (G |_| F |_| B'));
    rewrite <- (set_eqvUnionAssoc G F B');
      rewrite (set_eqvUnionAssoc (G |_| F) B' B);
	rewrite (set_eqvUnionAssoc G F (B' |_| B));
	  rewrite (set_eqvUnionSymm B' B).
  sconjLift3of3 Hp; apply objsMerge; scm.
  exact (objHpIsObjs Hp).
  apply objsMerge; scm.
  exact (objHpIsObjs Hp).
  assert (frToDisj : set_disj (setRange frSt frEnd) (setRange toSt toEnd)).
  apply disjObjsDisjRange with (heapOf s).
  assumption.
  assumption.
  clear K; simpl in Al; intuition; repeat (eapply aligned8Trans; eauto).
  constructor.
  assumption.
  constructor.
  rewrite WMeq; apply objRangeSetRangeSubset.
  constructor.
  subst; apply set_eqvRefl.
  constructor.
  intros y z Peq NSP.
  destruct Iso as [_ [Iso _]]; assert (Y := Iso _ _ Peq); clear Peq Iso.
  assert (I' : set_In z (rangeObjs toSt toEnd)).
  rewrite toEq.
  set_InElim.
  sui1; assumption.
  sui2; sui1; assumption.
  clear Y I; rename I' into I.
  destruct NSP.
  apply (not_even_and_odd z).
  apply ptrPeven.
  assert (Y : s |||- (z |->__ ** top) ** top).
  clear Hp; sconjLift2 HH; scm.
  hyr Hp (fasElim I Hp); hpunfoldIn Hp; scm.
  lle Y; apply (contElim1L Y).
  assumption.
  apply H; apply objRangeSetRangeSubset; assumption.
  constructor.
  clear K; simpl in Al; intuition;
    destruct H; destruct H1; destruct H0; omega.
  constructor.
  clear K; simpl in Al; intuition; destruct H; destruct H1; 
    destruct H0; destruct H3; omega.
  constructor.
  clear K; simpl in Al; intuition; destruct H; destruct H1; 
    destruct H0; destruct H3; omega.
  lle Hp; sconjLift5 Hp; sconjLift2 Hp; sconjLift5 Hp.
  apply (sconjImplR Hp); clear Hp; intros h Hp;
    apply (sconjImplR Hp); clear h Hp; intros h Hp;
      apply (sconjImplR Hp); clear h Hp; intros h Hp.
  clear rootEven Rok.

  assert (HQ : h ||- okFieldAt ((W |_| M) |_| setRange toSt toEnd) (k + rget s rroot) **
    okFieldAt
    (if set_mem (rget s rroot) G
      then (W |_| M) |_| B |_| G |_| B'
      else B |_| G |_| B') (k' + rget s rroot) **
    objHp (B |_| G |_| B') (set_remove (rget s rroot) (B |_| B')) **
    objHp ((W |_| M) |_| B |_| G |_| B') (set_remove (rget s rroot) G) **
  GCI ** A).
  apply sconjAssocL; apply sconjAssocL; apply sconjAssocL;
    apply sconjAssocR in Hp.
  apply (sconjImplL Hp); clear h Hp; intros h Hp.
  assert (I' : set_In (rget s rroot) ((B |_| B') |_| G)).
  rewrite (set_eqvUnionAssoc B B' G); rewrite (set_eqvUnionSymm B' G);
    assumption.
  clear I; destruct (set_union_elim I') as [I | I]; clear I'.
  (* case: root is black *)
  scm.
  hyr Hp (objHpElim I Hp); scm.
  unfold okObjAt in Hp; destruct K; subst k' k; [idtac | sconjLift2of2 Hp]; scm.
  apply (okFieldAtWeaken Hp); intros z Iz.
  sui2; apply objRangeSetRangeSubset; rewrite toEq.
  rewrite (set_eqvUnionSymm F B').
  rewrite <- (set_eqvUnionAssoc B G (B' |_| F));
    rewrite <- (set_eqvUnionAssoc (B |_| G) B' F); sui1.
  rewrite (set_eqvUnionAssoc B G B'); assumption.
  rewrite beq_true.
  rewrite set_mem_false.
  assumption.
  intro IX; apply (set_disj_elim disjBBG (rget s rroot)); assumption.
  apply (okFieldAtWeaken Hp); intros z Iz.
  sui2; apply objRangeSetRangeSubset; rewrite toEq.
  rewrite (set_eqvUnionSymm F B').
  rewrite <- (set_eqvUnionAssoc B G (B' |_| F));
    rewrite <- (set_eqvUnionAssoc (B |_| G) B' F); sui1.
  rewrite (set_eqvUnionAssoc B G B'); assumption.
  rewrite beq_false.
  rewrite set_mem_false.
  assumption.
  intro IX; apply (set_disj_elim disjBBG (rget s rroot)); assumption.
  omega.
  assert (NI : ~set_In (rget s rroot) G).
  intro I'; apply (set_disj_elim disjBBG (rget s rroot)); assumption.
  rewrite (set_removeEq NI); assumption.
  (* case: root is gray *)
  apply sconjAssocR; apply sconjGLift2'; scm.
  assert (NI : ~set_In (rget s rroot) (B |_| B')).
  intro I'; apply (set_disj_elim disjBBG (rget s rroot)); assumption.
  rewrite (set_removeEq NI); assumption.
  hyr Hp (objHpElim I Hp); scm.
  unfold okObjAt in Hp; destruct K; subst k' k; [idtac | sconjLift2of2 Hp]; scm.
  apply (okFieldAtWeaken Hp); intros z Iz.
  destruct (set_union_elim Iz) as [Iz' | Iz'].
  sui1; assumption.
  sui2; apply objRangeSetRangeSubset; rewrite toEq.
  rewrite (set_eqvUnionSymm F B').
  rewrite <- (set_eqvUnionAssoc B G (B' |_| F));
    rewrite <- (set_eqvUnionAssoc (B |_| G) B' F); sui1.
  rewrite (set_eqvUnionAssoc B G B'); assumption.
  rewrite beq_true.
  rewrite set_mem_true.
  assumption.
  assumption.
  apply (okFieldAtWeaken Hp); intros z Iz.
  destruct (set_union_elim Iz) as [Iz' | Iz'].
  sui1; assumption.
  sui2; apply objRangeSetRangeSubset; rewrite toEq.
  rewrite (set_eqvUnionSymm F B').
  rewrite <- (set_eqvUnionAssoc B G (B' |_| F));
    rewrite <- (set_eqvUnionAssoc (B |_| G) B' F); sui1.
  rewrite (set_eqvUnionAssoc B G B'); assumption.
  rewrite beq_false.
  rewrite set_mem_true.
  assumption.
  assumption.
  omega.

  clear Hp; rename HQ into Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  hyr Hp (sconjAssocR (sconjAssocR Hp)); scm.
  existsl h; apply aconjIntro.
  apply eqvPredEqv'; apply Heap.heap_eqvRefl.
  lle Hp; assumption.

  assert (SCok' : exists vh, scanFieldStOk phi W M F
           (okFieldAt ((W |_| M) |_| setRange toSt toEnd) (k + rget s rroot) **
            (Heap.heap_eqv vh &&
             okFieldAt
               (if set_mem (rget s rroot) G
                then (W |_| M) |_| B |_| G |_| B'
                else B |_| G |_| B') (k' + rget s rroot) **
             objHp (B |_| G |_| B') (set_remove (rget s rroot) (B |_| B')) **
             objHp ((W |_| M) |_| B |_| G |_| B')
               (set_remove (rget s rroot) G)) ** GCI ** A)
           (rset
              (rset
                 (rset
                    (rset
                       (rset
                          (rset
                             (rset
                                (rset
                                   (rset (rset s rfrStart frSt) rfrEnd frEnd)
                                   rtoStart toSt) rtoEnd toEnd) rfree free)
                          ralloc alloc) rdTmp1 (rget s ra)) a0
                    (k + rget s rroot)) rdTmp2 (k + rget s rroot)) ra
              BAKERREADRET)).
  clear K; unfold scanFieldStOk in *; intuition; lle H8;
    exists x; intuition.

  clear SCok; destruct SCok' as [bgh SCok].

  (* apply the guarantee to step past the first function call *)
  assert (X := Guar _ _ _ _ _ SCok); clear Guar.
  set (free' := rget s' rfree) in * .
  destruct X as [M' [W' [F' [G' [phi' [Weq [Feq [Iso' [alFr [Wcopy X]]]]]]]]]].
  generalize dependent X; repeat (rewrite <- Req; [stSimpl | regNotInListSolver]).
  intros [_ [_ [F'eq [_ [Le1 [Le2 [alFr' Hp'X]]]]]]].
  set (GCI1 := rget s rgcInfo |->, (frSt :: frEnd :: toSt :: toEnd :: scan :: nil)).
  assert (Hp' : s' |||- mapHp (M |_| M') (map_union phi phi') **
    objHp (W' |_| M |_| M') W' ** (foralls x in F', x |->__) **
    fieldRespMap (map_unionnn (spaceIdMap toSt toEnd) (map_union phi phi')) s
      (k + rget s rroot) **
    (foralls x in G', objCopiedPhi phi' s x) **
    (Heap.heap_eqv bgh &&
      okFieldAt
      (if set_mem (rget s rroot) G
        then (W |_| M) |_| B |_| G |_| B'
        else B |_| G |_| B') (k' + rget s rroot) **
      objHp (B |_| G |_| B') (set_remove (rget s rroot) (B |_| B')) **
      objHp ((W |_| M) |_| B |_| G |_| B') (set_remove (rget s rroot) G)) **
    (GCI1 ** 20 + rget s rgcInfo |-> free ** 24 + rget s rgcInfo |-> alloc) ** A).
  scm; subst GCI1; lli; assumption.
  clear Hp'X; unfold fieldRespMap in Hp'; lle Hp'.
  sconjLift6 Hp'; sconjLift6 Hp'; hyr Hp' (sconjAssocR Hp'); sconjLift3 Hp';
    hyr Hp' (sconjAssocR Hp').
  destruct (sconjExtract Hp') as [vh' [Hp2 Vhp']]; clear Hp'.
  destruct Guar' as [Guar' [Req' raEq']].
  generalize dependent Guar'; 
    rewrite <- Req; [stSimpl | regNotInListSolver];
      rewrite <- Req; [stSimpl | regNotInListSolver]; intro.
  sconjLift6 Hp2; sconjLift5 Hp2; destruct (Guar' _ _ _ Hp2) as [Hp'' retEq]; clear Guar'.
  generalize dependent raEq'; rewrite <- Req; [stSimpl | regNotInListSolver]; intro.

  set (GCI' := rget s rgcInfo
         |->, (frSt :: frEnd :: toSt :: toEnd :: scan :: free' :: alloc :: nil)).
  assert (Hp''2 : s''
         |||- k + rget s rroot |-> x ** Heap.heap_eqv vh' **
           mapHp (M |_| M') (map_union phi phi') ** (foralls x in F', x |->__) **
           GCI' ** A).
  apply (sconjImplR Hp''); clear Hp; intros h Hp.
  sconjLift2 Hp; apply (sconjImplR Hp); clear h Hp; intros h Hp.
  sconjLift2 Hp; apply (sconjImplR Hp); clear h Hp; intros h Hp.
  sconjLift2 Hp; apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply sconjComm; sconjLift4of4 Hp; apply (sconjImplR Hp); clear h Hp; intros h Hp.
  sconjLift2 Hp; subst GCI1 GCI'; lle Hp; assumption.

  (* now we've stepped through the states *)
  destruct H9 as [x0 [Hp0 Meq]].
  rename Wcopy into Wcopy0.
  assert (Wcopy : forall x, set_In x W' -> s' |||- objCopied x s x ** top).
  apply Wcopy0.
  clear Wcopy0.
  rename Hp2 into Hp'.

  assert (Hp1 : s |||- Heap.heap_eqv bgh ** top).
  unfold scanFieldStOk in SCok; clear K; intuition.
  sconjLift5 H10; apply (sconjImpl H10); clear Hp; intros h Hp.
  destruct (aconjElim Hp); assumption.
  lli.
  clear SCok.

  clear Hp''; rename Hp''2 into Hp''.
  rename vh into h0; rename vh' into h00.
  unfold okFieldAt in Vhp'; lle Vhp'.

  set (root := rget s rroot) in * .
  sconjLift3of3 Vhp'; assert (HpBG := heapEqvImplR Vhp'); 
    destruct (aconjElim (sconjDistrib2 Vhp')) as [VhpX _];
      clear Vhp'; rename VhpX into Vhp'.

  assert (HpX1 : s |||- k' + root |-> x1 ** top).
  hyr Hp1 (sconjInsert Hp1 HpBG); lle Hp1; scm.

  assert (Vhp'' : h00 ||- (foralls x in set_remove root (B |_| B' |_| G) |_| W',
    objCopied x s x) ** k' + root |-> x1 ** (foralls x in G', objCopiedPhi phi' s x)).

  assert (HpBG' : bgh ||- k' + root |-> x1 ** 
    (foralls x in set_remove root (B |_| B' |_| G), objCopied x s x)).
  clear Req VsEq Hp; apply (sconjImplS HpBG); intros h Sub Hp.
  assumption.
  unfold hpredTrue; rewrite <- (set_eqvUnionAssoc B B' G); 
    rewrite (set_removeUnionDistrib root (B |_| B') G).

  assert (SubSub : Heap.subHeap h (heapOf s)).
  eapply Heap.subHeapTrans; eauto.
  eapply sconjEqSubHeap; eauto.
  apply objCopiedMerge; apply (sconjImplS Hp); clear Hp; intros h' Sub' Hp.
  assert (SubX : Heap.subHeap h' (heapOf s)).
  eapply Heap.subHeapTrans; eauto.
  unfold objHp in Hp; apply (fasImplIS Hp); clear h SubSub Sub' Sub Hp; intros y h Iy Sub Hp.
  assert (Sub2 : Heap.subHeap h (heapOf s)).
  eapply Heap.subHeapTrans; eauto.
  hpunfoldIn Hp; hpunfold; lle Hp; apply (sconjImplS Hp); clear Hp; intros hx Sub' Hp.
  lli; [ | apply Hp].
  2:lli; [ | apply Hp].
  eapply subHeapImpl; eauto; eapply Heap.subHeapTrans; eauto.
  eapply subHeapImpl; eauto; eapply Heap.subHeapTrans; eauto.
  unfold objHp in Hp; apply (fasImplIS Hp); clear Sub Hp; intros y h'' Iy Sub'' Hp.
  assert (Sub2 : Heap.subHeap h'' (heapOf s)).
  eapply Heap.subHeapTrans; eauto.
  eapply Heap.subHeapTrans; eauto.
  hpunfoldIn Hp; hpunfold; lle Hp; apply (sconjImplS Hp); clear Hp; intros hx SubX Hp.
  lli; [idtac | apply Hp].
  2:lli; [idtac | apply Hp].
  eapply subHeapImpl; eauto; eapply Heap.subHeapTrans; eauto.
  eapply subHeapImpl; eauto; eapply Heap.subHeapTrans; eauto.

  hyr Vhp' (sconjInsert Vhp' HpBG'); clear HpBG' HpX1 HpBG; lle Vhp'.
  sconjLift3 Vhp'; sconjLift3 Vhp'; hyr Vhp' (sconjAssocR Vhp'); apply (sconjImplS Vhp').
  clear Hp Vhp'; intros h Sub Hp; apply objCopiedMerge.
  apply (sconjImplS Hp); clear Hp; intros h' Sub' Hp.
  assumption.
  unfold objHp in Hp; apply (fasImplIS Hp); clear Hp; intros y h'' Iy Sub'' Hp.
  assert (SubX : Heap.subHeap h'' (heapOf s')).
  eapply Heap.subHeapTrans; eauto.
  eapply Heap.subHeapTrans; eauto.
  eapply Heap.subHeapTrans; eauto.
  sconjLift3 Hp'; eapply sconjEqSubHeap; eauto.
  clear h Sub h' Sub' Sub''.
  hpunfoldIn Hp; lle Hp.
  assert (WC := Wcopy _ Iy); hpunfoldIn WC; lle WC.
  assert (HX := subHeapImpl Hp SubX); lle HX.
  assert (X1 := contContEq WC HX); sconjLift2 WC; sconjLift2 HX;
    assert (X2 := contContEq WC HX); subst x4 x5; clear WC HX.
  hpunfold; lli; eauto.
  auto.

  hyr Vhp' (sconjInsert Vhp' HpBG); clear HpBG bgh Hp1; lle Vhp'.
  clear Wcopy Hp'.
  set (h1 := Heap.uset h00 (k + root) x).
  set (mphi := (fun x => match map_union phi phi' x with Some y => y | None => x end)).
  set (vh' := Heap.HeapMap.map_mapRange mphi h1).
  exists (vh', rfileOf s'').

  assert (x0Ok : okFieldVal (if set_mem root G
    then (W |_| M) |_| B |_| G |_| B' else B |_| G |_| B') x0).
  destruct (aconjElim (sconjDistrib2 Hp)) as [_ HpA]; clear Hp; rename HpA into Hp; lle Hp.
  assert (I' : set_In root ((B |_| B') |_| G)).
  rewrite (set_eqvUnionAssoc B B' G); rewrite (set_eqvUnionSymm B' G); assumption.
  destruct (set_union_elim I') as [Ir | Ir].
  rewrite set_mem_false.
  assert (HpX : s |||- (okObjAt (B |_| G |_| B') root ** top) ** top).
  clear HpX1 Hp0; sconjLift2 Hp; scm; hyr Hp (objHpElim Ir Hp); scm.
  hpunfoldIn HpX; lle HpX.
  destruct K; subst h1 k' k; [idtac | sconjLift2 HpX];
    assert (E := contContEq HpX Hp0); subst; assumption.
  intro Ir'; apply (set_disj_elim disjBBG root); assumption.
  rewrite (set_mem_true Ir).
  assert (HpX : s |||- (okObjAt ((W |_| M) |_| B |_| G |_| B') root ** top) ** top).
  clear HpX1 Hp0; sconjLift3 Hp; scm; hyr Hp (objHpElim Ir Hp); scm.
  hpunfoldIn HpX; lle HpX.
  destruct K; subst h1 k' k; [idtac | sconjLift2 HpX];
    assert (E := contContEq HpX Hp0); subst; assumption.

  Notation objsHp' := (fun S => foralls x in S, x |->__).

  assert (HD : s |||- (objsHp' W ** objsHp' (B |_| B') ** objsHp' G) ** 
    objsHp' M ** objsHp' F ** top).
  clear Hp0 HpX1; scm.
  destruct (aconjElim Hp) as [_ Hp']; clear Hp; scm; exact (objHpIsObjs Hp).
  exact (mapHpIsObjs Hp).

  assert (frToDisj : set_disj (setRange frSt frEnd) (setRange toSt toEnd)).
  eapply disjObjsDisjRange.
  4:eauto.
  instantiate (1 := heapOf s).
  lle HD; sconjLift5 HD; sconjLift4 HD; hyr HD (sconjAssocR HD);
    sconjLift3 HD; hyr HD (sconjAssocR HD);
    sconjLift3 HD; sconjLift3 HD; hyr HD (sconjAssocR HD).
  scm.
  unfold hpredTrue; rewrite <- WMeq; apply objsMerge; assumption.
  unfold hpredTrue; rewrite toEq.
  rewrite (set_eqvUnionSymm F B'); rewrite <- (set_eqvUnionAssoc G B' F);
    rewrite (set_eqvUnionSymm G B'); rewrite <- (set_eqvUnionAssoc B (B' |_| G) F);
      rewrite <- (set_eqvUnionAssoc B B' G); rewrite (set_eqvUnionAssoc (B |_| B') G F).
  apply objsMerge; scm; apply objsMerge; assumption.
  assumption.
  simpl in Al; decompose [and] Al; repeat (eapply aligned8Trans; eauto).

  assert (x0Ok' : okFieldVal ((B |_| G |_| B') |_|
    (if set_mem root G then W |_| M else empty_set)) x0).
  set (X := rangeObjs scan free) in * .
  subst G; apply (okFieldValWeaken x0Ok); destruct (set_mem root X).
  intros z Iz; set_InElim; 
    [sui2; sui1 | sui2; sui2 | sui1; sui1 | sui1; sui2; sui1 | sui1; sui2; sui2]; assumption.
  intros z Iz; set_InElim; 
    [sui1; sui1 | sui1; sui2; sui1 | sui1; sui2; sui2]; assumption.

  assert (x1Ok' : okFieldVal ((B |_| G |_| B') |_|
    (if set_mem root G then W |_| M else empty_set)) x1).
  apply (okFieldValWeaken H9); intros z Iz.
  replace (rangeObjs scan free) with G by reflexivity.
    replace (rangeObjs toSt scan) with B by reflexivity;
      replace (rangeObjs alloc toEnd) with B' by reflexivity.
  set (TO := B |_| G |_| B') in * ; destruct (set_mem root G).
  rewrite (set_eqvUnionSymm TO (W |_| M)); assumption.
  sui1; assumption.

  assert (toG'disj : set_disj (rangeObjs toSt scan |_| rangeObjs alloc toEnd |_| 
    rangeObjs scan free) G').
  apply set_disjSymm; apply set_unionDisjElimL with F'; rewrite <- Feq.
  rewrite <- (set_eqvUnionAssoc (rangeObjs toSt scan) (rangeObjs alloc toEnd)
    (rangeObjs scan free)).
  lle HD; sconjLift3 HD; sconjLift3 HD; hyr HD (sconjAssocR HD); sconjLift4 HD.
  assert (HD' : s |||- fasd F ** fasd ((B |_| B') |_| G) ** top).
  scm.
  llMatch; lleim.
  apply fasdMerge; scm; llMatch; lleim.
  apply (fasFasDisj HD'); eauto.

  constructor.
  (* apply gcGuarOk *)
  subst vh' mphi h1 root; eapply gcGuarOk; try assumption.
  apply Req.
  assumption.
  apply VsEq.
  eauto.
  instantiate (1 := scan); apply Iso.
  assumption.
  scm.
  assumption.
  assumption.
  5:eauto.
  assumption.
  assumption.
  eauto.
  instantiate (1 := F'); assumption.
  assumption.
  assumption.
  eauto.
  assumption.
  scm.
  assumption.
  2:eauto.
  assumption.
  assumption.
  assumption.

  (* apply finalReadReprOk *)
  subst vh' mphi h1 root free'; stSimpl; eapply finalReadReprOk; try assumption.
  eauto.
  eauto.
  eauto.
  apply Al'.
  eauto.
  eauto.
  assumption.
  assumption.
  eauto.
  apply VsEq.
  eauto.
  assumption.
  eauto.
  eauto.
  assumption.
  assumption.
  assumption.
  assumption.
  assumption.
  eauto.
  assumption.
  assumption.
  eauto.
  assumption.
  assumption.
  assumption.
  assumption.
  assumption.
  assumption.
  assumption.
  assumption.

  (* finally, show register preservation *)
  clear K rootEven Hp' Guar scan phi.
  destruct Guar' as [_ [Req' raEq']].
  unfold BMGI.presReg; intros r [CR NeV0].
  destruct (regCompare r ra) as [Eq | NEra].
  subst r; rewrite raEq'; 
    rewrite <- Req; [stSimpl | regNotInListSolver];
      reflexivity.
  rewrite <- Req'; [clear raEq' Req' s'' | regNotInListSolver].
  rewrite <- Req; clear Req.
  stSimpl; unfold calleeSaved in CR; simpl in CR; decompose [or] CR;
    try (subst r; stSimpl; reflexivity); elf; assumption.
  unfold calleeSaved in CR; simpl in CR; decompose [or] CR;
    try (subst r; regNotInListSolver); elf; assumption.
  
Qed.


Lemma bakerReadRetOk : forall P,
  iseqOkS P bakerReadRetPre bakerReadRetGuar bakerReadRet.
Proof.
  intros P.
  unfold bakerReadRet.
  iseqWeakSolver fail.
  clear P.
  unfold bakerReadRetPre.
  intros s Hp; lle Hp; rename x0 into y.
  rewrite (lwStep Hp); try reflexivity.
  sconjLift2 Hp.
  rewrite (swStep Hp); try reflexivity.
  rewrite addiuStep0.
  constructor.
  trivial.
  intros s' S'eq; subst s'; unfold bakerReadRetGuar; stSimpl.
  constructor.
  intros A x' y' Hp'; sconjLift2 Hp'; assert (E1 := contContEq Hp Hp'); 
    sconjLift2 Hp; sconjLift2 Hp';
    assert (E2 := contContEq Hp Hp'); subst x' y'; clear Hp; rename Hp' into Hp.
  constructor.
  sconjLift2 Hp; hyr Hp (contUpdateSt Hp (rget s rfree)); sconjLift2 Hp; assumption.
  reflexivity.
  intuition.
  clear Hp; intros r E; simpl in E.
  intuition; stSimpl; reflexivity.
Qed.
