Require Export CheneyLemma.
Require Export ScanFieldB.
Require Export HeapDom.


Set Implicit Arguments.


Ltac hyr := hypReplace.


Lemma fasdMerge : forall h s s',
  h ||- fasd s ** fasd s' ->
  h ||- fasd (set_union s s').
Proof.
  intros h s s' Hp.
  apply (fasMerge _ Hp).
  apply (fasFasDisj' Hp); auto.
Qed.


Lemma fasdDisj : forall h X Y,
  h ||- fasd X ** fasd Y ** top ->
  set_disj X Y.
Proof.
  intros h X Y Hp.
  apply (fasFasDisj Hp); eauto.
Qed.



Lemma subsetInterEq : forall A B,
  set_subset A B ->
  set_eqv (set_inter A B) A.
Proof.
  intros A B S.
  constructor; intros I; assert (I' := S a); clear S.
  destruct (set_inter_elim I); assumption.
  apply set_inter_intro.
  assumption.
  auto.
Qed.


Lemma set_unionInterDistrib2 : forall A B C,
  set_eqv (set_inter A (set_union B C)) 
  (set_union (set_inter A B) (set_inter A C)).
Proof.
  intros A B C.
  constructor; intros I.
  destruct (set_inter_elim I) as [I1 I2]; clear I.
  destruct (set_union_elim I2); clear I2.
  apply set_union_intro1; apply set_inter_intro; assumption.
  apply set_union_intro2; apply set_inter_intro; assumption.

  destruct (set_union_elim I) as [I' | I']; clear I;
    destruct (set_inter_elim I'); clear I'; apply set_inter_intro.
  assumption.
  apply set_union_intro1; assumption.
  assumption.
  apply set_union_intro2; assumption.

Qed.

Lemma set_subsetUnionIntro1 : forall A B C,
  set_eqv (set_union A B) C ->
  set_subset A C.
Proof.
  intros A B C E x I.
  rewrite <- (E x).
  apply set_union_intro1; assumption.
Qed.
Implicit Arguments set_subsetUnionIntro1 [A B C].

Lemma set_subsetUnionIntro2 : forall A B C,
  set_eqv (set_union A B) C ->
  set_subset B C.
Proof.
  intros A B C E x I.
  rewrite <- (E x).
  apply set_union_intro2; assumption.
Qed.
Implicit Arguments set_subsetUnionIntro2 [A B C].


Add Morphism fieldRespToMap
  with signature mapEqvnn ==> state_eqv ==> eq ==> eq ==> iff
  as fieldRespToMapMorphism.
Proof.
  unfold fieldRespToMap.
  intros m m' E s s' SE n h; constructor; intros Hp;
    unfold spaceIdMap.
  repeat (rewrite <- (rgetEqv SE)).
  setoid_rewrite <- SE.
  setoid_rewrite <- E; assumption.
  repeat (rewrite (rgetEqv SE)).
  setoid_rewrite SE.
  setoid_rewrite E; assumption.
Qed.


Lemma set_addUnionDistrib : forall x A B,
  set_eqv (set_union (set_add x A) B) (set_add x (set_union A B)).
Proof.
  intros x A B.
  simpl.
  apply set_eqvRefl.
Qed.


Lemma set_removeInterDistrib : forall x A B,
  set_eqv (set_inter (set_remove x A) B) (set_remove x (set_inter A B)).
Proof.
  intros x A B.
  constructor; intros I.

  destruct (set_inter_elim I) as [I1 I2]; clear I.
  destruct (set_remove_elim _ _ _ I1) as [I1' Ne]; clear I1.
  apply set_remove_intro.
  apply set_inter_intro; assumption.
  assumption.

  destruct (set_remove_elim _ _ _ I) as [Ne I']; clear I.
  destruct (set_inter_elim I') as [I1 I2]; clear I'.
  apply set_inter_intro.
  apply set_remove_intro; assumption.
  assumption.
Qed.


Lemma set_removeCancel : forall x A,
  set_eqv (set_add x A) (set_add x (set_remove x A)).
Proof.
  intros x A.
  constructor; intros I.

  compare x a; intro E.
  apply set_add_intro1; auto.

  destruct (set_add_elim I) as [I' | I']; clear I.
  elimtype False; auto.
  apply set_add_intro2.
  apply set_remove_intro; auto.
  
  destruct (set_add_elim I) as [I' | I']; clear I.
  apply set_add_intro1; assumption.
  apply set_add_intro2.
  destruct (set_remove_elim _ _ _ I'); assumption.
Qed.


Lemma selfInterEqv : forall A B,
  set_eqv A (set_inter (set_union A B) A).
Proof.
  intros A B.
  constructor; intro I.

  apply set_inter_intro.
  apply set_union_intro1; assumption.
  assumption.

  destruct (set_inter_elim I) as [_ I']; assumption.

Qed.


Lemma set_unionInterDistrib1 : forall A B C,
  set_eqv (set_inter (set_union A B) C) 
  (set_union (set_inter A C) (set_inter B C)).
Proof.
  intros A B C.
  constructor; intro I.

  destruct (set_inter_elim I) as [I1 I2]; clear I.
  destruct (set_union_elim I1) as [I1' | I1'']; clear I1.
  apply set_union_intro1; apply set_inter_intro; assumption.
  apply set_union_intro2; apply set_inter_intro; assumption.

  destruct (set_union_elim I) as [I' | I']; clear I;
    destruct (set_inter_elim I') as [Ia Ib]; clear I';
      apply set_inter_intro.
  apply set_union_intro1; assumption.
  assumption.
  apply set_union_intro2; assumption.
  assumption.

Qed.


Lemma mapUnionAssoc : forall m m' m'',
  mapEqv (map_unionnn (map_unionnn m m') m'')
  (map_unionnn m (map_unionnn m' m'')).
Proof.
  intros m m' m''.
  intro a.
  unfold map_unionnn; unfold map_union.
  destruct (m a).
  reflexivity.
  reflexivity.
Qed.


Lemma mapSubTrans : forall m1 m2, mapSubnn m1 m2 ->
  forall m3, mapSubnn m2 m3 ->
    mapSubnn m1 m3.
Proof.
  intros m1 m2 S1 m3 S2 x y E.
  rewrite <- (S2 x).
  reflexivity.
  rewrite <- (S1 x).
  reflexivity.
  assumption.
Qed.


Lemma mapSub_isoDisj : forall p A A', iso_set p A A' ->
  forall p' B B', iso_set p' B B' ->
    set_disj A B ->
    mapSub p' (map_union p p').
Proof.
  intros p A A' I p' B B' I' D x y E.
  unfold map_union.
  replace (p x) with (None (A :=nat)).
  assumption.
  symmetry.
  destruct I' as [I2 _].
  destruct (I2 x) as [_ I3].
  rewrite E in I3; clear E I2.
  destruct I as [I2 _].
  destruct (I2 x) as [_ II].
  destruct (p x).
  elimtype False; eapply (set_disj_elim D).
  apply II; discriminate.
  apply I3; discriminate.
  reflexivity.
Qed.


Lemma selfInterEqv2 : forall A B,
  set_subset A B ->
  set_eqv (set_inter B A) A.
Proof.
  intros A B S.
  constructor; intros I.
  destruct (set_inter_elim I); assumption.
  apply set_inter_intro.
  apply S; assumption.
  assumption.
Qed.


Lemma set_disjInterIntro : forall A B,
  set_disj A B -> forall C,
    set_disj (set_inter A C) (set_inter B C).
Proof.
  intros A B Di C.
  apply set_disj_intro.
  intros x I1 I2.
  apply (set_disj_elim Di x).
  destruct (set_inter_elim I1); assumption.
  destruct (set_inter_elim I2); assumption.
Qed.


Section LL.

Import Heap.HeapMap.

Lemma topIR : forall h A,
  h ||- A ->
    h ||- A ** top.
Proof.
  intros h A Hp.
  eapply sconjIntro.
  apply Hp.
  instantiate (1 := Heap.empty_heap).
  linLogIntro.
  unfold Heap.heap_join.
  unfold Heap.empty_heap.
  apply mapJoinComm.
  apply joinUnitIntro.
Qed.


End LL.


Lemma fasMerge' : forall h s s' A,
  h ||- fas A s ** fas A s' ->
    (forall x h, h ||- A x -> h ||- x |->_ ** top) ->
    h ||- fas A (set_union s s').
Proof.
  intros h s s' A Hp F.
  unfold hpredTrue; setoid_rewrite (set_canonEqv s).
  assert (C := set_canonCanonP s).

  assert (H : h ||- fas A (set_canon s) ** fas A s').
  sconjMatch.
  unfold hpredTrue; setoid_rewrite <- (set_canonEqv s).
  assumption.
  clear Hp; rename H into Hp.

  generalize dependent s'.
  generalize dependent h.
  generalize dependent (set_canon s); clear s.
  induction 1; simpl; intros h s' Hp.

  apply unitEL; simpl.
  sconjMatch.

  assert (NI : ~set_In a s).
  intro X.
  assert (Q := H _ X).
  unfold POnat.POnat.LT in Q; omega.

  apply fasIntro.
  assert (I : set_In a (set_add a s)).
  left; reflexivity.

  assert (h ||- A a ** fas A s ** fas A s').
  apply sconjAssocL.
  sconjMatch.
  hypReplace Hp (fasElim I Hp).
  sconjMatch.

  simpl in Hp.
  destruct (POnat.POnat.eqDec a a).
  clear e.
  rewrite <- (set_removeEq NI).
  assumption.
  elimtype False; apply n; reflexivity.

  clear Hp I; rename H0 into Hp.
  sconjMatch.
  unfold hpredTrue; apply IHC.
  assumption.

  intro I.
  apply NI.
  destruct (set_union_elim I).
  assumption.
  clear I NI; elimtype False.

  assert (h ||- ((a |->_ ** top) ** top) ** ((a |->_ ** top) ** top)).
  sconjMatch.
  assert (I : set_In a (set_add a s)).
  apply set_add_intro1; reflexivity.
  hypReplace Hp (fasElim I Hp).
  sconjMatch.
  hypReplace Hp (fasElim H0 Hp).
  sconjMatch.
  linLogElim H1.
Qed.


Lemma okObjAtWeaken : forall h s x,
  h ||- okObjAt s x ->
    forall s', set_subset s s' ->
      h ||- okObjAt s' x.
Proof.
  unfold okObjAt; unfold okFieldAt.
  intros h s x Hp s' S.
  linLogElim Hp; linLogIntro; eauto.
  destruct H3; [left | right].
  assumption.
  apply S; assumption.
  destruct H4; [left | right].
  assumption.
  apply S; assumption.
Qed.



Definition objRespToMap phi s x : hpred :=
  (fieldRespToMap phi s x) ** (fieldRespToMap phi s (4+x)).


Add Morphism objRespToMap
  with signature mapEqvnn ==> state_eqv ==> eq ==> eq ==> iff 
  as objRespToMapMorphism.
Proof.
  unfold objRespToMap.
  intros m m' E s s' SE n h ; constructor; intros [ha [hb [J [HA HB]]]];
    exists ha; exists hb; intuition; unfold hpredTrue.
  setoid_rewrite <- E; setoid_rewrite <- SE; assumption.
  setoid_rewrite <- E; setoid_rewrite <- SE; assumption.
  setoid_rewrite E; setoid_rewrite SE; assumption.
  setoid_rewrite E; setoid_rewrite SE; assumption.
Qed.


Definition objIsToFwded phi phi' s : nat -> hpred :=
  objIsFwded (map_union (toSpaceIdMap s) phi) phi' s.

Fixpoint aligned8s (l:list nat) : Prop :=
  match l with
    | x::l' =>
      match l' with
	| y::nil => aligned8 x y
	| y::_ => aligned8 x y /\ aligned8s l'
	| nil => True
      end
    | _ => True
  end.

Definition fieldRespMapH phi h x : hpred :=
  lexists v0', 
    !(exists v0, h ||- x |-> v0 ** top /\ mapX phi v0 = Some v0') **
    x |-> v0'.

Definition objRespMapH phi h x : hpred :=
  (fieldRespMapH phi h x) ** (fieldRespMapH phi h (4+x)).

Definition fieldCopiedH x h y :=
  lexists v, !(h ||- x |-> v ** top) ** y |-> v.

Definition objCopiedH x h y :=
  fieldCopiedH x h y ** fieldCopiedH (4+x) h (4+y).


Lemma set_inter_split : forall x y,
  set_eqv x (set_inter x y |_| set_diff x y).
Proof.
  intros x y z; constructor; intro I.
  destruct (set_In_dec z y) as [I' | I'].
  apply set_union_intro1; apply set_inter_intro; assumption.
  apply set_union_intro2; apply set_diff_intro; assumption.
  destruct (set_union_elim I) as [I' | I']; clear I.
  destruct (set_inter_elim I'); assumption.
  destruct (set_diff_elim I'); assumption.
Qed.


Lemma interDiffDisj : forall x y,
  set_disj (set_inter x y) (set_diff x y).
Proof.
  intros x y; apply set_disj_intro; intros z I I'.
  destruct (set_inter_elim I); clear I.
  destruct (set_diff_elim I'); clear I'.
  auto.
Qed.


(*
Lemma objHp_interOk : forall h x y,
    h ||- objHp x x ** top ->
      h ||- objHp y y ** top ->
	h ||- objHp (set_inter x y) (set_inter x y) ** top.
Proof.
  intros h x y HpX HpY.
  assert (E : exists z, z = set_inter x y).
  econstructor; reflexivity.
  destruct E as [z ZE].

  cut (h ||- objHp (set_inter x y) z ** top).
  intro H; rewrite ZE in H; assumption.

  assert (set_subset z (set_inter x y)).
  rewrite ZE; apply set_subsetRefl.

  assert (Hp : h ||- objHp x (set_inter x y) ** top).
  clear HpY ZE H z.
  assert (Hp : h ||- objHp x (set_inter x y |_| set_diff x y) ** top).
  sconjMatch.
  unfold hpredTrue; rewrite <- (set_inter_split x y); assumption.
  cut (h ||- (objHp x (set_inter x y) ** top) ** top).
  intro Z; linLogElim Z; assumption.
  clear HpX; sconjMatch.

  hypReplace Hp (objHpSplit Hp (interDiffDisj x y)).
  sconjMatch.



Lemma interDiffDisj : forall x y,
  set_disj (set_inter x y) (set_diff x y).


Lemma set_inter_split : forall x y,
  set_eqv x (set_inter x y |_| set_diff x y).

  unfold hpredTrue; setoid_replace (set_inter x y |_| set_diff x y) with x.
  2:assumption.
  clear h Hp.
  apply objHpMerge.
  



  assert (S : set_subset z x).
  eapply set_subsetTrans; eauto.
  intros q I; destruct (set_inter_elim I); assumption.
  rewrite <- ZE; clear HpY ZE H.
  



  clear ZE.
  assert (exists h', h = h' /\ Heap.subHeap h' h).
  exists h; constructor.
  reflexivity.
  unfold Heap.subHeap; auto.
  destruct H0 as [h' [E Sub]].
  rewrite E; clear E; generalize dependent h'.
  induction z; intros h' S.

  assert (Hp : h' ||- unit ** top).
  linLogIntro.
  sconjMatch.

  assert (Sz : set_subset z (set_inter x y)).
  intros t It.
  apply (H t).
  right; assumption.
  hypReplace IHz (IHz Sz); clear Sz.
  
*)


(*
Lemma isoStateTrans : forall r h r' h',
  isoState' r h r' h' ->
  forall r'' h'',
    isoState' r' h' r'' h'' ->
    isoState' r h r'' h''.
Proof.
  intros r h r' h' [objs [objs' [phi [Hp [Hp' [iso [rOk Hiso]]]]]]].
  intros r'' h'' [objs'2 [objs'' [phi' [Hp'2 [Hp'' [iso' [rOk' Hiso']]]]]]].
  unfold isoState'.
  
*)


Lemma objRespToSubmap : forall h p s x,
  h ||- objRespToMap p s x -> 
    forall p', mapSub p p' ->
      h ||- objRespToMap p' s x.
Proof.
  unfold objRespToMap; unfold fieldRespToMap; unfold fieldRespMap.
  intros h p s x Hp p' S.
  linLogElim Hp.
  destruct H3 as [v0 [Hp0 Meq0]].
  destruct H4 as [v1 [Hp1 Meq1]].
  linLogIntro.
  exists v0; constructor.
  assumption.
  2:eauto.
  generalize dependent Meq0.
  unfold mapX.
  unfold map_unionnn; unfold map_union.
  destruct (isOdd v0).
  auto.
  destruct (spaceIdMap (rget s rtoStart) (rget s rtoEnd) v0).
  auto.
  apply S.
  
  exists v1; constructor.
  assumption.
  generalize dependent Meq1.
  unfold mapX.
  unfold map_unionnn; unfold map_union.
  destruct (isOdd v1).
  auto.
  destruct (spaceIdMap (rget s rtoStart) (rget s rtoEnd) v1).
  auto.
  apply S.
Qed.


Lemma objCopiedRespToMap : forall h p s' x,
  h ||- objRespToMap p s' x ->
    forall s,
      rget s rtoEnd = rget s' rtoEnd ->
      rget s rtoStart = rget s' rtoStart ->
      s' |||- objCopied x s x ** top ->
	h ||- objRespToMap p s x.
Proof.
  unfold objRespToMap; unfold fieldRespToMap; unfold fieldRespMap.
  unfold objCopied; unfold fieldCopied.
  intros h p s' x Hp s E1 E2 Hp'.
  replace (toSpaceIdMap s) with (toSpaceIdMap s').
  linLogElim Hp.
  linLogElim Hp'.
  destruct H3 as [v0 [Hp0 Meq0]].
  destruct H4 as [v1 [Hp1 Meq1]].
  assert (Q := contContEq Hp0 Hp').
  sconjLift2 Hp'.
  assert (Q' := contContEq Hp1 Hp').
  subst x2; subst x3.
  linLogIntro.
  exists v0; constructor.
  assumption.
  2:eauto.
  assumption.
  exists v1; constructor.
  assumption.
  assumption.
  unfold spaceIdMap; rewrite E1; rewrite E2; reflexivity.

Qed.


Lemma objCopiedFwded : forall h phi' phi s' x,
  h ||- objIsToFwded phi' phi s' x ->
    forall A B, iso_set phi A B ->
      forall s, 
	(forall x, set_In x A -> s' |||- objCopied x s x ** top) ->
	rget s rtoEnd = rget s' rtoEnd ->
	rget s rtoStart = rget s' rtoStart ->
	h ||- objIsToFwded phi' phi s x.
Proof.
  unfold objIsToFwded; unfold objIsFwded; unfold fieldIsFwded.
  intros h phi' phi s' x Hp A B iso s F E1 E2.
  replace (toSpaceIdMap s) with (toSpaceIdMap s').
  clear E1 E2.
  linLogElim Hp.

  assert (Hp' : s' |||- objCopied x0 s x0 ** top).
  apply F.
  destruct iso as [I _].
  destruct (I x0) as [_ Q].
  apply Q; rewrite H1; discriminate.

  clear F iso.
  destruct H6 as [v0 [Hp0 Meq0]].
  destruct H7 as [v1 [Hp1 Meq1]].
  unfold objCopied in Hp'; unfold fieldCopied in Hp'; linLogElim Hp'.
  assert (E1 := contContEq Hp' Hp0).
  sconjLift2 Hp'.
  assert (E2 := contContEq Hp' Hp1).
  clear Hp0 Hp1 Hp'.
  subst x3; subst x4.
  linLogIntro; eauto.
  
  unfold spaceIdMap; rewrite E1; rewrite E2; reflexivity.

Qed.


Lemma objIsToFwdedShift : forall h p p' p'' s x,
  h ||- objIsToFwded (map_union p p') p'' s x ->
    forall A A', iso_set p' A A' ->
      forall B B', iso_set p'' B B' ->
	set_disj A B ->
	h ||- objIsToFwded p (map_union p' p'') s x.
Proof.
  unfold objIsToFwded; unfold objIsFwded; unfold fieldIsFwded.
  intros h p p' p'' s x Hp A A' I1 B B' I2 D.
  linLogElim Hp.
  destruct H6 as [v0 [Hp0 Meq0]].
  destruct H7 as [v1 [Hp1 Meq1]].
  existsl x0; linLogIntro.
  3:eauto.

  assert (S : mapSub p'' (map_union p' p'')).
  eapply mapSub_isoDisj; eauto.
  
  apply S; assumption.

  exists v0; constructor.
  assumption.
  generalize dependent Meq0.
  unfold mapX.
  destruct (isOdd v0).
  auto.
  intro.
  replace (map_union (map_union (toSpaceIdMap s) p)) with
    (map_unionnn (map_unionnn (toSpaceIdMap s) p)); [idtac | reflexivity].
  replace (map_union p') with (map_unionnn p'); [idtac | reflexivity].
  rewrite (mapUnionAssoc (toSpaceIdMap s) p (map_unionnn p' p'')).
  unfold map_unionnn at 1; unfold map_union at 1.
  rewrite <- (mapUnionAssoc p p' p'').
  rewrite (mapUnionAssoc (toSpaceIdMap s) (map_union p p') p'') in Meq0.
  assumption.

  exists v1; constructor.
  assumption.
  generalize dependent Meq1.
  unfold mapX.
  destruct (isOdd v1).
  auto.
  intro.
  replace (map_union (map_union (toSpaceIdMap s) p)) with
    (map_unionnn (map_unionnn (toSpaceIdMap s) p)); [idtac | reflexivity].
  replace (map_union p') with (map_unionnn p'); [idtac | reflexivity].
  rewrite (mapUnionAssoc (toSpaceIdMap s) p (map_unionnn p' p'')).
  unfold map_unionnn at 1; unfold map_union at 1.
  rewrite <- (mapUnionAssoc p p' p'').
  rewrite (mapUnionAssoc (toSpaceIdMap s) (map_union p p') p'') in Meq1.
  assumption.

Qed.


Lemma MWmatch1 : forall h M W phi X Y Z X' Y',
  h ||- mapHp M phi ** objHp (W |_| M) W ** objHp X (Y |_| Z) ** objHp X' Y' ** top ->
    iso_set phi M (Y |_| Y') ->
    forall M' W' phi' X2 X2',
    h ||- mapHp M' phi' ** objHp (W' |_| M') W' ** objHp X2 (Y |_| Z) ** objHp X2' Y' ** top ->
      iso_set phi' M' (Y |_| Y') ->
      set_eqv (W |_| M) (W' |_| M') ->
      set_subset W W'.
Proof.
  intros h M W phi X Y Z X' Y' Hp iso M' W' phi' X2 X2' Hp' iso' WMeqv.
  intros x IW.
  assert (IWM : set_In x (W |_| M)).
  apply set_union_intro1; assumption.
  destruct (WMeqv x) as [WMeqv1 _]; clear WMeqv.
  hypReplace WMeqv1 (WMeqv1 IWM).
  destruct (set_union_elim WMeqv1).
  assumption.
  elimtype False.
  assert (Hp1 : h ||- top ** (okObjAt (W |_| M) x ** top) ** top).
  clear Hp'; sconjMatch.
  hypReplace Hp (objHpElim IW Hp); sconjMatch.
  unfold okObjAt in Hp1; unfold okFieldAt in Hp1; linLogElim Hp1.
  rename x0 into v0; rename x1 into v1; rename H4 into Ok0; rename H7 into Ok1.

  unfold mapHp in Hp'; linLogElim Hp'.
  rename H into IM; rename H2 into phiM.
  assert (Hp2 : h ||- (((lexists a, !(phi' x = Some a) ** x |-> a ** 4 + x |->_)) ** top) ** top).
  clear Hp Hp1; sconjMatch.
  hypReplace Hp (fasElim IM Hp); sconjMatch.

  linLogElim Hp2.
  assert (ZZZ := contContEq Hp1 Hp2); subst x0.
  clear Hp1 Hp2 Ok1 v1 x1.
  clear IW IWM WMeqv1 IM phiM.
  rename H3 into phiEq.
  destruct iso' as [_ [iso2 _]].
  assert (IY := iso2 _ _ phiEq); clear iso2 phiEq.
  destruct Ok0.

  apply (not_even_and_odd v0).
  2:assumption.
  assert (Hp'' : h ||- (((!ptrP v0 ** top) ** top) ** top) ** top).
  clear Hp.
  destruct (set_union_elim IY) as [I | I]; clear IY.

  sconjLift3 Hp'; sconjMatch.
  assert (I' : set_In v0 (Y |_| Z)).
  apply set_union_intro1; assumption.
  hypReplace Hp (objHpElim I' Hp).
  unfold okObjAt in Hp; sconjMatch.
  unfold okFieldAt in Hp; linLogElim Hp.
  linLogIntro.
  apply (contElim1 Hp).

  sconjLift4 Hp'; sconjMatch.
  hypReplace Hp (objHpElim I Hp).
  unfold okObjAt in Hp; sconjMatch.
  unfold okFieldAt in Hp; linLogElim Hp.
  linLogIntro.
  apply (contElim1 Hp).

  linLogElim Hp''; apply ptrPeven; assumption.
  clear Hp'.
  hypReplace Hp (sconjAssocR Hp).
  sconjLift3 Hp; sconjLift3 Hp; hypReplace Hp (sconjAssocR Hp).
  assert (h ||- ((v0 |->_ ** top) ** top) ** ((v0 |->_ ** top) ** top) ** top).
  apply (sconjImpl Hp); clear h Hp; intros h Hp.
  destruct (set_union_elim IY).
  sconjMatch.
  assert (II : set_In v0 (Y |_| Z)).
  apply set_union_intro1; assumption.
  hypReplace Hp (objHpElim II Hp); unfold okObjAt in Hp; unfold okFieldAt in Hp.
  linLogElim Hp; linLogIntro; sconjMatch.
  sconjLift2of2 Hp; sconjMatch.
  hypReplace Hp (objHpElim H0 Hp); unfold okObjAt in Hp; unfold okFieldAt in Hp.
  linLogElim Hp; linLogIntro; sconjMatch.
  apply (sconjImpl Hp); clear h Hp; intros h Hp.

  destruct (set_union_elim H) as [I | I]; clear IY H.
  sconjLift2of2 Hp; sconjMatch.
  hypReplace Hp (objHpElim I Hp); unfold okObjAt in Hp; unfold okFieldAt in Hp.
  linLogElim Hp; linLogIntro; sconjMatch.
  sconjMatch.
  hypReplace Hp (mapHpElim1 I Hp).
  linLogElim Hp; linLogIntro; sconjMatch.
  assumption.
  linLogElim H0.

Qed.


Lemma MWmatch2 : forall h M W phi X Y Z X' Y',
  h ||- mapHp M phi ** objHp (W |_| M) W ** objHp X (Y |_| Z) ** objHp X' Y' ** top ->
    iso_set phi M (Y |_| Y') ->
    forall M' W' phi' X2 X2',
    h ||- mapHp M' phi' ** objHp (W' |_| M') W' ** objHp X2 (Y |_| Z) ** objHp X2' Y' ** top ->
      iso_set phi' M' (Y |_| Y') ->
      set_eqv (W |_| M) (W' |_| M') ->
      set_eqv W W' /\ set_subset M M'.
Proof.
  intros h M W phi X Y Z X' Y' Hp iso M' W' phi' X2 X2' Hp' iso' WMeqv.

  assert (WWeq : set_eqv W W').
  assert (S1 : set_subset W W').
  eapply MWmatch1; eauto.
  assert (S2 : set_subset W' W).
  eapply MWmatch1; eauto.
  apply set_eqvSymm; assumption.
  constructor; eauto.

  constructor.
  assumption.

  assert (WMeqv' : set_eqv (W |_| M) (W |_| M')).
  rewrite WMeqv; rewrite WWeq; apply set_eqvRefl.

  clear WMeqv.
  intros a IM; destruct (WMeqv' a) as [I1 I2].
  assert (IWM : set_In a (W |_| M')).
  apply I1; apply set_union_intro2; assumption.
  clear I1 I2.
  destruct (set_union_elim IWM) as [I | I].
  2:assumption.
  unfold mapHp in Hp; unfold objHp in Hp; linLogElim Hp.
  elimtype False.
  assert (set_disj M W).
  apply (fasFasDisj Hp).
  intros h' x HpX; linLogElim HpX; linLogIntro; sconjMatch.
  unfold okObjAt; unfold okFieldAt.
  intros h' x HpX; linLogElim HpX; linLogIntro; sconjMatch.
  apply (set_disj_elim H a); assumption.

Qed.


Lemma MWmatch3 : forall h M W phi X Y Z X' Y',
  h ||- mapHp M phi ** objHp (W |_| M) W ** objHp X (Y |_| Z) ** objHp X' Y' ** top ->
    iso_set phi M (Y |_| Y') ->
    forall M' W' phi' X2 X2',
    h ||- mapHp M' phi' ** objHp (W' |_| M') W' ** objHp X2 (Y |_| Z) ** objHp X2' Y' ** top ->
      iso_set phi' M' (Y |_| Y') ->
      set_eqv (W |_| M) (W' |_| M') ->
      set_eqv W W' /\ set_eqv M M' /\ mapEqv phi phi'.
Proof.
  intros h M W phi X Y Z X' Y' Hp iso M' W' phi' X2 X2' Hp' iso' WMeqv.

  assert (ZZ : set_eqv W W' /\ set_eqv M M').
  assert (ZZ : set_eqv W W' /\ set_subset M M').
  eapply MWmatch2; eauto.
  assert (ZZ' : set_eqv W' W /\ set_subset M' M).
  eapply MWmatch2; eauto.
  apply set_eqvSymm; assumption.
  destruct ZZ; destruct ZZ'; constructor.
  assumption.
  constructor; eauto.

  destruct ZZ as [WWeq MMeq].
  constructor.
  assumption.
  constructor.
  assumption.
  
  intro x.
  destruct iso as [iso0 _].
  destruct iso' as [iso1 _].
  hypReplace iso0 (iso0 x).
  hypReplace iso1 (iso1 x).

  assert (D : phi x = None \/ exists v, phi x = Some v).
  destruct (phi x).
  right; econstructor; reflexivity.
  left; reflexivity.

  assert (D' : phi' x = None \/ exists v, phi' x = Some v).
  destruct (phi' x).
  right; econstructor; reflexivity.
  left; reflexivity.

  destruct D; destruct D'.
  rewrite H.
  symmetry; assumption.
  elimtype False.
  destruct H0 as [v Peq].
  destruct iso1 as [_ Q].
  assert (set_In x M').
  apply Q; rewrite Peq; discriminate.
  destruct iso0 as [R _].
  apply R.
  setoid_rewrite MMeq; assumption.
  assumption.

  elimtype False.
  destruct H as [v Peq].
  rewrite Peq in iso0; rewrite H0 in iso1.
  destruct iso1 as [Q _].
  destruct iso0 as [_ R].
  apply Q.
  rewrite <- MMeq.
  apply R.
  discriminate.
  reflexivity.

  destruct H as [v Peq].
  destruct H0 as [v' Peq'].

  assert (IM : set_In x M).
  destruct iso0 as [_ Q]; apply Q; rewrite Peq; discriminate.
  assert (IM' : set_In x M').
  rewrite <- MMeq; assumption.
  unfold mapHp in Hp; linLogElim Hp; clear H1.
  unfold mapHp in Hp'; linLogElim Hp'; clear H1.
  
  assert (Hp1 : h ||- ((lexists a, !(phi x = Some a) ** x |-> a ** top) ** top) ** top).
  clear Hp'; sconjMatch.
  hypReplace Hp (fasElim IM Hp); sconjMatch.
  linLogElim Hp; linLogIntro; sconjMatch.

  assert (Hp2 : h ||- ((lexists a, !(phi' x = Some a) ** x |-> a ** top) ** top) ** top).
  clear Hp Hp1; sconjMatch.
  hypReplace Hp (fasElim IM' Hp); sconjMatch.
  linLogElim Hp; linLogIntro; sconjMatch.
  
  clear Hp Hp'.
  linLogElim Hp1; linLogElim Hp2.
  assert (R := contContEq Hp1 Hp2); subst x0.
  rewrite H3; rewrite H4; reflexivity.

Qed.


Ltac swapHyps H1 H2 :=
  let Tmp := fresh in
    (rename H1 into Tmp;
     rename H2 into H1;
     rename Tmp into H2).


Lemma addrSetRangeDisj : forall x x' y y',
  set_disj (addrRange x x') (addrRange y y') ->
  ptrP x -> ptrP x' -> ptrP y -> ptrP y' ->
  aligned4 x x' -> aligned4 y y' ->
  set_disj (setRange x x') (setRange y y').
Proof.
  intros x x' y y' D Px Px' Py Py' Alx Aly.

  compare x x'; intro NEx.
  subst x'.
  setoid_rewrite <- (set_eqvEmptySetRange x).
  apply set_disj_intro; simpl; intros; assumption.

  compare y y'; intro NEy.
  subst y'.
  setoid_rewrite <- (set_eqvEmptySetRange y).
  apply set_disj_intro; simpl; intros; assumption.

  assert (C : x' <= y \/ y' <= x).
  assert (C : x' <= y \/ y < x').
  omega.
  destruct C as [Le | Lt1].
  left; assumption.
  assert (C : y' <= x \/ x < y').
  omega.
  destruct C as [Le | Lt2].
  right; assumption.

  elimtype False.
  hypReplace Px (ptrP_elim Px).
  hypReplace Px' (ptrP_elim Px').
  hypReplace Py (ptrP_elim Py).
  hypReplace Py' (ptrP_elim Py').
  unfold aligned4 in Alx.
  unfold aligned4 in Aly.
  destruct Px as [kx Px].
  destruct Px' as [kx' Px'].
  destruct Py as [ky Py].
  destruct Py' as [ky' Py'].
  destruct Alx as [k1 Alx].
  destruct Aly as [k2 Aly].

  assert (C : x < y \/ y <= x).
  omega.

  destruct C as [C | C].

  apply (set_disj_elim D y).
  eapply addrRangeIntro.
  omega.
  subst.
  instantiate (1 := ky - kx).
  omega.
  eapply addrRangeIntro.
  omega.
  instantiate (1 := 0).
  omega.

  apply (set_disj_elim D x).
  eapply addrRangeIntro.
  omega.
  instantiate (1 := 0).
  omega.
  eapply addrRangeIntro.
  omega.
  subst.
  instantiate (1 := kx - ky).
  omega.

  apply set_disj_intro; intros z Ix Iy.
  hypReplace Ix (setRangeElim Ix).
  hypReplace Iy (setRangeElim Iy).
  omega.

Qed.


Lemma disjObjsDisjRange : forall h x x' y y',
  h ||- (foralls z in rangeObjs x x', z |->__) **
    (foralls z in rangeObjs y y', z |->__) ** top ->
    aligned8 x x' ->
    aligned8 y y' ->
    set_disj (setRange x x') (setRange y y').
Proof.
  intros h x x' y y' Hp0 Alx Aly.

  assert (Hp : h ||- buffer x x' ** buffer y y' ** top).
  sconjMatch.
  apply objsIsBuffer; assumption.
  apply objsIsBuffer; assumption.

  clear Hp0.

  assert (set_disj (addrRange x x') (addrRange y y')).
  unfold buffer in Hp.
  apply (fasFasDisj Hp).
  intros; apply topIR; sconjMatch.
  intros; apply topIR; sconjMatch.

  compare x x'; intro NEx.
  subst x'.
  setoid_rewrite <- (set_eqvEmptySetRange x).
  apply set_disj_intro; simpl; intros; assumption.

  compare y y'; intro NEy.
  subst y'.
  setoid_rewrite <- (set_eqvEmptySetRange y).
  apply set_disj_intro; simpl; intros; assumption.

  unfold buffer in Hp.

  assert (Px : ptrP x).
  assert (Hp' : h ||- ((!ptrP x ** top) ** top) ** top).
  sconjMatch.
  assert (Ix : set_In x (addrRange x x')).
  eapply addrRangeIntro.
  destruct Alx; omega.
  instantiate (1 := 0).
  omega.
  hypReplace Hp (fasElim Ix Hp).
  sconjMatch.
  linLogElim Hp; linLogIntro; exact (contElim1 Hp).
  linLogElim Hp'; assumption.

  assert (Py : ptrP y).
  assert (Hp' : h ||- ((!ptrP y ** top) ** top) ** top).
  sconjLift2 Hp; sconjMatch.
  assert (Iy : set_In y (addrRange y y')).
  eapply addrRangeIntro.
  destruct Aly; omega.
  instantiate (1 := 0).
  omega.
  hypReplace Hp (fasElim Iy Hp).
  sconjMatch.
  linLogElim Hp; linLogIntro; exact (contElim1 Hp).
  linLogElim Hp'; assumption.

  destruct Alx.
  destruct Aly.
  apply addrSetRangeDisj; try assumption.

  destruct (ptrP_elim Px).
  eapply ptrP_intro.
  subst.
  instantiate (1 := x2 + 2 * x0).
  omega.

  destruct (ptrP_elim Py).
  eapply ptrP_intro.
  subst.
  instantiate (1 := x2 + 2 * x1).
  omega.

  subst; exists (2 * x0); omega.

  subst; exists (2 * x1); omega.

Qed.


Lemma rangeObjsSplit : forall x y z,
  aligned8 x y -> aligned8 y z ->
  set_eqv (rangeObjs x z) (rangeObjs x y |_| rangeObjs y z).
Proof.
  intros x y z [k E] [k' E'].
  constructor; intros I.
  destruct (rangeObjsElim I) as [I1 [k2 I2]]; clear I.

  assert (D : a < y \/ a >= y).
  omega.
  destruct D; [apply set_union_intro1 | apply set_union_intro2]; eapply rangeObjsIntro.
  omega.
  eauto.
  omega.
  instantiate (1 := k2 - k).
  omega.

  destruct (set_union_elim I) as [I' | I']; clear I;
    destruct (rangeObjsElim I') as [I1 [k2 I2]]; clear I'; eapply rangeObjsIntro.
  omega.
  eauto.
  omega.
  instantiate (1 := k + k2).
  omega.

Qed.
  


Lemma fieldRespMapAddRight : forall h phi s x,
  h ||- fieldRespMap phi s x ->
    forall phi',
      h ||- fieldRespMap (map_union phi phi') s x.
Proof.
  unfold fieldRespMap.
  intros h phi s x Hp phi'; linLogElim Hp.
  linLogIntro; eauto.
  destruct H1 as [v0 [X Meq]].
  exists v0; constructor.
  assumption.
  unfold mapX in * |- * .
  destruct (isOdd v0).
  assumption.
  unfold map_union; destruct (phi v0).
  assumption.
  discriminate Meq.
Qed.


Lemma fieldRespToMapAddRight : forall h phi s x,
  h ||- fieldRespToMap phi s x ->
    forall phi',
      h ||- fieldRespToMap (map_union phi phi') s x.
Proof.
  unfold fieldRespToMap; unfold fieldRespMap.
  intros h phi s x Hp phi'; linLogElim Hp.
  linLogIntro; eauto.
  destruct H1 as [v0 [X Meq]].
  exists v0; constructor.
  assumption.
  unfold mapX in * |- * .
  destruct (isOdd v0).
  assumption.
  unfold map_unionnn in * |- * .
  unfold map_union in * |- * .
  destruct (spaceIdMap (rget s rtoStart) (rget s rtoEnd) v0).
  assumption.
  rewrite Meq; reflexivity.
Qed.


Lemma objRespMapObjHp : forall h' X phi h A B,
  h' ||- (foralls x in X, objRespMapH phi h x) ->
  h ||- objHp A X ** top ->
  (forall x y, set_In x A -> phi x = Some y -> set_In y B) ->
  h' ||- objHp B X.
Proof.
  unfold objRespMapH; unfold fieldRespMapH; unfold objHp.
  intros h' X phi h A B Hp Hp' Imp.
  unfold objHp; apply (fasImplI Hp); clear h' Hp.
  intros x h' I Hp; lle Hp.
  assert (HpO : h ||- (okObjAt A x ** top) ** top).
  scm; hypReplace Hp0 (objHpElim I Hp0); scm.
  unfold okObjAt in HpO; unfold okFieldAt in HpO; lle HpO; clear Hp'.
  destruct H3 as [v0 [Hp0 Meq0]]; destruct H4 as [v1 [Hp1 Meq1]].
  assert (Z := contContEq Hp0 HpO); sconjLift2 HpO; 
    assert (Z' := contContEq Hp1 HpO); clear Hp0 Hp1 HpO; subst x3 x2.
  unfold okObjAt; unfold okFieldAt; lli.
  2:eauto.

  clear H8 Meq1 v1 Hp.
  assert (Q := isOddOk v0).
  unfold mapX in Meq0; destruct (isOdd v0).
  injection Meq0; intro; subst v0; left; assumption.
  destruct H5.
  elimtype False; apply (not_even_and_odd v0); assumption.
  right; eauto.

  clear v0 Meq0 H5; rename v1 into v0; rename Meq1 into Meq0; rename H8 into H5.
  assert (Q := isOddOk v0).
  unfold mapX in Meq0; destruct (isOdd v0).
  injection Meq0; intro; subst v0; left; assumption.
  destruct H5.
  elimtype False; apply (not_even_and_odd v0); assumption.
  right; eauto.

Qed.



Lemma fasInterSplit : forall h X P,
  h ||- fas P X ->
    forall Y,
      h ||- fas P (set_inter X Y) ** top.
Proof.
  intros h X P Hp Y.
  assert (Hp' : h ||- fas P (set_inter X Y) ** fas P (set_diff X Y)).
  apply fasSplit.
  unfold hpredTrue; rewrite <- (set_inter_split X Y); rewrite <- hpredTrueEq; assumption.
  apply interDiffDisj.
  scm.
Qed.


Lemma noFlipRespectsMap : forall M M' W' B' B'' G' G'' F' phi phi' B0 B1,
  let W := M' |_| W' in
  let F := ((B'' |_| G'') |_| F') in
  let G := B' |_| G' in
  let frSp := W |_| M in
  let B := B0 |_| B1 in
  let toSp := (B |_| G) |_| F in
  let cpObjs := B0 |_| G in
  let toObjs := cpObjs |_| B1 in
    iso_set phi M cpObjs ->
    iso_set phi' M' (B'' |_| G'') ->
    iso_set (map_union phi phi') (M |_| M') (cpObjs |_| B'' |_| G'') ->
    forall (s:state) h h' objs vs toStart toEnd root,
    s |||- top ** objHp frSp W ** objHp toObjs B ** objHp (frSp |_| toObjs) G ** top ->
    vs |||- (foralls x in B, objCopiedH x (heapOf s) x) **
      (foralls x in G |_| W, objRespMapH (map_union phi (Some (A:=nat))) (heapOf s) x) ->
    vs |||- (minObjHp (root :: nil) objs && Heap.heap_eqv h) ** top ->
    h' ||- (foralls x in W', objCopied x s x) **
      (foralls x in B', objRespToMap (map_union phi phi') (rset (rset s rtoStart toStart) rtoEnd toEnd) x) **
      (foralls x in G', objCopied x s x) **
      (foralls x in B'', objIsToFwded phi phi' (rset (rset s rtoStart toStart) rtoEnd toEnd) x) **
      (foralls x in G'', objCopiedPhi phi' s x) ** 
      (foralls x in B, objCopied x s x) ->
    let vh' := Heap.HeapMap.map_mapRange
      (fun x => match map_union phi phi' x with | Some y => y | None => x end) h' in
    let objs' := set_map (fun x => match phi' x with | Some y => y | None => x end) objs in
    let phi2 := fun x => 
      if set_mem x objs then 
        if set_mem x M' then 
          phi' x 
        else
          Some x
      else
        None (A:=nat) in
    set_subset objs (((B |_| G) |_| W') |_| M') ->
    set_subset objs' (W' |_| toObjs |_| B'' |_| G'') ->

    set_disj W' M' ->
    set_disj (B'' |_| G'') W' ->
    iso_set phi2 objs objs' ->
    (forall x, set_In x (M |_| M') -> ptrP x) ->
    (forall x, set_In x cpObjs -> ptrP x) ->
    set_disj cpObjs frSp ->
    set_disj (B'' |_| G'') B' ->
    set_disj frSp (setRange toStart toEnd) ->
    set_disj (B'' |_| G'') M ->
    set_disj (B'' |_| G'') M' ->
    set_disj (B'' |_| G'') G' ->
    set_disj (B'' |_| G'') B ->
    set_disj W M ->
    set_disj ((B0 |_| B' |_| G') |_| B1) frSp ->
    set_disj ((B |_| B' |_| G') |_| W') M' ->
    set_eqv toSp (rangeObjs toStart toEnd) ->

    vh' ||- (foralls x in objs', objIsFwdedH phi2 h x) ** top.
Proof.
  intros M M' W' B' B'' G' G'' F' phi phi' B0 B1.
  intros W F G frSp B toSp cpObjs toObjs.
  intros Iso1 Iso2 Iso12 s h h' objs vs toStart toEnd root Hp Vhp1 Vhp2 Hp' vh' objs' phi2 Sub Sub'.

  intros WMdisj BGWdisj Iso0 Mptr Cptr CMdisj BGBdisj frToDisj BGMdisj BGMdisj' BGGdisj BGBdisj''
    WMdisj0 toFrDisj BGGWMdisj toSpEqv.

  lle Hp.

  assert (X : h ||- objHp objs objs /\ vs |||- Heap.heap_eqv h ** top).
  constructor.
  eapply minObjHpElim1; eapply heapEqvImplL; apply Vhp2.
  destruct (aconjElim (sconjDistrib2 Vhp2)) as [_ HY]; exact HY.
  clear Vhp2; destruct X as [HpO Vhp2].

  assert (Hp'' : h' ||- ((foralls x in set_inter W' objs', objCopied x s x) ** top) **
    ((foralls x in set_inter B' objs', objRespToMap (map_union phi phi') 
      (rset (rset s rtoStart toStart) rtoEnd toEnd) x) ** top) **
    ((foralls x in set_inter G' objs', objCopied x s x) ** top) **
    ((foralls x in set_inter B'' objs', 
      objIsToFwded phi phi' (rset (rset s rtoStart toStart) rtoEnd toEnd) x) ** top) **
    ((foralls x in set_inter G'' objs',
      objCopiedPhi phi' s x) ** top) **
    ((foralls x in set_inter B objs', objCopied x s x)) ** top).
  clear Vhp1 vh' Sub Sub' Hp phi2 Iso0; scm;
    apply fasInterSplit; assumption.
  
  clear Hp'; rename Hp'' into Hp'; lle Hp'.
  sconjLift7of7 Hp'; apply sconjComm.
  subst vh'; apply (sconjMapRangeImpl Hp'); clear Hp'; intros hh Hp'.
  lli.
  unfold hpredTrue; rewrite <- (subsetInterEq Sub');
    setoid_rewrite (set_interSymm objs' (W' |_| toObjs |_| B'' |_| G''));
    rewrite (set_unionInterDistrib1 W' (toObjs |_| B'' |_| G'') objs').
  subst toObjs cpObjs.
  apply fasMerge.
  apply (sconjMapRangeImpl Hp'); clear hh Hp'; intros hh Hp'.
  (* W' *)
  apply (fasMapRangeImplI Hp'); clear hh Hp'; 
    unfold objCopied; unfold fieldCopied; unfold objIsFwdedH;
      intros hh x Ix Hp'; lle Hp'.
  lli.
  2:simpl; apply (sconjMapRangeImpl Hp'); intros hh' Hp3; apply (contMapRangeImpl Hp3).
  clear hh Hp'; destruct (set_inter_elim Ix) as [Ix1 Ix2]; clear Ix.
  assert (HX : vs |||- top ** (objRespMapH (map_union phi (Some (A:=nat))) (heapOf s) x ** top)).
  clear Vhp2; scm.
  assert (Ix' : set_In x (G |_| W)).
  apply set_union_intro2; subst W; apply set_union_intro2; assumption.
  hypReplace Hp0 (fasElim Ix' Hp0); scm.
  unfold objRespMapH in HX; unfold fieldRespMapH in HX; lle HX.
  destruct H5 as [v0 [Hp0 Meq0]]; destruct H8 as [v1 [Hp1 Meq1]].
  assert (X1 := contContEq H3 Hp0); assert (X2 := contContEq H4 Hp1); 
    subst x0 x1; clear H3 H4 Hp0 Hp1; rename x2 into v0'; rename x3 into v1'.
  exists x; exists v0'; exists v1'; subst phi2; cbv beta.
  assert (Ix : set_In x objs).
  subst objs'; destruct (set_map_elim Ix2) as [y [Xeq Iy]].
  destruct Iso2 as [_ [Iso2 _]]; hypReplace Iso2 (Iso2 y x); destruct (phi' y).
  subst a; elf; apply (set_disj_elim BGWdisj x); auto; apply Iso2; reflexivity.
  subst y; assumption.
  rewrite (set_mem_true Ix); rewrite set_mem_false.
  constructor.
  reflexivity.
  hypReplace HpO (objHpElim Ix HpO); unfold okObjAt in HpO; unfold okFieldAt in HpO;
    lle HpO.
  hypReplace Vhp2 (sconjInsert Vhp2 HpO); lle Vhp2; assert (X1 := contContEq Vhp2 HX);
    sconjLift2 Vhp2; sconjLift2 HX; assert (X2 := contContEq Vhp2 HX); clear HX Vhp2;
      subst x0 x1.
  constructor.
  lli; scm.
  clear x Ix1 Ix2 Ix HpO; constructor.
  (* first field *)
  clear Meq1 v1 v1' H6.
  unfold mapX in Meq0 |- * .
  assert (X0 := isOddOk v0); destruct (isOdd v0).
  destruct Iso12 as [Iso12 _]; destruct (Iso12 v0') as [_ Q].
  injection Meq0; clear Meq0; intro; subst v0; rewrite (isOddTrue X0).
  replace (map_union phi phi' v0') with (None (A := nat)).
  reflexivity.
  destruct (map_union phi phi' v0').
  elf; apply (not_even_and_odd v0'); auto; apply ptrPeven; apply Mptr; apply Q; discriminate.
  reflexivity.
  unfold map_union in Meq0 |- *; destruct Iso1 as [Iso1 [Iso1' _]]; destruct (Iso1 v0) as [_ Q];
    hypReplace Iso1' (Iso1' v0 v0'); destruct (phi v0).
  injection Meq0; clear Meq0; intro; subst a.
  assert (EE0 : even v0').
  subst B; apply ptrPeven; apply Cptr; apply Iso1'; reflexivity.
  rewrite (isOddFalse EE0).
  destruct H3.
  elf; apply (not_even_and_odd v0'); assumption.
  rewrite (set_mem_true H).
  rewrite set_mem_false.
  reflexivity.
  intro I; apply (set_disj_elim CMdisj v0'); auto; subst frSp;
    subst W; apply set_union_intro1; apply set_union_intro1; assumption.
  injection Meq0; clear Meq0; intro; subst v0'.
  rewrite (isOddFalse X0).
  destruct H3.
  elf; apply (not_even_and_odd v0); assumption.
  rewrite (set_mem_true H).
  clear Q; destruct Iso2 as [Iso2 _]; destruct (Iso2 v0) as [Q Q']; destruct (phi' v0).
  rewrite set_mem_true; [reflexivity | apply Q'; discriminate].
  rewrite set_mem_false.
  reflexivity.
  intro Z; apply Q; auto.
  (* second field *)
  clear Meq0 v0 v0' H3; rename Meq1 into Meq0; rename v1 into v0; rename v1' into v0'; 
    rename H6 into H3.
  unfold mapX in Meq0 |- * .
  assert (X0 := isOddOk v0); destruct (isOdd v0).
  destruct Iso12 as [Iso12 _]; destruct (Iso12 v0') as [_ Q].
  injection Meq0; clear Meq0; intro; subst v0; rewrite (isOddTrue X0).
  replace (map_union phi phi' v0') with (None (A := nat)).
  reflexivity.
  destruct (map_union phi phi' v0').
  elf; apply (not_even_and_odd v0'); auto; apply ptrPeven; apply Mptr; apply Q; discriminate.
  reflexivity.
  unfold map_union in Meq0 |- *; destruct Iso1 as [Iso1 [Iso1' _]]; destruct (Iso1 v0) as [_ Q];
    hypReplace Iso1' (Iso1' v0 v0'); destruct (phi v0).
  injection Meq0; clear Meq0; intro; subst a.
  assert (EE0 : even v0').
  subst B; apply ptrPeven; apply Cptr; apply Iso1'; reflexivity.
  rewrite (isOddFalse EE0).
  destruct H3.
  elf; apply (not_even_and_odd v0'); assumption.
  rewrite (set_mem_true H).
  rewrite set_mem_false.
  reflexivity.
  intro I; apply (set_disj_elim CMdisj v0'); auto;
    subst frSp; subst W; apply set_union_intro1; apply set_union_intro1; assumption.
  injection Meq0; clear Meq0; intro; subst v0'.
  rewrite (isOddFalse X0).
  destruct H3.
  elf; apply (not_even_and_odd v0); assumption.
  rewrite (set_mem_true H).
  clear Q; destruct Iso2 as [Iso2 _]; destruct (Iso2 v0) as [Q Q']; destruct (phi' v0).
  rewrite set_mem_true; [reflexivity | apply Q'; discriminate].
  rewrite set_mem_false.
  reflexivity.
  intro Z; apply Q; auto.
  
  intro I; apply (set_disj_elim WMdisj x); assumption.
  (* done with W' *)
  subst G.
  unfold hpredTrue; rewrite (set_eqvUnionSymm (B0 |_| B' |_| G') B1);
    rewrite <- (set_eqvUnionAssoc B1 B0 (B' |_| G'));
    rewrite (set_eqvUnionSymm B1 B0); replace (B0 |_| B1) with B; [idtac | reflexivity];
    rewrite (set_eqvUnionSymm B (B' |_| G')); rewrite (set_eqvUnionAssoc (B' |_| G') B (B'' |_| G''));
    rewrite (set_eqvUnionAssoc B' G' (B |_| B'' |_| G'')).
  rewrite (set_unionInterDistrib1 B' (G' |_| B |_| B'' |_| G'') objs').
  apply fasMerge.
  apply (sconjMapRangeImpl Hp'); clear hh Hp'; intros hh Hp'.
  (* B' *)
  apply (fasMapRangeImplI Hp'); clear hh Hp';
    unfold objRespToMap; unfold fieldRespToMap; unfold fieldRespMap; unfold objIsFwdedH;
      stSimpl; intros hh x Ix Hp'; lle Hp'; lli.
  2:simpl; apply (sconjMapRangeImpl Hp'); intros hh' Hp3; apply (contMapRangeImpl Hp3).
  clear hh Hp'; destruct (set_inter_elim Ix) as [Ix1 Ix2]; clear Ix.
  destruct H3 as [v0 [HpV0 Meq0]]; destruct H4 as [v1 [HpV1 Meq1]].
  subst objs'; destruct (set_map_elim Ix2) as [y [Xeq Iy]]; clear Ix2.
  subst phi2; cbv beta.
  assert (HX : vs |||- top ** (objRespMapH (map_union phi (Some (A:=nat))) (heapOf s) x ** top)).
  clear Vhp2; scm.
  assert (Ix' : set_In x ((B' |_| G') |_| W)).
  apply set_union_intro1; apply set_union_intro1; assumption.
  hypReplace Hp0 (fasElim Ix' Hp0); scm.
  unfold objRespMapH in HX; unfold fieldRespMapH in HX; lle HX.
  destruct H3 as [v0' [Hp0 Meq0']]; destruct H6 as [v1' [Hp1 Meq1']].
  exists y.
  assert (X1 := contContEq HpV0 Hp0); assert (X2 := contContEq HpV1 Hp1); 
    subst v0' v1'; clear HpV0 HpV1 Hp0 Hp1; rename x2 into v0'; rename x3 into v1'.
  rewrite (set_mem_true Iy).
  assert (E : x = y).
  destruct Iso2 as [_ [Iso2 _]]; hypReplace Iso2 (Iso2 y x); destruct (phi' y).
  subst a; elf; apply (set_disj_elim BGBdisj x); auto; apply Iso2; reflexivity.
  assumption.
  subst y; clear Xeq; rewrite set_mem_false.
  rename Iy into Ix.
  hypReplace HpO (objHpElim Ix HpO); unfold okObjAt in HpO; unfold okFieldAt in HpO;
    lle HpO.
  hypReplace Vhp2 (sconjInsert Vhp2 HpO); lle Vhp2; assert (X1 := contContEq Vhp2 HX);
    sconjLift2 Vhp2; sconjLift2 HX; assert (X2 := contContEq Vhp2 HX); clear HX Vhp2;
      subst x2 x3.
  exists v0'; exists v1'.
  constructor.
  reflexivity.
  constructor.
  lli; scm.
  clear x Ix1 Ix HpO; constructor.

  (* first field *)
  clear Meq1 Meq1' v1 v1' H6 x1.
  unfold mapX in Meq0, Meq0' |- * .
  assert (X0 := isOddOk v0); destruct (isOdd v0).
  injection Meq0; injection Meq0'; clear Meq0 Meq0'; intros; subst x0 v0'.
  rewrite (isOddTrue X0).
  destruct Iso12 as [Iso12 _]; destruct (Iso12 v0) as [_ Q].
  replace (map_union phi phi' v0) with (None (A := nat)).
  reflexivity.
  destruct (map_union phi phi' v0).
  elf; apply (not_even_and_odd v0); auto; apply ptrPeven; apply Mptr; apply Q; discriminate.
  reflexivity.
  unfold map_unionnn in Meq0; unfold map_union in Meq0, Meq0' |- * .
  assert (X := set_memSpaceIdMap toStart toEnd v0).
  destruct (set_In_dec v0 (setRange toStart toEnd)) as [I | NI].
  rewrite (set_mem_true I) in X; rewrite X in Meq0;
    injection Meq0; clear Meq0; intro; subst x0.
  assert (Peq : phi v0 = None).
  destruct Iso1 as [Iso1 _]; destruct (Iso1 v0) as [_ Q]; destruct (phi v0).
  elf; clear Meq0'; apply (set_disj_elim frToDisj v0); auto;
    subst frSp; apply set_union_intro2; apply Q; discriminate.
  reflexivity.
  rewrite Peq; rewrite Peq in Meq0'.
  injection Meq0'; clear Meq0'; intro; subst v0'.
  rewrite (isOddFalse X0).
  destruct H3.
  elf; apply (not_even_and_odd v0); assumption.
  rewrite (set_mem_true H).
  destruct Iso2 as [Iso2 _]; destruct (Iso2 v0) as [Q1 Q2].
  destruct (phi' v0).
  elf; apply (set_disj_elim frToDisj v0); auto; subst frSp; apply set_union_intro1;
    subst W; apply set_union_intro1; apply Q2; discriminate.
  rewrite set_mem_false.
  reflexivity.
  intro; apply Q1; auto.
  rewrite (set_mem_false NI) in X.
  rewrite X in Meq0.
  destruct Iso1 as [Iso1 [Iso1' _]]; destruct (Iso1 v0); hypReplace Iso1' (Iso1' v0);
    destruct (phi v0).
  injection Meq0; injection Meq0'; clear Meq0 Meq0'; intros; subst a x0.
  hypReplace Iso1' (Iso1' v0').
  assert (EE : even v0').
  apply ptrPeven; apply Cptr; auto.
  rewrite (isOddFalse EE).
  destruct H3.
  elf; apply (not_even_and_odd v0'); assumption.
  rewrite (set_mem_true H1).
  rewrite set_mem_false.
  destruct (Iso1 v0') as [_ X1]; destruct (phi v0').
  elf; apply (set_disj_elim CMdisj v0'); auto;
    subst frSp; subst W; apply set_union_intro2; apply X1; discriminate.
  destruct Iso2 as [Iso2 _]; destruct (Iso2 v0') as [_ X2]; destruct (phi' v0').
  elf; apply (set_disj_elim CMdisj v0'); auto;
      subst frSp; subst W; apply set_union_intro1; apply set_union_intro1; apply X2; discriminate.
  reflexivity.
  intro U; apply (set_disj_elim CMdisj v0'); auto;
      subst frSp; subst W; apply set_union_intro1; apply set_union_intro1; assumption.
  injection Meq0'; clear Meq0'; intro; subst v0'.
  rewrite (isOddFalse X0); destruct H3.
  elf; apply (not_even_and_odd v0); assumption.
  rewrite (set_mem_true H1).
  rewrite set_mem_true.
  rewrite Meq0; destruct Iso2 as [Iso2' [Iso2 _]]; hypReplace Iso2 (Iso2 _ _ Meq0);
    destruct (Iso1 x0); destruct (phi x0).
  elf; apply (set_disj_elim BGMdisj x0); trivial; apply H3; discriminate.
  clear H0 Iso1' H3.
  destruct (Iso2' x0); destruct (phi' x0).
  elf; apply (set_disj_elim BGMdisj' x0); trivial; apply H3; discriminate.
  reflexivity.
  destruct Iso2 as [Iso2 _]; destruct (Iso2 v0); apply H3; rewrite Meq0; discriminate.
  (* second field, same as the first *)
  clear Meq0 Meq0' v0 v0' H3 x0.
  rename Meq1 into Meq0; rename Meq1' into Meq0'; rename v1 into v0; rename v1' into v0';
    rename H6 into H3; rename x1 into x0.
  unfold mapX in Meq0, Meq0' |- * .
  assert (X0 := isOddOk v0); destruct (isOdd v0).
  injection Meq0; injection Meq0'; clear Meq0 Meq0'; intros; subst x0 v0'.
  rewrite (isOddTrue X0).
  destruct Iso12 as [Iso12 _]; destruct (Iso12 v0) as [_ Q].
  replace (map_union phi phi' v0) with (None (A := nat)).
  reflexivity.
  destruct (map_union phi phi' v0).
  elf; apply (not_even_and_odd v0); auto; apply ptrPeven; apply Mptr; apply Q; discriminate.
  reflexivity.
  unfold map_unionnn in Meq0; unfold map_union in Meq0, Meq0' |- * .
  assert (X := set_memSpaceIdMap toStart toEnd v0).
  destruct (set_In_dec v0 (setRange toStart toEnd)) as [I | NI].
  rewrite (set_mem_true I) in X; rewrite X in Meq0;
    injection Meq0; clear Meq0; intro; subst x0.
  assert (Peq : phi v0 = None).
  destruct Iso1 as [Iso1 _]; destruct (Iso1 v0) as [_ Q]; destruct (phi v0).
  elf; clear Meq0'; apply (set_disj_elim frToDisj v0); auto;
    subst frSp; apply set_union_intro2; apply Q; discriminate.
  reflexivity.
  rewrite Peq; rewrite Peq in Meq0'.
  injection Meq0'; clear Meq0'; intro; subst v0'.
  rewrite (isOddFalse X0).
  destruct H3.
  elf; apply (not_even_and_odd v0); assumption.
  rewrite (set_mem_true H).
  destruct Iso2 as [Iso2 _]; destruct (Iso2 v0) as [Q1 Q2].
  destruct (phi' v0).
  elf; apply (set_disj_elim frToDisj v0); auto; subst frSp; apply set_union_intro1;
    subst W; apply set_union_intro1; apply Q2; discriminate.
  rewrite set_mem_false.
  reflexivity.
  intro; apply Q1; auto.
  rewrite (set_mem_false NI) in X.
  rewrite X in Meq0.
  destruct Iso1 as [Iso1 [Iso1' _]]; destruct (Iso1 v0); hypReplace Iso1' (Iso1' v0);
    destruct (phi v0).
  injection Meq0; injection Meq0'; clear Meq0 Meq0'; intros; subst a x0.
  hypReplace Iso1' (Iso1' v0').
  assert (EE : even v0').
  apply ptrPeven; apply Cptr; auto.
  rewrite (isOddFalse EE).
  destruct H3.
  elf; apply (not_even_and_odd v0'); assumption.
  rewrite (set_mem_true H1).
  rewrite set_mem_false.
  destruct (Iso1 v0') as [_ X1]; destruct (phi v0').
  elf; apply (set_disj_elim CMdisj v0'); auto;
    subst frSp; subst W; apply set_union_intro2; apply X1; discriminate.
  destruct Iso2 as [Iso2 _]; destruct (Iso2 v0') as [_ X2]; destruct (phi' v0').
  elf; apply (set_disj_elim CMdisj v0'); auto;
      subst frSp; subst W; apply set_union_intro1; apply set_union_intro1; apply X2; discriminate.
  reflexivity.
  intro U; apply (set_disj_elim CMdisj v0'); auto;
      subst frSp; subst W; apply set_union_intro1; apply set_union_intro1; assumption.
  injection Meq0'; clear Meq0'; intro; subst v0'.
  rewrite (isOddFalse X0); destruct H3.
  elf; apply (not_even_and_odd v0); assumption.
  rewrite (set_mem_true H1).
  rewrite set_mem_true.
  rewrite Meq0; destruct Iso2 as [Iso2' [Iso2 _]]; hypReplace Iso2 (Iso2 _ _ Meq0);
    destruct (Iso1 x0); destruct (phi x0).
  elf; apply (set_disj_elim BGMdisj x0); trivial; apply H3; discriminate.
  clear H0 Iso1' H3.
  destruct (Iso2' x0); destruct (phi' x0).
  elf; apply (set_disj_elim BGMdisj' x0); trivial; apply H3; discriminate.
  reflexivity.
  destruct Iso2 as [Iso2 _]; destruct (Iso2 v0); apply H3; rewrite Meq0; discriminate.
  intro I; apply (set_disj_elim CMdisj x).
  apply set_union_intro2; apply set_union_intro1; assumption.
  subst frSp W; apply set_union_intro1; apply set_union_intro1; assumption.
  (* done with B' *)
  unfold hpredTrue; rewrite (set_unionInterDistrib1 G' (B |_| B'' |_| G'') objs').
  apply fasMerge.
  apply (sconjMapRangeImpl Hp'); clear hh Hp'; intros hh Hp'.

  (* G' *)
  apply (fasMapRangeImplI Hp'); clear hh Hp';
    unfold objCopied; unfold fieldCopied; unfold fieldRespMap; unfold objIsFwdedH;
      stSimpl; intros hh x Ix Hp'; lle Hp'; lli.
  2:simpl; apply (sconjMapRangeImpl Hp'); intros hh' Hp3; apply (contMapRangeImpl Hp3).
  clear hh Hp'; destruct (set_inter_elim Ix) as [Ix1 Ix2]; clear Ix.
  subst objs'; destruct (set_map_elim Ix2) as [y [Xeq Iy]]; clear Ix2.
  subst phi2; cbv beta.
  assert (HX : vs |||- top ** (objRespMapH (map_union phi (Some (A:=nat))) (heapOf s) x ** top)).
  clear Vhp2; scm.
  assert (Ix' : set_In x ((B' |_| G') |_| W)).
  apply set_union_intro1; apply set_union_intro2; assumption.
  hypReplace Hp0 (fasElim Ix' Hp0); scm.
  rename H3 into St1; rename H4 into St2.
  unfold objRespMapH in HX; unfold fieldRespMapH in HX; lle HX.
  destruct H3 as [v0' [Hp0 Meq0']]; destruct H6 as [v1' [Hp1 Meq1']].
  exists y.
  assert (X1 := contContEq St1 Hp0); assert (X2 := contContEq St2 Hp1); 
    subst v0' v1'; clear St1 St2 Hp0 Hp1; rename x2 into v0'; rename x3 into v1'.
  rewrite (set_mem_true Iy).
  assert (E : x = y).
  destruct Iso2 as [_ [Iso2 _]]; hypReplace Iso2 (Iso2 y x); destruct (phi' y).
  subst a; elf; apply (set_disj_elim BGGdisj x); auto; apply Iso2; reflexivity.
  assumption.
  subst y; clear Xeq; rewrite set_mem_false.
  rename Iy into Ix.
  hypReplace HpO (objHpElim Ix HpO); unfold okObjAt in HpO; unfold okFieldAt in HpO;
    lle HpO.
  hypReplace Vhp2 (sconjInsert Vhp2 HpO); lle Vhp2; assert (X1 := contContEq Vhp2 HX);
    sconjLift2 Vhp2; sconjLift2 HX; assert (X2 := contContEq Vhp2 HX); clear HX Vhp2;
      subst x2 x3.
  exists v0'; exists v1'.
  constructor.
  reflexivity.
  constructor.
  lli; scm.
  clear x Ix1 Ix HpO; constructor.
  (* first field *)
  clear Meq1' v1' H6 x1; rename Meq0' into Meq0; rename x0 into v0.
  unfold mapX in Meq0 |- * .
  assert (X0 := isOddOk v0); destruct (isOdd v0).
  injection Meq0; clear Meq0; intros; subst v0'.
  rewrite (isOddTrue X0).
  destruct Iso12 as [Iso12 _]; destruct (Iso12 v0) as [_ Q].
  replace (map_union phi phi' v0) with (None (A := nat)).
  reflexivity.
  destruct (map_union phi phi' v0).
  elf; apply (not_even_and_odd v0); auto; apply ptrPeven; apply Mptr; apply Q; discriminate.
  reflexivity.
  unfold map_unionnn in Meq0; unfold map_union in Meq0 |- * .
  assert (X0' : even v0').
  destruct Iso1 as [_ [Iso1 _]]; hypReplace Iso1 (Iso1 v0); destruct (phi v0).
  apply ptrPeven; apply Cptr; auto.
  injection Meq0; intro; subst v0'; assumption.
  rewrite (isOddFalse X0').
  destruct H3.
  elf; apply (not_even_and_odd v0'); assumption.
  rewrite (set_mem_true H).
  destruct Iso1 as [Iso1 [Iso1' _]]; destruct (Iso1 v0) as [Q Q']; assert (X := Iso1' v0);
    destruct (phi v0).
  injection Meq0; clear Meq0; intro; subst a.
  assert (NI : ~set_In v0' M').
  intros I; apply (set_disj_elim CMdisj v0'); auto; subst frSp.
  subst W; apply set_union_intro1; apply set_union_intro1; assumption.
  rewrite (set_mem_false NI).
  reflexivity.
  injection Meq0; clear Q' X Meq0 X0'; intro; subst v0'.
  destruct Iso2 as [Iso2 _]; destruct (Iso2 v0); destruct (phi' v0).
  rewrite set_mem_true; [reflexivity | apply H1; discriminate].
  rewrite set_mem_false; [reflexivity | intro Z; apply H0; auto].
  (* second field, same as the first *)
  clear Meq0' v0' H3 x0; rename Meq1' into Meq0'; rename v1' into v0'; rename H6 into H3;
    rename x1 into x0; rename Meq0' into Meq0; rename x0 into v0.
  unfold mapX in Meq0 |- * .
  assert (X0 := isOddOk v0); destruct (isOdd v0).
  injection Meq0; clear Meq0; intros; subst v0'.
  rewrite (isOddTrue X0).
  destruct Iso12 as [Iso12 _]; destruct (Iso12 v0) as [_ Q].
  replace (map_union phi phi' v0) with (None (A := nat)).
  reflexivity.
  destruct (map_union phi phi' v0).
  elf; apply (not_even_and_odd v0); auto; apply ptrPeven; apply Mptr; apply Q; discriminate.
  reflexivity.
  unfold map_unionnn in Meq0; unfold map_union in Meq0 |- * .
  assert (X0' : even v0').
  destruct Iso1 as [_ [Iso1 _]]; hypReplace Iso1 (Iso1 v0); destruct (phi v0).
  apply ptrPeven; apply Cptr; auto.
  injection Meq0; intro; subst v0'; assumption.
  rewrite (isOddFalse X0').
  destruct H3.
  elf; apply (not_even_and_odd v0'); assumption.
  rewrite (set_mem_true H).
  destruct Iso1 as [Iso1 [Iso1' _]]; destruct (Iso1 v0) as [Q Q']; assert (X := Iso1' v0);
    destruct (phi v0).
  injection Meq0; clear Meq0; intro; subst a.
  assert (NI : ~set_In v0' M').
  intros I; apply (set_disj_elim CMdisj v0'); auto; subst frSp.
  subst W; apply set_union_intro1; apply set_union_intro1; assumption.
  rewrite (set_mem_false NI).
  reflexivity.
  injection Meq0; clear Q' X Meq0 X0'; intro; subst v0'.
  destruct Iso2 as [Iso2 _]; destruct (Iso2 v0); destruct (phi' v0).
  rewrite set_mem_true; [reflexivity | apply H1; discriminate].
  rewrite set_mem_false; [reflexivity | intro Z; apply H0; auto].
  intro I; apply (set_disj_elim CMdisj x).
  apply set_union_intro2; apply set_union_intro2; assumption.
  subst frSp; subst W; apply set_union_intro1; apply set_union_intro1; assumption.
  (* done with G' *)
  sconjLift3of3 Hp'; unfold hpredTrue; rewrite (set_unionInterDistrib1 B (B'' |_| G'') objs').
  apply fasMerge.
  apply (sconjMapRangeImpl Hp'); clear hh Hp'; intros hh Hp'.

  (* B *)
  apply (fasMapRangeImplI Hp'); clear hh Hp';
    unfold objCopied; unfold fieldCopied; unfold fieldRespMap; unfold objIsFwdedH;
      stSimpl; intros hh x Ix Hp'; lle Hp'; lli.
  2:simpl; apply (sconjMapRangeImpl Hp'); intros hh' Hp3; apply (contMapRangeImpl Hp3).
  clear hh Hp'; destruct (set_inter_elim Ix) as [Ix1 Ix2]; clear Ix.
  subst objs'; destruct (set_map_elim Ix2) as [y [Xeq Iy]]; clear Ix2.
  subst phi2; cbv beta.
  assert (HX0 : s |||- top ** (okObjAt ((B0 |_| B' |_| G') |_| B1) x ** top) ** top).
  clear H3 H4; scm; hypReplace Hp (objHpElim Ix1 Hp); scm.
  unfold okObjAt in HX0; unfold okFieldAt in HX0; lle HX0.
  assert (X1 := contContEq HX0 H3); sconjLift2 HX0; assert (X2 := contContEq HX0 H4);
    subst x2 x3; clear HX0; rename H5 into F1ok; rename H8 into F2ok.
  assert (HX : vs |||- (objCopiedH x (heapOf s) x ** top) ** top). 
  clear Vhp2; scm.
  hypReplace Hp0 (fasElim Ix1 Hp0); scm.
  rename H3 into St1; rename H4 into St2.
  unfold objCopiedH in HX; unfold fieldCopiedH in HX; lle HX.
  exists y.
  assert (X1 := contContEq St1 H3); assert (X2 := contContEq St2 H6); 
    subst x2 x3; clear St1 St2 H3 H6; rename x0 into v0'; rename x1 into v1'.
  rewrite (set_mem_true Iy).
  assert (E : x = y).
  destruct Iso2 as [_ [Iso2 _]]; hypReplace Iso2 (Iso2 y x); destruct (phi' y).
  subst a; elf; apply (set_disj_elim BGBdisj'' x); auto; apply Iso2; reflexivity.
  assumption.
  subst y; clear Xeq; rewrite set_mem_false.
  rename Iy into Ix.
  hypReplace HpO (objHpElim Ix HpO); unfold okObjAt in HpO; unfold okFieldAt in HpO; lle HpO.
  hypReplace Vhp2 (sconjInsert Vhp2 HpO); lle Vhp2; assert (X1 := contContEq Vhp2 HX);
    sconjLift2 Vhp2; sconjLift2 HX; assert (X2 := contContEq Vhp2 HX); clear HX Vhp2;
      subst x0 x1.
  exists v0'; exists v1'.
  constructor.
  reflexivity.
  constructor.
  lli; scm.
  clear x Ix1 Ix HpO; constructor.
  (* first field *)
  clear v1' H6 F2ok; rename H3 into okV; unfold mapX; rename v0' into v0.
  assert (X0 := isOddOk v0); destruct (isOdd v0).
  destruct Iso12 as [Iso12 _]; destruct (Iso12 v0) as [_ Q].
  replace (map_union phi phi' v0) with (None (A := nat)).
  reflexivity.
  destruct (map_union phi phi' v0).
  elf; apply (not_even_and_odd v0); auto; apply ptrPeven; apply Mptr; apply Q; discriminate.
  reflexivity.
  destruct F1ok as [F1ok | F1ok].
  elf; apply (not_even_and_odd v0); assumption.
  destruct okV as [okV | okV].
  elf; apply (not_even_and_odd v0); assumption.
  rewrite (set_mem_true okV).
  rewrite set_mem_false.
  destruct Iso12 as [Iso12 _]; destruct (Iso12 v0) as [_ X]; 
    destruct (map_union phi phi' v0).
  elf; apply (set_disj_elim toFrDisj v0).
  auto.
  subst frSp W; rewrite (set_eqvUnionSymm M' W'); rewrite (set_eqvUnionAssoc W' M' M);
    rewrite (set_eqvUnionSymm M' M); apply set_union_intro2; apply X; discriminate.
  reflexivity.
  intro I; apply (set_disj_elim toFrDisj v0).
  auto.
  subst frSp W; apply set_union_intro1; apply set_union_intro1; assumption.
  (* field 2, same as 1 *)
  clear v0' H3 F1ok; rename H6 into okV; unfold mapX; rename v1' into v0; rename F2ok into F1ok.
  assert (X0 := isOddOk v0); destruct (isOdd v0).
  destruct Iso12 as [Iso12 _]; destruct (Iso12 v0) as [_ Q].
  replace (map_union phi phi' v0) with (None (A := nat)).
  reflexivity.
  destruct (map_union phi phi' v0).
  elf; apply (not_even_and_odd v0); auto; apply ptrPeven; apply Mptr; apply Q; discriminate.
  reflexivity.
  destruct F1ok as [F1ok | F1ok].
  elf; apply (not_even_and_odd v0); assumption.
  destruct okV as [okV | okV].
  elf; apply (not_even_and_odd v0); assumption.
  rewrite (set_mem_true okV).
  rewrite set_mem_false.
  destruct Iso12 as [Iso12 _]; destruct (Iso12 v0) as [_ X]; 
    destruct (map_union phi phi' v0).
  elf; apply (set_disj_elim toFrDisj v0).
  auto.
  subst frSp W; rewrite (set_eqvUnionSymm M' W'); rewrite (set_eqvUnionAssoc W' M' M);
    rewrite (set_eqvUnionSymm M' M); apply set_union_intro2; apply X; discriminate.
  reflexivity.
  intro I; apply (set_disj_elim toFrDisj v0).
  auto.
  subst frSp W; apply set_union_intro1; apply set_union_intro1; assumption.
  intro I; apply (set_disj_elim toFrDisj x).
  subst B; destruct (set_union_elim Ix1) as [I' | I'].
  apply set_union_intro1; apply set_union_intro1; assumption.
  apply set_union_intro2; assumption.
  subst frSp W; apply set_union_intro1; apply set_union_intro1; assumption.
  (* done with B *)
  unfold hpredTrue; rewrite (set_unionInterDistrib1 B'' G'' objs'); apply fasMerge.
  apply (sconjMapRangeImpl Hp'); clear hh Hp'; intros hh Hp'.
  (* B'' *)
  apply (fasMapRangeImplI Hp'); clear hh Hp';
    unfold objIsToFwded; unfold objIsFwded; unfold fieldIsFwded;
    unfold fieldRespMap; unfold objIsFwdedH;
      stSimpl; intros hh x Ix Hp'; lle Hp'; lli.
  2:simpl; apply (sconjMapRangeImpl Hp'); intros hh' Hp3; apply (contMapRangeImpl Hp3).
  clear hh Hp'; destruct (set_inter_elim Ix) as [Ix1 Ix2]; clear Ix.
  subst objs'; destruct (set_map_elim Ix2) as [y [Xeq Iy]]; clear Ix2.
  subst phi2; cbv beta.
  assert (Iy' := Sub _ Iy).
  destruct (set_union_elim Iy') as [Iy'' | Iy'']; clear Iy'.
  assert (x = y).
  destruct Iso2 as [Iso2 _]; destruct (Iso2 y) as [Iso2a Iso2b]; clear Iso2.
  destruct (phi' y).
  elf; apply (set_disj_elim BGGWMdisj y); auto; apply Iso2b; discriminate.
  assumption.
  clear Xeq; subst y; elf.
  destruct (set_union_elim Iy'') as [I | I]; clear Iy''.
  destruct (set_union_elim I) as [I' | I']; clear I.
  apply (set_disj_elim BGBdisj'' x); auto; apply set_union_intro1; assumption.
  destruct (set_union_elim I') as [I | I]; clear I'.
  apply (set_disj_elim BGBdisj x); auto; apply set_union_intro1; assumption.
  apply (set_disj_elim BGGdisj x); auto; apply set_union_intro1; assumption.
  apply (set_disj_elim BGWdisj x); auto; apply set_union_intro1; assumption.
  assert (Peq : phi' y = Some x).
  destruct Iso2 as [Iso2 _]; destruct (Iso2 y) as [Q _];
    hypReplace Q (Q Iy'').
  destruct (phi' y).
  subst a; reflexivity.
  elf; apply Q; reflexivity.
  clear Xeq.
  assert (x0 = y).
  destruct Iso2 as [_ [_ [Iso2 _]]]; hypReplace Iso2 (Iso2 x0 y x); auto.
  subst x0; clear H1; exists y.
  rewrite (set_mem_true Iy); rewrite (set_mem_true Iy''); rewrite Peq.
  assert (HX : vs |||- top ** (objRespMapH (map_union phi (Some (A:=nat))) (heapOf s) y ** top)).
  clear Vhp2; scm.
  assert (Ix' : set_In y ((B' |_| G') |_| W)).
  apply set_union_intro2; subst W; apply set_union_intro1; assumption.
  hypReplace Hp0 (fasElim Ix' Hp0); scm.
  rename H6 into St1; rename H7 into St2.
  unfold objRespMapH in HX; unfold fieldRespMapH in HX; lle HX.
  destruct St1 as [v0 [Hp0 Meq0]]; destruct St2 as [v1 [Hp1 Meq1]].
  destruct H3 as [v0' [Hp0' Meq0']]; destruct H6 as [v1' [Hp1' Meq1']].
  assert (X1 := contContEq Hp0 Hp0'); assert (X2 := contContEq Hp1 Hp1'); 
    subst v0' v1'; clear Hp0 Hp1 Hp0' Hp1'; rename x0 into v0'; rename x3 into v1'.
  hypReplace HpO (objHpElim Iy HpO); unfold okObjAt in HpO; unfold okFieldAt in HpO;
    lle HpO.
  hypReplace Vhp2 (sconjInsert Vhp2 HpO); lle Vhp2; assert (X1 := contContEq Vhp2 HX);
    sconjLift2 Vhp2; sconjLift2 HX; assert (X2 := contContEq Vhp2 HX);
      clear HX Vhp2; subst x0 x3.
  exists v0'; exists v1'; constructor.
  reflexivity.
  constructor.
  lli; scm.
  clear x y Iy Iy'' Ix1 HpO Peq; constructor.

  (* first field *)
  clear Meq1 Meq1' v1 v1' H6 x2; rename x1 into v0''.
  unfold mapX in Meq0, Meq0' |- * .
  assert (X0 := isOddOk v0); destruct (isOdd v0).
  injection Meq0; injection Meq0'; clear Meq0 Meq0'; intros; subst v0' v0''.
  rewrite (isOddTrue X0).
  destruct Iso12 as [Iso12 _]; destruct (Iso12 v0) as [_ Q].
  replace (map_union phi phi' v0) with (None (A := nat)).
  reflexivity.
  destruct (map_union phi phi' v0).
  elf; apply (not_even_and_odd v0); auto; apply ptrPeven; apply Mptr; apply Q; discriminate.
  reflexivity.
  unfold map_union in Meq0, Meq0' |- * .
  assert (D : phi v0 = Some v0' \/ v0 = v0').
  destruct (phi v0).
  left; assumption.
  right; injection Meq0'; auto.
  clear Meq0'.
  assert (X := set_memSpaceIdMap toStart toEnd v0).
  destruct D as [D | D].
  assert (I0 : set_In v0 M).
  destruct Iso1 as [Iso1 _]; destruct (Iso1 v0) as [_ Q]; apply Q; rewrite D; discriminate.
  assert (I0' : set_In v0' (B0 |_| B' |_| G')).
  destruct Iso1 as [_ [Iso1 _]]; eauto.
  assert (E' : even v0').
  apply ptrPeven; eauto.
  destruct H3 as [Y | I0'2].
  elf; apply (not_even_and_odd v0'); assumption.
  rewrite (isOddFalse E'); rewrite (set_mem_true I0'2).
  assert (NI : ~set_In v0' M').
  intro I; apply (set_disj_elim toFrDisj v0').
  apply set_union_intro1; assumption.
  subst frSp; subst W; apply set_union_intro1; apply set_union_intro1; assumption.
  rewrite (set_mem_false NI).
  rewrite set_mem_false in X.
  rewrite X in Meq0; clear X; rewrite D in Meq0.
  injection Meq0; clear Meq0; intro; subst v0''.
  assert (map_union phi phi' v0' = None).
  destruct Iso12 as [Iso12 _]; destruct (Iso12 v0') as [Y Z].
  destruct (map_union phi phi' v0').
  elf; apply (set_disj_elim toFrDisj v0').
  apply set_union_intro1; assumption.
  assert (U : set_In v0' (M |_| M')).
  apply Z; discriminate.
  subst frSp; destruct (set_union_elim U) as [Z' | Z']; clear U Z.
  apply set_union_intro2; assumption.
  apply set_union_intro1; subst W; apply set_union_intro1; assumption.
  reflexivity.
  unfold map_union in H; rewrite H; reflexivity.
  intro IX; elf; apply (set_disj_elim frToDisj v0);
    [subst frSp; apply set_union_intro2 | idtac]; assumption.

  subst v0'.
  rewrite (isOddFalse X0); destruct H3 as [Q | Iv0].
  elf; apply (not_even_and_odd v0); assumption.
  rewrite (set_mem_true Iv0).
  hypReplace Iv0 (Sub _ Iv0).
  assert (II' : set_In v0 ((B |_| B' |_| G') |_| (W' |_| M'))).
  rewrite <- (set_eqvUnionAssoc (B |_| B' |_| G') W' M'); assumption.
  clear X0 Iv0; destruct (set_union_elim II') as [II | II]; clear II'.
  rewrite set_mem_false.
  replace v0'' with v0.
  clear Meq0.
  assert (map_union phi phi' v0 = None).
  destruct Iso12 as [Iso12 _]; destruct (Iso12 v0) as [Y Z].
  destruct (map_union phi phi' v0).
  elf; apply (set_disj_elim toFrDisj v0).
  subst B; set_InElim.
  apply set_union_intro1; apply set_union_intro1; assumption.
  apply set_union_intro2; assumption.
  apply set_union_intro1; apply set_union_intro2; apply set_union_intro1; assumption.
  apply set_union_intro1; apply set_union_intro2; apply set_union_intro2; assumption.
  assert (U : set_In v0 (M |_| M')).
  apply Z; discriminate.
  subst frSp; destruct (set_union_elim U) as [Z' | Z']; clear U Z.
  apply set_union_intro2; assumption.
  apply set_union_intro1; subst W; apply set_union_intro1; assumption.
  reflexivity.
  unfold map_union in H; rewrite H; reflexivity.
  rewrite set_mem_true in X.
  rewrite X in Meq0.
  injection Meq0; intro; assumption.
  apply objRangeSetRangeSubset; rewrite <- toSpEqv; 
    subst toSp; apply set_union_intro1; assumption.
  intro III; apply (set_disj_elim BGGWMdisj v0); auto;
    apply set_union_intro1; assumption.
  rewrite set_mem_false in X.
  rewrite X in Meq0; clear X.
  assert (Peq : phi v0 = None).
  destruct Iso1 as [Iso1 _]; destruct (Iso1 v0) as [W1 W2].
  destruct (phi v0).
  assert (IM : set_In v0 M).
  apply W2; discriminate.
  elf; apply (set_disj_elim WMdisj0 v0); auto.
  subst W; rewrite (set_eqvUnionSymm M' W'); assumption.
  reflexivity.
  rewrite Peq in Meq0; clear Peq.
  rewrite Meq0.
  assert (I'' : set_In v0'' (B'' |_| G'')).
  destruct Iso2 as [_ [Iso2 _]]; exact (Iso2 _ _ Meq0).
  rename v0'' into v0'; assert (V : map_union phi phi' v0' = None).
  destruct Iso12 as [Iso12 _]; destruct (Iso12 v0') as [Y Z].
  destruct (map_union phi phi' v0').
  assert (U : set_In v0' (M |_| M')).
  apply Z; discriminate.
  elf; subst frSp; destruct (set_union_elim U) as [Z' | Z']; clear U Z.
  apply (set_disj_elim BGMdisj v0'); assumption.
  apply (set_disj_elim BGMdisj' v0'); assumption.
  reflexivity.
  unfold map_union in V; rewrite V; clear V.
  rewrite set_mem_true.
  reflexivity.
  destruct Iso2 as [Iso2 _]; destruct (Iso2 v0) as [_ X]; apply X; rewrite Meq0; discriminate.
  intro I; apply (set_disj_elim frToDisj v0); auto.
  subst frSp W; apply set_union_intro1;
    rewrite (set_eqvUnionSymm M' W'); assumption.

  (* second field *)
  clear Meq0 Meq0' v0 v0' H3 x1.
  rename Meq1 into Meq0; rename Meq1' into Meq0'; rename v1 into v0; 
    rename v1' into v0'; rename H6 into H3; rename x2 into v0''.
  unfold mapX in Meq0, Meq0' |- * .
  assert (X0 := isOddOk v0); destruct (isOdd v0).
  injection Meq0; injection Meq0'; clear Meq0 Meq0'; intros; subst v0' v0''.
  rewrite (isOddTrue X0).
  destruct Iso12 as [Iso12 _]; destruct (Iso12 v0) as [_ Q].
  replace (map_union phi phi' v0) with (None (A := nat)).
  reflexivity.
  destruct (map_union phi phi' v0).
  elf; apply (not_even_and_odd v0); auto; apply ptrPeven; apply Mptr; apply Q; discriminate.
  reflexivity.
  unfold map_union in Meq0, Meq0' |- * .
  assert (D : phi v0 = Some v0' \/ v0 = v0').
  destruct (phi v0).
  left; assumption.
  right; injection Meq0'; auto.
  clear Meq0'.
  assert (X := set_memSpaceIdMap toStart toEnd v0).
  destruct D as [D | D].
  assert (I0 : set_In v0 M).
  destruct Iso1 as [Iso1 _]; destruct (Iso1 v0) as [_ Q]; apply Q; rewrite D; discriminate.
  assert (I0' : set_In v0' (B0 |_| B' |_| G')).
  destruct Iso1 as [_ [Iso1 _]]; eauto.
  assert (E' : even v0').
  apply ptrPeven; eauto.
  destruct H3 as [Y | I0'2].
  elf; apply (not_even_and_odd v0'); assumption.
  rewrite (isOddFalse E'); rewrite (set_mem_true I0'2).
  assert (NI : ~set_In v0' M').
  intro I; apply (set_disj_elim toFrDisj v0').
  apply set_union_intro1; assumption.
  subst frSp; subst W; apply set_union_intro1; apply set_union_intro1; assumption.
  rewrite (set_mem_false NI).
  rewrite set_mem_false in X.
  rewrite X in Meq0; clear X; rewrite D in Meq0.
  injection Meq0; clear Meq0; intro; subst v0''.
  assert (map_union phi phi' v0' = None).
  destruct Iso12 as [Iso12 _]; destruct (Iso12 v0') as [Y Z].
  destruct (map_union phi phi' v0').
  elf; apply (set_disj_elim toFrDisj v0').
  apply set_union_intro1; assumption.
  assert (U : set_In v0' (M |_| M')).
  apply Z; discriminate.
  subst frSp; destruct (set_union_elim U) as [Z' | Z']; clear U Z.
  apply set_union_intro2; assumption.
  apply set_union_intro1; subst W; apply set_union_intro1; assumption.
  reflexivity.
  unfold map_union in H; rewrite H; reflexivity.
  intro IX; elf; apply (set_disj_elim frToDisj v0);
    [subst frSp; apply set_union_intro2 | idtac]; assumption.

  subst v0'.
  rewrite (isOddFalse X0); destruct H3 as [Q | Iv0].
  elf; apply (not_even_and_odd v0); assumption.
  rewrite (set_mem_true Iv0).
  hypReplace Iv0 (Sub _ Iv0).
  assert (II' : set_In v0 ((B |_| B' |_| G') |_| (W' |_| M'))).
  rewrite <- (set_eqvUnionAssoc (B |_| B' |_| G') W' M'); assumption.
  clear X0 Iv0; destruct (set_union_elim II') as [II | II]; clear II'.
  rewrite set_mem_false.
  replace v0'' with v0.
  clear Meq0.
  assert (map_union phi phi' v0 = None).
  destruct Iso12 as [Iso12 _]; destruct (Iso12 v0) as [Y Z].
  destruct (map_union phi phi' v0).
  elf; apply (set_disj_elim toFrDisj v0).
  subst B; set_InElim.
  apply set_union_intro1; apply set_union_intro1; assumption.
  apply set_union_intro2; assumption.
  apply set_union_intro1; apply set_union_intro2; apply set_union_intro1; assumption.
  apply set_union_intro1; apply set_union_intro2; apply set_union_intro2; assumption.
  assert (U : set_In v0 (M |_| M')).
  apply Z; discriminate.
  subst frSp; destruct (set_union_elim U) as [Z' | Z']; clear U Z.
  apply set_union_intro2; assumption.
  apply set_union_intro1; subst W; apply set_union_intro1; assumption.
  reflexivity.
  unfold map_union in H; rewrite H; reflexivity.
  rewrite set_mem_true in X.
  rewrite X in Meq0.
  injection Meq0; intro; assumption.
  apply objRangeSetRangeSubset; rewrite <- toSpEqv; 
    subst toSp; apply set_union_intro1; assumption.
  intro III; apply (set_disj_elim BGGWMdisj v0); auto;
    apply set_union_intro1; assumption.
  rewrite set_mem_false in X.
  rewrite X in Meq0; clear X.
  assert (Peq : phi v0 = None).
  destruct Iso1 as [Iso1 _]; destruct (Iso1 v0) as [W1 W2].
  destruct (phi v0).
  assert (IM : set_In v0 M).
  apply W2; discriminate.
  elf; apply (set_disj_elim WMdisj0 v0); auto.
  subst W; rewrite (set_eqvUnionSymm M' W'); assumption.
  reflexivity.
  rewrite Peq in Meq0; clear Peq.
  rewrite Meq0.
  assert (I'' : set_In v0'' (B'' |_| G'')).
  destruct Iso2 as [_ [Iso2 _]]; exact (Iso2 _ _ Meq0).
  rename v0'' into v0'; assert (V : map_union phi phi' v0' = None).
  destruct Iso12 as [Iso12 _]; destruct (Iso12 v0') as [Y Z].
  destruct (map_union phi phi' v0').
  assert (U : set_In v0' (M |_| M')).
  apply Z; discriminate.
  elf; subst frSp; destruct (set_union_elim U) as [Z' | Z']; clear U Z.
  apply (set_disj_elim BGMdisj v0'); assumption.
  apply (set_disj_elim BGMdisj' v0'); assumption.
  reflexivity.
  unfold map_union in V; rewrite V; clear V.
  rewrite set_mem_true.
  reflexivity.
  destruct Iso2 as [Iso2 _]; destruct (Iso2 v0) as [_ X]; apply X; rewrite Meq0; discriminate.
  intro I; apply (set_disj_elim frToDisj v0); auto.
  subst frSp W; apply set_union_intro1;
    rewrite (set_eqvUnionSymm M' W'); assumption.
  (* done with B'' *)

  (* G'' *)
  apply (fasMapRangeImplI Hp'); clear hh Hp';
    unfold objCopiedPhi; unfold objCopied; unfold fieldCopied; unfold objIsFwdedH;
      stSimpl; intros hh x Ix Hp'; lle Hp'; lli.
  2:simpl; apply (sconjMapRangeImpl Hp'); intros hh' Hp3; apply (contMapRangeImpl Hp3).
  clear hh Hp'; destruct (set_inter_elim Ix) as [Ix1 Ix2]; clear Ix.
  subst objs'; destruct (set_map_elim Ix2) as [y [Xeq Iy]]; clear Ix2.
  subst phi2; cbv beta.
  exists x0; rewrite H1.
  assert (D : phi' y = Some x \/ y = x).
  destruct (phi' y).
  subst x; left; reflexivity.
  right; auto.
  clear Xeq.
  assert (Ix : set_In x (B'' |_| G'')).
  destruct Iso2 as [_ [Iso2 _]]; eauto.
  assert (I0 : set_In x0 M').
  destruct Iso2 as [Iso2 _]; destruct (Iso2 x0) as [_ X]; apply X; rewrite H1; discriminate.
  assert (HX : vs |||- top ** (objRespMapH (map_union phi (Some (A:=nat))) (heapOf s) x0 ** top)).
  clear Vhp2; scm.
  assert (Ix' : set_In x0 ((B' |_| G') |_| W)).
  apply set_union_intro2; subst W; apply set_union_intro1; assumption.
  hypReplace Hp0 (fasElim Ix' Hp0); scm.
  rename H6 into St1; rename H7 into St2.
  unfold objRespMapH in HX; unfold fieldRespMapH in HX; lle HX.
  destruct H4 as [v0' [Hp0' Meq0']]; destruct H7 as [v1' [Hp1' Meq1']].
  rename x1 into v0; rename x2 into v1.
  assert (X1 := contContEq St1 Hp0'); assert (X2 := contContEq St2 Hp1'); 
    subst v0' v1'; clear St1 St2 Hp0' Hp1'; rename x3 into v0'; rename x4 into v1'.
  rename x0 into y0.
  hypReplace HpO (objHpElim Iy HpO); unfold okObjAt in HpO; unfold okFieldAt in HpO;
    lle HpO.
  hypReplace Vhp2 (sconjInsert Vhp2 HpO); lle Vhp2.
  destruct D as [Peq | Xeq].
  assert (y0 = y).
  destruct Iso2 as [_ [_ [Iso2 _]]]; eauto.
  subst y0; clear H1.
  assert (X1 := contContEq Vhp2 HX); sconjLift2 Vhp2; sconjLift2 HX; 
    assert (X2 := contContEq Vhp2 HX); clear HX Vhp2; subst x0 x1.
  rewrite (set_mem_true Iy); rewrite (set_mem_true I0).
  exists v0'; exists v1'; constructor.
  reflexivity.
  constructor.
  lli; scm.
  clear h HpO y I0 Peq Iy x Ix1 Ix.
  constructor.
  (* first field *)
  clear v1 v1' Meq1' H7; rename H4 into okF; rename Meq0' into Meq0.
  unfold mapX in Meq0 |- * .
  assert (IO := isOddOk v0); destruct (isOdd v0).
  injection Meq0; clear Meq0; intro; subst v0'; rewrite (isOddTrue IO).
  replace (map_union phi phi' v0) with (None (A := nat)).
  reflexivity.
  destruct Iso12 as [Iso12 _]; destruct (Iso12 v0) as [_ X];
    destruct (map_union phi phi' v0).
  elf; apply (not_even_and_odd v0).
  apply ptrPeven; apply Mptr; apply X; discriminate.
  assumption.
  reflexivity.
  assert (IO' := isOddOk v0'); destruct (isOdd v0').
  clear okF; unfold map_union in Meq0.
  unfold map_union; destruct (phi v0).
  symmetry; assumption.
  injection Meq0; intro; subst v0'; elf; apply (not_even_and_odd v0); assumption.
  destruct okF as [okF | okF].
  elf; apply (not_even_and_odd v0'); assumption.
  clear IO IO'.
  rewrite (set_mem_true okF).
  hypReplace okF (Sub _ okF).
  unfold map_union in Meq0 |- * .
  assert (D : phi v0 = Some v0' \/ v0 = v0').
  destruct (phi v0); [left | right]; auto.
  injection Meq0; auto.
  clear Meq0; destruct D as [Peq | Veq].
  rewrite Peq.
  assert (I' : set_In v0' (B0 |_| B' |_| G')).
  destruct Iso1 as [_ [Iso1 _]]; eauto.
  destruct Iso2 as [_ [Iso2b _]].
  rewrite set_mem_false.
  reflexivity.
  intro I; apply (set_disj_elim CMdisj v0').
  assumption.
  subst frSp; subst W; apply set_union_intro1; apply set_union_intro1; assumption.
  subst v0'.
  replace (phi v0) with (None (A := nat)).
  destruct (set_union_elim okF) as [I | I]; clear okF.
  cut (~set_In v0 M').
  intro NI; rewrite (set_mem_false NI).
  destruct Iso2 as [Iso2 _]; destruct (Iso2 v0) as [_ X]; destruct (phi' v0).
  elf; apply NI; apply X; discriminate.
  reflexivity.
  intro I'; apply (set_disj_elim BGGWMdisj v0); assumption.
  rewrite (set_mem_true I); destruct Iso2 as [Iso2 _]; destruct (Iso2 v0) as [X _].
  hypReplace X (X I); destruct (phi' v0).
  reflexivity.
  elf; apply X; reflexivity.
  destruct Iso1 as [Iso1 _]; destruct (Iso1 v0) as [_ X]; clear Iso1.
  destruct (phi v0).
  2:reflexivity.
  assert (IM : set_In v0 M).
  apply X; discriminate.
  clear X.
  assert (II' : set_In v0 ((B |_| B' |_| G') |_| (W' |_| M'))).
  rewrite <- (set_eqvUnionAssoc (B |_| B' |_| G') W' M'); assumption.
  clear okF; destruct (set_union_elim II') as [II | II]; clear II'.
  elf; apply (set_disj_elim toFrDisj v0).
  subst B; set_InElim.
  apply set_union_intro1; apply set_union_intro1; assumption.
  apply set_union_intro2; assumption.
  apply set_union_intro1; apply set_union_intro2; apply set_union_intro1; assumption.
  apply set_union_intro1; apply set_union_intro2; apply set_union_intro2; assumption.
  subst frSp; apply set_union_intro2; assumption.
  elf; apply (set_disj_elim WMdisj0 v0).
  subst W; rewrite (set_eqvUnionSymm M' W'); assumption.
  assumption.
  (* second field *)
  clear v0 v0' Meq0' H4; 
    rename v1 into v0; rename v1' into v0'; rename Meq1' into Meq0; rename
      H7 into okF.
  unfold mapX in Meq0 |- * .
  assert (IO := isOddOk v0); destruct (isOdd v0).
  injection Meq0; clear Meq0; intro; subst v0'; rewrite (isOddTrue IO).
  replace (map_union phi phi' v0) with (None (A := nat)).
  reflexivity.
  destruct Iso12 as [Iso12 _]; destruct (Iso12 v0) as [_ X];
    destruct (map_union phi phi' v0).
  elf; apply (not_even_and_odd v0).
  apply ptrPeven; apply Mptr; apply X; discriminate.
  assumption.
  reflexivity.
  assert (IO' := isOddOk v0'); destruct (isOdd v0').
  clear okF; unfold map_union in Meq0.
  unfold map_union; destruct (phi v0).
  symmetry; assumption.
  injection Meq0; intro; subst v0'; elf; apply (not_even_and_odd v0); assumption.
  destruct okF as [okF | okF].
  elf; apply (not_even_and_odd v0'); assumption.
  clear IO IO'.
  rewrite (set_mem_true okF).
  hypReplace okF (Sub _ okF).
  unfold map_union in Meq0 |- * .
  assert (D : phi v0 = Some v0' \/ v0 = v0').
  destruct (phi v0); [left | right]; auto.
  injection Meq0; auto.
  clear Meq0; destruct D as [Peq | Veq].
  rewrite Peq.
  assert (I' : set_In v0' (B0 |_| B' |_| G')).
  destruct Iso1 as [_ [Iso1 _]]; eauto.
  destruct Iso2 as [_ [Iso2b _]].
  rewrite set_mem_false.
  reflexivity.
  intro I; apply (set_disj_elim CMdisj v0').
  assumption.
  subst frSp; subst W; apply set_union_intro1; apply set_union_intro1; assumption.
  subst v0'.
  replace (phi v0) with (None (A := nat)).
  destruct (set_union_elim okF) as [I | I]; clear okF.
  cut (~set_In v0 M').
  intro NI; rewrite (set_mem_false NI).
  destruct Iso2 as [Iso2 _]; destruct (Iso2 v0) as [_ X]; destruct (phi' v0).
  elf; apply NI; apply X; discriminate.
  reflexivity.
  intro I'; apply (set_disj_elim BGGWMdisj v0); assumption.
  rewrite (set_mem_true I); destruct Iso2 as [Iso2 _]; destruct (Iso2 v0) as [X _].
  hypReplace X (X I); destruct (phi' v0).
  reflexivity.
  elf; apply X; reflexivity.
  destruct Iso1 as [Iso1 _]; destruct (Iso1 v0) as [_ X]; clear Iso1.
  destruct (phi v0).
  2:reflexivity.
  assert (IM : set_In v0 M).
  apply X; discriminate.
  clear X.
  assert (II' : set_In v0 ((B |_| B' |_| G') |_| (W' |_| M'))).
  rewrite <- (set_eqvUnionAssoc (B |_| B' |_| G') W' M'); assumption.
  clear okF; destruct (set_union_elim II') as [II | II]; clear II'.
  elf; apply (set_disj_elim toFrDisj v0).
  subst B; set_InElim.
  apply set_union_intro1; apply set_union_intro1; assumption.
  apply set_union_intro2; assumption.
  apply set_union_intro1; apply set_union_intro2; apply set_union_intro1; assumption.
  apply set_union_intro1; apply set_union_intro2; apply set_union_intro2; assumption.
  subst frSp; apply set_union_intro2; assumption.
  elf; apply (set_disj_elim WMdisj0 v0).
  subst W; rewrite (set_eqvUnionSymm M' W'); assumption.
  assumption.

  (* another case *)
  subst y; clear Ix; elf; hypReplace Iy (Sub _ Iy).
  destruct (set_union_elim Iy) as [I | I]; clear Iy.
  destruct (set_union_elim I) as [I' | I']; clear I.
  destruct (set_union_elim I') as [I | I]; clear I'.
  destruct (set_union_elim I) as [I' | I']; clear I.
  apply (set_disj_elim BGBdisj'' x).
  apply set_union_intro2; assumption.
  subst B; apply set_union_intro1; assumption.
  apply (set_disj_elim BGBdisj'' x).
  apply set_union_intro2; assumption.
  subst B; apply set_union_intro2; assumption.
  destruct (set_union_elim I) as [I' | I']; clear I.
  apply (set_disj_elim BGBdisj x).
  apply set_union_intro2; assumption.
  assumption.
  apply (set_disj_elim BGGdisj x).
  apply set_union_intro2; assumption.
  assumption.
  apply (set_disj_elim BGWdisj x).
  apply set_union_intro2; assumption.
  assumption.
  apply (set_disj_elim BGMdisj' x).
  apply set_union_intro2; assumption.
  assumption.

  (* now we have to prove all of the disjointedness lemmas *)

  apply (fasFasDisj' Hp').
  unfold objIsToFwded; unfold objIsFwded; unfold fieldIsFwded; cbv beta; illeim.
  unfold objCopiedPhi; unfold objCopied; unfold fieldCopied; cbv beta; illeim.
  assert (X : hh ||- (foralls x in set_inter B objs', x |->_ ** top) **
    (foralls x in set_inter (B'' |_| G'') objs', x |->_ ** top)).
  scm.
  apply (fasImplI Hp0); unfold objCopied; unfold fieldCopied; cbv beta; illeim.
  assert (X : h0 ||- (foralls x in set_inter B'' objs', x |->_ ** top) **
    (foralls x in set_inter G'' objs', x |->_ ** top)).
  scm; apply (fasImplI Hp0).
  unfold objIsToFwded; unfold objIsFwded; unfold fieldIsFwded; cbv beta; illeim.
  unfold objCopiedPhi; unfold objCopied; unfold fieldCopied; cbv beta; illeim.
  unfold hpredTrue; rewrite (set_unionInterDistrib1 B'' G'' objs'); 
    apply fasMerge.
  scm.
  apply (fasFasDisj' X); auto.
  apply (fasFasDisj' X); auto.
  rewrite (set_unionInterDistrib1 B (B'' |_| G'') objs');
    rewrite (set_unionInterDistrib1 B'' G'' objs').
  assert (X : hh ||- (foralls x in set_inter G' objs', x |->_ ** top) **
    (foralls x in (set_inter B objs' |_| set_inter B'' objs' |_| set_inter G'' objs'), 
      x |->_ ** top)).
  unfold objIsToFwded in Hp'; unfold objIsFwded in Hp'; unfold fieldIsFwded in Hp';
    unfold objCopiedPhi in Hp'; unfold objCopied in Hp'; unfold fieldCopied in Hp'.
  assert (X : hh ||- (foralls x in set_inter G' objs', x |->_ ** top) **
    (foralls x in set_inter B'' objs', x |->_ ** top) **
    (foralls x in set_inter G'' objs', x |->_ ** top) **
    (foralls x in set_inter B objs', x |->_ ** top)).
  scm; apply (fasImplI Hp0); cbv beta; illeim.
  clear Hp' Hp.
  scm; sconjLift3of3 Hp; apply fasMerge; scm.
  apply fasMerge; scm.
  apply (fasFasDisj' Hp); auto.
  assert (X : h0
       ||- (foralls x in set_inter B objs', x |->_ ** top) **
           (foralls x in (set_inter B'' objs') |_|
	     (set_inter G'' objs'), x |->_ ** top)).
  scm; apply fasMerge.
  scm.
  apply (fasFasDisj' Hp); auto.
  apply (fasFasDisj' X); auto.
  apply (fasFasDisj' X); auto.
  assert (X : hh ||- (foralls x in set_inter B' objs', x |->_ ** top) **
            (foralls x in set_inter G' objs', x |->_ ** top) **
            (foralls x in set_inter B'' objs', x |->_ ** top) **
            (foralls x in set_inter G'' objs', x |->_ ** top) **
            (foralls x in set_inter B objs', x |->_ ** top)).
  unfold objIsToFwded in Hp'; unfold objIsFwded in Hp'; unfold fieldIsFwded in Hp';
    unfold objCopiedPhi in Hp'; unfold objCopied in Hp'; unfold fieldCopied in Hp';
      unfold objRespToMap in Hp'; unfold fieldRespToMap in Hp'; unfold fieldRespMap in Hp'.
  scm; apply (fasImplI Hp0); cbv beta; illeim.
  clear Hp Hp'.
  assert (Y : hh ||- (foralls x in set_inter B' objs', x |->_ ** top) **
    (foralls x in (set_inter G' objs') |_| (set_inter B'' objs') 
      |_| (set_inter G'' objs') |_| (set_inter B objs'), x |->_ ** top)).
  scm; apply fasMerge.
  scm.
  apply fasMerge.
  scm.
  apply fasMerge.
  scm.
  apply (fasFasDisj' Hp); auto.
  eapply fasFasDisj'.
  instantiate (3 := h0); scm.
  apply fasMerge.
  scm.
  apply (fasFasDisj' Hp); auto.
  auto.
  auto.
  assert (Y : h0 ||- (foralls x in set_inter G' objs', x |->_ ** top) **
    (foralls x in (set_inter B'' objs') 
      |_| (set_inter G'' objs') |_| (set_inter B objs'), x |->_ ** top)).
  scm; apply fasMerge.
  scm.
  apply fasMerge.
  scm.
  apply (fasFasDisj' Hp); auto.
  assert (Y : h0 ||- (foralls x in set_inter B'' objs', x |->_ ** top) **
    (foralls x in (set_inter G'' objs') |_| (set_inter B objs'), x |->_ ** top)).
  scm.
  apply fasMerge.
  scm.
  apply (fasFasDisj' Hp); auto.
  apply (fasFasDisj' Y); auto.
  apply (fasFasDisj' Y); auto.
  setoid_replace (G' |_| B |_| B'' |_| G'') with
    (G' |_| B'' |_| G'' |_| B).

  rewrite (set_unionInterDistrib1 G' (B'' |_| G'' |_| B) objs');
    rewrite (set_unionInterDistrib1 B'' (G'' |_| B) objs');
      rewrite (set_unionInterDistrib1 G'' B objs').
  apply (fasFasDisj' Y); auto.

  constructor; intro I; set_InElim.
  apply set_union_intro1; assumption.
  apply set_union_intro2; apply set_union_intro2; apply set_union_intro2; assumption.
  apply set_union_intro2; apply set_union_intro1; assumption.
  apply set_union_intro2; apply set_union_intro2; apply set_union_intro1; assumption.
  apply set_union_intro1; assumption.
  apply set_union_intro2; apply set_union_intro2; apply set_union_intro1; assumption.
  apply set_union_intro2; apply set_union_intro2; apply set_union_intro2; assumption.  
  apply set_union_intro2; apply set_union_intro1; assumption.

  assert (X : hh ||- (foralls x in set_inter W' objs', x |->_ ** top) **
    (foralls x in set_inter B' objs', x |->_ ** top) **
    (foralls x in set_inter G' objs', x |->_ ** top) **
    (foralls x in set_inter B'' objs', x |->_ ** top) **
    (foralls x in set_inter G'' objs', x |->_ ** top) **
    (foralls x in set_inter B objs', x |->_ ** top)).
  unfold objIsToFwded in Hp'; unfold objIsFwded in Hp'; unfold fieldIsFwded in Hp';
    unfold objCopiedPhi in Hp'; unfold objCopied in Hp'; unfold fieldCopied in Hp';
      unfold objRespToMap in Hp'; unfold fieldRespToMap in Hp'; unfold fieldRespMap in Hp'.
  scm; apply (fasImplI Hp0); cbv beta; illeim.
  clear Hp Hp'.
  assert (X' : hh ||- (foralls x in set_inter W' objs', x |->_ ** top) **
    (foralls x in set_inter G objs', x |->_ ** top) **
    (foralls x in set_inter B'' objs', x |->_ ** top) **
    (foralls x in set_inter G'' objs', x |->_ ** top) **
    (foralls x in set_inter B objs', x |->_ ** top)).
  subst G.
  sconjLift3 X; sconjLift3 X; hypReplace X (sconjAssocR X);
  sconjLift2 X; scm.
  unfold hpredTrue; rewrite (set_unionInterDistrib1 B' G' objs'); apply fasMerge.
  scm.
  apply (fasFasDisj' Hp); auto.
  clear X; rename X' into X.
  setoid_replace (((B0 |_| G) |_| B1) |_| B'' |_| G'') with
    (G |_| B'' |_| G'' |_| B).

  rewrite (set_unionInterDistrib1 G (B'' |_| G'' |_| B) objs');
    rewrite (set_unionInterDistrib1 B'' (G'' |_| B) objs');
      rewrite (set_unionInterDistrib1 G'' B objs').
  eapply fasFasDisj' with (h := hh).
  scm.
  apply fasMerge.
  scm.
  apply fasMerge.
  scm.
  apply fasMerge.
  scm.
  apply (fasFasDisj' Hp); auto.
  eapply fasFasDisj' with (h := h0).
  scm.
  apply fasMerge.
  scm.
  apply (fasFasDisj' Hp); auto.
  auto.
  auto.
  eapply fasFasDisj' with (h := h0).
  scm.
  apply fasMerge.
  scm.
  apply fasMerge.
  scm.
  apply (fasFasDisj' Hp); auto.
  eapply fasFasDisj' with (h := h0).
  scm.
  apply fasMerge.
  scm.
  apply (fasFasDisj' Hp); auto.
  auto.
  auto.
  auto.
  auto.
  auto.
  auto.

  subst G B; constructor; intro I; set_InElim.
  sui2; sui2; sui2; sui1; auto.
  sui1; sui1; auto.
  sui1; sui2; auto.
  sui2; sui2; sui2; sui2; auto.
  sui2; sui1; auto.
  sui2; sui2; sui1; auto.
  sui1; sui1; sui2; sui1; auto.
  sui1; sui1; sui2; sui2; auto.
  sui2; sui1; auto.
  sui2; sui2; auto.
  sui1; sui1; sui1; auto.
  sui1; sui2; auto.

Qed.


Lemma grayObjExtract : forall B B' G' F' free scan alloc free' scan',
  let F := ((B' |_| G') |_| F') in
  let G := rangeObjs scan free in
  set_eqv F (rangeObjs free alloc) ->
  set_eqv F' (rangeObjs free' alloc) ->
  set_eqv (B |_| B') (rangeObjs scan scan') ->
  let G'' := rangeObjs scan' free' in
  aligned8 scan scan' ->
  aligned8 scan' free' ->
  aligned8 free free' ->
  aligned8 free' alloc ->
  aligned8 scan free ->
  set_disj G' (B |_| B') ->
  set_disj G' F' ->
  set_eqv G'' ((set_diff G (B |_| B')) |_| G').
Proof.
  intros B B' G' F' free scan alloc free' scan' F G Feq F'eq BBeq G'' 
    al1 al2 al3 al4 al5 Disj1 Disj2.
  constructor; intro Ia.
  
  subst G''; destruct (rangeObjsElim Ia) as [LeA [k EqA]]; clear Ia.
  assert (NI : ~set_In a (B |_| B')).
  intro I.
  assert (Ia : set_In a (rangeObjs scan scan')).
  rewrite <- BBeq; assumption.
  clear I.
  destruct (rangeObjsElim Ia) as [LeA' [k' EqA']]; clear Ia.
  omega.
  assert (I : set_In a (G |_| F)).
  subst G; rewrite Feq.
  assert (LE : free <= alloc).
  destruct al3; destruct al4; omega.
  rewrite (rangeObjsTransEqv al5 LE); clear LE.
  destruct al1 as [k' al1].
  eapply rangeObjsIntro.
  destruct al3; destruct al4; omega.
  subst a scan'; instantiate (1 := k' + k).
  omega.
  set_InElim.
  sui1; apply set_diff_intro; assumption.
  sui2; subst F; set_InElim.
  elf; apply NI; sui2; assumption.
  assert (Ia : set_In a (rangeObjs free' alloc)).
  rewrite <- F'eq; assumption.
  clear H0.
  destruct (rangeObjsElim Ia) as [LeA' [k' EqA']]; clear Ia.
  elf; omega.
  
  (* now the other direction *)
  subst G'' G; set_InElim.
  assert (D : scan' <= free \/ scan' > free).
  omega.
  destruct D.

  assert (alX : aligned8 scan' free).
  destruct al2 as [k' al2]; destruct al3 as [k'' al3].
  exists (k' - k''); omega.
  subst a; destruct al1 as [k' al1]; subst scan'.
  assert (D : k >= k' \/ k < k').
  omega.
  destruct D as [D | D].
  eapply rangeObjsIntro.
  destruct al3; omega.
  instantiate (1 := k - k').
  omega.
  elf.
  apply H1; clear H1; rewrite BBeq; clear BBeq.
  eapply rangeObjsIntro; eauto; omega.
  destruct al1 as [k' al1]; subst a scan'.
  eapply rangeObjsIntro.
  destruct al3; omega.
  instantiate (1 := k - k').
  assert (D : k >= k' \/ k < k').
  omega.
  destruct D as [D | D].
  omega.
  elf; apply H1; clear H1; rewrite BBeq; clear BBeq.
  eapply rangeObjsIntro; eauto; omega.
  
  (* case: set_In a G' *)
  assert (NIB : ~set_In a (rangeObjs scan scan')).
  intro I; apply (set_disj_elim Disj1 a); auto; rewrite BBeq; assumption.
  assert (NIF : ~set_In a (rangeObjs free' alloc)).
  intro I; apply (set_disj_elim Disj2 a); auto; rewrite F'eq; assumption.
  clear Disj1 Disj2.
  assert (Ia : set_In a (rangeObjs free alloc)).
  rewrite <- Feq; subst F;
    apply set_union_intro1; apply set_union_intro2; assumption.
  clear F Feq F'eq BBeq H B B' G' F'.
  assert (X1 := rangeObjsNegElim1 NIB al1); clear NIB.
  assert (X2 := rangeObjsNegElim1 NIF al4); clear NIF.
  destruct (rangeObjsElim Ia) as [LeA' [k' EqA']]; clear Ia.
  assert (LT : a < free').
  apply X2; clear X1 X2.
  destruct al3 as [k al3]; destruct al4 as [k'' al4].
  subst.
  exists (pred ((k + k'') - k')).
  omega.
  clear X2.
  assert (D : free < scan' \/ free >= scan').
  omega.
  destruct D as [D | D].

  assert (alX : aligned8 free scan').
  clear a X1 LeA' EqA' LT k'.
  destruct al1 as [k al1]; destruct al5 as [k' al5].
  subst.
  exists (k - k'); omega.
  clear D.
  destruct alX as [k alX].
  subst scan'.

  eapply rangeObjsIntro.
  destruct al3; omega.
  instantiate (1 := k' - k).
  assert (D : k <= k' \/ k > k').
  omega.
  destruct D as [D | D].
  omega.
  elf.
  assert (alX : aligned8 a (free + 8 * k)).
  exists (k - k'); omega.
  assert (LT' : a = free + 8 * k \/ a < scan).
  destruct alX as [k'' alX]; destruct k''.
  rewrite alX; left; auto.
  right; apply X1; clear X1; rewrite alX; exists k''; reflexivity.
  destruct LT'.
  omega.
  destruct al3; destruct al5; omega.

  assert (alX : aligned8 scan' free).
  clear a X1 LeA' k' EqA' LT.
  destruct al1 as [k al1]; destruct al5 as [k' al5].
  subst scan' free.
  exists (k' - k).
  omega.
  clear D.
  destruct alX as [k alX].
  eapply rangeObjsIntro.
  destruct al3; omega.
  instantiate (1 := k + k').
  omega.
  
Qed.



Ltac hpunfold := unfold objIsToFwded; unfold objIsFwded; unfold fieldIsFwded; 
  unfold objCopiedPhi; unfold objCopied; unfold fieldCopied;
    unfold objRespMapH; unfold fieldRespMapH;
      unfold objCopiedH; unfold fieldCopiedH; unfold objIsFwdedH;
	unfold okObjAt; unfold okFieldAt;
	  unfold objRespToMap; unfold fieldRespToMap; unfold fieldRespMap.

Ltac hpunfoldIn H := unfold objIsToFwded in H; unfold objIsFwded in H; unfold fieldIsFwded in H; 
  unfold objCopiedPhi in H; unfold objCopied in H; unfold fieldCopied in H;
    unfold objRespMapH in H; unfold fieldRespMapH in H;
      unfold objCopiedH in H; unfold fieldCopiedH in H; unfold objIsFwdedH in H;
	unfold okObjAt in H; unfold okFieldAt in H;
	  unfold objRespToMap in H; unfold fieldRespToMap in H; unfold fieldRespMap in H.

Ltac uilleim := hpunfold; illeim.


Lemma heap_eqvCopy : forall h h' A,
  h ||- Heap.heap_eqv h' ** A -> forall s,
  h' ||- (foralls x in s, x |->__) ->
  h' ||- (foralls x in s, objCopiedH x h x).
Proof.
  unfold fas.
  intros h h' A Hp s0 Hp'.
  assert (C := set_canonCanonP s0).
  set (s := set_canon s0) in * .
  generalize dependent h; generalize dependent h'; 
    generalize dependent A; generalize dependent s; clear s0.
  induction 1; simpl; intros.

  assumption.
  
  rename H into Lt.
  lle Hp'; assert (Hp'' := sconjInsert Hp Hp'); lle Hp''.
  unfold objCopiedH at 1; unfold fieldCopiedH at 1 2; lli.
  (* coq bug (?): without instantiate, this makes windows shut down Coq *)
  instantiate (1 := x); scm.
  instantiate (1 := x0).
  2:sconjLift2 Hp''; scm.
  hyr Hp' (sconjAssocR Hp'); apply sconjAssocL.
  apply (sconjImplS Hp'); clear Hp'; intros h'' Sub Hp3.
  assumption.
  apply IHC with top.
  assumption.
  clear Hp'' Hp3.
  assert (HH : h'' ||- Heap.heap_eqv h'').
  apply eqvPredEqv'; apply Heap.heap_eqvRefl.
  assert (HH' := subHeapImpl HH Sub).
  hyr Hp (sconjInsert Hp HH'); lle Hp; scm.
Qed.
Implicit Arguments heap_eqvCopy [h h' A].


Lemma objHpCopied : forall h objs h' A,
  h ||- (foralls x in objs, objCopiedH x h' x) ** A -> forall objs' A',
  h' ||- objHp objs' objs ** A' ->
  h ||- objHp objs' objs ** A.
Proof.
  hpunfold.
  intros h objs h' A Hp objs' A' Hp'.
  unfold objHp; llMatch; hpunfold.
  assert (Hp'' : h' ||- (okObjAt objs' x ** top) ** A').
  scm.
  hypReplace Hp0 (objHpElim I Hp0); scm.
  clear Hp'; hpunfoldIn Hp''; lle Hp''; lle Hp.
  assert (X1 := contContEq H4 Hp'');
    sconjLift2 Hp''; assert (X2 := contContEq H5 Hp'');
      clear H4 H5 Hp''; subst x3 x2.
  lli.
  2:apply Hp.
  assumption.
  assumption.
Qed.


Lemma contContCombine : forall h x x' v v',
  h ||- x |-> v ** top ->
  h ||- x' |-> v' ** top ->
  x <> x' ->
  h ||- x |-> v ** x' |-> v' ** top.
Proof.
  intros h x x' v v' Hp Hp' Ne.

  assert (Hp'' : h ||- (x |->_ ** x' |->_) ** top).  
  assert (Hp'' : h ||- (foralls x in (set_add x empty_set) |_| 
    (set_add x' empty_set), x |->_) ** top).
  apply heapSubDomToCont.
  assert (Sub1 : set_subset (set_add x empty_set) (Heap.heap_dom h)).
  apply (sconjDomSubL Hp); clear h Hp Hp'; intros h Hp.
  assert (X := contDom Hp).
  intros y I; destruct (X y); auto.
  assert (Sub2 : set_subset (set_add x' empty_set) (Heap.heap_dom h)).
  apply (sconjDomSubL Hp'); clear h Hp Hp' Sub1; intros h Hp.
  assert (X := contDom Hp).
  intros y I; destruct (X y); auto.
  intros y I; destruct (set_union_elim I) as [I' | I']; clear I; auto.
  unfold fas in Hp''; simpl in Hp''.
  scm.
  destruct (POnat.POnat.compareDec x x').
  destruct s.
  simpl in Hp0; lleim.
  elf; auto.
  simpl in Hp0; sconjLift2 Hp0; lleim.
  lle Hp''.
  assert (X1 := contContEq Hp Hp''); sconjLift2 Hp''; 
    assert (X2 := contContEq Hp' Hp''); sconjLift2 Hp''; 
      subst x0 x1; assumption.

Qed.


Lemma objHpIso' : forall h objs'' phi s,
  h ||- (foralls x in objs'', objCopiedPhi phi s x) ->
    forall objs0 objs objs' A,
      s |||- objHp objs0 objs ** A ->
	iso_set phi objs objs' ->
	set_subset objs'' objs' ->
	h ||- objHp objs0 objs''.
Proof.
  unfold objHp.
  intros h objs'' phi s Hp objs0 objs objs' A Hp0 Iso Sub.
  llMatch.
  hpunfoldIn Hp; lle Hp.
  assert (Hp' : s |||- (okObjAt objs0 x0 ** top) ** A).
  scm.
  assert (II : set_In x0 objs).
  destruct Iso as [Iso _]; destruct (Iso x0) as [_ Q]; apply Q; 
    rewrite H1; discriminate.
  hyr Hp0 (fasElim II Hp0); scm.
  clear Hp0 H1.
  hpunfoldIn Hp'; lle Hp'.
  assert (X1 := contContEq H6 Hp'); sconjLift2 Hp';
    assert (X2 := contContEq H7 Hp'); clear H6 H7 Hp'; subst x3 x4.
  hpunfold; lli.
  2:eauto.
  assumption.
  assumption.
Qed.


Lemma objHpFwded' : forall s A X h B phi phi' A' B',
  s |||- objHp A A ** X ->
  h ||- (foralls x in B, objIsFwded phi phi' s x) ->
  iso_set phi' A' B' ->
  set_subset A' A ->
  (forall x, set_In x A -> phi x = None) ->
  h ||- objHp B' B.
Proof.
  hpunfold.
  intros s A X h B phi phi' A' B' HpS Hp Iso Sub Peq.
  unfold objHp; llMatch; lle Hp.
  assert (HpS' : s |||- (okObjAt A x0 ** top) ** top).
  assert (I0 : set_In x0 A).
  apply Sub; destruct Iso as [Iso _]; destruct (Iso x0) as [_ Q]; apply Q;
    rewrite H1; discriminate.
  scm; hypReplace Hp0 (objHpElim I0 Hp0); scm.
  hpunfoldIn HpS'; lle HpS'; clear HpS Sub H1; hpunfold; lli.
  2:apply Hp.

  clear x I h Hp H7 H9; rename H4 into okF.
  destruct H6 as [v [HpV Meq]].
  assert (X1 := contContEq HpV HpS').
  subst x3; clear HpV HpS'.
  unfold mapX in Meq.
  assert (IO := isOddOk v); destruct (isOdd v); [left | right].
  injection Meq; intro; subst x1; assumption.
  destruct okF as [okF | okF].
  elf; apply (not_even_and_odd v); assumption.
  unfold map_union in Meq; rewrite (Peq _ okF) in Meq.
  destruct Iso as [_ [Iso _]]; eapply Iso; eauto.

  clear x I h Hp H6 H4 H; rename H9 into okF.
  destruct H7 as [v [HpV Meq]].
  sconjLift2 HpS'; assert (X1 := contContEq HpV HpS').
  subst x4; clear HpV HpS'.
  unfold mapX in Meq.
  assert (IO := isOddOk v); destruct (isOdd v); [left | right].
  injection Meq; intro; subst x2; assumption.
  destruct okF as [okF | okF].
  elf; apply (not_even_and_odd v); assumption.
  unfold map_union in Meq; rewrite (Peq _ okF) in Meq.
  destruct Iso as [_ [Iso _]]; eapply Iso; eauto.

Qed.




Lemma objHpIsFwdedH : forall h h' phi objs objs',
  h ||- objHp objs objs ->
  iso_set phi objs objs' ->
  h' ||- (foralls x in objs', objIsFwdedH phi h x) ->
  h' ||- objHp objs' objs'.
Proof.
  unfold objIsFwdedH.
  intros h h' phi objs objs' Hp Iso Hp'; unfold objHp; llMatch; lle Hp0.
  rename x into x'; rename I into I';
    destruct H1 as [x [v0 [v1 [Peq [Hp' [Meq0 Meq1]]]]]].
  assert (I : set_In x objs).
  destruct Iso as [Iso _]; destruct (Iso x) as [_ X]; apply X; rewrite Peq; discriminate.
  hyr Hp (objHpElim I Hp); hpunfoldIn Hp; lle Hp.
  lle Hp'.
  hpunfold; lli.
  2:simpl in Hp0; apply Hp0.

  clear Peq Hp0 I; clear x1 Meq1 H6; rename H3 into okF.
  assert (X := contContEq Hp Hp'); clear Hp Hp'; subst x2.
  clear x3 v1.
  unfold mapX in Meq0; assert (IO := isOddOk v0); destruct (isOdd v0).
  injection Meq0; intro; subst x0; left; assumption.
  destruct okF as [okF | okF].
  elf; apply (not_even_and_odd v0); assumption.
  right; clear IO.
  destruct Iso as [_ [Iso _]]; eapply Iso; eauto.

  clear Peq Hp0 I; clear x0 Meq0 H3; rename H6 into okF;
    rename x1 into x0; rename Meq1 into Meq0.
  sconjLift2 Hp; sconjLift2 Hp'; 
    assert (X := contContEq Hp Hp'); clear Hp Hp'; subst x3.
  clear x2 v0; rename v1 into v0.
  unfold mapX in Meq0; assert (IO := isOddOk v0); destruct (isOdd v0).
  injection Meq0; intro; subst x0; left; assumption.
  destruct okF as [okF | okF].
  elf; apply (not_even_and_odd v0); assumption.
  right; clear IO.
  destruct Iso as [_ [Iso _]]; eapply Iso; eauto.
  
Qed.







(*

Lemma bakerStViewUnique : forall vh A s,
  bakerStOk vh A s -> forall vh',
    bakerStOk vh' A s ->
    heap_eqv vh vh'.
Proof.
  intros vh A s [frStart [frEnd [toStart [toEnd [scan [free [alloc Hp]]]]]]]
    vh' [frStart' [frEnd' [toStart' [toEnd' [scan' [free' [alloc' Hp']]]]]]].
  destruct Hp as [phi [W [M [Al1 [Al2 [WMeqv [iso [rootOk [Hp Vhp]]]]]]]]].
  destruct Hp' as [phi' [W' [M' [Al1' [Al2' [WMeqv' [iso' [rootOk' [Hp' Vhp']]]]]]]]].
  intro x; unfold Heap.HeapMap.MapDef.A in x; unfold POnat.POnat.A in x.

  clear Al1' Al2' rootOk'.

  (* show all of the GC info is the same *)
  assert (HP := Hp).
  assert (HP' := Hp').
  sconjLift6of6 HP; linLogElim HP.
  sconjLift6of6 HP'; linLogElim HP'.
  assert (X1 := contContEq HP HP').
  sconjLift2 HP; sconjLift2 HP'; assert (X2 := contContEq HP HP').
  sconjLift3 HP; sconjLift3 HP'; assert (X3 := contContEq HP HP').
  sconjLift4 HP; sconjLift4 HP'; assert (X4 := contContEq HP HP').
  sconjLift5 HP; sconjLift5 HP'; assert (X5 := contContEq HP HP').
  sconjLift6 HP; sconjLift6 HP'; assert (X6 := contContEq HP HP').
  sconjLift7 HP; sconjLift7 HP'; assert (X7 := contContEq HP HP').
  clear HP HP'.
  subst frStart' frEnd' toStart' toEnd' scan' free' alloc'.

  assert (EE : set_eqv W W' /\ set_eqv M M' /\ mapEqv phi phi').
  eapply MWmatch3; eauto.
  instantiate (4 := heapOf s).
  sconjMatch.
  sconjMatch.
  setoid_rewrite WMeqv; setoid_rewrite WMeqv'; apply set_eqvRefl.

  destruct EE as [WWeq [MMeq phiEq]].

  assert (D : (exists y, set_In y 
    ((rangeObjs toStart scan |_| rangeObjs alloc toEnd) |_| (rangeObjs scan free |_| W')) /\
    (4 * x = y \/ 4 * x = y + 4)) \/ ~(exists y, set_In y 
    ((rangeObjs toStart scan |_| rangeObjs alloc toEnd) |_| (rangeObjs scan free |_| W')) /\
    (4 * x = y \/ 4 * x = y + 4))).
  destruct (set_In_dec (4 * x) ((rangeObjs toStart scan
         |_| rangeObjs alloc toEnd) |_| (rangeObjs scan free |_| W'))).
  left; exists (4 * x); auto.
  assert (D : 4 * x >= 4 \/ 4 * x < 4).
  omega.
  destruct D.
  destruct (set_In_dec (4 * x - 4) ((rangeObjs toStart scan
    |_| rangeObjs alloc toEnd) |_| (rangeObjs scan free |_| W'))).
  left; exists (4 * x - 4); constructor.
  assumption.
  omega.
  right; intros [y [Iy Ieq]].
  destruct Ieq; [subst y | idtac].
  auto.
  replace (4 * x - 4) with y in n0.
  auto.
  unfold POnat.POnat.A.
  omega.
  right; intros [y [Iy Ieq]].
  destruct Ieq.
  subst y; auto.
  omega.

  (* proved the disjunction: now use it *)
  destruct D as [[y [Iy Ey]] | NI].

  destruct (set_union_elim Iy) as [I | I]; clear Iy.
  assert (Hp1' : vh ||- ((objCopiedH y (heapOf s) y) ** top) ** top).
  sconjMatch.
  hypReplace Hp0 (fasElim I Hp0); sconjMatch.
  unfold objCopiedH in Hp1'; unfold fieldCopiedH in Hp1'.
  assert (Hp2' : vh' ||- ((objCopiedH y (heapOf s) y) ** top) ** top).
  sconjMatch.
  hypReplace Hp0 (fasElim I Hp0); sconjMatch.
  unfold objCopiedH in Hp2'; unfold fieldCopiedH in Hp2'.
  linLogElim Hp1'; linLogElim Hp2'.
  unfold Heap.HeapMap.MapDef.get.
  assert (Q := contContEq H3 H4); subst x2.
  assert (Q := contContEq H6 H8); subst x3.
  destruct Ey.

  (* case: pointer to first field *)
  subst y.
  assert (Py : ptrP (4 * x)).
  apply (ptrP_intro x); reflexivity.
  assert (R := isPtrEq Py).
  assert (X := contGetEqL Hp1'); unfold Heap.get in X.
  rewrite R in X.
  rewrite div44eq in X.
  rewrite X.
  clear X; assert (X := contGetEqL Hp2'); unfold Heap.get in X.
  rewrite R in X.
  rewrite div44eq in X.
  rewrite X; reflexivity.

  (* case: pointer to second field *)
  assert (Py : ptrP (4 + y)).
  apply (ptrP_intro x); rewrite H; intuition.
  assert (R := isPtrEq Py).
  sconjLift2 Hp1'; sconjLift2 Hp2'.
  assert (X := contGetEqL Hp1'); unfold Heap.get in X.
  rewrite R in X.
  assert (R' : 4 + y = 4 * x).
  intuition.
  rewrite R' in X; rewrite div44eq in X; rewrite X.
  clear X; assert (X := contGetEqL Hp2'); unfold Heap.get in X.
  rewrite R in X.
  rewrite R' in X; rewrite div44eq in X; rewrite X; reflexivity.

  (* case: y in G |_| W *)
  sconjLift2of2 Vhp; sconjLift2of2 Vhp'.
  assert (Vhp'2 : vh' ||- 
    (objRespMapH (map_union phi' (fun x0 : POnat.POnat.A => Some x0)) (heapOf s) y ** top) ** top).
  sconjMatch.
  hypReplace Hp0 (fasElim I Hp0).
  sconjMatch.
  hypReplace Vhp' Vhp'2; clear Vhp'2.
  unfold objRespMapH in Vhp'; linLogElim Vhp'.

  clear Hp Hp' rootOk.
  assert (Vhp2 : vh ||- 
    (objRespMapH (map_union phi (fun x0 : POnat.POnat.A => Some x0)) (heapOf s) y ** top) ** top).
  sconjMatch.
  assert (I' : set_In y (rangeObjs scan free |_| W)).
  rewrite WWeq; assumption.
  hypReplace Hp (fasElim I' Hp); sconjMatch.
  hypReplace Vhp Vhp2; clear Vhp2.
  unfold objRespMapH in Vhp; unfold fieldRespMapH in Vhp; linLogElim Vhp.
  destruct H3 as [v00 [Hp00 Meq00]].
  destruct H6 as [v01 [Hp01 Meq01]].
  unfold fieldRespMapH in Vhp'; linLogElim Vhp'.
  destruct H3 as [v10 [Hp10 Meq10]].
  destruct H6 as [v11 [Hp11 Meq11]].
  assert (R := contContEq Hp00 Hp10); subst v10; clear Hp00 Hp10.
  assert (R := contContEq Hp01 Hp11); subst v11; clear Hp01 Hp11.

  assert (E1 : x0 = x2).
  unfold mapX in Meq00; unfold map_union in Meq00.
  unfold mapX in Meq10; unfold map_union in Meq10.
  destruct (isOdd v00).
  rewrite Meq00 in Meq10; injection Meq10; intros; assumption.
  rewrite <- phiEq in Meq10.
  destruct (phi v00);
    rewrite Meq00 in Meq10; injection Meq10; intros; assumption.

  assert (E2 : x1 = x3).
  unfold mapX in Meq01; unfold map_union in Meq01.
  unfold mapX in Meq11; unfold map_union in Meq11.
  destruct (isOdd v01).
  rewrite Meq01 in Meq11; injection Meq11; intros; assumption.
  rewrite <- phiEq in Meq11.
  destruct (phi v01);
    rewrite Meq01 in Meq11; injection Meq11; intros; assumption.
  
  subst x2 x3.
  clear Meq00 Meq01 Meq10 Meq11.
  assert (Heap.get vh (4 * x) = Heap.get vh' (4 * x)).
  destruct Ey.

  subst y.
  rewrite (contGetEqL Vhp); rewrite (contGetEqL Vhp'); reflexivity.
  rewrite H; replace (y + 4) with (4 + y).
  sconjLift2 Vhp; sconjLift2 Vhp'.
  rewrite (contGetEqL Vhp'); rewrite (contGetEqL Vhp); reflexivity.
  intuition.

  unfold Heap.get in H.
  assert (xPtr : ptrP (4 * x)).
  apply (ptrP_intro x); reflexivity.
  rewrite (isPtrEq xPtr) in H.
  rewrite div44eq in H; assumption.

  (* None case *)
  clear Hp Hp' rootOk.
  
  assert (E1 : Heap.HeapMap.MapDef.get vh x = None).
  clear Vhp'.
  assert (Hp : vh ||- (foralls x in 
    (rangeObjs toStart scan |_| rangeObjs alloc toEnd) |_|
    (rangeObjs scan free |_| W), x |->__)).
  assert (Hp : vh ||- (foralls x in 
    rangeObjs toStart scan |_| rangeObjs alloc toEnd, x |->__) **
    (foralls x in rangeObjs scan free |_| W, x |->__)).
  sconjMatch.
  eapply fasImpl; [idtac | eauto].
  unfold objCopiedH; unfold fieldCopiedH; intros z h' Hp'; linLogElim Hp'; linLogIntro; sconjMatch.
  eapply fasImpl; [idtac | eauto].
  unfold objRespMapH; unfold fieldRespMapH; intros z h' Hp'; linLogElim Hp'; linLogIntro; sconjMatch.
  apply fasMerge'.
  assumption.
  intros x' hh HH; linLogElim HH; linLogIntro; sconjMatch.

  clear Vhp.
  unfold Heap.HeapMap.MapDef.get.
  cut (Heap.get vh (4 * x) = None).
  intro E.
  unfold Heap.get in E.
  rewrite div44eq in E.
  rewrite isPtrEq in E.
  assumption.
  eapply ptrP_intro; eauto.
  apply notInHeapDomNone.
  intro I.
  apply NI; clear NI.
  assert (I' : set_In (4 * x) (set_mmap
           (fun x : POnat.POnat.A => set_add x (set_add (x + 4) empty_set))
           ((rangeObjs toStart scan |_| rangeObjs alloc toEnd)
            |_| rangeObjs scan free |_| W))).
  rewrite <- (objBufferDom Hp); assumption.
  clear Hp I; rename I' into I.
  destruct (set_mmap_elim _ _ _ I) as [y [Iy Ix]]; clear I.
  exists y.
  constructor.
  rewrite <- WWeq; assumption.
  simpl in Ix; destruct Ix as [I | [I | I]].
  left; auto.
  right; auto.
  intuition.
  
  (* basically the same as E1 *)
  assert (E2 : Heap.HeapMap.MapDef.get vh' x = None).
  clear Vhp E1 Al1 Al2 WMeqv iso.
  assert (Hp : vh' ||- (foralls x in 
    (rangeObjs toStart scan |_| rangeObjs alloc toEnd) |_|
    (rangeObjs scan free |_| W'), x |->__)).
  assert (Hp : vh' ||- (foralls x in 
    rangeObjs toStart scan |_| rangeObjs alloc toEnd, x |->__) **
    (foralls x in rangeObjs scan free |_| W', x |->__)).
  sconjMatch.
  eapply fasImpl; [idtac | eauto].
  unfold objCopiedH; unfold fieldCopiedH; intros z h' Hp'; linLogElim Hp'; linLogIntro; sconjMatch.
  eapply fasImpl; [idtac | eauto].
  unfold objRespMapH; unfold fieldRespMapH; intros z h' Hp'; linLogElim Hp'; linLogIntro; sconjMatch.
  apply fasMerge'.
  assumption.
  intros x' hh HH; linLogElim HH; linLogIntro; sconjMatch.

  clear Vhp'.
  unfold Heap.HeapMap.MapDef.get.
  cut (Heap.get vh' (4 * x) = None).
  intro E.
  unfold Heap.get in E.
  rewrite div44eq in E.
  rewrite isPtrEq in E.
  assumption.
  eapply ptrP_intro; eauto.
  apply notInHeapDomNone.
  intro I.
  apply NI; clear NI.
  assert (I' : set_In (4 * x) (set_mmap
           (fun x : POnat.POnat.A => set_add x (set_add (x + 4) empty_set))
           ((rangeObjs toStart scan |_| rangeObjs alloc toEnd)
            |_| rangeObjs scan free |_| W'))).
  rewrite <- (objBufferDom Hp); assumption.
  clear Hp I; rename I' into I.
  destruct (set_mmap_elim _ _ _ I) as [y [Iy Ix]]; clear I.
  exists y.
  constructor.
  assumption.
  simpl in Ix; destruct Ix as [I | [I | I]].
  left; auto.
  right; auto.
  intuition.
  
  rewrite E1; rewrite E2; reflexivity.

Qed.
*)


