Require Import StateTheory.
Require Import VGC.
Require Import List.
Require Import GC.
Require Import Cheney.
Require Import CheneyLoop.
Require Import CheneyLemma.
Require Import SCAP.
Require Import ScanField.


Set Implicit Arguments.


Module CheneyGCImpl <: VIEW_GC with Module I := CopyMGI.

  Module I := CopyMGI.


  Definition repr := repr.


  Lemma atomRegEq : forall h vs rf rr A r,
    h ||- repr vs rf rr ** A ->
    In r rr ->
    I.atomP (RFile.get rf r) \/ I.atomP (rget vs r) ->
    RFile.get rf r = rget vs r.
  Proof.
    intros h vs rf rr A r Hp I D.
    unfold repr, Cheney.repr in Hp; lle Hp.
    destruct H3 as [_ [_ [objs [Req Vhp]]]].
    destruct (Req _ I) as [X _]; auto.
  Qed.

  Lemma setNonRootOk : forall h vs rf rr r v,
    h ||- repr vs rf rr ->
    ~In r (rgcInfo::rr) ->
    h ||- repr vs (RFile.set rf r v) rr.
  Proof.
    unfold repr, Cheney.repr.
    intros h vs rf rr r v Hp I; lle Hp.
    destruct H1 as [rrOk [alOk1 [alOk2 [rootsOk [endOk endOk']]]]].
    lli.

    repeat (apply conj).
    auto.
    apply alOk1.
    apply alOk2.
    intros r0 I0; destruct (rootsOk _ I0) as [X Y].
    constructor.
    replace (RFile.get (RFile.set rf r v) r0)
      with (rget (rset (h, rf) r v) r0) by auto.
    rewrite X.
    replace (RFile.get rf r0) with (rget (h, rf) r0)
      by reflexivity.
    destruct (regCompare r0 r).
    subst r; elf; apply I; right; auto.
    stSimpl; reflexivity.
    assumption.
    eauto.
    auto.
    sconjMatch.
    destruct (regCompare r rgcInfo).

    subst r.
    elf; apply I; left; reflexivity.

    replace (RFile.get (RFile.set rf r v) rgcInfo)
      with (rget (rset (h, rf) r v) rgcInfo) by auto.
    stSimpl; assumption.

  Qed.

  
  Lemma rootCopyOk : forall h vs rf rr rs rd,
    h ||- repr vs rf rr ->
    In rs rr -> 
    ~In rd (zero::rgcInfo::nil) ->
    h ||- repr (rset vs rd (rget vs rs)) (RFile.set rf rd (RFile.get rf rs)) (rd::rr).
  Proof.
    unfold repr; unfold Cheney.repr.
    intros h vs rf rr rs rd Hp I0 Nz.
    assert (NE1 : rd <> zero) by (intro E; subst rd; intuition).
    assert (NE2 : rd <> rgcInfo) by (intro E; subst rd; intuition).
    lle Hp.
    destruct H1 as [rrOk [al1 [al2 [rootOk [boundsOk boundsOk']]]]].
    lli.

    repeat (apply conj).
    simpl; intro X; destruct X.
      subst rd; intuition. auto.
    4:eauto. 4:assumption.
    eauto.
    auto.
    simpl; intros r I1.
    replace (RFile.get (RFile.set rf rd (RFile.get rf rs)) r) with
      (rget (rset (h, rf) rd (rget (h, rf) rs)) r) by auto.
    destruct (regCompare r rd).
    subst r; stSimpl.
    destruct (rootOk _ I0) as [X Y].
    rewrite X; constructor.
    reflexivity.
    rewrite X in Y; assumption.
    destruct I1 as [I | I].
    elf; auto.
    stSimpl.
    destruct (rootOk _ I) as [R Ok].
    rewrite R; rewrite R in Ok; constructor.
    reflexivity.
    assumption.

    stSimpl; scm.
    replace (RFile.get (RFile.set rf rd (RFile.get rf rs)) rgcInfo) with
      (rget (rset (h, rf) rd (rget (h, rf) rs)) rgcInfo)
      by auto.
    stSimpl; assumption.

  Qed.


  Lemma rootWeaken : forall h vs rf rr rr',
    h ||- repr vs rf rr ->
    incl rr' rr ->
    h ||- repr vs rf rr'.
  Proof.
    unfold repr, Cheney.repr.
    intros h vs rf rr rr' Hp IR; lle Hp.
    destruct H1 as [rrOk [al1 [al2 [rootOk [endOk endOk']]]]].
    lli.
    
    repeat apply conj.
    auto. 4:eauto. 4:eauto. eauto. auto.
    intros r I.
    apply rootOk.
    auto.

    assumption.

  Qed.


  Lemma addAtomRoot : forall h vs rf rr A r,
    h ||- repr vs rf rr ** A ->
    ~In r (zero::I.rgcInfo::rr) ->
    I.atomP (RFile.get rf r) ->
    h ||- repr (rset vs r (RFile.get rf r)) rf (r::rr) ** A.
  Proof.
    unfold repr, Cheney.repr.
    intros h vs rf rr A r Hp NI Atom.
    assert (NE1 : r <> zero) by (intro E; subst r; intuition).
    assert (NE2 : r <> rgcInfo) 
      by (intro E; subst r; apply NI; right; left; auto).
    lle Hp.
    destruct H3 as [rrOk [al1 [al2 [rootOk [endOk endOk']]]]].
    lli.

    repeat apply conj.
    simpl; intros [X | X]; auto.
    4:eauto. eauto. auto. 2:auto.
    simpl; intros r0 I.
    replace (RFile.get rf) with (rget (h, rf)) by reflexivity.
    destruct I.
    subst r0; stSimpl.
    constructor. reflexivity. left; assumption.
    destruct (rootOk _ H) as [Req OkF].
    replace (RFile.get rf) with (rget (h, rf)) in Req by auto.
    rewrite <- Req; clear Req.
    destruct (regCompare r0 r).
    subst r0; elf; apply NI; right; right; assumption.
    stSimpl. intuition.
    assumption.
  Qed.


  Definition phi n := 
    match Cheney.P n with
      | SomeT x => SomeT x
      | NoneT => ScanField.Pscanf n
    end.


  Ltac phiStep Ne f k :=
    destruct (natEqDec f k) as [R | R];
      [subst f; rewrite beq_true in Ne; apply setRangeIntro; omega |
       rewrite (beq_false R) in Ne; clear R].

  Lemma gcPhiDomOk : forall f,
    phi f <> NoneT _ -> set_In f I.gcPhiDom.
  Proof.
    unfold I.gcPhiDom; unfold phi; unfold P; unfold Pscanf; unfold Ploop.
    unfold SCANFIELD; unfold CHENEYLOOP; unfold CHENEYLOOP2;
      unfold CHENEYRETURN;
      unfold CHENEYLOOPHEADER; unfold CHENEYENTER; unfold CHENEYDOALLOC; unfold CHENEYALLOC;
      unfold CHENEYPOSTGC; unfold INFLOOP.
    unfold RETURN; unfold SCANPOINTER; unfold NOTTOSPACEPTR; 
      unfold SCANCOPIED; unfold SCANNOCOPY;
	unfold FWDOBJ; unfold RETURN.

    intros f Ne.

    phiStep Ne f 160.
    phiStep Ne f 64.
    phiStep Ne f 67.
    phiStep Ne f 87.
    phiStep Ne f 84.
    phiStep Ne f 71.
    phiStep Ne f 56.
    phiStep Ne f 10.
    phiStep Ne f 15.
    phiStep Ne f 89.
    phiStep Ne f 20.
    phiStep Ne f 30.
    phiStep Ne f 40.
    phiStep Ne f 50.
    phiStep Ne f 196.
    phiStep Ne f 169.
    phiStep Ne f 184.
    phiStep Ne f 173.
    phiStep Ne f 193.
    phiStep Ne f 176.

    elf; apply Ne; reflexivity.

  Qed.


  Ltac PhiOkSolver :=
    unfold phi; unfold P; unfold Ploop; rewrite beq_true;
    repeat (rewrite beq_false; [idtac | discriminate]);
    simpl; intuition.

  Lemma read1PhiOk :
    optPredT (phi I.read1Lbl)
      (specWeaker (read1Spec repr I.rroot I.gcGuar I.presReg)).
  Proof.
    PhiOkSolver.
  Qed.

  Lemma read2PhiOk :
    optPredT (phi I.read2Lbl)
      (specWeaker (read2Spec repr I.rroot I.gcGuar I.presReg)).
  Proof.
    PhiOkSolver.
  Qed.

  Lemma write1PhiOk : 
    optPredT (phi I.write1Lbl)
      (specWeaker (write1Spec repr I.rroot I.gcGuar I.presReg)).
  Proof.
    PhiOkSolver.
  Qed.

  Lemma write2PhiOk : 
    optPredT (phi I.write2Lbl)
      (specWeaker (write2Spec repr I.rroot I.gcGuar I.presReg)).
  Proof.
    PhiOkSolver.
  Qed.

  Lemma allocPhiOk : 
    optPredT (phi I.allocLbl)
      (specWeaker (allocSpec repr I.rroot I.gcGuar I.presReg I.NULL)).
  Proof.
    PhiOkSolver.
  Qed.


  Definition GCC' f :=
    if beq_nat f CHENEYREAD1 then
      Some (cheneyRead 0)
    else if beq_nat f CHENEYREAD2 then
      Some (cheneyRead 4)
    else if beq_nat f CHENEYWRITE1 then
      Some (cheneyWrite 0)
    else if beq_nat f CHENEYWRITE2 then
      Some (cheneyWrite 4)
    else if beq_nat f RETURN then
      Some CheneyLemma.returnBlock
    else if beq_nat f SCANPOINTER then
      Some scanPointer
    else if beq_nat f NOTTOSPACEPTR then
      Some isNotToSpacePtr
    else if beq_nat f SCANCOPIED then
      Some scanCopied
    else if beq_nat f SCANNOCOPY then
      Some scanNoCopy
    else if beq_nat f FWDOBJ then
      Some fwdObj
    else
      None.


  Definition GCC f :=
    if beq_nat f SCANFIELD then
      Some scanField
    else if beq_nat f CHENEYLOOP then
      Some cheneyLoop
    else if beq_nat f CHENEYLOOP2 then
      Some cheneyLoop2
    else if beq_nat f CHENEYRETURN then
      Some cheneyReturn
    else if beq_nat f CHENEYLOOPHEADER then
      Some cheneyLoopHeader
    else if beq_nat f CHENEYENTER then
      Some cheneyEnter
    else if beq_nat f CHENEYDOALLOC then
      Some cheneyDoAlloc
    else if beq_nat f CHENEYALLOC then
      Some cheneyAlloc
    else if beq_nat f CHENEYPOSTGC then
      Some cheneyPostGC
    else if beq_nat f INFLOOP then
      Some infLoop
    else 
      GCC' f.


  Ltac weakerStep f f' H :=
    destruct (natEqDec f f') as [R | R];
      [subst f; rewrite beq_true;
	injection 1; clear H; intros; subst;
	  econstructor; econstructor; constructor;
	    [unfold P; repeat (rewrite beq_false; [idtac | discriminate]) | 
	      constructor; [reflexivity | intros s s' Guar; apply Guar]] |
       rewrite (beq_false R); clear R].

  Lemma phiWeakerPscanf : forall f p0 g0,
    ScanField.Pscanf f = SomeT (bspeci p0 g0) ->
    exists p1, exists g1, phi f = SomeT (bspeci p1 g1) /\
      p0 = p1 /\ forall s s', g1 s s' -> g0 s s'.
  Proof.
    unfold Pscanf; unfold phi.
    intros f p g.

    weakerStep f RETURN H.
    reflexivity.
    weakerStep f SCANPOINTER H.
    reflexivity.
    weakerStep f NOTTOSPACEPTR H.
    reflexivity.
    weakerStep f SCANCOPIED H.
    reflexivity.
    weakerStep f SCANNOCOPY H.
    reflexivity.
    weakerStep f FWDOBJ H.
    reflexivity.
    weakerStep f SCANFIELD H.
    unfold Ploop; rewrite beq_true; reflexivity.
    discriminate 1.

  Qed.


  Lemma phiWeakerP : forall f p g,
    P f = SomeT (bspeci p g) ->
    exists p', exists g', phi f = SomeT (bspeci p' g') /\
      p = p' /\ forall s s', g' s s' -> g s s'.
  Proof.
    unfold phi.
    intros f p g Peq; destruct (P f).
    injection Peq; intro; subst b; clear Peq.
    econstructor; econstructor; constructor.
    reflexivity.
    intuition.
    discriminate.
  Qed.


  Lemma phiWeakerPloop : forall f p g,
    Ploop f = SomeT (bspeci p g) ->
    exists p', exists g', phi f = SomeT (bspeci p' g') /\
      p = p' /\ forall s s', g' s s' -> g s s'.
  Proof.
    unfold phi; unfold P.
    intros f p g Peq; destruct (Ploop f).
    injection Peq; intro; subst b; clear Peq.
    econstructor; econstructor; constructor.
    reflexivity.
    intuition.
    discriminate.
  Qed.


  Ltac GCCokStep' Peq f f' R0 :=
    destruct (natEqDec f f') as [R | R];
      [subst f; rewrite beq_true; injection 1; clear Peq; intros; subst |
	rewrite (beq_false R); rename R into R0].

  Ltac GCCokStep Peq f f' :=
    destruct (natEqDec f f') as [R | R];
      [subst f; rewrite beq_true; injection 1; clear Peq; intros; subst |
	rewrite (beq_false R); clear R].
    

  Lemma GCCok : cdHeapOkSC phi GCC phi.
  Proof.
    intros f [p g] Peq.
    unfold phi in Peq; unfold GCC; generalize dependent Peq; unfold P; unfold Ploop.
    
    GCCokStep' Peq f SCANFIELD RZ.
    2:GCCokStep Peq f CHENEYLOOP.
    3:GCCokStep Peq f CHENEYLOOP2.
    4:GCCokStep Peq f CHENEYRETURN.
    5:GCCokStep Peq f CHENEYLOOPHEADER.
    6:GCCokStep Peq f CHENEYENTER.
    7:GCCokStep Peq f CHENEYDOALLOC.
    8:GCCokStep Peq f CHENEYALLOC.
    9:GCCokStep Peq f CHENEYPOSTGC.
    10:GCCokStep Peq f INFLOOP.
    11:unfold GCC'; GCCokStep Peq f 20.

    apply (iseqOkWeakP _ _ _ _ scanFieldOk _ phiWeakerPscanf).
    apply (iseqOkWeakP _ _ _ _ cheneyLoopOk _ phiWeakerPloop).
    apply (iseqOkWeakP _ _ _ _ cheneyLoop2Ok _ phiWeakerPloop).
    apply (iseqOkWeakP _ _ _ _ cheneyReturnOk _ phiWeakerPloop).
    apply (iseqOkWeakP _ _ _ _ cheneyLoopHeaderOk _ phiWeakerP).
    apply (iseqOkWeakP _ _ _ _ cheneyEnterOk _ phiWeakerP).
    apply (iseqOkWeakP _ _ _ _ cheneyDoAllocOk _ phiWeakerP).
    apply (iseqOkWeakP _ _ _ _ cheneyAllocOk _ phiWeakerP).
    apply (iseqOkWeakP _ _ _ _ cheneyPostGCOk _ phiWeakerP).
    apply (iseqOkWeakP _ _ _ _ infLoopOk _ phiWeakerP).
    eapply iseqOkWeakP; [apply cheneyReadOk; auto | exact phiWeakerP].
    GCCokStep Peq f 30.
    2:GCCokStep Peq f 40.
    3:unfold Pscanf; GCCokStep Peq f 50.
    eapply iseqOkWeakP; [apply cheneyReadOk; auto | exact phiWeakerP].
    eapply iseqOkWeakP; [apply cheneyWriteOk; auto | exact phiWeakerP].
    eapply iseqOkWeakP; [apply cheneyWriteOk; auto | exact phiWeakerP].

    GCCokStep Peq f RETURN.
    2:GCCokStep Peq f SCANPOINTER.
    3:GCCokStep Peq f NOTTOSPACEPTR.
    4:GCCokStep Peq f SCANCOPIED.
    5:GCCokStep Peq f SCANNOCOPY.
    6:GCCokStep Peq f FWDOBJ.

    eapply iseqOkWeakP; [apply CheneyLemma.returnOk | discriminate 1].
    apply (iseqOkWeakP _ _ _ _ scanPointerOk _ phiWeakerPscanf).
    apply (iseqOkWeakP _ _ _ _ isNotToSpacePtrOk _ phiWeakerPscanf).
    apply (iseqOkWeakP _ _ _ _ scanCopiedOk _ phiWeakerPscanf).
    apply (iseqOkWeakP _ _ _ _ scanNoCopyOk _ phiWeakerPscanf).
    apply (iseqOkWeakP _ _ _ _ fwdObjOk _ phiWeakerPscanf).

    GCCokStep Peq f SCANFIELD.
    auto.
    discriminate 1.

  Qed.


End CheneyGCImpl.







(*

Require Import LinkedList.


Section Heapy.

  Require Import Heap.
  Import HeapMap.


  Lemma natListIsMinObjHp' : forall h l x A,
    h ||- NatListAt l x ** A ->
    h ||- (lexists objs, minObjHp (x::nil) objs) ** A.
  Proof.
    unfold NatListAt; induction l; simpl; intros x A Hp.

    linLogElim Hp.
    subst x.
    hypReplace Hp (unitIL Hp); sconjMatch.
    linLogIntro.
    instantiate (1 := empty_set).
    unfold minObjHp; apply aconjIntro.
    simpl; intros objs' OO; intuition.
    unfold okObjHp.
    linLogIntro.
    simpl; intros v [II | II].
    subst v; left; unfold NULL; constructor; constructor.
    intuition.
    sconjMatch.
    
    (* inductive case *)
    linLogElim Hp.
    rename H3 into NE; rename x0 into y.

    assert (NE' : x <> y).
    intro; subst y.
    destruct l; simpl in Hp; linLogElim Hp.
    auto.
    sconjLift3 Hp; linLogElim Hp.

    sconjLift3 Hp.
    assert (Hp' := IHl _ _ Hp); clear IHl Hp.
    linLogElim Hp'.
    rename x0 into objs.
    sconjLift4of4 Hp'; apply sconjComm; sconjMatch.
    existsl (set_add x objs).
    unfold minObjHp; apply aconjIntro.

    simpl; intros z I.
    exists x; constructor.
    simpl; auto.
    destruct I.
    subst z; constructor.
    econstructor 3.
    instantiate (1 := y).
    rewrite <- heap_canonEqv.
    sconjLift3of3 Hp; eapply contGetEqL; eauto.
    destruct Hp as [ha [hb [J [HA _]]]].
    unfold minObjHp in HA; destruct (aconjElim HA) as [X _]; clear HA.
    eapply reachableJoinLeft; eauto.
    unfold rootReachable in X.
    destruct (X _ H) as [v [Veq R]].
    simpl in Veq; destruct Veq.
    subst v.
    assumption.
    elimtype False; assumption.

    unfold okObjHp; linLogIntro.
    simpl; intros v [I | I].
    subst v; right; apply set_add_intro1; reflexivity.
    intuition.

    assert (X : h ||- (!okFieldVal objs y ** top) ** top).
    sconjMatch.
    linLogIntro.
    unfold minObjHp in Hp; destruct (aconjElim Hp) as [_ X]; clear Hp.
    unfold okObjHp in X; linLogElim X.
    apply H1; simpl; auto.

    linLogElim X; clear X; rename H1 into OkY.
    assert (Hp' : h ||- objHp objs objs ** okObjAt objs x).
    sconjMatch.
    unfold minObjHp in Hp; destruct (aconjElim Hp) as [_ Hp']; clear Hp.
    unfold okObjHp in Hp'; linLogElim Hp'; assumption.
    unfold okObjAt; unfold okFieldAt; sconjMatch.
    linLogIntro.
    2:eauto.
    left; unfold natEnc.
    unfold atomP; eapply oddIntro; reflexivity.
    linLogIntro; eauto.
    apply objHpIntro.
    apply sconjComm; sconjMatch.
    eapply objHpWeaken; eauto.
    intros q I; apply set_add_intro2; assumption.
    eapply okObjAtWeaken; eauto.
    intros q I; apply set_add_intro2; assumption.

  Qed.


End Heapy.


Lemma natListIsMinObjHp : forall h l x,
  h ||- NatListAt l x ->
  h ||- lexists objs, minObjHp (x::nil) objs.
Proof.
  intros h l x Hp.
  hypReplace Hp (unitIR Hp).
  assert (X := natListIsMinObjHp' _ _ Hp).
  linLogElim X.
  existsl x0; assumption.
Qed.


*)
