Require Import CheneyLemma.
Require Import ScanField.
Require Import Heap.
Require Import HeapDom.
Require Import VGC.


Set Implicit Arguments.


(* Definition and verification of the loop of the Cheney collector. *)


(* code address constants *)
Definition CHENEYLOOP := 64.
Definition CHENEYLOOP2 := 67.
Definition CHENEYRETURN := 87.


(* local register conventions *)
Definition rscan := t3.
Definition raSave := v1.


(* code blocks *)

Definition cheneyLoop :=
  beq rscan rfree CHENEYRETURN;;
  addiu a0 rscan 0;
  jal SCANFIELD CHENEYLOOP2.

Definition cheneyLoop2 :=
  addiu a0 rscan WORDSIZE;
  addiu rscan rscan OBJSIZE;
  jal SCANFIELD CHENEYLOOP.

Definition cheneyReturn :=
  addiu ra raSave 0;
  jr ra.


(* state specifications *)

Definition cheneyLoopStOk phi G W M F A s :=
  set_eqv G (rangeObjs (rget s rscan) (rget s rfree)) /\
  rget s rtoStart <= rget s rscan /\
  aligned8 (rget s rscan) (rget s rfree) /\
  scanFieldStOk phi W M F A s.

(* This specifies what happens when we scan an entire object. *)
Definition cheneyScanObjGuar s s' :=
  (forall phi G W M F A,
    cheneyLoopStOk phi G W M F (okObjAt (W |_| M) (rget s rscan) ** A) s ->
    exists phi', exists W', exists M', exists F', exists G',
      set_eqv W (M' |_| W') /\
      set_eqv F (G' |_| F') /\
      iso_set phi' M' G' /\
      (forall x, set_In x G' -> 
	get (heapOf s') (rget s rscan) = Some x \/ 
	get (heapOf s') (4 + rget s rscan) = Some x) /\
      aligned8 (rget s rfree) (rget s' rfree) /\
      (forall x, set_In x W' -> s' |||- objCopied x s x ** top) /\
      cheneyLoopStOk (map_union phi phi') (set_remove (rget s rscan) G |_| G')
        W' (M |_| M') F'
	(objRespMap (map_union phi phi') s (rget s rscan) **
	  (foralls x in G', objCopiedPhi phi' s x) ** A) s') /\
  regPres 
    (regNotInList (rscan::ra::rtemp::temp0::temp1::t2::t4::t7::a0::rfree::v0::nil)) s s' /\
  rget s' rscan = 8 + rget s rscan /\
  rget s' ra = CHENEYLOOP.

(* low level version of cheneyScanObjGuar that uses scanFieldStOk. *)
Definition cheneyScanObjGuar' s s' :=
  (forall phi W M F A,
    scanFieldStOk phi W M F (okObjAt (W |_| M) (rget s rscan) ** A) s ->
    exists phi', exists W', exists M', exists F', exists G',
      set_eqv W (M' |_| W') /\
      set_eqv F (G' |_| F') /\
      iso_set phi' M' G' /\
      (forall x, set_In x G' ->
	get (heapOf s') (rget s rscan) = Some x \/ 
	get (heapOf s') (4 + rget s rscan) = Some x) /\
      aligned8 (rget s rfree) (rget s' rfree) /\
      (forall x, set_In x W' -> s' |||- objCopied x s x ** top) /\
      scanFieldStOk (map_union phi phi') W' (M |_| M') F'
	(objRespMap (map_union phi phi') s (rget s rscan) **
	  (foralls x in G', objCopiedPhi phi' s x) ** A) s') /\
  regPres 
    (regNotInList (rscan::ra::rtemp::temp0::temp1::t2::t4::t7::a0::rfree::v0::nil)) s s' /\
  rget s' rscan = 8 + rget s rscan /\
  rget s' ra = CHENEYLOOP.


(* block specifications *)

Definition cheneyLoopPre s :=
  exists phi, exists G, exists W, exists M, exists F,
    cheneyLoopStOk phi G W M F (objHp (W |_| M) G ** top) s.

Notation cheneyLoopRegGuar :=
  (fun s s' => regPres 
    (regNotInList (ra::rscan::rtemp::temp0::temp1::t2::t4::t7::a0::rfree::v0::nil)) s s' /\
    rget s' ra = rget s raSave).

Definition cheneyLoopGuar s s' :=
  (forall phi G W M F A,
    cheneyLoopStOk phi G W M F (objHp (W |_| M) G ** A) s ->
    exists phi', exists W', exists M', exists F', exists G',
      set_eqv W (M' |_| W') /\
      set_eqv F (G' |_| F') /\
      iso_set phi' M' G' /\
      (forall x', set_In x' G' -> exists x, set_In x G /\ reachable (heapOf s') x x') /\
      aligned8 (rget s rfree) (rget s' rfree) /\
      cheneyLoopStOk (map_union phi phi') empty_set W' (M |_| M') F'
      ((foralls x in G, objRespMap (map_union phi phi') s x) ** 
	(foralls x in G', objIsFwded phi phi' s x) ** A) s') /\
  cheneyLoopRegGuar s s'.

Definition ch2Step s :=
  rset (rset (rset s a0 (4+rget s rscan)) rscan (8+rget s rscan)) ra CHENEYLOOP.

Definition cheneyLoop2Pre s :=
  scanFieldPre (ch2Step s) /\
  (forall s', scanFieldGuar (ch2Step s) s' -> cheneyLoopPre s').

Definition cheneyLoop2Guar s s'' :=
  exists s', 
    scanFieldGuar (ch2Step s) s' /\
    cheneyLoopGuar s' s''.

Definition cheneyReturnPre : spred := fun s => True.

Definition cheneyReturnGuar s s' :=
  s' = rset s ra (rget s raSave).



(* Code heap type for the loop *)
Definition Ploop : chspec := fun l => 
  if beq_nat l SCANFIELD then
    SomeT (bspeci scanFieldPre scanFieldGuar)
  else if beq_nat l CHENEYLOOP then
    SomeT (bspeci cheneyLoopPre cheneyLoopGuar)
  else if beq_nat l CHENEYLOOP2 then
    SomeT (bspeci cheneyLoop2Pre cheneyLoop2Guar)
  else if beq_nat l CHENEYRETURN then
    SomeT (bspeci cheneyReturnPre cheneyReturnGuar)
  else
    NoneT _.

Ltac PloopSimpl := 
  unfold Ploop; 
  rewrite beq_true; 
  repeat (rewrite beq_false; [idtac | discriminate]); 
  reflexivity.


(* Some facts about cheneyLoopStOk *)

Lemma cheneyLoopStOkImpl : forall phi G W M F A s,
  cheneyLoopStOk phi G W M F A s -> forall A',
    (A ==> A') ->
  cheneyLoopStOk phi G W M F A' s.
Proof.
  unfold cheneyLoopStOk.
  intros phi G W M F A s X A' Aimp; intuition.
  apply (scanFieldStOkImpl H3); auto.
Qed.


Lemma cheneyLoopStOkPhiEqv : forall phi G W M F A s,
  cheneyLoopStOk phi G W M F A s -> forall phi',
    mapEqv phi phi' ->
  cheneyLoopStOk phi' G W M F A s.
Proof.
  unfold cheneyLoopStOk; intuition.
  apply (scanFieldStOkPhiEqv H3 H2).
Qed.


Add Morphism cheneyLoopStOk
  with signature eq ==> set_eqv ==> set_eqv ==> set_eqv ==> set_eqv ==> eq ==>
    state_eqv ==> iff
  as cheneyLoopStOkMorphism.
Proof.
  unfold cheneyLoopStOk; intuition.
  rewrite <- H; rewrite <- (rgetEqv H3); rewrite <- (rgetEqv H3); assumption.
  rewrite <- (rgetEqv H3); rewrite <- (rgetEqv H3); assumption.
  rewrite <- (rgetEqv H3); rewrite <- (rgetEqv H3); assumption.
  setoid_rewrite <- H0; setoid_rewrite <- H1; setoid_rewrite <- H2; 
    setoid_rewrite <- H3; assumption.
  rewrite H; rewrite (rgetEqv H3); rewrite (rgetEqv H3); assumption.
  rewrite (rgetEqv H3); rewrite (rgetEqv H3); assumption.
  rewrite (rgetEqv H3); rewrite (rgetEqv H3); assumption.
  setoid_rewrite H0; setoid_rewrite H1; setoid_rewrite H2; 
    setoid_rewrite H3; assumption.
Qed.


(* Verification of the simple blocks. *)

Lemma cheneyReturnOk : iseqOkS Ploop cheneyReturnPre cheneyReturnGuar cheneyReturn.
Proof.
  unfold cheneyReturn.
  iseqWeakSolver PloopSimpl.
  unfold cheneyReturnPre; unfold cheneyReturnGuar.
  intros s _.
  rewrite addiuStep.
  intuition.
Qed.


Lemma cheneyLoop2Ok : iseqOkS Ploop cheneyLoop2Pre cheneyLoop2Guar cheneyLoop2.
Proof.
  unfold cheneyLoop2.
  iseqWeakSolver PloopSimpl.

  apply scanFieldGuarPresRa.

  intros s [H1 H2].
  rewrite addiuStep.
  rewrite addiuStep.
  stSimpl.
  constructor.
  constructor.
  apply H1.
  apply H2.

  intros s''' [s' [H' H''']].
  hypReplace H2 (H2 s' H').
  unfold cheneyLoop2Guar.
  exists s'.
  constructor; assumption.
Qed.


(* Verification of the main loop block, broken down into a number of steps. *)


(* The low level object scanning guarantee is stronger than the high level one. *)
Lemma cheneyScanObjGuarImpl : forall s s',
  cheneyScanObjGuar' s s' ->
  rget s rscan <> rget s rfree ->
  cheneyScanObjGuar s s'.
Proof.
  intros s s' [Guar [Req [ReqSc RaEq]]] NE.
  constructor; [idtac | intuition].

  intros phi G W M F A [Geq [LE [Al scOk]]].
  destruct (Guar phi W M F A) as [phi' [W' [M' [F' [G' Guar']]]]]; clear Guar.
  unfold scanFieldStOk in scOk |- * ; intuition.

  exists phi'; exists W'; exists M'; exists F'; exists G'.
  destruct Guar' as [Weq [Feq [Iso' [G'eq [Al' [W'copy scOk']]]]]].
  repeat (constructor; [assumption | idtac]).

  constructor.
  rewrite Geq; rewrite ReqSc.
  setoid_rewrite (rangeRemoveEqv (rget s rscan) (rget s rfree)).
  setoid_replace G' with (rangeObjs (rget s rfree) (rget s' rfree)).

  apply rangeObjsTransEqv.
  apply aligned8_succ; auto.
  destruct Al'; omega.

  assert (Feq' : set_eqv F (rangeObjs (rget s rfree) (rget s rtoEnd))).
  unfold scanFieldStOk in scOk; intuition.
  assert (F'eq : set_eqv F' (rangeObjs (rget s' rfree) (rget s rtoEnd))).
  unfold scanFieldStOk in scOk'; intuition;
    rewrite <- (Req rtoEnd) in H0; [exact H0 | regNotInListSolver].
  apply rangeSplit with (rget s rtoEnd).
  rewrite <- F'eq.
  assert (scOk2 : scanFieldStOk (map_union phi phi') W' (M |_| M') F'
            ((foralls x in G', objCopiedPhi phi' s x) ** top) s').
  apply (scanFieldStOkImpl scOk').
  intros h Hp; sconjLift2 Hp; scm.
  apply set_disjSymm; apply (scanFieldStOkFGDisj scOk2).
  hpunfold; intros; lleim.
  unfold scanFieldStOk in scOk'; intuition.
  rewrite <- (Req rtoEnd) in H6; [exact H6 | regNotInListSolver].
  apply set_eqvSymm; rewrite <- Feq'; rewrite <- F'eq; assumption.
  assumption.

  constructor.
  rewrite ReqSc; rewrite <- Req; [omega | regNotInListSolver].

  constructor.
  rewrite ReqSc; unfold scanFieldStOk in scOk'; intuition.
  apply aligned8_succ.
  apply aligned8Trans with (rget s rfree); auto.
  destruct Al'; destruct Al; omega.

  assumption.

Qed.


(* If we scan two fields, we scan an object. *)
Lemma scanFieldsScanObj : forall s s' s'',
  rget s rscan <> rget s rfree ->
  scanFieldGuar (rset (rset s a0 (rget s rscan)) ra CHENEYLOOP2) s' ->
  scanFieldGuar (ch2Step s') s'' ->
  cheneyScanObjGuar s s''.
Proof.
  intros s s' s'' NE Guar1 Guar2.
  cut (cheneyScanObjGuar' s s'').
  intros; apply cheneyScanObjGuarImpl; auto.

  destruct Guar1 as [Guar1 Req1].
  constructor; [idtac | clear Guar1; destruct Guar2 as [_ Req2]].
  
  intros phi W M F A scOk.

  assert (scOk' : exists x, scanFieldStOk phi W M F
     (okFieldAt (W |_| M) (rget s rscan) **
       4 + rget s rscan |-> x ** A) s /\
     okFieldVal (W |_| M) x).
  clear Guar1.
  unfold scanFieldStOk in scOk |- * ; intuition.
  unfold okObjAt in H5; unfold okFieldAt in H5; lle H5; exists x0; intuition.
  scm.
  unfold okFieldAt; existsl x; lli; assumption.

  clear scOk; destruct scOk' as [x [scOk xok]].
  destruct (Guar1 phi W M F (4 + rget s rscan |-> x ** A))
    as [phi' [W' [M' [F' [G' Guar1']]]]]; clear Guar1.
  unfold scanFieldStOk in scOk |- *; stSimpl; intuition.

  generalize dependent Guar1'; stSimpl.
  intros [Weq [Feq [Iso [G'eq [Al [W'copy scOk']]]]]].
  destruct Guar2 as [Guar2 Req2].

  assert (scOk2 : exists v0', scanFieldStOk (map_union phi phi') W' (M |_| M') F'
            (rget s rscan |-> v0' **
              (foralls x in G', 
		objCopiedPhi phi' (rset (rset s a0 (rget s rscan)) ra CHENEYLOOP2) x) **
             4 + rget s rscan |-> x ** A) s' /\
	    exists v0 : nat, s |||- rget s rscan |-> v0 ** top /\ 
	      mapX (map_union phi phi') v0 = Some v0').
  clear Guar2 Req1 Req2 scOk.
  unfold scanFieldStOk in scOk' |- * ; intuition; unfold fieldRespMap in H5; lle H5.
  exists x0; intuition.

  clear scOk'; destruct scOk2 as [sv [scOk' svOk]].
  destruct (Guar2 (map_union phi phi') W' (M |_| M') F' 
    (rget s rscan |-> sv **
      (foralls x in G', 
	objCopiedPhi phi' (rset (rset s a0 (rget s rscan)) ra CHENEYLOOP2) x) ** A)) as
    [phi'' [W'' [M'' [F'' [G'' Guar2']]]]]; clear Guar2.

  unfold scanFieldStOk in scOk' |- *; unfold ch2Step; stSimpl; intuition.
  permConj H5 (1::2::3::6::4::5::7::nil); scm.
  rewrite <- Req1; [stSimpl | regNotInListSolver].
  unfold okFieldAt; lli; [idtac | apply Hp].
  rewrite (set_eqvUnionSymm M M').
  rewrite <- (set_eqvUnionAssoc W' M' M).
  rewrite (set_eqvUnionSymm W' M').
  rewrite <- Weq.  
  assumption.

  (* the final step *)
  exists (map_union phi' phi''); exists W''; exists (M' |_| M''); 
    exists F''; exists (G' |_| G'').
  generalize dependent Guar2'; unfold ch2Step; 
    rewrite <- Req1; [stSimpl | regNotInListSolver].
  intros [W'eq [F'eq [Iso'' [G''eq [Al'' [W''copy scOk'']]]]]].

  constructor.
  setoid_rewrite (set_eqvUnionAssoc M' M'' W''); setoid_rewrite <- W'eq; assumption.
  constructor.
  setoid_rewrite (set_eqvUnionAssoc G' G'' F''); setoid_rewrite <- F'eq; assumption.

  constructor.
  apply iso_union; auto.
  apply set_unionDisjElimR with M;
    apply set_disjSymm; apply set_unionDisjElimL with W''; setoid_rewrite <- W'eq;
      apply (scanFieldStOkWMDisj scOk').
  apply set_disjSymm; apply set_unionDisjElimL with F''; setoid_rewrite <- F'eq.
  clear scOk scOk'' W''copy W'copy scOk Req1 Req2.
  assert (scOk : scanFieldStOk (map_union phi phi') W' (M |_| M') F'
    ((foralls x in G', 
      objCopiedPhi phi' (rset (rset s a0 (rget s rscan)) ra CHENEYLOOP2) x) ** top) s').
  apply (scanFieldStOkImpl scOk'); clear scOk'.
  intros h Hp; sconjLift2 Hp; scm.
  apply (scanFieldStOkFGDisj scOk).
  hpunfold; stSimpl; intros; lleim.

  constructor.
  intros y Iy; destruct (set_union_elim Iy) as [Iy' | Iy']; clear Iy; [left | right].
  (* case: y in G' *)
  assert (Heq := G'eq _ Iy'); clear G'eq Iy'.
  assert (Hp'' : s'' |||- rget s rscan |-> sv ** top).
  unfold scanFieldStOk in scOk''; intuition; sconjLift6 H5; scm.
  assert (Hp' : s' |||- rget s rscan |-> sv ** top).
  unfold scanFieldStOk in scOk'; intuition; sconjLift4 H5; scm.
  rewrite (contGetEqL Hp'').
  rewrite (contGetEqL Hp') in Heq; assumption.
  (* case: y in G'' *)
  auto.

  constructor.
  apply aligned8Trans with (rget s' rfree); assumption.
  constructor.

  intros y Iy''.
  clear Req1 Req2 scOk scOk'.
  assert (C1 : s'' |||- objCopied y s' y ** top).
  unfold objCopied in W''copy |- *; unfold fieldCopied in W''copy |- *;
    generalize dependent W''copy; stSimpl; auto.
  assert (C2 : s' |||- objCopied y s y ** top).
  unfold objCopied in W'copy |- *; unfold fieldCopied in W'copy |- *;
    generalize dependent W'copy; stSimpl; intro; apply W'copy.
  rewrite W'eq; apply set_union_intro2; assumption.
  eapply objCopiedTrans; eauto.

  assert (Peqv : mapEqv (map_union (map_union phi phi') phi'')
     (map_union phi (map_union phi' phi''))).
  intro z; unfold map_union; destruct (phi z); auto.

  generalize dependent scOk''; unfold scanFieldStOk; stSimpl.
  intuition.
  setoid_rewrite <- (set_eqvUnionAssoc M M' M''); auto.
  apply (H2 x0 y); auto.
  rewrite Peqv; assumption.
  permConj H5 (7::5::1::2::3::6::4::8::nil).
  hyr H5 (sconjAssocR H5).
  permConj H5 (2::3::4::5::6::1::7::nil).
  clear Req1 Req2 W''copy.
  unfold objRespMap; lli; scm.

  unfold hpredTrue; rewrite <- (set_eqvUnionAssoc M M' M'');
    apply (mapHpPhiEqv Hp); auto.

  unfold hpredTrue; rewrite <- (set_eqvUnionAssoc M M' M''); assumption.

  unfold fieldRespMap; lli; [idtac | apply Hp].
  destruct svOk as [v0 [HpOk Meq]]; exists v0; constructor; auto.
  unfold mapX in Meq |- * .
  rewrite <- Peqv.
  destruct (isOdd v0); [auto | unfold map_union at 1].
  destruct (map_union phi phi' v0); [auto | discriminate Meq].

  generalize dependent Hp; unfold fieldRespMap; unfold mapX; stSimpl; intro; lleim.
  destruct H7 as [v0 [Hp0 X]].
  assert (E : x = v0).
  unfold scanFieldStOk in scOk, scOk' ; intuition.
  sconjLift6 H17; apply (contContEq H17 Hp0).
  subst v0.
  exists x; constructor.
  unfold scanFieldStOk in scOk; intuition.
  sconjLift5 H11; scm.
  rewrite <- Peqv; assumption.

  apply fasMerge; llMatch.

  (* show objects in G' are properly copied and forwarded *)
  apply objCopiedPhiSubmap with phi'.
  apply (objCopiedPhiHeq Hp).
  reflexivity.
  unfold map_union; intros a b E; rewrite E; reflexivity.

  (* show objects in G'' are properly copied and forwarded *)
  eapply objCopiedPhiUnionPhi.
  2:eauto.
  2:eauto.
  eapply objCopiedPhiCopied; eauto.
  stSimpl; intros y Iy.
  assert (Iy' : set_In y W').
  setoid_rewrite W'eq; apply set_union_intro1; assumption.
  assert (X := W'copy _ Iy'); scm.
  apply set_unionDisjElimR with M.
  apply set_disjSymm; apply set_unionDisjElimL with W''; setoid_rewrite <- W'eq.
  apply (scanFieldStOkWMDisj scOk').
  eapply fasFasDisj.
  instantiate (1 := unit).
  lli; apply Hp.
  cbv beta; hpunfold; intros; lleim.
  cbv beta; hpunfold; intros; lleim.

  (* show the registers are okay *)
  unfold ch2Step in Req2.
  constructor.
  intros r NL; simpl in NL; intuition.
  rewrite <- Req2; [stSimpl | regNotInListSolver].
  rewrite <- Req1; [stSimpl; reflexivity | regNotInListSolver].
  repeat (rewrite <- Req2; [stSimpl | regNotInListSolver]).
  rewrite <- Req1; [stSimpl | regNotInListSolver].
  intuition.

Qed.


(* If the loop test fails, we satisfy the loop guarantee *)
Lemma cheneyLoopExitGuarOk : forall s,
  rget s rscan = rget s rfree ->
  cheneyLoopGuar s (rset s ra (rget s raSave)).
Proof.
  intros s scFrEq.
  unfold cheneyLoopGuar; stSimpl.
  
  constructor.
  intros phi G W M F A H.
  exists (empty_map nat nat); exists W; exists empty_set; exists F; exists empty_set.
  constructor.
  apply set_eqvUnionEmptyUnit.
  constructor.
  apply set_eqvUnionEmptyUnit.
  constructor.
  apply empty_mapIso.
  constructor.
  simpl; intros; elf; auto.
  constructor.
  apply aligned8Refl.

  destruct H as [Geq [LtSt [Al H]]].
  unfold cheneyLoopStOk; stSimpl.
  rewrite scFrEq; rewrite scFrEq in Geq; 
    rewrite scFrEq in LtSt; clear scFrEq.

  assert (Geq' : set_eqv G empty_set).
  setoid_rewrite (set_eqvEmptyRangeObjs (rget s rfree)); assumption.
  clear Geq; rename Geq' into Geq.

  constructor.
  apply set_eqvEmptyRangeObjs.
  constructor.
  assumption.
  constructor.
  apply aligned8Refl.
  destruct H as [WFlt [WMdisj [Feq [PhiOk [StLt H]]]]].
  unfold scanFieldStOk; stSimpl.

  assert (Meq : set_eqv (set_union M empty_set) M).
  setoid_replace (set_union M empty_set) with (set_union empty_set M).
  apply set_eqvSymm; apply set_eqvUnionEmptyUnit.
  apply set_eqvUnionSymm.

  constructor.
  assumption.
  constructor.
  setoid_rewrite Meq; assumption.
  constructor.
  assumption.

  assert (PhiEq : forall x, map_union phi (empty_map nat nat) x = phi x).
  unfold map_union; intro x; destruct (phi x); reflexivity.

  constructor.
  intros x y; rewrite PhiEq; apply PhiOk.
  constructor.
  assumption.

  clear LtSt WFlt WMdisj Feq StLt.
  scm.

  unfold hpredTrue; setoid_rewrite Meq.
  apply (mapHpPhiEqv Hp).
  intros x; auto.

  unfold hpredTrue; setoid_rewrite Meq; auto.

  unfold hpredTrue; setoid_rewrite Geq; rewrite <- hpredTrueEq.
  apply fasIntroE.
  unfold hpredTrue in Hp; rewrite Geq in Hp; unfold objHp in Hp; exact (fasElimE Hp).

  lli; auto.

  unfold regPres; simpl; intuition.
  stSimpl; reflexivity.

Qed.


(* It is safe to perform the first field scan in the loop *)
Lemma cheneyLoopFirstScanSafe : forall s,
  cheneyLoopPre s ->
  rget s rscan <> rget s rfree -> 
  scanFieldPre (rset (rset s a0 (rget s rscan)) ra CHENEYLOOP2).
Proof.
  intros s [phi [G [W [M [F [Geq [stLe [Al scOk]]]]]]]] NE.
  exists phi; exists W; exists M; exists F.
  destruct scOk as [leWF [WMdisj [Feq [PhiOk [LE Hp]]]]].
  unfold scanFieldStOk; stSimpl.
  constructor.
  trivial.
  constructor.
  trivial.
  constructor.
  trivial.
  constructor.
  trivial.
  constructor.
  trivial.
  assert (Hp' : s |||- mapHp M phi ** objHp (W |_| M) W ** (foralls x in F, x |->__) **
    (okFieldAt (W |_| M) (rget s rscan) ** top) ** top).
  2:lle Hp'; assumption.
  scm.
  
  assert (IG : set_In (rget s rscan) G).
  apply (set_eqv_elim2 Geq); clear Geq Hp PhiOk.
  destruct Al as [k Al].
  apply rangeObjsIntro with 0; omega.

  hyr Hp (objHpElim IG Hp).
  unfold okObjAt in Hp; lle Hp.
  scm.

Qed.


(* It is safe to perform the second field scan in the loop *)
Lemma cheneyLoopSecondScanSafe : forall s s',
  cheneyLoopPre s ->
  rget s rscan <> rget s rfree ->
  scanFieldGuar (rset (rset s a0 (rget s rscan)) ra CHENEYLOOP2) s' ->
  scanFieldPre (ch2Step s').
Proof.
  intros s s' [phi [G [W [M [F [Geq [stLe [Al scOk]]]]]]]] NE [Guar Req].
  hyr Guar (Guar phi W M F (okFieldAt (W |_| M) (rget (ch2Step s') a0) ** top)).
  destruct Guar as [phi' [W' [M' [F' [G' Guar]]]]].
  unfold ch2Step; stSimpl; rewrite <- Req; [stSimpl | regNotInListSolver].
  unfold scanFieldStOk; stSimpl.
  destruct scOk as [leWF [WMdisj [Feq [PhiOk [LE Hp]]]]].
  intuition.
  clear H H0.
  assert (Hp' : s |||- mapHp M phi ** objHp (W |_| M) W ** (foralls x in F, x |->__) **
    (okFieldAt (W |_| M) (rget s rscan) ** okFieldAt (W |_| M) (4 + rget s rscan) ** top) **
    top).
  2:lle Hp'; assumption.
  scm.

  assert (IG : set_In (rget s rscan) G).
  apply (set_eqv_elim2 Geq); clear Geq Hp PhiOk.
  destruct Al as [k Al].
  apply rangeObjsIntro with 0; omega.

  hyr Hp (objHpElim IG Hp); apply sconjAssocL; scm.
  exists (map_union phi phi'); exists W'; exists (M |_| M'); exists F'.
  unfold scanFieldStOk; unfold ch2Step; stSimpl.

  generalize dependent Guar; unfold ch2Step; unfold scanFieldStOk; stSimpl.
  rewrite <- (Req rscan); [idtac | regNotInListSolver];
    rewrite <- (Req rtoStart); [idtac | regNotInListSolver];
      rewrite <- (Req rtoEnd); [idtac | regNotInListSolver]; 
	stSimpl; clear Req.
  intros [Weq [Feq [Iso' [G'eq [Al' [W'copy scfOk]]]]]].
  intuition.
  permConj H5 (1::2::3::6::5::4::7::nil); scm.
  unfold hpredTrue.
  rewrite (set_eqvUnionSymm M M').
  rewrite <- (set_eqvUnionAssoc W' M' M).
  rewrite (set_eqvUnionSymm W' M').
  rewrite <- Weq.
  assumption.

Qed.


(* it is safe to reenter the loop once we have scanned an object *)
Lemma scanObjLoopPreOk : forall s s',
  cheneyLoopPre s ->
  rget s rscan <> rget s rfree ->
  cheneyScanObjGuar s s' ->
  cheneyLoopPre s'.
Proof.
  intros s s' [phi [G [W [M [F stOk]]]]] NE Guar.
  
  assert (stOk' : cheneyLoopStOk phi G W M F 
    (okObjAt (W |_| M) (rget s rscan) ** objHp (W |_| M) (set_remove (rget s rscan) G) ** top) s).
  apply (cheneyLoopStOkImpl stOk).
  intros h Hp; apply sconjAssocL; scm.
  apply objHpElim; auto.
  destruct stOk as [Geq [_ [Al _]]]; setoid_rewrite Geq.
  apply rangeObjsIntro with 0; [idtac | auto].
  destruct Al; omega.

  clear stOk; rename stOk' into stOk.
  destruct Guar as [Guar [Req [scEq _]]].
  destruct (Guar _ _ _ _ _ _ stOk)
    as [phi' [W' [M' [F' [G' [Weq [Feq [Iso' [G'eq [Al' [W'copy stOk']]]]]]]]]]]; clear Guar.
  unfold cheneyLoopPre.

  exists (map_union phi phi'); exists (set_remove (rget s rscan) G |_| G');
    exists W'; exists (M |_| M'); exists F'.
  apply (cheneyLoopStOkImpl stOk'); clear G'eq.

  intros h Hp; sconjLift2 Hp; sconjLift3 Hp; hyr Hp (sconjAssocR Hp).
  scm.
  unfold hpredTrue; setoid_rewrite (set_eqvUnionSymm M M');
    setoid_rewrite <- (set_eqvUnionAssoc W' M' M); 
    setoid_rewrite (set_eqvUnionSymm W' M'); setoid_rewrite <- Weq; rewrite <- hpredTrueEq.
  apply objHpMerge; scm.
  apply (objHpIso _ _ Hp) with M' (objHp (W |_| M) W' ** top); auto.

  unfold cheneyLoopStOk in stOk; unfold scanFieldStOk in stOk; intuition.
  apply sconjAssocL; sconjLift2 H8; scm.
  apply objHpSplit.
  unfold hpredTrue; rewrite <- Weq; assumption.
  destruct stOk' as [_ [_ [_ stOk']]].
  apply set_unionDisjElimR with M; apply set_disjSymm;
    apply (scanFieldStOkWMDisj stOk').
  
Qed.


(* The "inductive step" of the loop guarantee: 
   running the loop once (by scanning an object), then running it k times,
   behaves the same as running it k+1 times. *)
Lemma scanObjLoopGuarOk : forall s s' s'',
  rget s rscan <> rget s rfree ->
  cheneyScanObjGuar s s' ->
  cheneyLoopGuar s' s'' -> 
  cheneyLoopGuar s s''.
Proof.
  intros s s' s'' NE Guar1 Guar2.
  constructor; [idtac | destruct Guar1 as [_ Req1]; destruct Guar2 as [_ Req2]].

  intros phi G W M F A stOk.

  (* step through the object scan *)
  assert (stOk' : exists h,
    cheneyLoopStOk phi G W M F (okObjAt (W |_| M) (rget s rscan) ** heap_eqv h ** A) s /\
    h ||- objHp (W |_| M) (set_remove (rget s rscan) G)).
  clear Guar1 Guar2; destruct stOk as [Geq [Le [Al1 scOk]]];
    unfold cheneyLoopStOk; unfold scanFieldStOk in scOk |- *; intuition.
  assert (Hp : s |||- mapHp M phi ** objHp (W |_| M) W **
    (foralls x in F, x |->__) ** (okObjAt (W |_| M) (rget s rscan) **
    objHp (W |_| M) (set_remove (rget s rscan) G)) ** A).
  scm; apply objHpElim; auto.
  rewrite Geq.
  destruct Al1; apply rangeObjsIntro with 0; omega.
  clear H5; lle Hp; sconjLift5 Hp; destruct (sconjExtract Hp) as [h [Hp' hOk]];
    exists h; constructor; unfold scanFieldStOk; intuition.
  permConj Hp' (2::3::4::5::1::6::nil); assumption.

  clear stOk; destruct stOk' as [h [stOk hOk]].
  destruct Guar1 as [Guar1 [Req1 [scEq _]]].
  destruct (Guar1 _ _ _ _ _ _ stOk) as
    [phi' [W' [M' [F' [G' [Weq [Feq [Iso' [G'eq [Al' [W'copy stOk']]]]]]]]]]];
    clear Guar1.

  (* step through the rest of the iterations of the loop *)
  destruct Guar2 as [Guar2 [Req2 raEq]].
  destruct (Guar2 (map_union phi phi') (set_remove (rget s rscan) G |_| G') W' (M |_| M') F'
    (objRespMap (map_union phi phi') s (rget s rscan) ** A)) as
    [phi'' [W'' [M'' [F'' [G'' Guar2']]]]]; clear Guar2.
  apply (cheneyLoopStOkImpl stOk'); clear Req2.
  intros h' Hp.
  permConj Hp (3::2::1::4::nil); hyr Hp (sconjAssocR Hp); scm.
  hyr Hp (sconjInsert Hp hOk).
  unfold hpredTrue; setoid_rewrite (set_eqvUnionSymm M M');
    setoid_rewrite <- (set_eqvUnionAssoc W' M' M); 
    setoid_rewrite (set_eqvUnionSymm W' M'); setoid_rewrite <- Weq; rewrite <- hpredTrueEq.
  apply objHpMerge; scm.
  apply (objHpIso _ _ Hp) with M' (objHp (W |_| M) W' ** top); auto.

  unfold cheneyLoopStOk in stOk; unfold scanFieldStOk in stOk; intuition.
  apply sconjAssocL; sconjLift2 H8; scm.
  apply objHpSplit.
  unfold hpredTrue; rewrite <- Weq; assumption.
  destruct stOk' as [_ [_ [_ scOk']]].
  apply set_unionDisjElimR with M; apply set_disjSymm;
    apply (scanFieldStOkWMDisj scOk').

  (* put it together *)
  destruct Guar2' as [W'eq [F'eq [Iso'' [G''eq [Al'' stOk'']]]]].
  exists (map_union phi' phi''); exists W''; exists (M' |_| M''); exists F''; exists (G' |_| G'').

  constructor.
  setoid_rewrite (set_eqvUnionAssoc M' M'' W''); setoid_rewrite <- W'eq; assumption.
  constructor.
  setoid_rewrite (set_eqvUnionAssoc G' G'' F''); setoid_rewrite <- F'eq; assumption.
  constructor.
  apply iso_union; auto.
  apply set_unionDisjElimR with M; apply set_disjSymm; 
    apply set_unionDisjElimL with W''; rewrite <- W'eq.
  destruct stOk' as [_ [_ [_ X]]]; apply (scanFieldStOkWMDisj X).
  apply set_disjSymm; apply set_unionDisjElimL with F''; rewrite <- F'eq.
  assert (Z : scanFieldStOk (map_union phi phi') W' (M |_| M') F'
    ((foralls x in G', objCopiedPhi phi' s x) ** top) s').
  destruct stOk' as [_ [_ [_ X]]]; apply (scanFieldStOkImpl X); 
    intros h' Hp; sconjLift2 Hp; scm.
  apply (scanFieldStOkFGDisj Z).
  hpunfold; illeim.

  constructor.
  (* everything in G' is reachable from something in G *)
  assert (G'rch : forall x', set_In x' G' -> exists x, set_In x G /\ reachable (heapOf s'') x x').
  intros x' Ix'.
  exists (rget s rscan); constructor.
  destruct stOk as [Geq [_ [[k Al] _]]]; setoid_rewrite Geq; apply rangeObjsIntro with 0; omega.
  assert (Heq := G'eq _ Ix'); clear G'eq Ix'.
  destruct stOk' as [_ [_ [_ stOk']]].
  destruct stOk'' as [_ [_ [_ stOk'']]].
  unfold scanFieldStOk in stOk', stOk''; unfold objRespMap in stOk', stOk''.
  destruct stOk' as [_ [_ [_ [_ [_ stOk']]]]].
  destruct stOk'' as [_ [_ [_ [_ [_ stOk'']]]]].
  lle stOk'; lle stOk''.
  destruct Heq as [Heq | Heq].
  econstructor 2.
  2:constructor 1.
  2:econstructor 3.
  3:constructor 1.
  unfold POnat.POnat.A in * |- *; rewrite <- Heq.
  sconjLift4 stOk'; sconjLift6 stOk''; eapply fieldRespMapEq; scm.
  unfold POnat.POnat.A in * |- *; rewrite <- Heq.
  sconjLift5 stOk'; sconjLift7 stOk''; eapply fieldRespMapEq; scm.

  intros x' Ix; destruct (set_union_elim Ix) as [Ix' | Ix']; clear Ix.
  auto.
  (* everything in G'' is reachable from something in G *)
  destruct (G''eq _ Ix') as [x [Ix Rchx]]; clear G''eq Ix'.
  destruct (set_union_elim Ix) as [Ix' | Ix']; clear Ix.
  destruct (set_remove_elim _ _ _ Ix') as [_ Ix]; exists x; constructor; assumption.
  destruct (G'rch _ Ix') as [z [Iz Rchz]]; exists z; constructor; auto.
  apply reachableTrans with x; assumption.

  constructor.
  apply aligned8Trans with (rget s' rfree); assumption.
  apply cheneyLoopStOkPhiEqv with (map_union (map_union phi phi') phi'').
  (* updating morphisms broke this somehow so now we have to be explicit *)
  rewrite <- (cheneyLoopStOkMorphism (map_union (map_union phi phi') phi'')
    (set_eqvRefl empty_set) (set_eqvRefl W'') (set_eqvUnionAssoc M M' M'')
    (set_eqvRefl F'') ((foralls x in G, objRespMap (map_union phi (map_union phi' phi'')) s x) **
      (foralls x in G' |_| G'', objIsFwded phi (map_union phi' phi'') s x) **
      A) _ _ (state_eqvRefl s'')).
  apply (cheneyLoopStOkImpl stOk''); clear stOk''.
  intros h' Hp.
  assert (Hp' : h' ||- 
    ((foralls x in set_remove (rget s rscan) G,
      objRespMap (map_union (map_union phi phi') phi'') s' x) **
    (foralls x in G', objRespMap (map_union (map_union phi phi') phi'') s' x)) **
    (foralls x in G'', objIsFwded (map_union phi phi') phi'' s' x) **
    objRespMap (map_union phi phi') s (rget s rscan) ** A).
  scm; apply fasSplit; auto.
  (* set_disj (set_remove (rget s rscan) G) G' *)
  apply set_disjSymm; apply set_unionDisjElimL with F'; rewrite <- Feq.
  assert (X : scanFieldStOk phi W M F
    ((foralls x in set_remove (rget s rscan) G, x |->__) ** top) s).
  destruct stOk as [_ [_ [_ X]]]; apply (scanFieldStOkImpl X); clear X.
  intros h' Hp'.
  sconjLift2 Hp'; hyr Hp' (sconjInsert Hp' hOk); scm.
  hyr Hp0 (objHpIsObjs Hp0); assumption.
  apply (scanFieldStOkFGDisj X); illeim.
  
  lle Hp'; clear Hp Req2.
  aspConj Hp' ([| [|4, 1|], [|[|2, 3|], 5|]|]); scm.
  assert (II : set_In (rget s rscan) G).
  destruct stOk as [Geq [_ [[k Al] _]]]; rewrite Geq; apply rangeObjsIntro with 0; omega.
  unfold hpredTrue; rewrite <- (addRemoveEqv II).
  rewrite <- hpredTrueEq; apply fasIntro; scm.

  apply objRespMapSubPhi with (map_union phi phi'); auto.
  unfold map_union; intros x y E; destruct (phi x); auto; destruct (phi' x); auto;
    discriminate E.
  llMatch.
  apply objRespMapSubPhi with (map_union (map_union phi phi') phi'').
  2:unfold map_union; intros x' y' E; destruct (phi x'); auto; destruct (phi' x'); auto;
    discriminate E.

  apply (objRespMapCopied Hp).
  hyr hOk (objHpElim I hOk); hpunfoldIn hOk; lle hOk.
  clear Hp; assert (Hp : s |||- (x |-> x0 ** 4 + x |-> x1) ** top).
  destruct stOk as [_ [_ [_ X]]]; unfold scanFieldStOk in X; intuition;
    sconjLift5 H7; hyr H7 (sconjInsert H7 hOk); lleim.
  assert (Hp' : s' |||- (x |-> x0 ** 4 + x |-> x1) ** top).
  destruct stOk' as [_ [_ [_ X]]]; unfold scanFieldStOk in X; intuition;
    sconjLift6 H7; hyr H7 (sconjInsert H7 hOk); lleim.
  lle Hp; scm; hpunfold; lli; eauto.
  scm.
  sconjLift2 Hp; scm.

  intros IX; destruct (set_remove_elim _ _ _ IX) as [Q _]; auto.
  
  (* show that G' and G'' are copied and forwarded *)
  apply fasMerge; llMatch.

  (* G' *)
  eapply copyMapFwded.
  2:instantiate (1 := s').
  instantiate (1 := top ** top); apply sconjAssocL.
  destruct stOk' as [_ [_ [_ X]]]; unfold scanFieldStOk in X; intuition.
  sconjLift5 H5; scm.
  hyr Hp0 (fasElim I Hp0); scm.
  apply (objCopiedPhiSubmap Hp0).
  unfold map_union; intros z; destruct (phi' z); [auto | discriminate 1].
  apply (objRespMapSubPhi Hp).
  unfold map_union; intros x' y' E; destruct (phi x'); auto; destruct (phi' x'); 
    auto; discriminate E.

  (* G'' *)
  apply objIsFwdShiftRight.
  apply (objCopiedFwded1 Hp Iso'').
  assumption.
  intros y Iy; apply W'copy; rewrite W'eq; sui1; assumption.

  clear x h0 I Hp; intros x y E.
  assert (Ix : set_In x W').
  rewrite W'eq; sui1.
  destruct Iso'' as [X _]; destruct (X x) as [_ Y]; apply Y; rewrite E; discriminate.
  clear y E.
  destruct Iso' as [Iso' _]; destruct (Iso' x) as [_ Y]; destruct (phi' x); auto.
  elf; apply set_disj_elim with W' (M |_| M') x.
  destruct stOk' as [_ [_ [_ X]]]; apply (scanFieldStOkWMDisj X).
  auto.
  sui2.
  apply Y; discriminate.

  apply (fasFasDisj' Hp).
  unfold objRespMap; unfold fieldRespMap; illeim.
  hpunfold; illeim.

  unfold map_union; intros x; destruct (phi x); auto; destruct (phi' x); auto.


  (* register preservation *)
  destruct Req1 as [Req1 [scEq _]]; destruct Req2 as [Req2 raEq].
  constructor.
  intros x; simpl; intuition.
  rewrite <- Req2; [idtac | regNotInListSolver];
    rewrite <- Req1; [reflexivity | regNotInListSolver].
  rewrite raEq; rewrite <- Req1; [auto | regNotInListSolver].

Qed.


(* Verification of the cheney loop, by combining the previously defined lemmaas. *)
Lemma cheneyLoopOk : iseqOkS Ploop cheneyLoopPre cheneyLoopGuar cheneyLoop.
Proof.
  unfold cheneyLoop.
  iseqWeakSolver PloopSimpl.

  intros; apply scanFieldGuarPresRa; auto.

  cbv beta; intros s Pre; bsteps.

  (* return case *)
  constructor.
  unfold cheneyReturnPre; trivial.
  unfold cheneyReturnGuar; intros s' E; subst s';
    apply cheneyLoopExitGuarOk; auto.

  (* loop case *)
  apply callChain.
  constructor.
  apply cheneyLoopFirstScanSafe; auto.

  unfold cheneyLoop2Pre.
  intros s' G'.
  unfold cheneyLoop2Guar.
  apply callChain.
  constructor.
  eapply cheneyLoopSecondScanSafe; eauto.

  intros s'' G''.
  assert (G12 := scanFieldsScanObj Bt G' G''); clear G' G''.
  constructor.
  eapply scanObjLoopPreOk; eauto.
  intros; eapply scanObjLoopGuarOk; eauto.
Qed.



