Require ListMap.
Require POnat.
Require Import OptUtil.
Require Import NatUtil.

Module HeapMap := ListMap.ListMap POnat.POnat POnat.OneTypeNat.
Import HeapMap.

Notation addr := nat (only parsing).
Notation hval := nat (only parsing).

Definition heap := map.

Definition empty_heap : heap := empty_map.

Definition get (h:heap) (a:nat) : option hval := 
  if isPtr a then
    map_get h (div4 a)
  else
    None.

Definition uset (h:heap) (l:nat) : hval -> heap := 
  map_add h (div4 l).

Definition set (h:heap) (a:addr) (v:hval) :=
  do v0 <-- get h a;
    ret (uset h a v).


(* heap specifications *)

Definition heap_canon : heap -> heap := map_canon.
Definition heap_canonP : heap -> Prop := map_canonP.

Definition heap_eqv : heap -> heap -> Prop := map_eqv.

Definition heap_join : heap -> heap -> heap -> Prop := mapJoin.

Definition heap_dom (h:heap) : _ := ASet.set_map (fun x => 4 * x) (map_dom h).

Definition heap_remove (h:heap) (x:nat) : heap :=
  map_remove h (div4 x).

Definition singleton_heap (l v:nat) : heap := 
  singleton_map (div4 l) v.

Definition heap_union : heap -> heap -> heap := map_union.

Definition subHeap (h h':heap) : Prop :=
  forall l v, get h l = Some v -> get h' l = Some v.


(* this should go somewhere else, like a heap theory.  along with
   the heap specs stuff *)

Set Implicit Arguments.


(* should these be in a different file? *)

Lemma heap_joinSub : forall h1 h2 h,
  heap_join h1 h2 h ->
  subHeap h1 h.
Proof.
  intros h1 h2 h [J] x y G.
  assert (J' := J (div4 x)).
  unfold get in G |- * .
  destruct (isPtr x).
  unfold HeapMap.MapDef.get in J'.
  rewrite G in J'.
  inversion J'; auto.
  discriminate G.
Qed.


Lemma subHeapTrans : forall s1 s2,
  subHeap s1 s2 -> forall s3,
    subHeap s2 s3 ->
    subHeap s1 s3.
Proof.
  intros s1 s2 S12 s3 S23 x v G.
  rewrite (S23 x v).
  reflexivity.
  rewrite (S12 x v); auto.
Qed.



Import ASet.

Lemma subHeapSplit : forall ha h,
  subHeap ha h -> 
  exists hb,
    heap_join ha hb h.
Proof.
  intros ha h S.
  exists (map_diff h ha).
  constructor; intros k.

  assert (S' : forall v, map_get ha k = Some v -> map_get h k = Some v).
  assert (S' := S (4 * k)); clear S; unfold get in S'.
  rewrite div44eq in S'.
  assert (Q := isPtrOk (4 * k)).
  destruct (isPtr (4 * k)).
  assumption.
  elimtype False; apply Q.
  apply (ptrP_intro k); reflexivity.

  clear S; rename S' into S.
  assert (X := map_mem_In_dom k h).
  unfold MapDef.get.
  assert (D : (exists v, map_get h k = Some v) \/ map_get h k = None).
  destruct (map_get h k); [left | right].
  econstructor; reflexivity.
  reflexivity.

  destruct D as [[b E] | E]; rewrite E; rewrite E in S.

  assert (Y := map_mem_In_dom k ha).
  assert (Z := map_mem_In_dom k (map_diff h ha)).

  destruct (map_get ha k).
  replace b0 with b.
  replace (map_get (map_diff h ha) k) with (None (A := nat)).
  constructor.

  destruct (map_get (map_diff h ha) k).
  elimtype False.
  assert (Idiff : set_In k (set_diff (map_dom h) (map_dom ha))).
  assert (Q := map_domDiffEqv h ha).
  setoid_rewrite <- Q.
  assumption.
  destruct (set_diff_elim Idiff) as [HA HB].
  exact (HB Y).
  reflexivity.

  assert (Some b = Some b0).
  intuition.
  injection H; auto.

  replace (map_get (map_diff h ha) k) with (Some b).
  constructor.
  symmetry.
  rewrite <- (map_get_diffNotInEq _ _ Y); assumption.

  rewrite E in X.
  replace (map_get ha k) with (None (A := nat)).
  assert (NI : ~set_In k (map_dom ha)).
  assert (Y := map_mem_In_dom k ha).
  destruct (map_get ha k).
  assert (None = Some b).
  eauto.
  discriminate H.
  assumption.
  rewrite <- (map_get_diffNotInEq _ _ NI).
  rewrite E.
  constructor.

  destruct (map_get ha k).
  assert (None = Some b).
  eauto.
  assumption.
  reflexivity.

Qed.


Lemma heapDom_eqv : forall h h',
  heap_eqv h h' ->
  set_eqv (heap_dom h) (heap_dom h').
Proof.
  intros h h' Heq.
  unfold heap_dom.
  assert (E := mapDomEqv Heq); clear Heq.
  rewrite (set_map_eqv (fun x : POnat.POnat.A => 4 * x) E).
  apply set_eqvRefl.
Qed.



Lemma heap_canonEqv : forall h x,
  get h x = get (heap_canon h) x.
Proof.
  intros h x.
  unfold get.
  destruct (isPtr x).
  2:reflexivity.
  rewrite <- map_canonEqv.
  reflexivity.
Qed.



Lemma heap_eqvRefl : forall h,
  heap_eqv h h.
Proof.
  intros h x; reflexivity.
Qed.

Lemma heap_eqvSymm : forall h h',
  heap_eqv h h' ->
  heap_eqv h' h.
Proof.
  intros h h' E x; auto.
Qed.


Lemma heap_eqvTrans : forall h h' h'',
  heap_eqv h h' ->
  heap_eqv h' h'' ->
  heap_eqv h h''.
Proof.
  intros h h' h'' E1 E2 x; rewrite E1; rewrite E2; reflexivity.
Qed.
