Require Import OptUtil.
Require Import StateTheory.
Require Import EqNat.
Require Import WeakSCAP.
Require Import SpecUtil.
Require Import VGC.
Require Import LinkedList.


Set Implicit Arguments.





Module ListSumGuar <: BASIC_MGI_ARG.

  Definition gcGuarR vs vs' r : Prop :=
    (rget vs r = NULL -> rget vs' r = NULL) /\
    (forall m n y,
      vs |||- rget vs r |->, (m::y::nil) ** ListSumAt n y ** top ->
	atomP m ->
	exists y', vs' |||- rget vs' r |->, (m::y'::nil) ** ListSumAt n y' ** top).

  Definition gcGuar vs vs' rr : Prop :=
    rr = t0::nil ->
    gcGuarR vs vs' t0.

End ListSumGuar.


Module ListSumMGI := BasicMGI (ListSumGuar).

Definition rlist := ListSumMGI.rroot.

Definition rsum := s0.


Definition MENTER  := 0.
Definition MLOOP   := 1.
Definition MLOOP2  := 2.
Definition MLOOP3  := 3.
Definition MRET    := 4.

Notation READ1 := ListSumMGI.read1Lbl.
Notation READ2 := ListSumMGI.read2Lbl.


Definition menter :=
  addiu rsum zero 1;
  j MLOOP.

Definition mloop :=
  addiu ast zero NULL;
  beq rlist ast MRET;;
  jal READ1 MLOOP2.

Definition mloop2 :=
  addiu ast zero 1;
  subu v0 v0 ast;
  addu rsum rsum v0;
  jal READ2 MLOOP3.

Definition mloop3 :=
  addiu rlist v0 0;
  j MLOOP.

Definition mret :=
  jr ra.


Module ListSumVerif (V : VIEW_GC with Module I := ListSumMGI).

  Notation weakPre := (fun s => True).
  Notation trivGuar := (fun s s' => True).

  Definition menterPre s :=
    exists n, exists vs,
      s |||- V.repr vs s (rlist::nil) ** top /\
      vs |||- ListSumAt n (rget vs rlist) ** top.

  Definition menterGuar s s' :=
    forall vs A n,
      s |||- V.repr vs s (rlist::nil) ** A ->
      vs |||- ListSumAt n (rget vs rlist) ** top ->
	exists vs',
	  s' |||- V.repr vs' s' (rlist::nil) ** A /\
	  rget s' rsum = natEnc n.

  Definition mloopGuar s s' :=
    forall vs m n A,
      s |||- V.repr vs s (rlist::nil) ** A ->
      vs |||- ListSumAt m (rget vs rlist) ** top ->
      rget s rsum = natEnc n ->
      exists vs',
	s' |||- V.repr vs' s' (rlist::nil) ** A /\
	rget s' rsum = natEnc (n + m).


  Definition mloop3Pre s :=
    menterPre (rset s rlist (rget s v0)).
  
  Definition mloop3Guar s s' :=
    mloopGuar (rset s rlist (rget s v0)) s'.


  Notation mloop2Step :=
    (fun s => 
      rset (rset (rset s ast 1) v0 (rget s v0 - 1)) rsum (rget s rsum + (rget s v0 - 1))).

  Notation readPre := (readPre V.repr ListSumMGI.rroot).
  Notation argSpecAdder := (fun sp =>
    sp V.repr ListSumMGI.rroot ListSumMGI.gcGuar ListSumMGI.presReg).
  Notation readGuar := (argSpecAdder readGuar).
  Notation read1Spec := (argSpecAdder read1Spec).
  Notation read2Spec := (argSpecAdder read2Spec).

  Definition mloop2Pre s :=
    readPre 4 (mloop2Step s) /\
    forall s', readGuar 4 (rset (mloop2Step s) ra MLOOP3) s' -> mloop3Pre s'.

  Definition mloop2Guar s s'' :=
    exists s',
      readGuar 4 (rset (mloop2Step s) ra MLOOP3) s' /\
      mloop3Guar s' s''.
      
  Definition menterSpec := bspeci menterPre menterGuar.
  Definition mloopSpec := bspeci menterPre mloopGuar.
  Definition mloop2Spec := bspeci mloop2Pre mloop2Guar.
  Definition mloop3Spec := bspeci mloop3Pre mloop3Guar.
  Definition mretSpec := bspeci weakPre (fun s s' => s = s').


  Definition mPhi (f:nat) :=
    if beq_nat MENTER f then
      SomeT menterSpec
    else if beq_nat MLOOP f then
      SomeT mloopSpec
    else if beq_nat MLOOP2 f then
      SomeT mloop2Spec
    else if beq_nat MLOOP3 f then
      SomeT mloop3Spec
    else if beq_nat MRET f then
      SomeT mretSpec
    else if beq_nat READ1 f then
      SomeT read1Spec
    else if beq_nat READ2 f then
      SomeT read2Spec
    else
      NoneT _.


  Ltac specUnfolder := 
    unfold menterSpec;
    unfold mloopSpec;
    unfold mloop2Spec;
    unfold mloop3Spec;
    unfold mretSpec;
    unfold VGC.read1Spec;
    unfold VGC.read1Spec.

  Ltac Psimpl :=
    unfold mPhi;
    unfold VGC.read1Spec;
    unfold VGC.read2Spec;
    rewrite beq_true;
    repeat (rewrite beq_false; [idtac | discriminate]);
    specUnfolder; reflexivity.


  Ltac nonRootSolver :=
    sconjMatch;
    apply V.setNonRootOk;
      [try assumption |
	let X := fresh in
	  (simpl; intro X; destruct X as [X | X]; [discriminate X | intuition])].
  

  Lemma menterOk : iseqOkS mPhi menterPre menterGuar menter.
  Proof.
    unfold menter.
    iseqWeakSolver Psimpl.
    
    cbv beta.
    intros s [n [vs [V Hp]]].
    rewrite addiuStep.
    stSimpl; simpl.
    unfold menterGuar, menterGuar, menterPre, mloopGuar.
    simpl; stSimpl.
    
    constructor.

    exists n; exists vs; constructor.
    nonRootSolver.
    discriminate H0.
    assumption.
    
    intros s' F vs' A m V' Hp'.
    destruct (F vs' m 0 A) as [vs'' [V'' E]]; clear F.
    nonRootSolver.
    discriminate H0.
    assumption.
    reflexivity.
    rewrite E; clear E.
    exists vs''; constructor.
    assumption.
    reflexivity.
  
  Qed.


  Lemma mloop3Ok : iseqOkS mPhi mloop3Pre mloop3Guar mloop3.
  Proof.
    unfold mloop3.
    iseqWeakSolver Psimpl.
    
    cbv beta.
    intro s.
    rewrite addiuStep.
    unfold mloop3Pre; unfold mloop3Guar.
    auto.
  Qed.


  Lemma readRApres : forall k s s', 
    readPre k s -> readGuar k s s' ->
    rget s ra = rget s' ra.
  Proof.
    unfold VGC.readGuar.
    intros k s s' _ [_ G].
    rewrite (regPresEq G).
    reflexivity.
    constructor.
    constructor; reflexivity.
    discriminate.
  Qed.


  Lemma mloop2Ok : iseqOkS mPhi mloop2Pre mloop2Guar mloop2.
  Proof.
    unfold mloop2.
    iseqWeakSolver Psimpl.
    apply readRApres.

    cbv beta.
    intro s.
    rewrite addiuStep; stSimpl; replace (1 + 0) with 1; [idtac | auto].
    rewrite subuStep; stSimpl.
    rewrite adduStep; stSimpl.
    unfold mloop2Pre; unfold mloop2Guar.
    intuition.
    destruct H0 as [vs [Hp Vhp]]; exists vs; constructor.
    simpl; simpl in Hp; nonRootSolver.
    discriminate H0.
    assumption.

  Qed.


  Lemma rootNullEq : forall vs A s,
    s |||- V.repr vs s (rlist::nil) ** A ->
    if beq_nat (rget s rlist) NULL then
      rget vs rlist = NULL
    else
      rget vs rlist <> NULL.
  Proof.
    assert (NP : V.I.atomP NULL).
    unfold NULL; constructor; constructor.

    intros vs h s V.
    assert (X := V.atomRegEq rlist V).
    compare (rget s rlist) NULL.

    intro E.
    rewrite E; rewrite beq_true.
    rewrite <- X.
    assumption.
    simpl; auto.
    left; replace (RFile.get (rfileOf s) rlist) with (rget s rlist).
    unfold V.I.atomP.
    unfold atomP.

    rewrite E; apply NP.
    reflexivity.
    
    intro Ne; rewrite beq_false.
    2:assumption.
    intro R.
    apply Ne.
    replace (rget s rlist) with (RFile.get (rfileOf s) rlist).
    rewrite X.
    assumption.
    left; reflexivity.
    right; rewrite R; assumption.
    reflexivity.

  Qed.


  Lemma gcHeapStep : forall vs vs' m n y, 
    vs |||- rget vs rlist |-> m ** 4 + rget vs rlist |-> y ** ListSumAt n y ** top->
      V.I.gcGuar vs vs' (rlist::nil) ->
      atomP m ->
      exists y',
	vs' |||- rget vs' rlist |-> m ** 4 + rget vs' rlist |-> y' ** ListSumAt n y' ** top.
  Proof.
    intros vs vs' m n y Hp G mAtom.
    destruct G as [_ Gptr]. auto.
    destruct (Gptr m n y) as [y' Vhp']; clear Gptr.
    linLogIntro; assumption.
    assumption.
    linLogElim Vhp'.
    econstructor; sconjMatch.
  Qed.


  Lemma natEncAtom : forall m,
    atomP (natEnc m).
  Proof.
    intros m; unfold natEnc; constructor; intuition.
  Qed.

  Lemma mloopOk : iseqOkS mPhi menterPre mloopGuar mloop.
  Proof.
    unfold mloop.
    iseqWeakSolver Psimpl.
    apply readRApres.
    
    intros s [n [vs [V Hp]]].
    rewrite addiuStep.
    stSimpl; replace (NULL + 0) with NULL.
    2:reflexivity.
    simpl.

    compare (rget s rlist) NULL.

    (* case: list is empty *)
    intro R; rewrite R; rewrite beq_true.
    constructor.
    trivial.
    intros s' E; subst s'.
    unfold mloopGuar; simpl; stSimpl.
    intros vs' m n' A V' Hp' Seq.
    exists vs'; constructor.
    simpl; nonRootSolver.
    discriminate H0.
    stSimpl; rewrite Seq; clear Seq.
    assert (NE := rootNullEq _ V');
      rewrite R in NE; rewrite beq_true in NE.
    rewrite NE in Hp'.
    assert (XX : vs' |||- (!(m = 0) ** top) ** top).
    sconjMatch.
    linLogIntro.
    apply (ListSumAtElimNil1 Hp0).
    linLogElim XX; subst m; clear XX.
    unfold natEnc; omega.

  
    (* case: list isn't empty *)
    intro Ne; rewrite (beq_false Ne).
    assert (Ne' := rootNullEq _ V); rewrite (beq_false Ne) in Ne'.
    destruct (ListSumAtElimConsSt Hp Ne') as [m [n' [x' [E Hp']]]].
    clear Hp; rename Hp' into Hp.
    linLogElim Hp; subst n.

    apply callChain.
    constructor.

    exists vs; constructor.
    replace ListSumMGI.rroot with rlist; [idtac | reflexivity].
    simpl.
    scm.
    nonRootSolver.
    nonRootSolver.
    discriminate H0.
    discriminate H0.
    linLogIntro; sconjMatch.

    intros s' [F Rp].
    destruct (F vs top (rlist::nil)) as [vs' [G1 [Hp' Vhp']]].
    simpl; sconjMatch.
    nonRootSolver.
    nonRootSolver.
    discriminate H0.
    discriminate H0.
    simpl; auto.
    simpl; intros r U; destruct U as [U | U].
    auto.
    elf; assumption.

    destruct (gcHeapStep Hp G1) as [x'' Hp'']; clear G1.
    apply natEncAtom.
    simpl in Vhp'.
    assert (X := contContEq Vhp' Hp'').
    clear Vhp'; rename Hp'' into Vhp'.
    
    constructor.
    constructor.
    
    exists vs'.
    constructor.
    replace ListSumMGI.rroot with rlist; [idtac | reflexivity].
    stSimpl.
    scm.
    replace (rfileOf (rset (rset (rset s' ast 1) v0 (rget s' v0 - 1)) rsum
            (rget s' rsum + (rget s' v0 - 1)))) with
    (RFile.set (RFile.set (RFile.set (rfileOf s') ast 1) v0 (rget s' v0 - 1)) rsum
      (rget s' rsum + (rget s' v0 - 1))).
    nonRootSolver.
    nonRootSolver.
    nonRootSolver.
    apply V.rootWeaken with (v0::rlist::nil); auto.
    unfold incl; intuition.
    discriminate H0.
    discriminate H0.
    discriminate H0.
    reflexivity.
    sconjLift2 Vhp'; lli; scm.

    intros s'' [G2 Rp2].
    destruct (G2 vs' top (rlist::nil)) as [vs3 [G3 [Hp3 Vhp3]]]; clear G2.
    simpl.
    sconjMatch.
    apply V.setNonRootOk.
    apply V.setNonRootOk.
    apply V.setNonRootOk.
    apply V.setNonRootOk.
    eapply V.rootWeaken; eauto.
    intuition.
    simpl; intro Z; destruct Z as [Z | [Z | Z]]; try assumption; discriminate Z.
    simpl; intro Z; destruct Z as [Z | [Z | Z]]; try assumption; discriminate Z.
    simpl; intro Z; destruct Z as [Z | [Z | Z]]; try assumption; discriminate Z.
    simpl; intro Z; destruct Z as [Z | [Z | Z]]; try assumption; discriminate Z.
    simpl; auto.
    simpl; intros r Z; destruct Z as [Z | Z].
    rewrite <- Z; left; reflexivity.
    elf; assumption.

    destruct (gcHeapStep Vhp' G3) as [y Hp'']; clear G3.
    apply natEncAtom.
    assert (y = rget vs3 v0).
    sconjLift2 Hp''; exact (contContEq Hp'' Vhp3).
    subst y; clear Vhp3.

    exists n'; exists (rset vs3 rlist (rget vs3 v0)); constructor.
    simpl; sconjMatch.
    eapply V.rootWeaken.
    instantiate (1 := (rlist::v0::nil)).
    replace (rget s'' v0) with (RFile.get (rfileOf s'') v0).
    eapply V.rootCopyOk.
    eapply V.rootWeaken; eauto.
    unfold incl; simpl; intuition.
    intuition.
    simpl; intro Z; destruct Z as [Z | [Z | Z]]; try assumption; discriminate Z.
    reflexivity.
    intuition.

    stSimpl.
    sconjLift3 Hp''; sconjMatch.
    
    clear V Ne' Hp vs vs' Hp' m n' x'' Vhp' X.
    intros s''' [s'' [[G1 Rp1] G2]].
    intros vs m n A V Hp Enc.
    destruct (ListSumAtElimConsSt Hp) as [y [z [yz [Seq Hp2]]]].
    intro E.
    replace (rget s rlist) with (RFile.get (rfileOf s) rlist) in Ne.
    rewrite (V.atomRegEq rlist V) in Ne.
    auto.
    simpl; auto.
    right; rewrite E; unfold NULL; constructor; constructor.
    reflexivity.

    simpl in Hp2; linLogElim Hp2; clear Hp; rename Hp2 into Hp.
    subst m.
    destruct (F vs A (rlist::nil)) as [vs' [GG1 [Hp' Vhp']]]; clear F.
    simpl; repeat nonRootSolver.
    discriminate H0.
    discriminate H0.
    simpl; auto.
    simpl; intros r [E | E]; [rewrite <- E; left; reflexivity | elf; assumption].
    
    destruct (gcHeapStep Hp GG1) as [x'' Hp'']; clear GG1.
    apply natEncAtom.
    simpl in Hp'.
    assert (Y := contContEq Vhp' Hp'').

    clear Hp Vhp'; rename Hp'' into Vhp'.
    clear V.

    destruct (G1 vs' A (rlist::nil)) as [vs'' X]; clear G1.
    simpl; repeat nonRootSolver.
    eapply V.rootWeaken; eauto; simpl; intuition.
    discriminate H0.
    discriminate H0.
    discriminate H0.
    discriminate H0.
    simpl; auto.
    simpl; intros r [E | E]; [rewrite <- E; left; reflexivity | elf; auto].
    
    destruct X as [GG [HP VV1]].
    destruct (gcHeapStep Vhp' GG) as [x3 Hp'']; clear GG.
    apply natEncAtom.

    assert (x3 = rget vs'' v0).
    sconjLift2 Hp''; apply (contContEq Hp'' VV1).
    subst x3; clear VV1.

    destruct (G2 (rset vs'' rlist (rget vs'' v0)) z (n + y) A).
    simpl; sconjMatch.
    eapply V.rootWeaken.
    instantiate (1 := (rlist::v0::nil)).
    replace (rget s'' v0) with (RFile.get (rfileOf s'') v0).
    eapply V.rootCopyOk.
    eapply V.rootWeaken; eauto.
    unfold incl; simpl; intuition.
    intuition.
    simpl.
    intuition.
    discriminate.
    discriminate.
    reflexivity.
    simpl; intuition.

    stSimpl.
    sconjLift3 Hp''; sconjMatch.
    stSimpl.
    rewrite <- Rp1; stSimpl.
    rewrite <- Rp; stSimpl.
    rewrite Enc.
    replace (rget s' v0) with (natEnc y).
    unfold natEnc; omega.
    assert (X := V.atomRegEq v0 Hp').
    replace (rget s' v0) with (RFile.get (rfileOf s') v0).
    rewrite X; auto.
    simpl; auto.
    rewrite Y.
    unfold natEnc; right.
    unfold V.I.atomP; unfold atomP.
    constructor; apply even_mult_l; constructor; constructor; constructor.
    reflexivity.
    constructor; [unfold ListSumMGI.presReg; unfold calleeSaved; simpl; auto | discriminate].
    constructor; [unfold ListSumMGI.presReg; unfold calleeSaved; simpl; auto | discriminate].
    
    destruct H; exists x; constructor.
    assumption.
    rewrite H0.
    unfold natEnc; intuition.

  Qed.


  Lemma mretOk : iseqOkS mPhi weakPre (fun s s' => s = s') mret.
  Proof.
    unfold mretSpec; unfold mret.
    iseqWeakSolver Psimpl.
    simpl; intuition.
  Qed.


End ListSumVerif.


Module ListPresGuar <: BASIC_MGI_ARG.

  Definition gcGuar vs vs' rr :=
    rr = rlist::nil ->
    forall l,
      vs |||- NatListAt l (rget vs rlist) ** top ->
      vs' |||- NatListAt l (rget vs' rlist) ** top.

End ListPresGuar.

Module ListPresMGI := BasicMGI (ListPresGuar).

Require Import NullGC.
Require Import CheneyGC.
Require Import CheneyLemma.


(* this is copied or modified from ListRev *)
Module CheneyGCCompat <: MG_SUB_INTERFACE.
  Module I := CopyMGI.
  Module I' := ListPresMGI.

  Require Import FnMap.


  Lemma natListAtFwd : forall l h h' x x' phi objs objs',
    h ||- NatListAt l x ->
    h ||- objHp objs objs ->
    h' ||- (foralls x in objs', objIsFwdedH phi h x) ->
    iso_set phi objs objs' ->
    mapX phi x = Some x' ->
    h' ||- NatListAt l x'.
  Proof.
    assert (NullOdd : odd NULL).
    unfold NULL; constructor; constructor.

    unfold NatListAt.
    induction l; intros h h' x x' phi objs objs' Hp HpO Hp' Iso Meq.
    
    (* base case *)
    simpl in Hp; lle Hp; subst x.
    unfold mapX in Meq; rewrite isOddTrue in Meq.
    injection Meq; clear Meq; intro; subst x'.
    destruct objs.
    destruct objs'.
    simpl; lli.
    apply (fasElimE Hp').
    reflexivity.
    elimtype False; clear Hp HpO Hp'.
    destruct Iso as [Iso [_ [_ Iso']]].
    destruct (Iso' a) as [x Peq]; clear Iso'.
    left; reflexivity.
    destruct (Iso x) as [_ Q]; apply Q; rewrite Peq; discriminate.
    elimtype False; clear Hp' Iso.

    assert (I : set_In a (a::objs)).
    left; reflexivity.

    hypReplace HpO (fasElim I HpO); unfold okObjAt in HpO; unfold okFieldAt in HpO; lle HpO.
    eapply unitContElim; eauto.
    assumption.


    (* inductive case *)
    simpl in Hp; lle Hp; clear H1.
    unfold mapX in Meq; rewrite isOddFalse in Meq.
    assert (I' : set_In x' objs').
    destruct Iso as [_ [Iso _]].
    eapply Iso; apply Meq.

    hypReplace Hp' (fasElim I' Hp'); cbv beta in Hp'; unfold objIsFwdedH at 1 in Hp'; lle Hp'.
    destruct H3 as [y [v0 [v1 [Peq [HpY [Meq0 Meq1]]]]]].
    lle HpY.
    assert (x = y).
    destruct Iso as [_ [_ [X _]]]; eapply X; try (instantiate (1 := x')); assumption.
    subst y.
    assert (U := contContEq HpY Hp).
    assert (v1 = x0).
    sconjLift2 HpY; sconjLift2 Hp; exact (contContEq HpY Hp).
    subst v0 v1; clear HpY.
    unfold mapX in Meq0; rewrite isOddTrue in Meq0.
    injection Meq0; clear Meq0; intro; subst x1.
    simpl; lli.
    intro; subst x'.
    apply (not_even_and_odd NULL).
    apply ptrPeven; eapply contElim1L; apply Hp'.
    assumption.

    instantiate (1 := x2).
    clear Peq.
    rename x0 into y; rename x2 into y'.
    hypReplace Hp (sconjAssocR Hp).

    assert (I : set_In x objs).
    destruct Iso as [Iso _]; destruct (Iso x).
    apply H0; rewrite Meq; discriminate.
    
    hypReplace HpO (objHpElim I HpO); unfold okObjAt in HpO; unfold okFieldAt in HpO;
      lle HpO; hypReplace HpO (sconjAssocR HpO).
    clear H3 H6.
    sconjLift2of2 Hp; hypReplace Hp (sconjExtract Hp); destruct Hp as [ha [Hp HpA]].
    sconjLift2of2 HpO; hypReplace HpO (sconjExtract HpO); destruct HpO as [ha' [HpO HpA']].
    sconjLift2of2 Hp; sconjLift2of2 HpO.

    assert (Iso' : iso_set (map_remove _ _ eqDecB phi x) (set_remove x objs) (set_remove x' objs')).
    apply iso_setRemove; assumption.

    cut (Heap.heap_eqv ha ha').
    intro HeqA.
    
    cut (ha ||- objHp (set_remove x objs) (set_remove x objs)).
    intro HpA''.

    cut (h' ||- x' |-> natEnc a ** 4 + x' |-> y' ** 
      (foralls z in set_remove x' objs', objIsFwdedH (map_remove _ _ eqDecB phi x) ha z)).

    intro Hp'B; clear Hp'; rename Hp'B into Hp'.
    sconjMatch.
    eapply IHl.
    apply HpA.
    apply HpA''.
    apply Hp0.
    assumption.
    unfold mapX; unfold mapX in Meq1; destruct (isOdd y).
    assumption.
    unfold map_remove; unfold eqDecB; destruct (POnat.POnat.eqDec x y).
    elimtype False.
    2:assumption.
    subst y.
    assert (X : h ||- (x |-> (natEnc a) ** top) ** (x |->__ ** top)).
    sconjMatch.
    destruct l.
    simpl in HpA; lle HpA; subst x; lle HpO.
    elimtype False; apply (not_even_and_odd NULL).
    apply ptrPeven; eapply contElim1L; eauto.
    assumption.
    simpl in HpA; lle HpA.
    eapply heapEqvTrue.
    apply Heap.heap_eqvSymm; apply eqvPredEqv; eauto.
    lli; sconjMatch.
    lle X.

    (* forward on the smaller phi *)
    sconjMatch.
    apply (fasImplI Hp0); clear Hp0.
    clear IHl.

    unfold objIsFwdedH; intros x2 h1 Ix Hp0.
    lle Hp0.
    rename Meq1 into Meq1'.
    destruct H1 as [z [v0 [v1 [Peq' [HpZ [Meq0 Meq1]]]]]].
    lli.
    2:eauto.
    exists z; exists v0; exists v1.
    clear h1 Hp0.

    assert (Iz : set_In z (set_remove x objs)).
    destruct Iso' as [Iso' [_ [_ Iso'']]].
    destruct (Iso'' _ Ix) as [z' PeqZ].
    assert (z' = z).
    destruct Iso as [_ [_ [X _]]]; eapply X.
    2:eauto.
    unfold map_remove in PeqZ; unfold eqDecB in PeqZ;
      destruct (POnat.POnat.eqDec x z').
    discriminate PeqZ.
    assumption.
    subst z'.
    destruct (Iso' z) as [_ Q]; apply Q.
    rewrite PeqZ; discriminate.
    
    unfold map_remove; rewrite Peq'.

    assert (NE : z <> x).
    exact (set_remove_elim2 _ Iz).

    unfold eqDecB; destruct (POnat.POnat.eqDec x z).
    elimtype False; auto.
    constructor.
    reflexivity.
    lle HpZ.

    cut (ha ||- z |->, (v0 :: v1 :: nil) ** top).
    intro HA.
    constructor.
    assumption.
    lle HA.
    hypReplace HpA'' (objHpElim Iz HpA'').
    unfold okObjAt in HpA''; unfold okFieldAt in HpA''; lle HpA''.
    assert (U := contContEq HA HpA'').
    assert (v1 = x6).
    sconjLift2 HA; sconjLift2 HpA''; exact (contContEq HA HpA'').
    subst x5 x6.
    unfold mapX in Meq0, Meq1 |- * .
    constructor.
    assert (U := isOddOk v0).
    destruct (isOdd v0).
    assumption.
    destruct H3.
    elimtype False; apply (not_even_and_odd v0); assumption.
    destruct (POnat.POnat.eqDec x v0).
    2:assumption.
    assert (X := set_remove_elim2 _ H); elimtype False; auto.
    assert (U := isOddOk v1).
    destruct (isOdd v1).
    assumption.
    destruct H6.
    elimtype False; apply (not_even_and_odd v1); assumption.
    destruct (POnat.POnat.eqDec x v1).
    2:assumption.
    assert (X := set_remove_elim2 _ H); elimtype False; auto.

    (* show that this part of ha is the same as h *)
    lli.
    lle HpO.

    hypReplace HpA'' (objHpElim Iz HpA''); unfold okObjAt in HpA'';
      unfold okFieldAt in HpA''; lle HpA''.
    assert (Sub : Heap.subHeap ha h).
    sconjLift2of2 Hp; exact (sconjEqSubHeap Hp).
    assert (Q := subHeapContEq Sub HpZ HpA'').
    sconjLift2 HpZ; sconjLift2 HpA''.
    assert (Q' := subHeapContEq Sub HpZ HpA'').
    subst x5 x6.
    sconjLift2 HpA''; sconjMatch.
    hypReplace HpA (natListIsMinObjHp _ _ _ HpA).
    lle HpA.
    unfold minObjHp in HpA; destruct (aconjElim HpA) as [_ HpA''].
    unfold okObjHp in HpA''; lle HpA''.
    
    assert (H11 : ha ||- objHp objs (set_remove x objs)).
    eapply heapEqvTrue; eauto.

    assert (X := objHpEqObjs _ _ _ _ _ HpA'' H11).
    unfold hpredTrue; rewrite <- X; assumption.

    clear IHl Hp' HpA' HpA Iso' Meq1 Meq I I'.
    lle Hp; lle HpO.
    assert (X := contContEq Hp HpO).
    subst x0.
    assert (y = x1).
    sconjLift2 Hp; sconjLift2 HpO; exact (contContEq Hp HpO).
    subst x1.
    hypReplace Hp (sconjAssocR Hp).
    hypReplace HpO (sconjAssocR HpO).
    clear Iso y' objs objs' phi NullOdd l h' x'.
    rename HpO into Hp'; rename ha into h1; rename ha' into h2.
    destruct (sconjElim Hp) as [ha [hb [J [HA HB]]]]; clear Hp.
    destruct (sconjElim Hp') as [ha' [hb' [J' [HA' HB']]]]; clear Hp'.

    assert (E : Heap.heap_eqv ha ha').
    clear hb hb' J J' HB HB'.
    hypReplace HA (unitIR HA).
    hypReplace HA' (unitIR HA').
    assert (HA2 : ha ||- (x |-> natEnc a ** S (S (S (S x))) |-> y) ** 
      (Heap.heap_eqv Heap.empty_heap)).
    sconjMatch.
    assert (HA2' : ha' ||- (x |-> natEnc a ** S (S (S (S x))) |-> y) ** 
      (Heap.heap_eqv Heap.empty_heap)).
    sconjMatch.
    clear HA HA'.
    lle HA2; lle HA2'.
    sconjLift3of3 HA2; sconjLift2 HA2.
    sconjLift3of3 HA2'; sconjLift2 HA2'.
    hypReplace HA2 (sconjComm (unitIR (heapEqvAdd HA2))).
    hypReplace HA2' (sconjComm (unitIR (heapEqvAdd HA2'))).
    sconjLift2 HA2; sconjLift3of3 HA2.
    sconjLift2 HA2'; sconjLift3of3 HA2'.
    hypReplace HA2 (heapEqvAdd HA2).
    hypReplace HA2' (heapEqvAdd HA2').
    lle HA2; lle HA2'.
    eapply Heap.heap_eqvTrans.
    apply Heap.heap_eqvSymm; apply eqvPredEqv; apply HA2.
    apply eqvPredEqv; assumption.
    eapply Heap.heap_eqvTrans.
    instantiate (1 := hb).
    apply eqvPredEqv; assumption.
    eapply Heap.heap_eqvTrans.
    instantiate (1 := hb').
    2:apply Heap.heap_eqvSymm; apply eqvPredEqv; assumption.
    eapply Heap.HeapMap.ABJoin.mapJoinJoinEqvR.
    apply J.
    eapply Heap.HeapMap.ABJoin.mapJoinEqv.
    apply J'.
    apply Heap.heap_eqvSymm; assumption.
    apply Heap.heap_eqvRefl.
    apply Heap.heap_eqvRefl.
    unfold natEnc.
    constructor.
    apply even_mult_l.
    constructor; constructor; constructor; constructor.
    
    apply ptrPeven.
    eapply contElim1L.
    apply Hp.

  Qed.


  Lemma gcGuarSub : forall s s' rr,
    I.gcGuar s s' rr -> I'.gcGuar s s' rr.
  Proof.
    unfold I.gcGuar, CopyGCGuar.gcGuar.
    intros s s' rr G.
    unfold I'.gcGuar, ListPresGuar.gcGuar.
    intros E l Hp. subst rr.
    destruct (sconjExtract Hp) as [h [Hp' OHp]]; clear Hp.

    assert (OHp' : h ||- lexists objs, minObjHp (rget s rlist::nil) objs).
    eapply natListIsMinObjHp; eauto.

    lle OHp'; rename x into objs.
    assert (Hp : s |||- (minObjHp (rget s rlist :: nil) objs && Heap.heap_eqv h) ** top).
    sconjMatch.
    apply aconjIntro.
    eapply heapEqvTrue.
    2:eauto.
    eapply Heap.heap_eqvTrans.
    2:apply Heap.heap_eqvSymm; apply Hp.
    apply Heap.HeapMap.map_canonEqv; assumption.
    assumption.

    clear Hp'.
    destruct (G _ _ Hp) as [h' [Hp' Iso]]; clear G Hp objs OHp'.
    scm.
    eapply heapEqvTrue.
    eapply Heap.heap_eqvTrans.
    2:apply Heap.heap_eqvSymm; apply Hp.
    apply Heap.HeapMap.map_canonEqv; assumption.
    
    clear h0 Hp; rename OHp into Hp1.
    destruct Iso as [objs [objs' [phi X]]].
    generalize dependent X; stSimpl; intro.
    destruct X as [Hp1' [Hp2 [Iso [Req HpF]]]].

    assert (E1 : mapX phi (rget s rlist) = Some (rget s' rlist)).
    rewrite Req; auto; left; auto.

    eapply natListAtFwd; eauto.

  Qed.
    
  Lemma rootEq : I.rroot = I'.rroot.
  Proof.
    reflexivity.
  Qed.

  Lemma gcInfoEq : I.rgcInfo = I'.rgcInfo.
  Proof.
    reflexivity.
  Qed.

  Lemma nullEq : I.NULL = I'.NULL.
  Proof.
    reflexivity.
  Qed.

  Lemma presRegSub : forall r,
    I'.presReg r -> I.presReg r.
  Proof.
    intros; assumption.
  Qed.

  Lemma atomPSub : forall n,
    I'.atomP n -> I.atomP n.
  Proof.
    intros; assumption.
  Qed.

  Lemma gcPhiDomSub : set_subset I.gcPhiDom I'.gcPhiDom.
  Proof.
    intro; auto.
  Qed.

  Lemma allocLblEq : I.allocLbl = I'.allocLbl.
  Proof.
    reflexivity.
  Qed.

  Lemma read1LblEq : I.read1Lbl = I'.read1Lbl.
  Proof.
    reflexivity.
  Qed.

  Lemma read2LblEq : I.read2Lbl = I'.read2Lbl.
  Proof.
    reflexivity.
  Qed.

  Lemma write1LblEq : I.write1Lbl = I'.write1Lbl.
  Proof.
    reflexivity.
  Qed.

  Lemma write2LblEq : I.write2Lbl = I'.write2Lbl.
  Proof.
    reflexivity.
  Qed.


End CheneyGCCompat.


Module ListPresGCCompat <: MG_SUB_INTERFACE.
  Module I := ListPresMGI.
  Module I' := ListSumMGI.

  Lemma rootEq : I.rroot = I'.rroot.
  Proof.
    reflexivity.
  Qed.

  Lemma gcInfoEq : I.rgcInfo = I'.rgcInfo.
  Proof.
    reflexivity.
  Qed.

  Lemma nullEq : I.NULL = I'.NULL.
  Proof.
    reflexivity.
  Qed.


  Lemma oddEvenElim : forall x,
    (odd x -> exists y, x = (S (2 * y))) /\
    (even x -> exists y, x = 2 * y).
  Proof.
    induction x.

    constructor.
    inversion 1.
    intro; exists 0; reflexivity.
    
    destruct IHx as [IH1 IH2].
    constructor; inversion 1; subst; clear H.
    destruct (IH2 H1) as [y X]; subst; exists y; reflexivity.
    destruct (IH1 H1) as [y X]; subst; exists (S y); omega.

  Qed.

  Lemma oddElim : forall x,
    odd x -> exists y, x = (S (2 * y)).
  Proof.
    intros x OX; destruct (oddEvenElim x) as [X _]; exact (X OX).
  Qed.

  Lemma gcGuarSub : forall s s' rr,
    I.gcGuar s s' rr -> I'.gcGuar s s' rr.
  Proof.
    unfold I.gcGuar; unfold ListPresGuar.gcGuar.
    intros s s' rr G; constructor.

    (* NULL list case *)
    hyr G (G H nil). intros E; subst rr.
    replace NullGCImpl.rroot with rlist in * by auto.
    rewrite E in G.
    assert (Hp : s |||- NatListAt nil NULL ** top).
    assert (Hp : s |||- unit ** top).
    lli.
    scm.
    apply NatListAtIntroNil; assumption.
    hyr Hp (G Hp); clear G.
    unfold NatListAt in Hp; simpl in Hp; lle Hp; assumption.
    
    (* CONS list case *)
    unfold ListSumAt.
    intros m n y Hp mAtom; lle Hp.
    destruct (oddElim mAtom); subst m.
    assert (Hp' : s' |||- NatListAt (x0::x) (rget s' rlist) ** top).
    apply G; clear G.
    auto.
    unfold NatListAt; simpl; lli.
    replace NullGCImpl.rroot with rlist in * by auto.
    intro X; rewrite X in Hp; assert (Y := contElim1L Hp).
    inversion Y.
    scm.
    unfold NatListAt in Hp'; simpl in Hp'; lle Hp'.
    exists x1.
    lli.
    2:eauto.
    scm.
  Qed.

  Lemma presRegSub : forall r,
    I'.presReg r -> I.presReg r.
  Proof.
    auto.
  Qed.

  Lemma atomPSub : forall n,
    I'.atomP n -> I.atomP n.
  Proof.
    auto.
  Qed.

  Lemma gcPhiDomSub : set_subset I.gcPhiDom I'.gcPhiDom.
  Proof.
    intros s X; assumption.
  Qed.

  Lemma allocLblEq : I.allocLbl = I'.allocLbl.
  Proof.
    reflexivity.
  Qed.

  Lemma read1LblEq : I.read1Lbl = I'.read1Lbl.
  Proof.
    reflexivity.
  Qed.

  Lemma read2LblEq : I.read2Lbl = I'.read2Lbl.
  Proof.
    reflexivity.
  Qed.

  Lemma write1LblEq : I.write1Lbl = I'.write1Lbl.
  Proof.
    reflexivity.
  Qed.

  Lemma write2LblEq : I.write2Lbl = I'.write2Lbl.
  Proof.
    reflexivity.
  Qed.

End ListPresGCCompat.


Module CoerceGCTrans
  (S1 : MG_SUB_INTERFACE)
  (S2 : MG_SUB_INTERFACE with Module I := S1.I')
    <: MG_SUB_INTERFACE with Module I := S1.I with Module I' := S2.I'.

  Module I := S1.I.
  Module I' := S2.I'.

  Lemma rootEq : I.rroot = I'.rroot.
  Proof.
    transitivity S1.I'.rroot.
    apply S1.rootEq.
    apply S2.rootEq.
  Qed.

  Lemma gcInfoEq : I.rgcInfo = I'.rgcInfo.
  Proof.
    transitivity S1.I'.rgcInfo.
    apply S1.gcInfoEq.
    apply S2.gcInfoEq.
  Qed.

  Lemma nullEq : I.NULL = I'.NULL.
  Proof.
    transitivity S1.I'.NULL.
    apply S1.nullEq.
    apply S2.nullEq.
  Qed.

  Lemma gcGuarSub : forall s s' rr,
    I.gcGuar s s' rr -> I'.gcGuar s s' rr.
  Proof.
    intros s s' rr G; apply S2.gcGuarSub; apply S1.gcGuarSub; assumption.
  Qed.

  Lemma presRegSub : forall r,
    I'.presReg r -> I.presReg r.
  Proof.
    intros r P; apply S1.presRegSub; apply S2.presRegSub; assumption.
  Qed.

  Lemma atomPSub : forall n,
    I'.atomP n -> I.atomP n.
  Proof.
    intros n A; apply S1.atomPSub; apply S2.atomPSub; assumption.
  Qed.

  Lemma gcPhiDomSub : set_subset I.gcPhiDom I'.gcPhiDom.
  Proof.
    apply set_subsetTrans with S1.I'.gcPhiDom.
    exact S1.gcPhiDomSub.
    exact S2.gcPhiDomSub.
  Qed.
  
  Lemma allocLblEq : I.allocLbl = I'.allocLbl.
  Proof.
    transitivity S1.I'.allocLbl.
    apply S1.allocLblEq.
    apply S2.allocLblEq.
  Qed.

  Lemma read1LblEq : I.read1Lbl = I'.read1Lbl.
  Proof.
    transitivity S1.I'.read1Lbl.
    apply S1.read1LblEq.
    apply S2.read1LblEq.
  Qed.

  Lemma read2LblEq : I.read2Lbl = I'.read2Lbl.
  Proof.
    transitivity S1.I'.read2Lbl.
    apply S1.read2LblEq.
    apply S2.read2LblEq.
  Qed.

  Lemma write1LblEq : I.write1Lbl = I'.write1Lbl.
  Proof.
    transitivity S1.I'.write1Lbl.
    apply S1.write1LblEq.
    apply S2.write1LblEq.
  Qed.

  Lemma write2LblEq : I.write2Lbl = I'.write2Lbl.
  Proof.
    transitivity S1.I'.write2Lbl.
    apply S1.write2LblEq.
    apply S2.write2LblEq.
  Qed.

End CoerceGCTrans.


Module CheneyCoercer := CoerceGCTrans CheneyGCCompat ListPresGCCompat.

Module ListSumCheneyGC := CoerceGC CheneyGCImpl CheneyCoercer.

Module ListRevCheneyGCed := ListSumVerif ListSumCheneyGC.


