Require Export State.
Require Export NatUtil.

(** **** Basic definitions *)

Inductive btype : Set :=
  | bteq 
  | btne.


Inductive iseq : Set :=
  | seq : comm -> iseq -> iseq
  | br : btype -> rnum -> rnum -> nat -> iseq -> iseq
  | j : nat -> iseq
  | jal : nat -> nat -> iseq
  | jr : rnum -> iseq.

Definition beq := br bteq.
Definition bne := br btne.


Definition cdHeap := nat -> option iseq.

Inductive prog : Set :=
  progi : cdHeap -> state -> iseq -> prog.


(** **** Semantics *)

Definition doBtest (bt:btype) (m n:nat) : bool :=
  match bt with
    | bteq => EqNat.beq_nat m n
    | btne => negb (EqNat.beq_nat m n)
  end.

Definition step (p:prog) : option prog :=
  let (C, s, i) := p in
    match i with 
      | seq c i' =>
	do s' <-- cstep c s;
	  ret (progi C s' i')
      | br bt rs rt f i' =>
	if doBtest bt (rget s rs) (rget s rt) then
	  do i'' <-- C f;
	    ret (progi C s i'')
	else
	  ret (progi C s i')
      | j f =>
	do i' <-- C f;
	  ret (progi C s i')
      | jal f fret =>
	do i' <-- C f;
	  ret (progi C (rset s r31 fret) i')
      | jr r =>
	do i' <-- C (rget s r);
	  ret (progi C s i')
    end.


Definition predSound (P:prog -> Prop) :=
  forall s, P s -> exists s', (step s) = Some s' /\ P s'.


Notation "i ; I" := (seq i I) (at level 30, right associativity).
Notation "i ;; I" := (i (I:iseq)) (at level 30, right associativity).
