Require Import StateTheory.
Require Import VGC.
Require Import List.
Require Import GC.
Require Import WeakSCAP.


Set Implicit Arguments.


Module NullGCGuar <: BASIC_MGI_ARG.
  Definition gcGuar (s s':state) (rr:list rnum) := 
    (forall objs h rr',
      incl rr' rr ->
      s |||- (minObjHp (List.map (rget s) rr') objs && Heap.heap_eqv h) ** top ->
      s' |||- Heap.heap_eqv h ** top) /\
    forall r, In r rr -> rget s r = rget s' r.


End NullGCGuar.


Module NullGCMGI := BasicMGI (NullGCGuar).


Module NullGCImpl <: VIEW_GC with Module I := NullGCMGI.

  Module I := NullGCMGI.

  Notation rgcInfo := s7.

  Definition repr vs rf rr :=
    lexists freeStart, lexists freeEnd,
    !(~In rgcInfo rr /\ 
      aligned8 freeStart freeEnd /\
      exists objs, 
	(forall r, In r rr -> rget vs r = RFile.get rf r /\ okFieldVal objs (RFile.get rf r)) /\
	vs |||- objHp objs objs) **
    Heap.heap_eqv (heapOf vs) ** 
    (foralls x in rangeObjs freeStart freeEnd, x |->__) **
    RFile.get rf rgcInfo |->, (freeStart::freeEnd::nil).

  Lemma atomRegEq : forall h vs rf rr A r,
    h ||- repr vs rf rr ** A ->
    In r rr ->
    I.atomP (RFile.get rf r) \/ I.atomP (rget vs r) ->
    RFile.get rf r = rget vs r.
  Proof.
    intros h vs rf rr A r Hp I D.
    unfold repr in Hp; linLogElim Hp.
    destruct H3 as [_ [_ [objs [Req Vhp]]]].
    destruct (Req _ I) as [X _].
    auto.
  Qed.

  Lemma setNonRootOk : forall h vs rf rr r v,
    h ||- repr vs rf rr ->
    ~In r (rgcInfo::rr) ->
    h ||- repr vs (RFile.set rf r v) rr.
  Proof.
    unfold repr.
    intros h vs rf rr r v Hp I.
    linLogElim Hp.
    destruct H1 as [rrOk [alOk [objs [Req Vhp]]]].
    linLogIntro.
    constructor.
    assumption.
    constructor.
    eauto.
    exists objs; constructor.
    intros r' I'; destruct (Req _ I') as [X Y].
    destruct (regCompare r r') as [R | R].
    elimtype False; subst r; intuition.
    replace (RFile.get (RFile.set rf r v) r') with (RFile.get rf r').
    auto.
    unfold RFile.set; unfold RFile.get.
    destruct (r0compare r').
    subst r'; rewrite beq_rnum_refl; reflexivity.
    rewrite beq_rnum_neq.
    2:auto.
    rewrite (beq_rnum_neq R).
    reflexivity.
    assumption.
    replace (RFile.get (RFile.set rf r v) rgcInfo) with (RFile.get rf rgcInfo).
    assumption.
    unfold RFile.set; unfold RFile.get.
    assert (r <> rgcInfo).
    intro E; subst r; intuition.
    rewrite (beq_rnum_neq H); reflexivity.
  Qed.

  
  Lemma rootCopyOk : forall h vs rf rr rs rd,
    h ||- repr vs rf rr ->
    In rs rr -> 
    ~In rd (zero::rgcInfo::nil) ->
    h ||- repr (rset vs rd (rget vs rs)) (RFile.set rf rd (RFile.get rf rs)) (rd::rr).
  Proof.
    unfold repr.
    intros h vs rf rr rs rd Hp I0 Nz.
    assert (NE1 : rd <> zero).
    intro E; subst rd; intuition.
    assert (NE2 : rd <> rgcInfo).
    intro E; subst rd; intuition.
    linLogElim Hp.
    destruct H1 as [rrOk [alOk [objs [Req Vhp]]]].
    linLogIntro.

    constructor.
    simpl; intro X; destruct X.
    subst rd; intuition.
    auto.
    constructor.
    eauto.
    exists objs.
    constructor.
    2:assumption.
    intros r I.
    simpl in I; destruct I as [R | R].
    subst r; stSimpl.
    replace (RFile.get (RFile.set rf rd (RFile.get rf rs)) rd) with
      (RFile.get rf rs).
    apply Req; apply I0; simpl; auto.
    unfold RFile.set; unfold RFile.get.
    rewrite beq_rnum_refl.
    rewrite (beq_rnum_neq (r:=zero) (r':=rd)).
    reflexivity.
    auto.

    destruct (regCompare rd r) as [R' | R'].
    
    subst r; stSimpl.
    replace (RFile.get (RFile.set rf rd (RFile.get rf rs)) rd) with
      (RFile.get rf rs).
    auto.
    unfold RFile.set; unfold RFile.get.
    rewrite (beq_rnum_neq (r:=zero) (r':=rd)).
    rewrite beq_rnum_refl.
    reflexivity.
    auto.

    stSimpl.
    replace (RFile.get (RFile.set rf rd (RFile.get rf rs)) r) with
      (RFile.get rf r).
    auto.
    unfold RFile.get; unfold RFile.set.
    rewrite (beq_rnum_neq R'); reflexivity.

    replace (RFile.get (RFile.set rf rd (RFile.get rf rs)) rgcInfo)
      with (RFile.get rf rgcInfo).
    assumption.
    unfold RFile.get; unfold RFile.set.
    rewrite (beq_rnum_neq NE2); reflexivity.

  Qed.


  Lemma rootWeaken : forall h vs rf rr rr',
    h ||- repr vs rf rr ->
    incl rr' rr ->
    h ||- repr vs rf rr'.
  Proof.
    unfold repr.
    intros h vs rf rr rr' Hp IR.
    linLogElim Hp.
    destruct H1 as [rrOk [alOk [objs [Req Vhp]]]].
    linLogIntro.
    constructor.
    auto.
    constructor.
    eauto.
    exists objs.
    constructor.
    intros r I; apply Req.
    auto.
    assumption.
    assumption.
  Qed.

  Lemma addAtomRoot : forall h vs rf rr A r,
    h ||- repr vs rf rr ** A ->
    ~In r (zero::I.rgcInfo::rr) ->
    I.atomP (RFile.get rf r) ->
    h ||- repr (rset vs r (RFile.get rf r)) rf (r::rr) ** A.
  Proof.
    unfold repr.
    intros h vs rf rr A r Hp NI Atom.
    assert (NE1 : r <> zero).
    intro E; subst r; intuition.
    assert (NE2 : r <> rgcInfo).
    intro E; subst r; apply NI; right; left; reflexivity.
    linLogElim Hp.
    destruct H3 as [rrOk [alOk [objs [Req Vhp]]]].
    linLogIntro.

    constructor.
    simpl; intros [X | X]; auto.
    constructor.
    eauto.

    exists objs.
    constructor.
    intros r' I.
    simpl in I; destruct I as [R | R].
    subst r'; stSimpl.
    constructor.
    reflexivity.
    left; assumption.

    assert (r' <> r).
    intro X; subst r'; simpl in NI; auto.
    stSimpl; apply Req; assumption.

    assumption.

    assumption.

  Qed.


  Definition phi n :=
    if beq_nat n I.allocLbl then
      SomeT (allocSpec repr I.rroot I.gcGuar I.presReg I.NULL)
    else if beq_nat n I.read1Lbl then
      SomeT (read1Spec repr I.rroot I.gcGuar I.presReg)
    else if beq_nat n I.read2Lbl then
      SomeT (read2Spec repr I.rroot I.gcGuar I.presReg)
    else if beq_nat n I.write1Lbl then
      SomeT (write1Spec repr I.rroot I.gcGuar I.presReg)
    else if beq_nat n I.write2Lbl then
      SomeT (write2Spec repr I.rroot I.gcGuar I.presReg)
    else
      NoneT _.


  Lemma natEqDec : forall m n : nat, m = n \/ m <> n.
  Proof.
    intros m n; compare m n; auto.
  Qed.

  
  Lemma gcPhiDomOk : forall f,
    phi f <> NoneT _ -> set_In f I.gcPhiDom.
  Proof.
    unfold I.gcPhiDom.
    unfold phi.
    intros f Ne.
    unfold I.allocLbl in Ne.
    unfold I.read1Lbl in Ne.
    unfold I.read2Lbl in Ne.
    unfold I.write1Lbl in Ne.
    unfold I.write2Lbl in Ne.

    destruct (natEqDec f 10) as [R | R].
    subst f; rewrite beq_true in Ne.
    apply setRangeIntro; omega.

    rewrite (beq_false R) in Ne; clear R.
    destruct (natEqDec f 20) as [R | R].
    subst f; rewrite beq_true in Ne.
    apply setRangeIntro; omega.

    rewrite (beq_false R) in Ne; clear R.
    destruct (natEqDec f 30) as [R | R].
    subst f; rewrite beq_true in Ne.
    apply setRangeIntro; omega.

    rewrite (beq_false R) in Ne; clear R.
    destruct (natEqDec f 40) as [R | R].
    subst f; rewrite beq_true in Ne.
    apply setRangeIntro; omega.

    rewrite (beq_false R) in Ne; clear R.
    destruct (natEqDec f 50) as [R | R].
    subst f; rewrite beq_true in Ne.
    apply setRangeIntro; omega.

    rewrite (beq_false R) in Ne; clear R.
    elimtype False; apply Ne; reflexivity.

  Qed.


  Ltac PhiOkSolver :=
    unfold phi; rewrite beq_true;
    repeat (rewrite beq_false; [idtac | discriminate]);
    simpl; intuition.

  Lemma read1PhiOk :
    optPredT (phi I.read1Lbl)
      (specWeaker (read1Spec repr I.rroot I.gcGuar I.presReg)).
  Proof.
    PhiOkSolver.
  Qed.

  Lemma read2PhiOk :
    optPredT (phi I.read2Lbl)
      (specWeaker (read2Spec repr I.rroot I.gcGuar I.presReg)).
  Proof.
    PhiOkSolver.
  Qed.

  Lemma write1PhiOk : 
    optPredT (phi I.write1Lbl)
      (specWeaker (write1Spec repr I.rroot I.gcGuar I.presReg)).
  Proof.
    PhiOkSolver.
  Qed.

  Lemma write2PhiOk : 
    optPredT (phi I.write2Lbl)
      (specWeaker (write2Spec repr I.rroot I.gcGuar I.presReg)).
  Proof.
    PhiOkSolver.
  Qed.

  Lemma allocPhiOk : 
    optPredT (phi I.allocLbl)
      (specWeaker (allocSpec repr I.rroot I.gcGuar I.presReg I.NULL)).
  Proof.
    PhiOkSolver.
  Qed.


  Lemma gcGuarRefl : forall vs rr,
    I.gcGuar vs vs rr.
  Proof.
    intros vs rr.
    constructor.
    intros objs h rr' _ HpOk.
    destruct (aconjElim (sconjDistrib2 HpOk)) as [_ X].
    assumption.
    reflexivity.
  Qed.


  Lemma gcGuarSetNonRoot : forall vs rr r v,
    ~In r rr ->
    I.gcGuar vs (rset vs r v) rr.
  Proof.
    intros vs rr r v NI.
    constructor.
    intros objs h rr' _ HpOk.
    destruct (aconjElim (sconjDistrib2 HpOk)) as [_ X].
    assumption.
    intros r' IR.
    assert (r <> r').
    intro E; subst r'; auto.
    stSimpl; reflexivity.
  Qed.


  Notation rroot := t0.

  Notation readkImpl := (fun k => lw v0 k rroot; jr ra).
  Notation writekImpl := (fun k => sw a0 k rroot; jr ra).


  Notation readPre' := (readPre repr rroot).
  Notation readGuar' := (readGuar repr rroot I.gcGuar I.presReg).


  Lemma readkImplOk : 
    forall k, k = 0 \/ k = 4 ->
      iseqOkS phi (readPre' k) (readGuar' k) (readkImpl k).
  Proof.
    intros k D.
    iseqWeakSolver idtac.
    unfold readPre; unfold repr.
    intros s [vs [Hp Vhp']].
    linLogElim Hp.
    destruct H3 as [rrOk [alOk [objs [Req0 Vhp]]]].
    linLogElim Vhp'.
    assert (Hp2 := sconjInsert Hp Vhp'); linLogElim Hp2.
    destruct (Req0 rroot) as [Req rootOk]; clear Req0.
    simpl; auto.
    assert (RX : rgcInfo <> rroot).
    simpl in rrOk; auto.
    clear rrOk.

    assert (HH : s |||- (!even(rget s rroot) ** top) ** top).
    clear Hp; sconjMatch.
    linLogIntro; apply ptrPeven; assert (P : ptrP (k + rget s rroot)).
    rewrite Req in Hp; eapply contElim1; eauto.
    destruct (ptrP_elim P) as [y Y]; destruct D; subst k;
      [apply (ptrP_intro y); auto | apply (ptrP_intro (pred y)); omega].

    linLogElim HH; clear HH; rename H1 into Ev.
    destruct rootOk.

    elimtype False; eapply not_even_and_odd; eauto.

    hypReplace Vhp (objHpElim H Vhp).
    unfold okObjAt in Vhp; unfold okFieldAt in Vhp; linLogElim Vhp.
    assert (Hp3 := sconjInsert Hp Vhp); linLogElim Hp3; clear Vhp.
    
    assert (Dstep : exists v,
      cstep (lw v0 k rroot) s = Some (rset s v0 v) /\ 
      ((k = 0 /\ v = x2) \/ (k = 4 /\ v = x3))).
    destruct D; [exists x2 | exists x3]; subst k; constructor.
    eapply lwStep; eauto.
    auto.
    sconjLift2 Hp3; eapply lwStep; eauto.
    auto.

    destruct Dstep as [v [Ceq Veq]].
    rewrite Ceq.
    constructor.
    trivial.

    clear Hp Vhp' Hp2 alOk.
    assert (Hp0 : s |||- k + rget s rroot |-> v ** top).
    destruct Veq as [[X Y] | [X Y]].
    subst k v; simpl; sconjMatch.
    subst k v; simpl; sconjLift2 Hp3; sconjMatch.

    clear vs Req objs H H4 H7 Hp3 x x0 x1 x2 x3 Veq Ceq.
    intros s' X; subst s'.
    unfold readGuar.
    constructor.

    intros vs A rr Hp Irr NIrr.
    linLogElim Hp.
    destruct H3 as [rrOk [alOk [objs [Req0 Vhp]]]].
    destruct (Req0 _ Irr) as [Req rootOk].
    destruct rootOk as [H | I].
    
    elimtype False; eapply not_even_and_odd; eauto.

    exists (rset vs v0 v).
    assert (Vhp0 := Vhp).
    hypReplace Vhp (objHpElim I Vhp).
    unfold okObjAt in Vhp; unfold okFieldAt in Vhp; linLogElim Vhp.
    assert (Hp2 := sconjInsert Hp Vhp); linLogElim Hp2.

    assert (NIv0 : ~ In v0 rr).
    intro Iv0; destruct (NIrr v0 Iv0) as [E | CS].
    discriminate E.
    unfold SpecUtil.calleeSaved in CS.
    simpl in CS.
    repeat (destruct CS as [E | CS]; [discriminate E | idtac]); assumption.

    constructor.
    apply gcGuarSetNonRoot; assumption.
    stSimpl.
    constructor.
    linLogIntro.
    constructor.
    simpl; intro DZ; destruct DZ.
    discriminate H.
    auto.
    constructor.
    eauto.
    exists objs.
    constructor.
    simpl.
    intros r [E | II].
    subst r; stSimpl.
    constructor.
    reflexivity.
    destruct D; subst k.
    rewrite (contContEq Hp0 Hp2); assumption.
    sconjLift2 Vhp; sconjLift2 Hp2; 
      rewrite (contContEq Hp0 Hp2); assumption.
    destruct (regCompare r v0) as [X | X].
    subst r; elimtype False; auto.
    generalize dependent Req0; stSimpl; intro;
      apply Req0; assumption.
    assumption.
    assumption.
    
    rewrite Req.
    destruct D; subst k.
    rewrite (contContEq Hp0 Hp2).
    scm.
    sconjLift2 Hp2; rewrite (contContEq Hp0 Hp2).
    sconjLift2 Vhp; scm.

    intros r [A B]; stSimpl; reflexivity.

  Qed.


  Lemma read1ImplOk : iseqOkS phi (readPre' 0) (readGuar' 0) (readkImpl 0).
  Proof.
    apply (readkImplOk (k := 0)); auto.
  Qed.

  Lemma read2ImplOk : iseqOkS phi (readPre' 4) (readGuar' 4) (readkImpl 4).
  Proof.
    apply (readkImplOk (k := 4)); auto.
  Qed.


  Notation writePre' := (writePre repr rroot).
  Notation writeGuar' := (writeGuar repr rroot I.gcGuar I.presReg).


  Lemma writekImplOk : 
    forall k, k = 0 \/ k = 4 ->
      iseqOkS phi (writePre' k) (writeGuar' k) (writekImpl k).  
  Proof.
    intros k D.
    iseqWeakSolver idtac.

    intros s [vs [Hp Vhp]].
    unfold repr in Hp; linLogElim Hp.
    destruct H3 as [rrOk [alOk [objs [Req0 Vhp']]]].
    clear rrOk.

    linLogElim Vhp.
    assert (Hp2 := sconjInsert Hp Vhp); linLogElim Hp2.

    destruct (Req0 rroot) as [Req rootOk].
    simpl; auto.
    rewrite (swStep Hp2); auto.
    constructor.
    trivial.
    intros s' Seq; subst s'.
    rewrite Req in Hp2; rewrite Req.

    assert (HH : s |||- (!even(rget s rroot) ** top) ** top).
    clear Hp; sconjMatch.
    linLogIntro; apply ptrPeven; assert (P : ptrP (k + rget s rroot)).
    eapply contElim1; eauto.
    destruct (ptrP_elim P) as [y Y]; destruct D; subst k;
      [apply (ptrP_intro y); auto | apply (ptrP_intro (pred y)); omega].

    linLogElim HH; clear HH; rename H1 into Ev.
    clear Hp Vhp' Vhp vs Req0 Req x Hp2 objs rootOk alOk x0 x1.
    constructor.
    intros vs A rr Hp Irr.
    assert (I1 : In rroot rr).
    apply Irr; simpl; auto.
    assert (I2 : In a0 rr).
    apply Irr; simpl; auto.
    clear Irr.
    unfold repr in Hp; linLogElim Hp.
    destruct H3 as [rrOk [alOk [objs [Req0 Vhp]]]].
    exists (uhset vs (k + rget vs rroot) (rget vs a0)).
    constructor.
    apply gcGuarRefl.

    destruct (Req0 rroot) as [Req rootOk].
    assumption.
    destruct rootOk.

    elimtype False; eapply not_even_and_odd; eauto.

    clear Ev.
    hypReplace Vhp (objHpElim H Vhp).
    unfold okObjAt in Vhp; unfold okFieldAt in Vhp; linLogElim Vhp.
    unfold repr.
    linLogIntro.
    constructor.
    auto.
    constructor.
    eauto.
    exists objs; stSimpl.
    constructor.
    auto.

    destruct D; subst k.

    hypReplace Vhp (contUpdateSt Vhp (rget vs a0)).
    replace (Heap.uset (heapOf vs) (RFile.get (rfileOf s) rroot) (rget vs a0)) with
      (heapOf (uhset vs (0 + rget vs rroot) (rget vs a0))) in Vhp.
    unfold hpredTrue; rewrite <- (addRemoveEqv H).
    apply objHpIntro.
    unfold okObjAt; unfold okFieldAt.
    destruct (Req0 a0) as [ReqA rootOkA].
    assumption.

    hyr Vhp (sconjAssocR Vhp); scm.
    lli. 2:eauto.
    rewrite (addRemoveEqv H); rewrite ReqA; assumption.
    lli. 2:eauto.
    rewrite (addRemoveEqv H); assumption.

    unfold hpredTrue; rewrite (addRemoveEqv H); assumption.
    rewrite Req.
    reflexivity.

    sconjLift2 Vhp; hypReplace Vhp (contUpdateSt Vhp (rget vs a0)).
    replace (Heap.uset (heapOf vs) (4 + RFile.get (rfileOf s) rroot) (rget vs a0)) with
      (heapOf (uhset vs (4 + rget vs rroot) (rget vs a0))) in Vhp.
    sconjMatch.
    unfold hpredTrue; rewrite <- (addRemoveEqv H).
    apply objHpIntro.
    unfold okObjAt; unfold okFieldAt.
    destruct (Req0 a0) as [ReqA rootOkA].
    assumption.

    sconjLift2 Vhp.
    hyr Vhp (sconjAssocR Vhp). scm.
    lli. 2:eassumption.
    rewrite (addRemoveEqv H); assumption.
    lli. 2:eassumption.
    rewrite (addRemoveEqv H); rewrite ReqA; assumption.
    unfold hpredTrue; rewrite (addRemoveEqv H); assumption.
    rewrite Req; reflexivity.

    simpl; eapply sconjHpSet; eauto.
    destruct (Req0 a0).
    assumption.
    rewrite H0.
    rewrite Req.
    destruct D; subst k.
    eapply contSetEqL; eauto.
    sconjLift2 Vhp; eapply contSetEqL; eauto.

    intros r _; stSimpl; reflexivity.

  Qed.


  Lemma write1ImplOk : iseqOkS phi (writePre' 0) (writeGuar' 0) (writekImpl 0).
  Proof.
    apply (writekImplOk (k := 0)); auto.
  Qed.

  Lemma write2ImplOk : iseqOkS phi (writePre' 4) (writeGuar' 4) (writekImpl 4).
  Proof.
    apply (writekImplOk (k := 4)); auto.
  Qed.


  Notation allocImpl :=
    (lw v0 0 rgcInfo;
     lw ast 4 rgcInfo;
     beq v0 ast I.allocLbl;;
     addiu ast zero 1;
     sw ast 0 v0;
     sw ast 4 v0;
     addiu ast v0 8;
     sw ast 0 rgcInfo;
     jr ra).

  Notation allocPre' := (allocPre repr rroot).
  Notation allocGuar' := (allocGuar repr rroot I.gcGuar I.presReg I.NULL).

  Require Import Heap.


  Lemma allocImplOk : iseqOkS phi allocPre' allocGuar' allocImpl.
  Proof.
    iseqWeakSolver fail.

    unfold phi; rewrite beq_true.
    unfold allocSpec; reflexivity.

    intros s [vs Hp].
    unfold repr in Hp; linLogElim Hp.
    destruct H3 as [rrOk [alOk [objs [Req0 Vhp]]]].
    clear rrOk.

    assert (Hp0 := Hp).
    sconjLift3 Hp0.
    rewrite (lwStep Hp0); auto.
    sconjLift4 Hp0.
    rewrite (lwStep Hp0); auto.
    clear Hp0.

    stSimpl; unfold doBtest.
    compare x x0; intro R.

    subst x0; rewrite beq_true.
    constructor.
    exists vs; stSimpl; unfold repr; linLogIntro.
    constructor.
    simpl; intuition.
    discriminate H0.
    constructor.
    eauto.
    exists objs.
    constructor.
    2:assumption.
    intros r X; hypReplace Req0 (Req0 r X).
    destruct Req0 as [E1 Ok2].
    rewrite E1.
    replace (RFile.get (rfileOf (rset (rset s v0 x) ast x)) r)
      with (RFile.get (rfileOf s) r).
    intuition.
    simpl; unfold RFile.set; unfold RFile.get.
    rewrite (beq_rnum_neq (r := ast) (r' := r)).
    rewrite (beq_rnum_neq (r := v0) (r' := r)).
    reflexivity.
    intro Q; subst r.
    simpl in X; intuition.
    discriminate H.
    intro Q; subst r.
    simpl in X; intuition.
    discriminate H.
    assumption.

    clear objs Req0 alOk Hp vs Vhp.
    intros s' [G Rp]; constructor.
    intros vs A Hp.
    apply (G vs A); clear G.
    stSimpl.
    sconjMatch.
    simpl; apply setNonRootOk.
    apply setNonRootOk.
    assumption.
    simpl; intro H.
    repeat (destruct H; [discriminate H | idtac]); assumption.
    simpl; intro H.
    repeat (destruct H; [discriminate H | idtac]); assumption.

    intros r X.
    assert (R1 := Rp r X).
    rewrite <- R1.
    unfold I.presReg in X; unfold SpecUtil.calleeSaved in X; simpl in X.
    intuition; subst; stSimpl; reflexivity.


    (* non-infinite loop case *)
    rewrite (beq_false R).

    assert (I : set_In x (rangeObjs x x0)).
    eapply rangeObjsIntro.
    destruct alOk; omega.
    instantiate (1 := 0).
    omega.

    rewrite addiuStep; stSimpl.
    rename R into Ne.
    assert (Hp2 := sconjInsert Hp Vhp).
    assert (Hp' : s |||- (x |->__ ** top) ** 
      RFile.get (rfileOf s) rgcInfo |-> x ** 4 + RFile.get (rfileOf s) rgcInfo |-> x0 ** top).
    sconjLift4 Hp2; sconjLift4 Hp2; sconjLift4 Hp2.
    clear Hp; sconjMatch.
    hypReplace Hp (fasElim I Hp); sconjMatch.
    clear Hp Hp2; rename Hp' into Hp.

    replace (1 + 0) with 1; [idtac | reflexivity].
    linLogElim Hp.
    rewrite (swStep Hp); auto.
    stSimpl.
    assert (Hp00 := Hp).
    hypReplace Hp (contUpdateSt Hp 1).
    sconjLift2 Hp; rewrite (swStep Hp); auto.
    hypReplace Hp (contUpdateSt Hp 1).
    rewrite addiuStep.
    sconjLift3 Hp.
    rewrite (swStep Hp); auto.
    clear Req0 objs Vhp Hp.
    constructor.
    trivial.
    stSimpl.
    intros; subst s'.

    rename vs into vs0.

    constructor.
    unfold repr; intros vs A Hp.
    linLogElim Hp.
    destruct H3 as [_ [_ [objs [Req Vhp]]]].

    assert (x3 = x).
    sconjLift3 Hp; sconjLift3 Hp00; exact (contContEq Hp Hp00).
    assert (x4 = x0).
    sconjLift4 Hp; sconjLift4 Hp00; exact (contContEq Hp Hp00).
    subst x3; subst x4; clear vs0.

    exists vs; exists x.
    stSimpl.
    
    assert (Hp' : s |||- (x |->__ ** (foralls x in rangeObjs (8 + x) x0, x |->__)) **
      Heap.heap_eqv (heapOf vs) **
      RFile.get (rfileOf s) rgcInfo |-> x ** 4 + RFile.get (rfileOf s) rgcInfo |-> x0 ** A).
    sconjLift2 Hp; sconjMatch.
    hypReplace Hp0 (fasElim I Hp); sconjMatch.
    unfold hpredTrue; rewrite <- (rangeRemoveEqv x x0); assumption.
    clear Hp; rename Hp' into Hp; linLogElim Hp.

    assert (X1 := contContEq Hp Hp00).
    assert (x4 = x2).
    sconjLift2 Hp00; sconjLift2 Hp; exact (contContEq Hp Hp00).
    subst x3 x4; clear Hp00.

    constructor.
    apply gcGuarRefl.

    constructor.

    assert (Q : forall s, RFile.get (rfileOf s) = rget s).
    intros; reflexivity.

    repeat (rewrite Q); repeat (rewrite Q in Hp); repeat (rewrite Q in R).
    stSimpl.

    hypReplace Hp (contUpdateSt Hp 1).
    sconjLift2 Hp.
    hypReplace Hp (contUpdateSt Hp 1).
    sconjLift5 Hp.
    hypReplace Hp (contUpdateSt Hp (8 + x)).
    assert (Hp0 := Hp).
    sconjLift5 Hp; sconjLift4 Hp; hypReplace Hp (heapEqvAdd Hp).
    assert (Hp1 := Hp).
    sconjLift3 Hp; hypReplace Hp (heapEqvAdd Hp).
    linLogIntro.

    constructor.
    simpl; intros [D | [D | D]]; auto; discriminate D.
    constructor.
    instantiate (1 := x0).
    instantiate (1 := (8 + x)).
    destruct alOk as [k QQ]; subst x0; destruct k.
    elimtype False; omega.
    exists k; omega.

    exists (set_add x objs).
    constructor.
    intros r IR.
    simpl in IR; destruct IR as [D | [D | D]].
    subst r.
    stSimpl.
    constructor.
    reflexivity.
    right; apply set_add_intro1; reflexivity.
    subst r; stSimpl.
    destruct (Req rroot).
    simpl; auto.
    constructor.
    rewrite H; reflexivity.
    destruct H0; [left | right]; auto.
    apply set_add_intro2; assumption.
    elimtype False; assumption.
    apply objHpIntro.
    unfold okObjAt; unfold okFieldAt.
    replace I.NULL with 1 by reflexivity.
    assert (HXX : heapOf (uhset (uhset vs x 1) (4 + x) 1)
      ||- x |-> 1 ** 4 + x |-> 1 **
       objHp (set_add x objs) objs).
    apply sconjGLift2'; simpl.
    eapply heapAddOk.
    eapply heapAddOk.
    eapply objHpWeaken; eauto.
    intros z II; apply set_add_intro2; assumption.

    assert (HH : uset (uset (uset (heapOf s) x 1) (4 + x) 1)
      (rget s rgcInfo) (8 + x) ||- (!(ptrP x) ** top) ** top).
    sconjLift3 Hp0; clear Hp Hp1; sconjMatch.
    linLogIntro; exact (contElim1 Hp).

    linLogElim HH; assumption.
    intro Z; linLogElim Z.
    clear Hp; sconjLift5 Hp0; hypReplace Hp0 (sconjInsert Hp0 Z).
    sconjLift4 Hp0; linLogElim Hp0.
    clear Hp Hp0.

    assert (HH : uset (uset (uset (heapOf s) x 1) (4 + x) 1)
      (rget s rgcInfo) (8 + x) ||- (!(ptrP (4 + x)) ** top) ** top).
    sconjLift3 Hp1; sconjMatch.
    linLogIntro; exact (contElim1 Hp).

    linLogElim HH; assumption.

    intro Z; linLogElim Z.
    hypReplace Hp1 (sconjInsert Hp1 Z).
    simpl in Hp1; sconjLift3 Hp1; linLogElim Hp1.

    clear - HXX.
    apply sconjAssocR; scm.
    lli; eauto. left; constructor; constructor.
    lli; eauto. left; constructor; constructor.

    clear Hp0 Hp1.
    sconjLift2 Hp; sconjLift3 Hp; sconjLift3 Hp.
    sconjMatch.

    intros B Vhp'; simpl.
    assert (X : s |||- (!(ptrP x) ** top) ** top).
    sconjMatch.
    linLogIntro; eapply contElim1; eauto.
    linLogElim X; clear X; rename H1 into PtrX.
    apply sconjComm; linLogIntro; apply sconjComm; linLogIntro.
    apply heapAddOk.
    apply sconjComm; apply heapAddOk.
    assumption.
    assumption.
    clear Vhp'; intro Vhp'; linLogElim Vhp'.
    sconjLift4 Hp; hypReplace Hp (sconjInsert Hp Vhp').
    linLogElim Hp.
    destruct (ptrP_elim PtrX) as [k E]; apply (ptrP_intro (S k)).
    omega.
    clear Vhp'; intro Vhp'; linLogElim Vhp'.
    hypReplace Hp (contUpdateSt Hp 1).
    sconjLift4 Hp; sconjLift2 Hp.
    hypReplace Hp (heapEqvAdd Hp).
    replace (S (S (S (S x)))) with (4 + x) in Vhp'.
    hypReplace Hp (sconjInsert Hp Vhp'); linLogElim Hp.
    reflexivity.

    intros r X; stSimpl.
    unfold I.presReg in X; unfold SpecUtil.calleeSaved in X; simpl in X.
    intuition; subst r; stSimpl; reflexivity.

  Qed.


  Definition GCC f :=
    if beq_nat f I.allocLbl then
      Some allocImpl
    else if beq_nat f I.read1Lbl then
      Some (readkImpl 0)
    else if beq_nat f I.read2Lbl then
      Some (readkImpl 4)
    else if beq_nat f I.write1Lbl then
      Some (writekImpl 0)
    else if beq_nat f I.write2Lbl then
      Some (writekImpl 4)
    else
      None.


  Lemma GCCok : cdHeapOkSC phi GCC phi.
  Proof.
    intros x [p g] Peq.
    unfold phi in Peq; unfold GCC; generalize dependent Peq.
    compare x I.allocLbl; intro.

    subst x; rewrite beq_true.
    unfold allocSpec; injection 1; clear Peq; intros; subst p g.
    exact allocImplOk.

    rewrite (beq_false n); clear n.
    compare x I.read1Lbl; intro.

    subst x; rewrite beq_true.
    unfold read1Spec; injection 1; clear Peq; intros; subst p g.
    exact read1ImplOk.

    rewrite (beq_false n); clear n.
    compare x I.read2Lbl; intro.

    subst x; rewrite beq_true.
    unfold read2Spec; injection 1; clear Peq; intros; subst p g.
    exact read2ImplOk.

    rewrite (beq_false n); clear n.
    compare x I.write1Lbl; intro.

    subst x; rewrite beq_true.
    unfold write1Spec; injection 1; clear Peq; intros; subst p g.
    exact write1ImplOk.

    rewrite (beq_false n); clear n.
    compare x I.write2Lbl; intro.

    subst x; rewrite beq_true.
    unfold write2Spec; injection 1; clear Peq; intros; subst p g.
    exact write2ImplOk.

    rewrite (beq_false n); clear n.
    discriminate 1.

  Qed.


End NullGCImpl.


Lemma gcGuarShrinkR : forall s s' rr,
  NullGCGuar.gcGuar s s' rr ->
  forall rr', incl rr' rr ->
    NullGCGuar.gcGuar s s' rr'.
Proof.
  unfold NullGCGuar.gcGuar; unfold NullGCGuar.gcGuar.
  intros s s' rr [G Req] rr' I.
  constructor.
  
  2:intros r I'; apply Req; apply I; assumption.
  
  intros objs h rr3 I' Hp.
  clear Req.
  eapply G.
  2:eauto.
  intros x II; auto.
Qed.



Require Import LinkedList.


Section Heapy.

  Require Import Heap.
  Import HeapMap.

  Lemma natListIsMinObjHp' : forall h l x A,
    h ||- NatListAt l x ** A ->
    h ||- (lexists objs, minObjHp (x::nil) objs) ** A.
  Proof.
    unfold NatListAt; induction l; simpl; intros x A Hp.

    linLogElim Hp.
    subst x.
    hypReplace Hp (unitIL Hp); sconjMatch.
    linLogIntro.
    instantiate (1 := empty_set).
    unfold minObjHp; apply aconjIntro.
    simpl; intros objs' OO; intuition.
    unfold okObjHp.
    linLogIntro.
    simpl; intros v [II | II].
    subst v; left; unfold NULL; constructor; constructor.
    intuition.
    sconjMatch.
    
    (* inductive case *)
    linLogElim Hp.
    rename H3 into NE; rename x0 into y.

    assert (NE' : x <> y).
    intro; subst y.
    destruct l; simpl in Hp; linLogElim Hp.
    auto.
    sconjLift3 Hp; linLogElim Hp.

    sconjLift3 Hp.
    assert (Hp' := IHl _ _ Hp); clear IHl Hp.
    linLogElim Hp'.
    rename x0 into objs.
    sconjLift4of4 Hp'; apply sconjComm; sconjMatch.
    existsl (set_add x objs).
    unfold minObjHp; apply aconjIntro.

    simpl; intros z I.
    exists x; constructor.
    simpl; auto.
    destruct I.
    subst z; constructor.
    econstructor 3.
    instantiate (1 := y).
    rewrite <- heap_canonEqv.
    sconjLift3of3 Hp; eapply contGetEqL; eauto.
    destruct Hp as [ha [hb [J [HA _]]]].
    unfold minObjHp in HA; destruct (aconjElim HA) as [X _]; clear HA.
    eapply reachableJoinLeft; eauto.
    unfold rootReachable in X.
    destruct (X _ H) as [v [Veq R]].
    simpl in Veq; destruct Veq.
    subst v.
    assumption.
    elimtype False; assumption.

    unfold okObjHp; linLogIntro.
    simpl; intros v [I | I].
    subst v; right; apply set_add_intro1; reflexivity.
    intuition.

    assert (X : h ||- (!okFieldVal objs y ** top) ** top).
    sconjMatch.
    linLogIntro.
    unfold minObjHp in Hp; destruct (aconjElim Hp) as [_ X]; clear Hp.
    unfold okObjHp in X; linLogElim X.
    apply H1; simpl; auto.

    linLogElim X; clear X; rename H1 into OkY.
    assert (Hp' : h ||- objHp objs objs ** okObjAt objs x).
    sconjMatch.
    unfold minObjHp in Hp; destruct (aconjElim Hp) as [_ Hp']; clear Hp.
    unfold okObjHp in Hp'; linLogElim Hp'; assumption.
    unfold okObjAt; unfold okFieldAt; sconjMatch.
    linLogIntro.
    2:eauto.
    left; unfold natEnc.
    unfold atomP; eapply oddIntro; reflexivity.
    linLogIntro; eauto.
    apply objHpIntro.
    apply sconjComm; sconjMatch.
    eapply objHpWeaken; eauto.
    intros q I; apply set_add_intro2; assumption.
    eapply okObjAtWeaken; eauto.
    intros q I; apply set_add_intro2; assumption.

  Qed.


End Heapy.


Lemma natListIsMinObjHp : forall h l x,
  h ||- NatListAt l x ->
  h ||- lexists objs, minObjHp (x::nil) objs.
Proof.
  intros h l x Hp.
  hypReplace Hp (unitIR Hp).
  assert (X := natListIsMinObjHp' _ _ Hp).
  linLogElim X.
  existsl x0; assumption.
Qed.

