Require Import CAP.
Require Export Machine.
Require StateTheory.

(* definitions and rules taken from Xinyu's SCAP paper *)

Definition spred := state -> Prop.
Definition guar := state -> state -> Prop.

Inductive bspec : Type := 
  | bspeci : spred -> guar -> bspec.

Notation chspec := (chspec bspec).


Inductive iseqOkS (P:chspec) : spred -> guar -> iseq -> Prop :=
  | retOk : forall (p:spred) (g:guar),
    (forall s, p s -> g s s) ->
    iseqOkS P p g (jr ra)
  | tailCallOk : forall (p p':spred) (g g':guar) f,
    P f = (SomeT (bspeci p' g')) ->
    (forall s, p s -> p' s /\ forall s', g' s s' -> g s s') ->
    iseqOkS P p g (j f)
  | callOk : forall (p p' p'':spred) (g g' g'':guar) f fret,
    P f = (SomeT (bspeci p' g')) ->
    P fret = (SomeT (bspeci p'' g'')) ->
    (forall s, p s -> p' (rset s ra fret) /\
      forall s', g' (rset s ra fret) s' -> p'' s' /\
	forall s'', g'' s' s'' -> g s s'') ->
    (forall s s', p' s -> g' s s' -> rget s ra = rget s' ra) ->
    iseqOkS P p g (jal f fret)
  | brOk : forall (p p' p'':spred) (g g' g'':guar) bt rs rt f i,
    P f = (SomeT (bspeci p' g')) ->
    iseqOkS P p'' g'' i ->
    (forall s, p s ->
      (btPred bt (rget s rs) (rget s rt) -> p' s /\ (forall s', g' s s' -> g s s')) /\
      (btPred (btNeg bt) (rget s rs) (rget s rt) -> p'' s /\ (forall s', g'' s s' -> g s s'))) ->
    iseqOkS P p g (br bt rs rt f i)
  | scapSeqOk : forall (p p':spred) (g g':guar) c i,
    iseqOkS P p' g' i ->
    (forall s, p s -> optPred (cstep c s)
      (fun s' => p' s' /\ forall s'', g' s' s'' -> g s s'')) ->
    iseqOkS P p g (c; i).


Definition cdHeapOkSC (P : chspec) (C : cdHeap) (P' : chspec) : Prop :=
  forall f sp, P' f = SomeT sp ->
    let (p, g) := sp in
      optPred (C f) (iseqOkS P p g).

Fixpoint WFST (n:nat) (q:spred) (P:chspec) {struct n} : Prop :=
  forall s, q s ->
    match n with
      | 0 => False
      | S n' => optPredT (P (rget s ra)) 
	(fun sp => let (p, g) := sp in p s /\ WFST n' (g s) P)
    end.

Definition interp (sp:bspec) : chstpred bspec :=
  let (p, g) := sp in
    fun P => fun s => p s /\ exists n, WFST n (g s) P.


Lemma WFSTweak : forall n (p p':spred) P,
  WFST n p P ->
  (forall s, p' s -> p s) ->
  WFST n p' P.
induction n; simpl.
eauto.
intros p p' P jOk pWeak s p'Ok.
exact (jOk _ (pWeak _ p'Ok)).
Qed.


Lemma scapToCap : forall P p g i,
  iseqOkS P p g i ->
  iseqOk interp (lift (interp (bspeci p g)) P) i.
unfold lift.
induction 1 as
  [AA BB rOk | AA p' BB g' f fOk pOk gOk |
    p p' p'' g g' g'' f fret fOk fretOk cOk raOk |
    p p' p'' g g' g'' bt rs rt f i fOk iOk bOk |
    p p' g g' c i' iOk IH Hweak ];
  subst; simpl.

(* ret *)
constructor.
intros.
destruct H as [Pweak [pOk [n wfst]]].
assert (gOk := rOk _ pOk); clear rOk pOk.

destruct n; simpl in wfst.

elimtype False; eauto.

assert (Q := wfst _ gOk); clear wfst gOk.
destruct (P0 (rget s ra)).
destruct b.
simpl.
destruct Q.
constructor.
assumption.
econstructor.
eauto.
assumption.

(* tail call *)
constructor.
intros P' s [Pweak [psOk [n WF]]].
rewrite (Pweak _ _ fOk); clear Pweak fOk.
simpl.
destruct (pOk _ psOk) as [Hp Hg]; clear pOk psOk.
constructor.
assumption.
exists n.
eapply WFSTweak; eauto.

(* call *)
constructor.
intros P' s [Pweak [pOk [n WF]]].
rewrite (Pweak _ _ fOk); clear fOk f.
destruct (cOk _ pOk) as [p'Ok Q]; clear cOk pOk.
simpl.
constructor.
assumption.

exists (S n).
simpl.
intros s' g'Ok.
assert (raOk' := raOk _ _ p'Ok g'Ok); clear raOk.
destruct (Q _ g'Ok) as [p''Ok gWeak]; clear Q g'Ok.
rewrite StateTheory.rgetRsetEq in raOk'.
rewrite <- raOk'; rewrite (Pweak _ _ fretOk); clear raOk' Pweak fretOk.
constructor.
assumption.
eapply WFSTweak.
eauto.
assumption.
discriminate.

(* br *)
econstructor.
apply bOk.
simpl.
intros P' s [Pweak [pOk [n WF]]].
destruct (H _ pOk) as [bOk1 bOk2]; clear H pOk.
constructor; [clear bOk1 | clear bOk2].
intro Ne.
assert (Q := bOk2 Ne); clear bOk2 Ne.
intuition.
exists n; eapply WFSTweak; eauto.
intro Eq.
assert (Q := bOk1 Eq); clear bOk1 Eq.
rewrite (Pweak _ _ fOk); clear Pweak fOk.
simpl.
intuition.
exists n.
eapply WFSTweak; eauto.

(* scap seq *)
eapply seqOk.
apply IH.
intros P' s [Pweak [pOk [n WF]]].
assert (PP := Hweak _ pOk); clear Hweak pOk.
destruct (cstep c s).
constructor.
assumption.
destruct PP.
simpl.
constructor.
assumption.
exists n.
eapply WFSTweak.
eauto.
eauto.
assumption.

Qed.


Theorem iseqOkWeak : forall P p g ii,
  iseqOkS P p g ii -> forall (p':spred) (g':guar),
  (forall s, p' s -> p s /\ forall s', g s s' -> g' s s') ->
  iseqOkS P p' g' ii.
Proof.
  induction 1; econstructor; eauto.

  intro s.
  assert (Q := H s); assert (Z := H0 s); clear H H0.
  intuition.

  intro s.
  assert (Q := H0 s); assert (Z := H1 s); clear H0 H1.
  intuition.

  intro s.
  assert (Q := H1 s); assert (Z := H3 s); clear H1 H3.
  intuition; assert (A := H7 s'); intuition.

  intro s.
  assert (Q := H1 s); assert (Z := H2 s); clear H1 H2.
  intuition.
  simpl.

  intro s.
  assert (Q := H0 s); assert (Z := H1 s); clear H0 H1.
  intuition.
  destruct (cstep c s).
  intuition.
  assumption.

Qed.



Theorem iseqOkWeakP : forall P p g ii,
  iseqOkS P p g ii ->
  forall P',
    (forall f p0 g0, P f = SomeT (bspeci p0 g0) ->
      exists p1, exists g1, P' f = SomeT (bspeci p1 g1) /\
	p0 = p1 /\ forall s s', g1 s s' -> g0 s s') ->
    iseqOkS P' p g ii.
Proof.
  intros P p g ii I P' Psub.
  induction I.
  econstructor; auto.

  destruct (Psub _ _ _ H) as [p'' [g'' [P'eq [X Y]]]]; clear H Psub.
  subst p''.
  econstructor.
  eauto.
  intros s Pok.
  destruct (H0 _ Pok) as [P'ok Gok].
  intuition.

  destruct (Psub _ _ _ H) as [p''1 [g''1 [P'eq1 [X1 Y1]]]].
  destruct (Psub _ _ _ H0) as [p''2 [g''2 [P'eq2 [X2 Y2]]]]; clear H H0 Psub.
  subst p''1 p''2.
  econstructor.
  eauto.
  eauto.
  clear P'eq1 P'eq2.
  intros s Pok.
  destruct (H1 _ Pok) as [P'ok Q]; clear H1 Pok.
  constructor.
  intuition.
  intros s' g11ok.
  destruct (Q _ (Y1 _ _ g11ok)) as [p''ok G2ok]; clear g11ok Y1 Q.
  intuition.
  auto.

  rename i into ii.
  destruct (Psub _ _ _ H) as [p''1 [g''1 [P'eq1 [X1 Y1]]]]; clear Psub H.
  subst p''1.
  econstructor; eauto.
  intros s Pok; destruct (H0 _ Pok) as [A B]; clear H0 Pok.
  constructor.
  intro U; destruct (A U) as [Z Z']; intuition.
  intro U; destruct (B U) as [Z Z']; intuition.

  econstructor; eauto.

Qed.


