Require Import CheneyLemma.


Set Implicit Arguments.


(* code address constants *)
Definition SCANFIELD   := 89.
Definition SCANFIELD2  := 21.
Definition RRETURN     := 22.
Definition FWDOBJ      := 23.
Definition NOTTOSPACEPTR := 24.
Definition NOTFROMSPACEPTR := 25.
Definition SCANPOINTER := 26.
Definition SCANCOPIED  := 27.
Definition SCANNOCOPY  := 28.
Definition INFLOOP     := 29.
Definition RETURN      := 88.


(* local register conventions *)
Definition rfree := a3.
Definition rtoStart := t7.
Definition rtoEnd := t8.
Definition rfrStart := a1.
Definition rfrEnd := a2.
Definition ralloc := t2.

Definition rtemp0 := v1.
Definition rtemp1 := t9.
Definition rtemp2 := t4.
Definition raSave0 := t5.


(* code block definitions *)

Definition infLoop :=
  j INFLOOP.

Definition fwdObj :=
  (* if no space left, give up *)
  beq rfree ralloc INFLOOP;;

  (* copy fields from src to free *)
  lw ast 0 a0;
  sw ast 0 rfree;
  lw ast WORDSIZE a0;
  sw ast WORDSIZE rfree;
  (* store forwarding pointer in src *)
  sw rfree 0 a0;
  (* return value of free *)
  addiu v0 rfree 0;
  (* increment free *)
  addiu rfree rfree OBJSIZE;
  jr ra.


Notation isNotSpacePtr :=
  (fun rstart rend =>
    andi1 v0 a0;
    bne v0 zero RETURN;;
    sltu v0 a0 rstart;
    bne v0 zero RETURN;;
    addiu v0 a0 1;
    sltu v0 rend v0;
    bne v0 zero RETURN;;
    addiu v0 zero 0;
    jr ra).

Definition isNotToSpacePtr := isNotSpacePtr rtoStart rtoEnd.
Definition isNotFromSpacePtr := isNotSpacePtr rfrStart rfrEnd.


Definition scanField :=
   lw rtemp1 0 a0;           (* rtemp1 := [a0] *)
   addiu rtemp0 a0 0;        (* rtemp0 := a0   *)
   addiu raSave0 ra 0;       (* raSave0 := ra  *)
   addiu a0 rtemp1 0;        (* a0 := rtemp1   *)
   jal NOTFROMSPACEPTR SCANFIELD2.

Definition scanField2 :=
   bne v0 zero RRETURN;;  (* if v0 <> 0 return *)
   lw rtemp2 0 rtemp1;    (* rtemp2 := [rtemp1] *)
   addiu a0 rtemp2 0;     (* a0 := rtemp2 *)
   jal NOTTOSPACEPTR SCANPOINTER.

Definition scanPointer :=
  beq v0 zero SCANNOCOPY;;
  addiu a0 rtemp1 0;
  jal FWDOBJ SCANCOPIED.

Definition scanNoCopy :=
  sw rtemp2 0 rtemp0;
  addiu ra raSave0 0;
  jr ra.

Definition scanCopied :=
  sw v0 0 rtemp0;
  addiu ra raSave0 0;
  jr ra.

Definition rReturn :=
  addiu ra raSave0 0;
  jr ra.


(* specs *)

Notation rReturnPre := (fun s => True).
Notation rReturnGuar := (fun s s' => s' = rset s ra (rget s raSave0)).

Notation infLoopPre := rReturnPre.
Notation infLoopGuar := (fun s s' => False).


Definition fwdObjPre s :=
  rget s rfree <> rget s ralloc ->
  s |||- rget s a0 |->__ ** rget s rfree |->__ ** top.

Definition fwdObjGuar s s' :=
  (forall A x y x' y',
    s |||- rget s a0 |->, (x::y::nil) ** 
      rget s rfree |->, (x'::y'::nil) ** A ->
	heapOf s' ||- rget s a0 |->, (rget s rfree::y::nil) ** 
	  rget s rfree |->, (x::y::nil) ** A) /\
  (rget s rfree = rget s ralloc -> False) /\
  rget s' rfree = OBJSIZE + rget s rfree /\
  rget s' v0 = rget s rfree /\
  regPres (regNotInList (rfree::ast::v0::nil)) s s'.


Definition isNotSpacePtrPre : spred := fun s => True.

Notation isNotSpacePtrGuar := (fun rstart rend s s' =>
  sbool (rget s' v0) (notSpacePtrP (rget s a0) (rget s rstart) (rget s rend)) /\
  heapOf s' = heapOf s /\
  forall r, r <> v0 -> rget s' r = rget s r).

Definition isNotToSpacePtrGuar := isNotSpacePtrGuar rtoStart rtoEnd.
Definition isNotFromSpacePtrGuar := isNotSpacePtrGuar rfrStart rfrEnd.

Definition scanNoCopyPre s :=
  exists v, heapOf s ||- rget s rtemp0 |-> v ** top.

Definition scanNoCopyGuar s s' :=
  (forall v A, heapOf s ||- rget s rtemp0 |-> v ** A ->
    heapOf s' ||- rget s rtemp0 |-> rget s rtemp2 ** A) /\
  rfileEqv s' (rset s ra (rget s raSave0)).


Definition scanCopiedPre s :=
  exists v, heapOf s ||- rget s rtemp0 |-> v ** top.

Definition scanCopiedGuar s s' :=
  (forall v A, heapOf s ||- rget s rtemp0 |-> v ** A ->
    heapOf s' ||- rget s rtemp0 |-> rget s v0 ** A) /\
  rfileEqv s' (rset s ra (rget s raSave0)).


Definition scanPointerPre s :=
  if beq_nat (rget s v0) 0 then
    scanNoCopyPre s
  else
    rget s rfree <> rget s ralloc ->
    exists v, 
      heapOf s ||- rget s rtemp0 |-> v ** rget s rtemp1 |->__ ** rget s rfree |->__ ** top.

Definition scanPointerGuar s s' :=
  if beq_nat (rget s v0) 0 then
    scanNoCopyGuar s s'
  else
    exists s'',
      fwdObjGuar (rset (rset s a0 (rget s rtemp1)) ra SCANCOPIED) s'' /\ 
      scanCopiedGuar s'' s'.


Definition scanField2Pre s :=
  rget s v0 = 0 -> 
  exists v, s |||- rget s rtemp1 |-> v ** top /\
    forall v', 
      sbool v' (notSpacePtrP v (rget s rtoStart) (rget s rtoEnd)) ->
      scanPointerPre (rset (rset (rset s rtemp2 v) a0 v) v0 v').


Definition scanField2Guar s s'' :=
  if beq_nat (rget s v0) 0 then
    (exists v, s |||- rget s rtemp1 |-> v ** top /\
      exists v',
	sbool v' (notSpacePtrP v (rget s rtoStart) (rget s rtoEnd)) /\
	scanPointerGuar 
	(rset (rset (rset (rset s rtemp2 v) a0 v) ra SCANPOINTER) v0 v') s'')
  else
    rReturnGuar s s''.


Definition fromSpace s :=
  setRange (rget s rfrStart) (rget s rfrEnd).

Definition toSpace s :=
  setRange (rget s rtoStart) (rget s rtoEnd).

Definition scanFieldStOk phi W M F A s :=
  set_disj (fromSpace s) (toSpace s) /\
  set_subset (W |_| M) (fromSpace s) /\
  set_eqv F (rangeObjs (rget s rfree) (rget s ralloc)) /\
  (forall x y, phi x = Some y -> 
    ~notSpacePtrP y (rget s rtoStart) (rget s rtoEnd)) /\
  rget s rtoStart <= rget s rfree <= rget s ralloc /\
  rget s ralloc <= rget s rtoEnd /\
  aligned8 (rget s rfree) (rget s ralloc) /\
  heapOf s ||- mapHp M phi ** objHp (W |_| M) W **
    (foralls x in F, x |->__) ** A.

Definition spaceIdMap a b x :=
  if blt_nat x b then
    if blt_nat a (S x) then
      Some x
    else
      None
  else
    None.

Notation toSpaceIdMap :=
  (fun s => spaceIdMap (rget s rtoStart) (rget s rtoEnd)).

Definition fieldRespToMap (phi:mapnn) s v :=
  fieldRespMap (map_unionnn (toSpaceIdMap s) phi) s v.


Definition scanFieldPre s :=
  exists phi, exists W, exists M, exists F,
    scanFieldStOk phi W M F 
    (okFieldAt ((W |_| M) |_| toSpace s) (rget s a0) ** top) s.

Definition scanFieldGuar s s' :=
  (forall phi W M F A,
    scanFieldStOk phi W M F
    (okFieldAt ((W |_| M) |_| toSpace s) (rget s a0) ** A) s ->
    (exists M'', exists W', exists F', exists G, exists phi',
      set_eqv W (M'' |_| W') /\
      set_eqv F (G |_| F') /\
      iso_set phi' M'' G /\
      aligned8 (rget s rfree) (rget s' rfree) /\
      (forall x, set_In x W' -> heapOf s' ||- objCopied x s x ** top) /\
      scanFieldStOk (map_union phi phi') W' (M |_| M'') F'
      (fieldRespToMap (map_union phi phi') s (rget s a0) **
	(foralls x in G, objCopiedPhi phi' s x) ** A) s')) /\
    regPres 
    (regNotInList (rfree::a0::v0::ast::rtemp0::rtemp1::rtemp2::raSave0::nil)) s s'.


(* code heap spec *)

Definition Pscanf : chspec := fun l => 
  if beq_nat l RETURN then
    SomeT (bspeci returnPre returnGuar)
  else if beq_nat l SCANPOINTER then
    SomeT (bspeci scanPointerPre scanPointerGuar)
  else if beq_nat l NOTTOSPACEPTR then
    SomeT (bspeci isNotSpacePtrPre isNotToSpacePtrGuar)
  else if beq_nat l NOTFROMSPACEPTR then
    SomeT (bspeci isNotSpacePtrPre isNotFromSpacePtrGuar)
  else if beq_nat l SCANCOPIED then
    SomeT (bspeci scanCopiedPre scanCopiedGuar)
  else if beq_nat l SCANNOCOPY then
    SomeT (bspeci scanNoCopyPre scanNoCopyGuar)
  else if beq_nat l FWDOBJ then
    SomeT (bspeci fwdObjPre fwdObjGuar)
  else if beq_nat l SCANFIELD2 then
    SomeT (bspeci scanField2Pre scanField2Guar)
  else if beq_nat l SCANFIELD then
    SomeT (bspeci scanFieldPre scanFieldGuar)
  else if beq_nat l RRETURN then
    SomeT (bspeci rReturnPre rReturnGuar)
  else if beq_nat l INFLOOP then
    SomeT (bspeci infLoopPre infLoopGuar)
  else
    NoneT _.



Ltac PscanfSimpl := 
  unfold Pscanf; 
  rewrite beq_true; 
  repeat (rewrite beq_false; [idtac | discriminate]); 
  reflexivity.


(* lemmas related to the scanField invariant *)


Ltac destrSF H :=
  let WMdisj := fresh "WMdisj" in
  let WMsub := fresh "WMsub" in
  let Feq := fresh "Feq" in
  let NSP := fresh "NSP" in
  let Le1 := fresh "Le1" in
  let Le2 := fresh "Le2" in
  let Le3 := fresh "Le3" in
  let AlF := fresh "AlF" in
  let Hp := fresh "Hp" in
    destruct H as [WMdisj [WMsub [Feq [NSP [[Le1 Le2] [Le3 [AlF Hp]]]]]]].

Lemma phiWEq : forall phi W M F P s,
  scanFieldStOk phi W M F P s ->
  forall x, set_In x W -> phi x = None.
Proof.
  intros phi W M F P s H x IW.
  destrSF H.
  assert (NI : ~set_In x M).
  intro IM.
  hypReplace Hp (sconjImplL Hp _ (mapHpElim1 IM (phi:=phi))).
  linLogElimOld Hp.
  sconjLift3 Hp.
  hypReplace Hp (sconjImplL Hp _ 
    (objHpElim IW (s1 := W |_| M))).
  sconjLift3 Hp.
  unfold okObjAt in Hp; unfold okFieldAt in Hp.
  linLogElimOld Hp.
  sconjLift5 Hp; linLogElim Hp.
  refine (sconjifierL (mapHpElim2 NI (phi:=phi)) Hp).
Qed.


Lemma scanFieldPtrW : forall phi W M F A s,
  scanFieldStOk phi W M F A s ->
  forall x, set_In x W ->
    ptrP x.
Proof.
  intros phi W M F A s H x IW.
  destrSF H.

  assert (Ne : hget s x <> None).
  intro Eq.
  sconjLift2 Hp.
  hypReplace Hp (sconjImplL Hp _ (objHpElim IW (s1 := _))).
  unfold okObjAt in Hp; unfold okFieldAt in Hp; linLogElim Hp.
  assert (X := contGetEqL Hp).
  intuition.
  unfold hget in Eq.
  rewrite Eq in X; discriminate X.

  unfold hget in Ne; unfold Heap.get in Ne.
  assert (U := isPtrOk x).
  destruct (isPtr x).
  assumption.
  elimtype False; auto.
    
Qed.


Lemma scanFieldGuarPresRa : forall s s',
  scanFieldPre s -> 
  scanFieldGuar s s' -> 
  rget s ra = rget s' ra.
Proof.
  unfold scanFieldPre; unfold scanFieldGuar.
  intros s s' _ [_ Rp].
  rewrite (regPresEq Rp).
  reflexivity.
  regNotInListSolver.
Qed.


Lemma scanFieldStOkTop : forall phi W M F P P' s,
  scanFieldStOk phi W M F (P ** P') s ->
  scanFieldStOk phi W M F (P ** top) s.
Proof.
  intros phi W M F P P' s H.
  destrSF H.
  unfold scanFieldStOk; intuition.
  sconjMatch.
Qed.


Lemma scanFieldStOkImpl : forall phi W M F P s ,
  scanFieldStOk phi W M F P s ->
  forall (P':hpred), (forall h, h ||- P -> h ||- P') ->
  scanFieldStOk phi W M F P' s.
Proof.
  intros phi W M F P s H P' Pimp.
  destrSF H.
  unfold scanFieldStOk; intuition.
  sconjMatch.
Qed.


Lemma scanFieldStOkWMDisj : forall phi W M F P s,
  scanFieldStOk phi W M F P s -> set_disj W M.
Proof.
  destruct s; unfold scanFieldStOk; simpl.
  unfold mapHp; unfold objHp.
  intro H; destrSF H.
  clear WMdisj WMsub Feq NSP Le1 Le2 Le3 AlF.
  linLogElim Hp; clear H1.
  apply set_disjSymm.
  apply (fasFasDisj Hp); clear Hp P; rename h into h0; 
    intros x h Hp.
  clear h0.
  linLogElim Hp.
  linLogIntro; sconjMatch.
  unfold okObjAt in Hp; unfold okFieldAt in Hp; linLogElim Hp.
  linLogIntro; sconjMatch.
Qed.


Lemma scanFieldStOkFGDisj : forall phi W M F B G A s,
  scanFieldStOk phi W M F (fas B G ** A) s -> 
  (forall h x, h ||- B x -> h ||- x |->_ ** top) ->
  set_disj F G.
Proof.
  unfold scanFieldStOk.
  intros phi W M F B G A.
  destruct s.
  intros [_ [_ [_ [_ [_ [_ [_  H]]]]]]] Imp.
  eapply fasFasDisj.
  assert (Q := H).
  sconjLift3 Q; apply (sconjImplR Q); clear Q; intros h' Q.
  sconjLift3 Q; apply (sconjImplR Q); clear h' Q; intros h' Q.
  apply Q.
  simpl; intros h' x H'; apply (sconjImplR H'); intros; linLogIntro.
  assumption.
Qed.


Lemma scanFieldStOkMorphHalf : forall u s1 s1' s2 s2' s3 s3' A s s',
  set_eqv s1 s1' ->
  set_eqv s2 s2' ->
  set_eqv s3 s3' ->
  state_eqv s s' ->
  scanFieldStOk u s1 s2 s3 A s ->
  scanFieldStOk u s1' s2' s3' A s'.
Proof.
  intros until s'; intros E1 E2 E3 Seq H.
  destrSF H.

  unfold scanFieldStOk.
  unfold fromSpace; unfold toSpace.
  replace (rget s' rfrStart) with (rget s rfrStart).
  replace (rget s' rfrEnd) with (rget s rfrEnd).
  replace (rget s' rtoStart) with (rget s rtoStart).
  replace (rget s' rtoEnd) with (rget s rtoEnd).
  replace (rget s' rfree) with (rget s rfree).
  replace (rget s' ralloc) with (rget s ralloc).
  intuition.

  setoid_rewrite <- E1.
  setoid_rewrite <- E2.
  assumption.

  setoid_rewrite <- E3.
  assumption.
  
  clear WMdisj WMsub Feq NSP Le1 Le2 Le3.
  eapply heapEqvTrue.
  apply Heap.heap_eqvSymm; apply stateHeapEqv; eauto.
  apply (sconjImpl Hp); clear Hp; intros h H.
  unfold hpredTrue; setoid_rewrite <- E2.
  assumption.

  apply (sconjImpl H); clear h H; intros h H.
  unfold hpredTrue.
  setoid_rewrite <- E1.
  setoid_rewrite <- E2.
  assumption.
 
  apply (sconjImpl H); clear h H; intros h H.
  unfold hpredTrue.
  setoid_rewrite <- E3.
  assumption.
  assumption.

  apply rgetEqv; assumption.
  apply rgetEqv; assumption.
  apply rgetEqv; assumption.
  apply rgetEqv; assumption.
  apply rgetEqv; assumption.
  apply rgetEqv; assumption.

Qed.


Add Morphism scanFieldStOk
  with signature eq ==> set_eqv ==> set_eqv ==> set_eqv ==> eq ==> state_eqv ==> iff
  as scanFieldStOkMorphism.
  intros u s1 s1' E1 s2 s2' E2 s3 s3' E3 A s s' Seq; constructor; intro H;
    eapply scanFieldStOkMorphHalf; eauto; try (apply set_eqvSymm; auto).
  apply state_eqvSymm; auto.
Qed.


Lemma scanFieldStOkPhiEqv : forall phi W M F A s,
  scanFieldStOk phi W M F A s ->
  forall phi', mapEqv phi phi' ->
    scanFieldStOk phi' W M F A s.
Proof.
  intros phi W M F A s H phi' Meq.
  destrSF H.
  unfold scanFieldStOk; intuition.
  assert (Q := NSP x y).
  rewrite (Meq x) in Q; auto.
  apply (sconjImplL Hp); clear Hp; intros h Hp.
  exact (mapHpPhiEqv Hp Meq).
Qed.


Notation heap_eqv := Heap.heap_eqv.

Lemma scanFieldStOkExtract : forall phi W M F A B s,
  scanFieldStOk phi W M F (A ** B) s ->
  exists h, scanFieldStOk phi W M F (A ** heap_eqv h) s /\ h ||- B.
Proof.
  intros phi W M F A B s H; destrSF H.
  sconjLift5of5 Hp.
  destruct (sconjExtract Hp) as [h [H' HB]].
  exists h.
  clear Hp.
  unfold scanFieldStOk; intuition.
  hypReplace H' (sconjComm H'); linLogElim H'.
  assumption.
Qed.


Lemma scanPhiDom : forall phi W M F A s,
  scanFieldStOk phi W M F A s ->
  forall x, ~set_In x M -> 
    phi x = None.
Proof.
  intros phi W M F A s H x NI; destrSF H.
  assert (Hp' : heapOf s ||- (!(phi x = None) ** top) ** top).
  sconjMatch.
  hypReplace Hp (mapHpElim2 NI Hp).
  linLogIntro.
  linLogElim Hp'.
  assumption.
Qed.


(* spaceIdMap soundness *)

Lemma set_memSpaceIdMap : forall x y z,
  if set_mem z (setRange x y) then
    spaceIdMap x y z = Some z
  else
    spaceIdMap x y z = None.
Proof.
  intros x y z.
  unfold spaceIdMap.
  assert (X := blt_natSound z y).
  destruct (blt_nat z y).
  assert (Y := blt_natSound x (S z)).
  destruct (blt_nat x (S z)).
  rewrite set_mem_true.
  reflexivity.
  eapply setRangeIntro; omega.
  rewrite set_mem_false.
  reflexivity.
  intro I; destruct (setRangeElim I); clear I; omega.
  rewrite set_mem_false.
  reflexivity.
  intro I; destruct (setRangeElim I); clear I; omega.
Qed.


Lemma spaceIdMapSome : forall x y z,
  set_In z (setRange x y) ->
  spaceIdMap x y z = Some z.
Proof.
  intros x y z I.
  assert (X := set_memSpaceIdMap x y z).
  rewrite (set_mem_true I) in X; assumption.
Qed.


Lemma spaceIdMapNone : forall x y z,
  ~set_In z (setRange x y) ->
  spaceIdMap x y z = None.
Proof.
  intros x y z NI.
  assert (X := set_memSpaceIdMap x y z).
  rewrite (set_mem_false NI) in X; assumption.
Qed.


Lemma spaceIdMapEq : forall x y b,
  spaceIdMap x y b = if set_mem b (setRange x y) then Some b else None.
Proof.
  intros x y b.
  destruct (set_In_dec b (setRange x y)) as [I | NI].
  rewrite (set_mem_true I); apply spaceIdMapSome; assumption.
  rewrite (set_mem_false NI); apply spaceIdMapNone; assumption.
Qed.


(* block verification *)

Lemma rReturnOk : iseqOkS Pscanf rReturnPre rReturnGuar rReturn.
Proof.
  unfold rReturn.
  iseqWeakSolver PscanfSimpl.
  intuition.
Qed.

Lemma infLoopOk : iseqOkS Pscanf infLoopPre infLoopGuar infLoop.
Proof.
  unfold infLoop.
  iseqWeakSolver PscanfSimpl.
  intuition.
Qed.


Lemma scanNoCopyOk : iseqOkS Pscanf scanNoCopyPre scanNoCopyGuar scanNoCopy.
Proof.
  unfold scanNoCopyPre; unfold scanNoCopyGuar; unfold scanNoCopy.
  iseqWeakSolver PscanfSimpl.
  intros s [v H].
  rewrite (swStep H); try reflexivity.
  rewrite addiuStep.
  constructor.
  trivial.
  stSimpl; simpl.
  intros; subst.
  stSimpl.
  constructor.
  intros v0 h0 H0.
  simpl.
  apply (contUpdateSt H0).
  intro r; destruct r; stSimpl; reflexivity.
Qed.


Lemma scanCopiedOk : iseqOkS Pscanf scanCopiedPre scanCopiedGuar scanCopied.
Proof.
  unfold scanCopiedPre; unfold scanCopiedGuar; unfold scanCopied.
  iseqWeakSolver PscanfSimpl.
  intros s [v H].
  rewrite (swStep H); [idtac | reflexivity | reflexivity].
  rewrite addiuStep.
  constructor.
  trivial.
  intros; subst.
  stSimpl; simpl.
  constructor.
  intros v0 h0 H0.
  apply (contUpdateSt H0).
  intro r; destruct r; stSimpl; reflexivity.
Qed.


Lemma scanPointerOk : iseqOkS Pscanf scanPointerPre scanPointerGuar scanPointer.
Proof.
  unfold scanPointerPre; unfold scanPointerGuar; unfold scanPointer.
  iseqWeakSolver PscanfSimpl.

  unfold fwdObjGuar.
  intros s s' _ [_ [_ [_ [_ H]]]]; apply H; regNotInListSolver.

  intros s.
  compare (rget s v0) 0.

  intro R; rewrite R; rewrite beq_true.
  rewrite bteqEq; stSimpl; rewrite beq_true.
  auto.

  intro NeqV0.
  rewrite (beq_false NeqV0).
  rewrite addiuStep; stSimpl.
  rewrite bteqEq.
  rewrite (beq_false NeqV0).
  unfold fwdObjPre; unfold fwdObjGuar; stSimpl.
  replace (0 + rget s rtemp1) with (rget s rtemp1).
  2:reflexivity.
  replace (0 + rget s rfree) with (rget s rfree).
  2:reflexivity.
  intro Pre.
  constructor.
  constructor.
  intro Ne.
  destruct (Pre Ne) as [v Pre']; clear Pre; rename Pre' into Pre.
  sconjLift2 Pre.
  apply (sconjImplR Pre); clear Pre; intros h Hp.
  sconjLift2 Hp.
  apply (sconjImplR Hp); intros; linLogIntro.

  intros s' [G [Neq [Feq [V0eq Cso]]]].

  destruct (Pre Neq) as [v Hp]; clear Pre.
  linLogElim Hp.
  simpl in G.
  assert (Gx := G (rget s rtemp0 |-> v ** top) x x0 x1 x2); clear G.
  exists v.
  replace (rget s' rtemp0) with (rget s rtemp0).

  assert (X : ((rget s rtemp1 |-> rget s rfree ** S (S (S (S (rget s rtemp1)))) |-> x0) **
    (rget s rfree |-> x ** S (S (S (S (rget s rfree)))) |-> x0) **
    rget s rtemp0 |-> v ** top) (Heap.heap_canon (heapOf s'))).
  apply Gx.
  clear Gx Cso V0eq Feq.
  sconjGLift3.
  linLogIntro.
  sconjMatch.

  clear Hp Gx.
  sconjLift3 X; sconjMatch.
  rewrite (regPresEq Cso rtemp0).
  stSimpl; reflexivity.
  regNotInListSolver.

  intros.
  destruct H as [s'' [[G [Neq [Feq [V0eq Cso]]] SCG]]].
  exists s''.
  rewrite Feq.
  intuition.
Qed.


Section INSP.

Variable rstart rend : rnum.

Variable rstartOk : rstart <> v0.
Variable rendOk : rend <> v0.


Lemma isNotSpacePtrOk : iseqOkS Pscanf isNotSpacePtrPre 
  (isNotSpacePtrGuar rstart rend) (isNotSpacePtr rstart rend).
Proof.
  iseqWeakSolver PscanfSimpl.

  unfold isNotSpacePtrPre.
  unfold returnPre; unfold returnGuar.
  intros s _.
  rewrite andi1Step; stSimpl.
  rewrite sltuStep; stSimpl.
  rewrite addiuStep; stSimpl.
  rewrite sltuStep; stSimpl.
  rewrite addiuStep; stSimpl.
  repeat (rewrite btneEq).
  repeat (rewrite beqSBoolEq).
  repeat (rewrite Bool.negb_elim).
  
  assert (R : isOdd (rget s a0) = true \/ isOdd (rget s a0) = false).
  destruct (isOdd (rget s a0)); auto.
  destruct R.

  rewrite H.
  constructor.
  trivial.
  intros; subst.
  stSimpl.
  constructor.
  simpl; left.
  assert (Q := isOddOk (rget s a0)).
  rewrite H in Q; auto.
  constructor.
  reflexivity.
  intros; stSimpl; reflexivity.

  rewrite H.
  assert (rget s a0 < rget s rstart \/ rget s a0 >= rget s rstart).
  omega.
  destruct H0.
  rewrite blt_true.
  2:assumption.
  constructor.
  trivial.
  intros; subst.
  stSimpl; constructor.
  simpl; right.
  intro.
  assert (X := setRangeElim H1).
  omega.

  constructor.
  reflexivity.
  intros; stSimpl; reflexivity.

  rewrite blt_false.
  2:assumption.
  assert (rget s rend < 1 + rget s a0  \/ rget s rend >= 1 + rget s a0).
  omega.
  destruct H1.
  rewrite blt_true.
  2:assumption.
  constructor.
  trivial.
  intros; subst.
  stSimpl; simpl.
  constructor.
  2:intuition.
  right.
  intros Y.
  assert (X := setRangeElim Y).
  omega.
  
  stSimpl; reflexivity.

  rewrite blt_false.
  2:assumption.
  constructor.
  trivial.
  intros; subst; stSimpl; simpl.
  constructor.
  intro Z; destruct Z.
  assert (Q := isOddOk (rget s a0)).  
  rewrite H in Q.
  eapply not_even_and_odd; eauto.
  apply H2; clear H2.
  apply setRangeIntro.
  omega.

  constructor.
  reflexivity.
  intros; stSimpl; reflexivity.

Qed.

End INSP.


Lemma isNotToSpacePtrOk : iseqOkS Pscanf isNotSpacePtrPre isNotToSpacePtrGuar isNotToSpacePtr.
Proof.
  unfold isNotToSpacePtrGuar; unfold isNotToSpacePtr.
  apply isNotSpacePtrOk; discriminate.
Qed.


Lemma isNotFromSpacePtrOk : 
  iseqOkS Pscanf isNotSpacePtrPre isNotFromSpacePtrGuar isNotFromSpacePtr.
Proof.
  unfold isNotFromSpacePtrGuar; unfold isNotFromSpacePtr.
  apply isNotSpacePtrOk; discriminate.
Qed.


Lemma fwdObjOk : iseqOkS Pscanf fwdObjPre fwdObjGuar fwdObj.
Proof.
  unfold fwdObjPre; unfold fwdObjGuar; unfold fwdObj.
  iseqWeakSolver PscanfSimpl.

  intros s H.
  unfold doBtest.
  compare (rget s rfree) (rget s ralloc); [intro Eq0 | intro Ne0].

  rewrite Eq0; rewrite beq_true.
  constructor.
  trivial.
  intros s' H'; elimtype False; assumption.

  hypReplace H (H Ne0).
  rewrite (beq_false Ne0).
  simpl in H.
  linLogElimOld H.
  assert (QQ := H).
  rewrite (lwStep H); try reflexivity.
  sconjLift3 H.
  rewrite (swStep H); try reflexivity.
  stSimpl.
  unfold WORDSIZE.
  assert (Z := contUpdateSt H x); clear H.
  assert (Q := Z); sconjLift3 Q.
  rewrite (lwStep Q); try reflexivity.
  clear Q.
  sconjLift4 Z.
  rewrite (swStep Z); try reflexivity.
  stSimpl.
  assert (Z2 := contUpdateSt Z x0).
  sconjLift3 Z2.
  rewrite (swStep Z2); try reflexivity.
  assert (ZQ := contUpdateSt Z (rget s rfree)).
  clear Z Z2.
  rewrite addiuStep.
  rewrite addiuStep.

  constructor.
  trivial.

  stSimpl.
  intros.
  subst s'.
  constructor.

  clear ZQ.
  simpl.
  intros.
  linLogIntroOld.
  linLogElimOld H.

  assert (R1 := contContEq QQ H).
  assert (x0 = y).
  sconjLift2 QQ; sconjLift2 H.
  exact (contContEq QQ H).
  assert (x1 = x').
  sconjLift3 QQ; sconjLift3 H.
  exact (contContEq QQ H).
  assert (x2 = y').
  sconjLift4 QQ; sconjLift4 H. 
  exact (contContEq QQ H).
  subst.

  sconjLift3 H.
  hypReplace H (contUpdateSt H x3).
  sconjLift4 H.
  hypReplace H (contUpdateSt H y).
  sconjLift3 H.
  hypReplace H (contUpdateSt H (rget s rfree)).
  apply (sconjImplR H); clear H; intros h H.
  sconjLift3 H.
  apply (sconjImplR H); clear h H; intros h H.
  sconjLift2 H.
  assumption.

  stSimpl.
  intuition.

  unfold regPres.
  intros; simpl in H.
  intuition.
  stSimpl.
  reflexivity.

Qed.


Lemma isNotToSpaceRaOk : forall s s' : state,
  isNotToSpacePtrGuar s s' -> rget s ra = rget s' ra.
Proof.
  unfold isNotToSpacePtrGuar.
  intros s s' [_ [_ Rp]].
  rewrite Rp; stSimpl.
  reflexivity.
  discriminate.
Qed.


(* move elsewhere *)
Lemma regEqDec : forall (r1 r2:rnum), r1 = r2 \/ r1 <> r2.
Proof.
  intros r1 r2.
  compare r1 r2; auto.
Qed.


(* not quite what i want *)
Ltac rSimpl r :=
  match goal with
    | Req : forall r, r <> ?r0 -> rget ?s' r = rget ?s r |- _ =>
      rewrite (Req r); [stSimpl | discriminate]
  end.


Lemma scanField2Ok : iseqOkS Pscanf scanField2Pre scanField2Guar scanField2.
Proof.
  unfold scanField2.
  iseqWeakSolver PscanfSimpl.
  
  intros; apply isNotToSpaceRaOk; assumption.

  unfold scanField2Pre.
  intros s F.
  unfold doBtest.
  stSimpl.
  compare (rget s v0) 0; intro E.

  rewrite E; rewrite beq_true.
  replace (negb true) with false; [idtac | reflexivity].
  destruct (F E) as [v [Hp Pr]]; clear F.
  rewrite (lwStep Hp); try reflexivity.
  rewrite addiuStep0; stSimpl.
  constructor.
  constructor.
  unfold isNotSpacePtrPre; trivial.
  unfold isNotToSpacePtrGuar; stSimpl.
  intros s' [B [Heq Req]].
  hypReplace Pr (Pr _ B).
  generalize dependent Pr.
  unfold scanPointerPre; stSimpl.
  unfold scanNoCopyPre; stSimpl.
  rewrite Heq.
  rSimpl rtemp0.
  rSimpl rtemp1.
  rSimpl rfree.
  rSimpl ralloc.
  auto.

  intros s'' [s' [G1 G2]].
  destruct G1 as [B [Heq Req]].
  simpl in Heq.
  generalize dependent B; stSimpl; intro.
  hypReplace Pr (Pr _ B).
  unfold scanField2Guar.
  rewrite E; rewrite beq_true.
  exists v; constructor.
  assumption.
  exists (rget s' v0).
  constructor.
  assumption.
  clear B Pr.
  generalize dependent G2; unfold scanPointerGuar; stSimpl.

  compare (rget s' v0) 0; intro R.
  rewrite R; rewrite beq_true.
  unfold scanNoCopyGuar; stSimpl.
  rSimpl rtemp0.
  rSimpl rtemp2.
  rewrite Heq.
  intros [G2 Req'].
  constructor.
  assumption.

  intro r.
  rewrite Req'; clear Req'.
  replace (rget s' raSave0) with (rget s raSave0).
  destruct (regEqDec r ra) as [Y | Y].
  subst r; stSimpl; reflexivity.
  stSimpl.
  destruct (regEqDec r v0) as [Z | Z].
  subst r; stSimpl; auto.
  stSimpl.
  rewrite Req.
  stSimpl.
  reflexivity.
  assumption.
  rSimpl raSave0; reflexivity.

  rewrite beq_false.
  2:assumption.
  intros [sz [G1 G2]]; exists sz.
  constructor.
  2:assumption.
  generalize dependent G1; unfold fwdObjGuar; stSimpl.
  simpl in Heq; rewrite Heq; clear Heq.
  rewrite (Req rtemp1); stSimpl.
  rewrite (Req rfree); stSimpl.
  rewrite (Req ralloc); stSimpl.
  intuition.
  clear H.
  simpl; intros r [Ne1 [Ne2 [Ne3 _]]].
  rewrite <- H4.
  stSimpl.
  2:simpl; intuition.
  destruct (regEqDec ra r) as [Y | Y].
  subst r; stSimpl; reflexivity.
  destruct (regEqDec a0 r) as [X | X].
  subst r; stSimpl; reflexivity.
  symmetry; stSimpl.
  rewrite Req; stSimpl.
  reflexivity.
  assumption.
  discriminate.
  discriminate.
  discriminate.

  unfold scanField2Guar.
  rewrite beq_false.
  2:assumption.
  replace (negb false) with true.
  2:reflexivity.
  clear F.
  constructor.
  unfold returnPre; trivial.
  unfold returnGuar; auto.
Qed.


(* lower level version of scanFieldGuar that 
    should be easier to verify scanField using *)
Definition scanFieldGuar0 s s' :=
  (forall phi W M F A,
    scanFieldStOk phi W M F 
    (okFieldAt ((W |_| M) |_| toSpace s) (rget s a0) ** A) s ->
    ((rget s rfree = rget s' rfree /\
      (forall x, set_In x W -> (heapOf s' ||- objCopied x s x ** top)) /\
      scanFieldStOk phi W M F 
      (fieldRespToMap phi s (rget s a0) ** A) s') \/
    (rget s rfree <> rget s ralloc /\
      8 + rget s rfree = rget s' rfree /\
      exists v, heapOf s ||- rget s a0 |-> v ** top /\ set_In v W /\
	(forall x, set_In x (set_remove v W) -> (heapOf s' ||- objCopied x s x ** top)) /\
	scanFieldStOk (map_add phi v (rget s rfree))
        (set_remove v W) (set_add v M) 
        (set_remove (rget s rfree) F) 
        (rget s a0 |-> rget s rfree ** 
	  objCopied v s (rget s rfree) ** A) s'))) /\
  regPres 
    (regNotInList (rfree::a0::v0::ast::rtemp0::rtemp1::rtemp2::raSave0::nil)) s s'.


(* scanFieldGuar0 is stronger than scanFieldGuar *)
Lemma scanFieldGuar0Stronger : forall s s',
  scanFieldGuar0 s s' ->
  scanFieldGuar s s'.
Proof.
  intros s s' [X Rp].
  constructor.
  intros phi W M F A H.
  assert (Y := X _ _ _ _ _ H); clear X; rename H into UU.

  assert (RM : set_eqv (M |_| empty_set) M).
  setoid_replace (M |_| empty_set) with (empty_set |_| M).
  simpl; apply set_eqvRefl.
  apply set_eqvUnionSymm.

  destruct Y.

  destruct H as [FrEq H].
  exists empty_set; exists W; exists F; exists empty_set;
    exists (fun n:nat => None (A := nat)).
  constructor.
  apply set_eqvUnionEmptyUnit.
  constructor.
  apply set_eqvUnionEmptyUnit.
  constructor.
  unfold iso_set.
  simpl.
  intuition.
  discriminate H.
  discriminate H.
  destruct H as [Wcopy H]; destrSF H.
  unfold scanFieldStOk.
  intuition.
  rewrite FrEq; apply aligned8Refl.
  
  setoid_rewrite RM.
  assumption.
  apply (NSP x y).
  unfold map_union in H.
  destruct (phi x); assumption.
  assumption.

  clear UU FrEq Wcopy WMdisj WMsub Feq NSP Le1 Le2 Le3.
  apply (sconjImpl Hp); clear Hp; intros h H.
  unfold hpredTrue.
  setoid_rewrite RM.
  unfold mapHp; unfold mapHp in H.
  linLogElimOld H.
  rewrite <- hpredTrueEq.
  linLogIntroOld.
  unfold map_union.
  intros x y Z.
  apply (H2 x y).
  destruct (phi x); assumption.
  eapply fasImpl.
  apply H.
  simpl; clear h H; intros x h H.
  linLogElim H.
  linLogIntro.
  2:eauto.
  unfold map_union; destruct (phi x); eauto.

  sconjMatch.
  unfold hpredTrue; setoid_rewrite RM; assumption.
  unfold fieldRespToMap in Hp |- * .
  unfold fieldRespMap in Hp |- * .
  linLogElim Hp.
  destruct H1 as [v [Heq Meq]].
  linLogIntro.
  exists v; intuition.
  unfold mapX in * |- * .
  destruct (isOdd v).
  eauto.
  rewrite <- Meq.
  unfold map_unionnn; unfold map_union.
  destruct (spaceIdMap (rget s rtoStart) (rget s rtoEnd) v).
  reflexivity.
  destruct (phi v); reflexivity.
  assumption.
  hypReplace Hp (unitIL Hp).
  sconjMatch.


  (* other case *)
  destruct H as [Ne [FrEq [v [vEq [vIn [Y0 H]]]]]].
  exists (set_add v empty_set); exists (set_remove v W).
  exists (set_remove (rget s rfree) F); exists (set_add (rget s rfree) empty_set).
  exists (fun x => if beq_nat x v then Some (rget s rfree) else None).

  assert (Pvn := phiWEq UU _ vIn).
  assert (vptr := scanFieldPtrW UU _ vIn).

  constructor.
  simpl; apply set_eqvSymm; apply addRemoveEqv; assumption.
  constructor.

  assert (I : set_In (rget s rfree) F).
  destrSF UU.
  setoid_rewrite Feq.
  destruct AlF as [k AlF].
  eapply rangeObjsIntro.
  omega.
  instantiate (1 := 0).
  omega.

  simpl; apply set_eqvSymm; apply addRemoveEqv; assumption.

  constructor.
  unfold iso_set.
  simpl; intuition.
  subst; rewrite beq_true in H1; discriminate H1.
  compare x v; intro R.
  subst; left; reflexivity.
  rewrite (beq_false R) in H0.
  right; apply H0; reflexivity.
  compare x v; intro R.
  subst; rewrite beq_true in H0; injection H0; clear H0; intro; subst.
  left; reflexivity.
  rewrite beq_false in H0; [discriminate H0 | assumption].

  compare x v; intro R.
  compare x' v; intro R'.
  rewrite R; rewrite R'; reflexivity.
  rewrite beq_false in H1; [discriminate H1 | assumption].
  rewrite beq_false in H0; [discriminate H0 | assumption].
  exists v; rewrite H1; rewrite beq_true; reflexivity.
  2:assumption.

  unfold scanFieldStOk in H |- * .

  assert (MvEq : set_eqv (set_union M (set_add v empty_set)) (set_add v M)).
  setoid_rewrite (set_eqvUnionSymm M (set_add v empty_set)).
  simpl; apply set_eqvRefl.

  destrSF H.
  intuition.

  rewrite <- FrEq; exists 1; omega.

  setoid_rewrite MvEq; assumption.

  unfold map_unionnn in H; unfold map_union in H.
  apply (NSP x y); unfold FnMap.map_add.
  compare v x; intro R.
  subst v; rewrite Pvn in H.
  rewrite beq_true in H; rewrite beq_true.
  assumption.
  rewrite beq_false in H.
  rewrite beq_false.
  destruct (phi x); assumption.
  assumption.
  auto.
  assumption.

  clear Y0 UU FrEq Feq NSP Le1 Le2 Le3 AlF.
  
  apply (sconjImpl Hp); clear Hp; intros h Hp.
  unfold hpredTrue; setoid_rewrite MvEq.
  rewrite <- hpredTrueEq; unfold mapHp in Hp |- * .
  linLogElim Hp.
  linLogIntro.
  intros.
  unfold map_unionnn in H; unfold map_union in H.
  compare x v; intro R.
  rewrite R; apply set_add_intro1; reflexivity.
  apply (H1 x y).
  rewrite beq_false in H.
  unfold FnMap.map_add.
  rewrite beq_false.
  destruct (phi x); assumption.
  auto.
  assumption.

  eapply fasImpl.
  apply Hp.
  clear h Hp; intros x h Hp.
  simpl in Hp; linLogElim Hp.
  existsl x0.
  linLogIntroOld.
  unfold FnMap.map_add in H2; unfold map_union.
  compare x v; intro R.
  subst v; rewrite beq_true; rewrite beq_true in H2.
  rewrite Pvn; assumption.
  rewrite beq_false; rewrite beq_false in H2.
  rewrite H2; reflexivity.
  auto.
  assumption.
  auto.

  linLogIntro; sconjMatch.

  apply (sconjImpl Hp); clear h Hp; intros h Hp.
  unfold hpredTrue; setoid_rewrite MvEq; assumption.

  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply (sconjImpl Hp); clear h Hp; intros h Hp.
  unfold fieldRespToMap; unfold fieldRespMap.
  linLogIntro.
  exists v.
  constructor.
  assumption.

  unfold map_unionnn; unfold map_union; unfold mapX.
  rewrite beq_true.
  2:eauto.

  rewrite Pvn.
  replace (isOdd v) with false.
  assert (NITo : ~(rget s rtoStart <= v < rget s rtoEnd)).
  intro NI.

  assert (IFr : set_In v (fromSpace s)).
  unfold fromSpace in WMsub.
  rewrite <- Rp in WMsub.
  rewrite <- Rp in WMsub.
  unfold toSpace; apply WMsub.
  setoid_rewrite <- (unionRemoveAddEqv _ _ M vIn).
  apply set_union_intro1; assumption.
  regNotInListSolver.
  regNotInListSolver.
  unfold fromSpace in WMdisj; unfold toSpace in WMdisj.
  rewrite <- Rp in WMdisj; [idtac | regNotInListSolver].
  rewrite <- Rp in WMdisj; [idtac | regNotInListSolver].
  rewrite <- Rp in WMdisj; [idtac | regNotInListSolver].
  rewrite <- Rp in WMdisj; [idtac | regNotInListSolver].
  apply (set_disj_elim WMdisj v).
  assumption.
  apply setRangeIntro; assumption.
  unfold spaceIdMap.
  assert (D : v >= rget s rtoEnd \/ rget s rtoStart >= S v).
  omega.
  destruct D.
  rewrite blt_false; auto.
  destruct (blt_nat v (rget s rtoEnd)).
  rewrite blt_false; auto.
  reflexivity.

  assert (Ev := isOddOk v).
  destruct (isOdd v).
  clear Rp; elimtype False.
  eapply not_even_and_odd.
  2:apply Ev.
  apply ptrPeven.
  assert (X := isPtrOk v).
  assumption.
  reflexivity.

  apply (sconjImplL Hp); clear h Hp; intros h Hp.
  apply fasIntro.
  hypReplace Hp (unitIR Hp).
  apply (sconjImpl Hp); clear h Hp; intros h Hp.
  unfold objCopiedPhi; existsl v.
  rewrite beq_true.
  linLogIntro.
  assumption.
  apply fasIntroE; assumption.
  simpl; auto.
Qed.


Lemma scanFieldOk : iseqOkS Pscanf scanFieldPre scanFieldGuar scanField.
Proof.
  cut (iseqOkS Pscanf scanFieldPre scanFieldGuar0 scanField).
  intros H.
  apply (iseqOkWeak _ _ _ _ H); clear H.
  intros s H.
  constructor.
  assumption.
  intros s' H'.
  exact (scanFieldGuar0Stronger H').
  
  unfold scanFieldPre; unfold scanFieldGuar; unfold scanField.
  iseqWeakSolver PscanfSimpl.

  (* ra is preserved *)
  intros s s' _ [_ [_ C]].
  rewrite C.
  reflexivity.
  discriminate.

  (* proving the actual thing *)
  intros s [phi [W [M [F SF]]]].
  destrSF SF.
  unfold okFieldAt in Hp.
  linLogElimOld Hp; sconjLift4 Hp.
  rewrite (lwStep Hp); try reflexivity.
  rewrite addiuStep; stSimpl.
  rewrite addiuStep; stSimpl.
  rewrite addiuStep; stSimpl.
  simpl.

  assert (E : exists s', s' = 
    rset (rset (rset (rset (rset s rtemp1 x) rtemp0 (rget s a0)) raSave0 (rget s ra))
      a0 x) ra SCANFIELD2).
  econstructor; reflexivity.

  destruct E as [s' E].
  rename H9 into Xok.

  cut (isNotSpacePtrPre s' /\
    forall s'', isNotFromSpacePtrGuar s' s'' ->
      scanField2Pre s'' /\
      forall s3, scanField2Guar s'' s3 ->
	scanFieldGuar0 s s3).
  intros [Pre1 Guar1].
  rewrite <- E.
  constructor.
  constructor.
  assumption.
  intros s'' G2.
  destruct (Guar1 _ G2) as [HH _]; exact HH.
  intros s3 [s2 [G2 G3]].
  destruct (Guar1 _ G2) as [_ Guar2].
  exact (Guar2 _ G3).

  subst s'.
  constructor.
  unfold isNotSpacePtrPre; trivial.

  unfold isNotFromSpacePtrGuar; stSimpl.
  intros s'' [B [Heq Req]].

  constructor.

  (* proving the precondition *)
  unfold scanField2Pre.
  intro E.
  rewrite E in B; simpl in B.
  rewrite Req; stSimpl.
  destruct Xok as [Xok | Xok].
  elimtype False; apply B; left; assumption.
  rewrite Heq.
  destruct (set_union_elim Xok) as [IWM | ITo]; clear Xok.

  destruct (set_union_elim IWM) as [IW | IM]; clear IWM.

  assert (Q : s |||- (okFieldAt (W |_| M) x ** top) ** top).
  sconjLift3 Hp.
  apply (sconjImpl Hp); clear Hp; intros h Hp.
  hypReplace Hp (objHpElim IW Hp).
  unfold okObjAt in Hp; hypReplace Hp (sconjAssocL Hp).
  apply (sconjImplR Hp); intros; linLogIntroOld.
  linLogIntroOld.
  unfold okFieldAt in Q; linLogElimOld Q.
  exists x0; constructor.
  assumption.
  unfold scanPointerPre; stSimpl.
  intro v'; stSimpl.

  intro B'.
  compare v' 0; intro R.
  subst v'.
  simpl in B'; rewrite beq_true.

  unfold scanNoCopyPre.
  stSimpl.
  rewrite Req; [stSimpl | discriminate].
  exists x; rewrite Heq.
  apply (sconjImplR Hp); intros; linLogIntroOld.

  destruct v'.
  elimtype False; auto.
  simpl in B'; rewrite beq_false.
  2:discriminate.
  clear R v'.

  rewrite Req; [stSimpl | discriminate].
  rewrite Req; [stSimpl | discriminate].
  rewrite Req; [stSimpl | discriminate].
  rewrite Heq.
  exists x.
  apply (sconjImplR Hp); clear Hp; intros h Hp.

  assert (Hp' : h ||- (x |->__ ** top) ** (rget s rfree |->__ ** top) ** top).
  sconjLift2 Hp; apply (sconjImpl Hp); clear h Hp; intros h Hp.
  hypReplace Hp (objHpElim IW Hp).
  sconjMatch.
  unfold okObjAt in Hp; unfold okFieldAt in Hp; linLogElimOld Hp.
  linLogIntro; eauto.
  sconjLift2 Hp.
  apply (sconjImpl Hp); clear h Hp; intros h Hp.

  assert (InF : set_In (rget s rfree) F).
  setoid_rewrite Feq.
  destruct AlF as [k AlF].
  eapply rangeObjsIntro.
  omega.
  instantiate (1 := 0).
  omega.

  assert (HF := fasElim InF Hp).
  apply (sconjImpl HF); clear h HF Hp; intros h Hp.
  auto.
  linLogIntroOld.
  linLogIntroOld.

  hypReplace Hp' (sconjAssocL Hp').
  rewrite Req; [stSimpl | discriminate].
  apply (sconjImplR Hp'); clear h Hp' Hp; intros h Hp.

  sconjLift2 Hp; hypReplace Hp (sconjAssocL Hp).
  apply (sconjImplR Hp); intros; linLogIntroOld.

  assert (Hp' : s |||- ((lexists z, 
    !(~notSpacePtrP z (rget s rtoStart) (rget s rtoEnd)) ** x |-> z) ** top) ** top).
  sconjLift2 Hp.
  apply (sconjImpl Hp); clear Hp; intros h Hp.
  hypReplace Hp (mapHpElim1 IM Hp).
  linLogElimOld Hp; linLogIntroOld.
  existsl x0; linLogIntroOld.
  eapply NSP; eauto.
  apply Hp.
  linLogIntroOld.
  linLogElimOld Hp'.
  exists x0.
  constructor.
  assumption.
  intros v' B'.
  unfold scanPointerPre; stSimpl.
  destruct v'; [rewrite beq_true | rewrite beq_false].
  unfold scanNoCopyPre; stSimpl.
  rewrite Req; stSimpl.
  rewrite Heq.
  exists x; apply (sconjImplR Hp); intros; linLogIntroOld.
  discriminate.
  simpl in B'; elimtype False.
  replace (rget s'' rtoStart) with (rget s rtoStart) in B'.
  replace (rget s'' rtoEnd) with (rget s rtoEnd) in B'.
  auto.
  rewrite Req; [stSimpl; reflexivity | discriminate].
  rewrite Req; [stSimpl; reflexivity | discriminate].
  discriminate.

  elimtype False; apply B; right; intro IFr.
  apply (set_disj_elim WMdisj _ IFr ITo).

  discriminate.

  (* proving the guarantee *)
  intros s'''.
  assert (Hp' : s |||- rget s a0 |-> x ** top).
  apply (sconjImplR Hp); intros; linLogIntroOld.
  clear Hp; rename Hp' into Hp.

  intro G1.

  unfold scanFieldGuar0.

  constructor.
  clear phi W M F WMdisj WMsub Feq NSP Xok Le1 Le2 Le3 AlF.
  intros phi W M F A Hp'.

  unfold scanField2Guar in G1.
  compare (rget s'' v0) 0; intro E.
  rewrite E in G1; rewrite beq_true in G1.
  rewrite E in B; simpl in B.
  rewrite Heq in G1.
  generalize dependent G1.
  rewrite Req; [stSimpl | discriminate].
  rewrite Req; [stSimpl | discriminate].
  rewrite Req; [stSimpl | discriminate].

  rename Hp into Hp0.
  destrSF Hp'.
  sconjLift4 Hp.
  unfold okFieldAt in Hp; linLogElimOld Hp.
  assert (Z := contContEq Hp Hp0).
  subst x0.
  destruct H3 as [atom | Xin].
  elimtype False; apply B; left; assumption.

  destruct (set_union_elim Xin) as [IWM | ITo]; clear Xin.
  destruct (set_union_elim IWM) as [IW | IM]; clear IWM; [right | left].

  assert (Q : s |||- (okObjAt (W |_| M) x ** 
    objHp (set_union  W M) (set_remove x W)) **
    rget s a0 |-> x ** mapHp M phi ** (foralls x in F, x |->__) ** A).
  sconjLift3 Hp.
  apply (sconjImpl Hp); clear Hp; intros h Hp.
  hypReplace Hp (objHpElim IW Hp).
  assumption.
  assumption.

  clear Hp; rename Q into Hp.
  unfold okObjAt in Hp; unfold okFieldAt at 1 in Hp.
  linLogElimOld Hp.
  rename x0 into f0.
  assert (Hp' : s |||- x |-> f0 ** top).
  apply (sconjImplR Hp); intros; linLogIntroOld.
  destruct G1 as [vx [Hpx G1]].
  assert (Q := contContEq Hpx Hp).
  subst vx; clear Hpx.
  destruct G1 as [v' [B' G2]]; clear Hp'.
  generalize dependent G2; unfold scanPointerGuar; stSimpl.
  compare v' 0; intro E'.
  elimtype False; subst v'; simpl in B'; apply B'; clear B' Req Hp.
  destruct H3 as [Z | IWM]; [left | right].
  assumption.
  intro ITo.
  apply (set_disj_elim WMdisj f0).
  apply WMsub; assumption.
  assumption.
  
  rewrite beq_false.
  2:assumption.
  destruct v'.
  elimtype False; apply E'; reflexivity.
  simpl in B'; clear E'.
  clear B'.
  unfold fwdObjGuar; stSimpl.
  rewrite Heq.
  rewrite Req; [stSimpl | discriminate].
  rewrite Req; [stSimpl | discriminate].

  intros [sb [[G1 [AlEq [FrEq [v0Eq Rp']]] G2]]].
  unfold okFieldAt in Hp; linLogElimOld Hp; rename x0 into f1.
  simpl in G1.

  assert (Ne : rget s rfree <> rget s ralloc).
  replace (rget s ralloc) with (rget s'' ralloc).
  assumption.
  rewrite Req; [stSimpl; reflexivity | discriminate].

  constructor.
  assumption.

  assert (InF : set_In (rget s rfree) F).
  setoid_rewrite Feq.
  destruct AlF as [k AlF].
  eapply rangeObjsIntro.
  omega.
  instantiate (1 := 0).
  omega.

  assert (Hp' : s |||- (x |-> f0 ** 4+x |-> f1) **
    rget s rfree |->__ ** (foralls x in (set_remove (rget s rfree) F), x |->__) **
      objHp (W |_| M) (set_remove x W) ** rget s a0 |-> x ** mapHp M phi ** A).
  clear G1 Req Rp' NSP WMsub WMdisj.
  apply (sconjImplR (sconjAssocR Hp)); clear Hp; intros h Hp.
  apply sconjAssocL.
  sconjLift4 Hp.
  apply (sconjImplL Hp); clear h Hp; intros h Hp.
  
  exact (fasElim InF Hp).
  clear Hp; rename Hp' into Hp.
  simpl in Hp; linLogElimOld Hp.

  assert (Hp' : s |||- (x |->, (f0::f1::nil)) **
    (rget s rfree |->, (x0::x1::nil)) ** 
    (foralls x in (set_remove (rget s rfree) F), x |->__) **
    (objHp (W |_| M) (set_remove x W) &&
      foralls x in (set_remove x W), objCopied x s x) **
    rget s a0 |-> x ** mapHp M phi ** A).
  clear Req G1 NSP Rp' Feq Le1 Le2 Le3.
  simpl; linLogIntroOld.
  sconjLift6 Hp.
  hypReplace Hp (objCopiedRefl Hp).
  sconjLift6 Hp; sconjLift6 Hp; sconjLift6 Hp; sconjLift6 Hp; sconjLift6 Hp.
  assumption.
  
  clear Hp; rename Hp' into Hp.
  assert (Hp' := G1 _ _ _ _ _ Hp); clear G1.
  destruct G2 as [G2 Rf3].
  assert (E00 : rget sb rtemp0 = rget s a0).
  rewrite <- Rp'.
  stSimpl.
  rewrite Req; [stSimpl | discriminate].
  reflexivity.
  simpl.
  intuition; discriminate.

  rewrite E00 in G2.
  sconjLift5 Hp'; simpl in Hp'; linLogElimOld Hp'.
  hypReplace Hp' (G2 _ _ Hp'); clear G2.
  constructor.
  rewrite Rf3; stSimpl.
  symmetry; assumption.
  exists x.
  constructor.
  sconjLift5 Hp; apply (sconjImplR Hp); intros; linLogIntroOld.
  constructor.
  assumption.

  constructor.
  intros y IWy.
  assert (Hp2 : s''' |||- (objCopied y s y ** top) ** top).
  sconjLift7 Hp'.
  apply (sconjImpl Hp'); clear Hp Hp'; intros h Hp.
  destruct Hp as [_ Hp].
  hypReplace Hp (fasElim IWy Hp).
  apply (sconjImplR Hp); intros; linLogIntroOld.
  linLogIntroOld.
  linLogElimOld Hp2; assumption.

  (* show the final state is well-formed when we copy *)
  unfold scanFieldStOk.
  
  unfold fromSpace; unfold toSpace;
    rewrite Rf3; rewrite Rf3; rewrite Rf3; rewrite Rf3; stSimpl.
  rewrite <- Rp'.
  rewrite <- Rp'.
  rewrite <- Rp'.
  rewrite <- Rp'.
  stSimpl.
  rewrite Req; [stSimpl | discriminate].
  rewrite Req; [stSimpl | discriminate].
  rewrite Req; [stSimpl | discriminate].
  rewrite Req; [stSimpl | discriminate].

  constructor.
  assumption.
  constructor.
  setoid_rewrite <- (unionRemoveAddEqv _ _ M IW).
  assumption.

  rewrite Rf3; stSimpl.
  rewrite FrEq.

  constructor.
  setoid_rewrite Feq.
  rewrite Rf3; stSimpl.
  rewrite <- Rp'; [stSimpl | regNotInListSolver].
  rewrite Req; [stSimpl | discriminate].
  apply rangeRemoveEqv.

  assert (IF' : set_In (rget s rfree) (rangeObjs (rget s rfree) (rget s ralloc))).
  setoid_rewrite <- Feq; assumption.

  constructor.
  intros x2 y Meq Nsp.
  unfold FnMap.map_add in Meq.
  compare x x2; intro R.
  subst x2; rewrite beq_true in Meq.
  injection Meq; intros; subst; clear Meq.
  destruct Nsp.
  assert (HH : s |||- (!(ptrP (rget s rfree)) ** top) ** top).
  simpl in Hp; linLogElimOld Hp; sconjLift3 Hp.
  apply (sconjImpl Hp); intros hx HX.
  linLogIntroOld; exact (contElim1 HX).
  linLogIntroOld.
  linLogElimOld HH.
  eapply not_even_and_odd.
  eapply ptrPeven; eauto.
  eauto.
  apply H; clear H.
  destruct (rangeObjsElim IF') as [Hz [k Hz']].
  eapply setRangeIntro.
  omega.

  rewrite beq_false in Meq.
  apply (NSP x2 y); assumption.
  assumption.

  constructor.
  destruct (rangeObjsElim IF') as [D [k D']].
  constructor.
  omega.
  rewrite Rf3; stSimpl.
  rewrite <- Rp'; [stSimpl | regNotInListSolver].
  rewrite Req; [stSimpl | discriminate].
  assumption.

  constructor.
  rewrite Rf3; stSimpl.
  rewrite <- Rp'; [stSimpl | regNotInListSolver].
  rewrite Req; [stSimpl | discriminate].
  assumption.

  constructor.
  rewrite Rf3; stSimpl.
  rewrite <- Rp'; [stSimpl | regNotInListSolver].
  rewrite Req; [stSimpl | discriminate].
  destruct AlF as [k AlF].
  destruct k.
  elimtype False; omega.
  exists k.
  unfold OBJSIZE; omega.

  rewrite v0Eq in Hp'.
  clear Rf3 FrEq v0Eq sb Rp' E00.
  clear s'' Heq Req E AlEq.
  clear Hp0; rename Hp into Hp0; rename Hp' into Hp.
  sconjLift3 Hp; sconjLift3 Hp.
  hypReplace Hp (sconjAssocR Hp).
  sconjLift7 Hp.
  hypReplace Hp (sconjAssocR Hp).
  apply (sconjImpl Hp); clear s''' Hp; intros h Hp.
  eapply mapHpIntro; simpl; eauto.
  sconjLift5 Hp.
  apply (sconjImpl Hp); clear h Hp; intros h Hp.
  destruct Hp as [Hp _].
  unfold hpredTrue; setoid_rewrite <- (unionRemoveAddEqv _ _ M IW).
  assumption.
  sconjLift4 Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  hypReplace Hp (sconjAssocR Hp).
  apply (sconjImplL Hp); clear h Hp; intros h Hp.
  unfold objCopied; unfold fieldCopied.
  linLogIntroOld; existsl f0; linLogIntroOld; existsl f1.
  simpl in Hp0; linLogElimOld Hp0; linLogIntroOld.
  apply (sconjImplR Hp0); intros; linLogIntroOld.
  assumption.
  sconjLift2 Hp0; apply (sconjImplR Hp0); intros; linLogIntroOld.

  regNotInListSolver.
  regNotInListSolver.
  regNotInListSolver.
  regNotInListSolver.

  (* set_In x M *)
  sconjLift2 Hp.

  assert (Hp' : s |||- (lexists v, !(phi x = Some v) ** x |-> v ** top) ** top).
  apply (sconjImpl Hp); clear Req Hp; intros h Hp.
  exact (mapHpElim1 IM Hp).
  linLogIntroOld.

  linLogElimOld Hp'.
  destruct G1 as [vx [Hpx G1]].
  assert (Q := contContEq Hpx Hp').
  subst vx; clear Hpx.
  destruct G1 as [b' [B' G2]]; clear Hp'.
  generalize dependent G2; unfold scanPointerGuar; stSimpl.
  destruct b'; simpl in B'; [rewrite beq_true | rewrite beq_false].
  unfold scanNoCopyGuar; stSimpl.
  rewrite Req; [stSimpl | discriminate].
  rewrite Heq.
  intros [G1 Rf'].

  constructor.
  rewrite Rf'; stSimpl.
  rewrite Req.
  stSimpl; reflexivity.
  discriminate.

  sconjLift3 Hp.
  hypReplace Hp (objCopiedRefl Hp).
  sconjLift3 Hp.
  assert (Hp' := G1 _ _ Hp); clear G1.

  constructor.
  intros y IWy.
  sconjLift2 Hp'; destruct (sconjDistrib2 Hp') as [_ Hp3]; clear Hp'.
  clear Req Rf' Hp; rename Hp3 into Hp.
  assert (Hp' : s''' |||- (objCopied y s y ** top) ** top).
  apply (sconjImpl Hp); clear Hp; intros h Hp.
  hypReplace Hp (fasElim IWy Hp).
  apply (sconjImplR Hp); intros; linLogIntroOld.
  linLogIntroOld.
  linLogElimOld Hp'; assumption.

  unfold scanFieldStOk.
  unfold fromSpace; unfold toSpace;
    rewrite Rf'; rewrite Rf'; rewrite Rf'; rewrite Rf'; rewrite Rf'; rewrite Rf'; stSimpl.
  clear Rf'.
  rewrite Req; [stSimpl | discriminate].
  rewrite Req; [stSimpl | discriminate].
  rewrite Req; [stSimpl | discriminate].
  rewrite Req; [stSimpl | discriminate].
  rewrite Req; [stSimpl | discriminate].
  rewrite Req; [stSimpl | discriminate].
  clear s'' Req E Heq.
  constructor.
  assumption.
  constructor.
  assumption.
  constructor.
  assumption.
  constructor.
  assumption.
  constructor.
  omega.
  constructor.
  assumption.
  constructor.
  assumption.
  clear Hp; rename Hp' into Hp.
  sconjLift3 Hp; apply (sconjImplR Hp); clear Hp; intros h Hp.
  sconjLift2 Hp; apply (sconjImpl Hp); clear h Hp; intros h Hp.
  destruct Hp; assumption.
  sconjLift2 Hp; apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply (sconjImplL Hp); clear h Hp; intros h Hp.
  unfold fieldRespToMap; unfold fieldRespMap.
  existsl x0; linLogIntroOld.
  exists x; constructor.
  assumption.
  unfold mapX.
  assert (IO := isOddOk x); destruct (isOdd x).
  elimtype False; apply B; left; assumption.
  clear IO; unfold map_unionnn; unfold map_union.
  rewrite H3.
  
  assert (NITo : ~(rget s rtoStart <= x < rget s rtoEnd)).
  intro NI.
  assert (IFr : set_In x (fromSpace s)).
  apply WMsub; apply set_union_intro2; assumption.
  apply (set_disj_elim WMdisj x).
  assumption.
  unfold toSpace; apply setRangeIntro; assumption.
  unfold spaceIdMap.
  assert (D : x >= rget s rtoEnd \/ rget s rtoStart >= S x).
  omega.
  destruct D.
  rewrite blt_false; auto.
  destruct (blt_nat x (rget s rtoEnd)).
  rewrite blt_false; auto.
  reflexivity.
  assumption.

  elimtype False; apply (NSP _ _ H3); assumption.
  
  discriminate.

  elimtype False; apply B; right.
  intro IFr; apply (set_disj_elim WMdisj x); assumption.
  
  (* intial field value is not a to-space pointer *)
  left.
  rewrite beq_false in G1.
  2:assumption.
  subst s'''; stSimpl.
  destruct (rget s'' v0).
  elimtype False; auto.
  clear E.
  simpl in B.
  constructor.
  rewrite Req; [stSimpl | discriminate]; reflexivity.
  rename Hp into Hp0; rename Hp' into Hp.
  destrSF Hp.
  constructor.
  rewrite Heq; clear Req Heq.
  intros y IWy.
  sconjLift2 Hp1; hypReplace Hp1 (objCopiedRefl Hp1).
  destruct (sconjDistrib2 Hp1) as [_ Hp']; clear Hp1.
  
  assert (Hp2 : s |||- (objCopied y s y ** top) ** top).
  apply (sconjImpl Hp'); clear Hp'; intros h Hp.
  hypReplace Hp (fasElim IWy Hp).
  apply (sconjImplR Hp); intros; linLogIntroOld.
  linLogIntroOld.
  linLogElimOld Hp2; assumption.

  unfold scanFieldStOk; stSimpl.
  unfold fromSpace; unfold toSpace.
  repeat (rewrite Req; [stSimpl | discriminate]).
  rewrite Heq.
  clear Req Heq.
  constructor.
  assumption.
  constructor.
  assumption.
  constructor.
  assumption.
  constructor.
  assumption.
  constructor.
  omega.
  constructor.
  assumption.
  constructor.
  assumption.

  rename Hp1 into Hp.
  unfold okFieldAt in Hp; linLogElimOld Hp.
  sconjLift4 Hp.
  assert (Q := contContEq Hp Hp0); subst x0.
  sconjLift4 Hp; sconjLift4 Hp; sconjLift4 Hp.
  apply (sconjImplR Hp); clear Hp; intros h Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply (sconjImplL Hp); clear h Hp; intros h Hp.
  unfold fieldRespToMap; unfold fieldRespMap.
  existsl x; linLogIntroOld.
  exists x; constructor.
  assumption.
  unfold mapX.
  assert (IO := isOddOk x).
  destruct B as [Oddx | Itox].
  destruct (isOdd x).
  reflexivity.
  elimtype False; apply (not_even_and_odd x); assumption.
  destruct (isOdd x).  
  reflexivity.
  unfold map_unionnn; unfold map_union.
  destruct H9 as [IO' | IWMT].
  elimtype False; apply (not_even_and_odd x); assumption.
  clear IO.
  destruct (set_union_elim IWMT) as [IWM | ITo]; clear IWMT.
  elimtype False; apply Itox.
  apply WMsub; assumption.
  unfold toSpace in ITo; destruct (setRangeElim ITo).
  unfold spaceIdMap; rewrite blt_true.
  rewrite blt_true.
  reflexivity.
  omega.
  assumption.
  assumption.

  (* regPres solving *)
  clear WMdisj WMsub Feq NSP Le1 Le2 Le3 AlF W M F Xok.
  unfold regPres; simpl.
  generalize dependent G1.
  unfold scanField2Guar.
  rewrite Heq; clear Heq.
  rewrite (Req rtemp1); [stSimpl | discriminate] .
  rewrite (Req rtoStart); [stSimpl | discriminate].
  rewrite (Req rtoEnd); [stSimpl | discriminate].
  destruct (beq_nat (rget s'' v0)).
  intros [v [_ [v' [_ G2]]]].
  generalize dependent G2; unfold scanPointerGuar; stSimpl.
  rewrite (Req rtemp1); [stSimpl | discriminate].
  destruct (beq_nat v' 0).
  unfold scanNoCopyGuar; stSimpl.
  intros [_ Re3] r Ne.
  intuition.
  rewrite Re3; clear Re3.
  destruct (regEqDec r ra).
  subst r; stSimpl.
  rewrite Req; [stSimpl | discriminate]; reflexivity.
  stSimpl.
  rewrite Req; [stSimpl | auto]; reflexivity.

  unfold fwdObjGuar; stSimpl.
  intros [s''b [[_ [_ [_ [_ Rp]]]] G]].
  generalize dependent G; unfold scanCopiedGuar.
  intros [_ Re].
  intuition.
  rewrite Re; clear Re.
  destruct (regEqDec r ra).
  subst r; stSimpl.
  rewrite <- Rp; [stSimpl | regNotInListSolver].
  rewrite Req; [stSimpl | discriminate]; reflexivity.
  stSimpl.
  rewrite <- Rp; [stSimpl | regNotInListSolver]; clear Rp s''b Hp v v'.
  rewrite Req; stSimpl; trivial.
  intro; subst s'''; clear Hp.
  intuition.
  rewrite Req.
  destruct (regEqDec r ra).
  subst r; stSimpl; reflexivity.
  stSimpl.
  rewrite Req; [stSimpl; reflexivity | assumption].
  discriminate.

Qed.
