Require Export SCAP.
Require Import List.
Require Import StateTheory.


Set Implicit Arguments.

Definition sbool (n:nat) (P:Prop) :=
  match n with
    | O => ~P
    | _ => P
  end.

Definition emptyP : chspec := fun _ => NoneT _.

Fixpoint regInList (l:list rnum) {struct l} : rnum -> Prop :=
  fun r =>
    match l with
      | nil => False
      | r'::l' => r = r' \/ regInList l' r
    end.

Fixpoint regNotInList (l:list rnum) {struct l} : rnum -> Prop :=
  fun r =>
    match l with
      | nil => True
      | r'::l' => r <> r' /\ regNotInList l' r
    end.

Ltac regNotInListSolver' :=
  solve [trivial | constructor; [auto ; discriminate | regNotInListSolver']].
Ltac regNotInListSolver :=
  simpl; regNotInListSolver'.

Definition regPres (P:rnum -> Prop) s s' :=
  forall r, P r -> rget s r = rget s' r.


Ltac discr :=
  match goal with
    | X : Some ?x = None |- _ => discriminate X
    | X : ?x = ?y |- _ => discriminate X 
  end.


(* MIPS register conventions *)

Definition calleeSaved :=
  regInList (ra::s0::s1::s2::s3::s4::s5::s6::s7::nil).

Definition calleeSavedOk s s' :=
  regPres calleeSaved s s'.


(* this is out of date.. OR IS IT??? *)
Ltac calleeSavedOkSolver :=
  intro r; destruct r; simpl; trivial.

Hint Extern 5 (calleeSavedOk ?X ?Y) => calleeSavedOkSolver.

Ltac calleeSavedSolver Y :=
  unfold calleeSaved in Y; simpl in Y; decompose sum Y; discr.



Lemma regPresEq : forall P s s',
  regPres P s s' -> 
  forall r, P r ->
  rget s' r = rget s r.
Proof.
  intros P s s' X r Pok.
  symmetry.
  apply X; assumption.
Qed.


Ltac regPresEqs Rp rr :=
  match rr with
    | nil => stSimpl
    | ?r::?rr' => (rewrite (regPresEq Rp r); 
      [regPresEqs Rp rr' | regNotInListSolver])
  end.


(* simplifies a common pattern that arises when using WeakSCAP *)
Lemma callChain : forall s0 s (p p':spred) (g g' g'':guar),
  (p s /\ forall s', g s s' ->
    p' s' /\ forall s'', g' s' s'' ->
      g'' s0 s'') ->
  (p s /\ forall s', g s s' -> p' s') /\
  (forall s'', (exists s', g s s' /\ g' s' s'') ->
    g'' s0 s'').
Proof.
  intros s0 s p p' g g' g'' [Pre G].
  constructor.
  constructor.
  assumption.
  intros s' G'; destruct (G _ G'); assumption.
  intros s'' [s' [G' G'']].
  destruct (G _ G'); eauto.
Qed.


