Require Export OptUtil.
Require RFile.
Require Heap.
Require Export RNum32.
Require Export NatUtil.


Set Implicit Arguments.

Definition rfile := RFile.rfile.
Definition heap := Heap.heap.
Definition state := prod Heap.heap rfile.

Definition heapOf (s:state) : heap := fst s.
Definition rfileOf (s:state) : rfile := snd s.

Definition rget (s:state) (r:rnum) : nat := RFile.get (rfileOf s) r.
Definition rset (s:state) (r:rnum) (v:nat) : state :=
  (heapOf s, RFile.set (rfileOf s) r v).

Definition hget (s:state) (a:nat) : option nat := Heap.get (heapOf s) a.
Definition hset (s:state) (a:nat) (hv:nat) : option state :=
  do h <-- Heap.set (heapOf s) a hv;
    ret (h, rfileOf s).


Inductive comm : Set :=
  | addu : rnum -> rnum -> rnum -> comm
  | addiu : rnum -> rnum -> nat -> comm
  | subu : rnum -> rnum -> rnum -> comm
  | andi1 : rnum -> rnum -> comm
  | sltu : rnum -> rnum -> rnum -> comm
  | lw : rnum -> nat -> rnum -> comm
  | sw : rnum -> nat -> rnum -> comm.


Definition cstep (c:comm) (s:state) : option state :=
  match c with
    | addu rd rs rt => 
      ret (rset s rd (rget s rs + rget s rt))
    | addiu rd rs i => 
      ret (rset s rd (i + rget s rs))
    | subu rd rs rt =>
      ret (rset s rd (rget s rs - rget s rt))
    | andi1 rd rs =>
      ret (rset s rd (if isOdd (rget s rs) then 1 else 0))
    | sltu rd rs rt =>
      ret (rset s rd (if blt_nat (rget s rs) (rget s rt) then 1 else 0))
    | lw rd i rs =>
      do v <-- hget s (i + rget s rs);
	ret (rset s rd v)
    | sw rs i rd =>
      hset s (i + rget s rd) (rget s rs)
  end.


Coercion heapOf : state >-> heap.
